# Copyright start
# Copyright (C) 2008 - 2023 Fortinet Inc.
# All rights reserved.
# FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
# Copyright end
source_docker_utils() {
    local fsr_docker_utils_sh="/opt/cyops/scripts/docker/utils.sh"
    source $fsr_docker_utils_sh
}

init() {
    source_docker_utils
    prefix="/opt/cyops-api"
}

validate_input() {
    local context=$1
    local version=$2
    if [ "X$context" = "X" ] || [ "X$version" = "X" ]; then
        echo "$0 upgrade/common <version-here>"
        exit 1
    fi
}

update_proxy() {
    local conf="/etc/php-fpm.d/cyops-api.conf"
    if grep -q "PROXY" $conf; then
        # Since proxy is already added, return
        return
    fi
    echo "" >>$conf
    echo "env[HTTPS_PROXY]= \$https_proxy" >>$conf
    echo "env[HTTP_PROXY]= \$http_proxy" >>$conf
    echo "env[NO_PROXY]= \$no_proxy" >>$conf
}

handle_docker_upgrade() {
    local previous_version=$1

    # Run cache:clear so removed entities do not cause any issue in schema update command.
    sudo -u nginx php /opt/cyops-api/bin/console cache:clear

    # Running generate models to create custom entities back that were prior to upgrade
    sudo -u nginx php /opt/cyops-api/bin/console app:metadata:generate:models

    # Run cache:clear so removed entities do not cause any issue in schema update command.
    sudo -u nginx php /opt/cyops-api/bin/console cache:clear

    # Run schema update to ensure newly added system module changes are reflected back on database
    sudo -u nginx php /opt/cyops-api/bin/console doctrine:schema:update --force

    # Running Migrations for upgrade
    sudo -u nginx php /opt/cyops-api/bin/console doctrine:migrations:migrate --no-interaction

    sudo -u nginx php /opt/cyops-api/bin/console cache:clear
    sudo -u nginx php /opt/cyops-api/bin/console cache:warmup --no-debug

    # Added c in below command to avoid clearing php memory cache
    # Added b to avoid broadcasting to HA during upgrade.
    sudo -u nginx php /opt/cyops-api/bin/console app:system:update -lafcb
}
handle_rpm_upgrade() {

    sudo -u nginx php /opt/cyops-api/bin/console cache:clear

    # Running generate models to create custom entities back that were prior to upgrade
    # This is prominently required from centos to rl migration as file may have uuid type as base and data might be old
    sudo -u nginx php /opt/cyops-api/bin/console app:metadata:generate:models

    # Run cache:clear so removed entities do not cause any issue in schema update command.
    sudo -u nginx php /opt/cyops-api/bin/console cache:clear

    sudo -u nginx php /opt/cyops-api/bin/console doctrine:schema:update --force
    sudo -u nginx php /opt/cyops-api/bin/console doctrine:migrations:migrate --no-interaction
    sudo -u nginx php /opt/cyops-api/bin/console cache:warmup --no-debug
    sudo -u nginx php /opt/cyops-api/bin/console app:system:update -lackfb
}

handle_upgrade_steps() {
    local new_version=$1
    local previous_version=$2
    rm -rf /opt/cyops-api/var/cache/prod
    # Often observe that the above command does not run successfully giving following error.
    # rm: cannot remove '/opt/cyops-api/var/cache/prod': Directory not empty
    # This generally occurs if service is stuck or serving any active request.
    # On safer side add below code. This will check if command does not run successfully
    # It will restart the service and then reattempt to delete cache directory.
    if [ $? -eq 0 ]; then
        echo "Cache Deleted Successfully"
    else
        systemctl restart php-fpm fsr-api-consumer
        rm -rf /opt/cyops-api/var/cache/prod
    fi
    rm -rf /opt/cyops-api/config/packages/twig.yaml

    # Write post upgrade action here

    ln -sf /opt/cyops/configs/keys/APPLIANCE_PUBLIC_KEY ${prefix}/src/keys/appliance_public.key
    ln -sf /opt/cyops/configs/keys/APPLIANCE_PRIVATE_KEY ${prefix}/src/keys/appliance_private.key

    python3 /opt/cyops/scripts/confUtil.py -f /opt/cyops/configs/cyops-api/mergeConfig.yml

    # Enable Cache Clear for Drop MetaData Command to be Available in 5.2
    sudo -u nginx php /opt/cyops-api/bin/console cache:clear
    sudo -u nginx php /opt/cyops-api/bin/console app:drop:metadata:copy

    sleep 8
    restorecon -R ${prefix}

    # Run full commands for cyops-api packages to generate new classes and tables
    sudo chown -R nginx:nginx /opt/cyops-api/src/Entity/Custom/

    if [[ "$previous_version" < "7.4.2" ]]; then
        # Update Staging and Model Metadata fields with latest upgrade
        sudo -u nginx php /opt/cyops-api/bin/console doctrine:migrations:execute --up App\\Migrations\\Version20230616100128 -n
    fi
    if is_docker; then
        handle_docker_upgrade "$previous_version"
    else
        firewall-cmd --reload
        handle_rpm_upgrade
    fi
    f_nginx_conf=/etc/nginx/conf.d/cyops-api.conf
    if grep -Pzoq "index app.php;\n *add_header Cache-Control 'no-store, no-cache, max-age=0';" $f_nginx_conf; then
        echo "Skipping: no cache header already added"
    else
        echo "Updating nginx conf file to add no cache headers"
        sed -i '\/location \/auth\/ {/a \       \ add_header Cache-Control \"no-store, no-cache, max-age=0\"\;' /etc/nginx/conf.d/cyops-api.conf
        sed -i '\/location \/gateway\/ {/a \       \ add_header Cache-Control \"no-store, no-cache, max-age=0\"\;' /etc/nginx/conf.d/cyops-api.conf
        sed -i '\/location \/wf {/a \       \ add_header Cache-Control \"no-store, no-cache, max-age=0\"\;' /etc/nginx/conf.d/cyops-api.conf
        sed -i '\/location @siteupcrud {/a \       \ add_header Cache-Control \"no-store, no-cache, max-age=0\"\;' /etc/nginx/conf.d/cyops-api.conf
        echo "Added no cache headers to nginx conf file"
    fi
    
    restorecon -R ${prefix}
    # restart services
    systemctl daemon-reload
    systemctl restart nginx.service
}

handle_common_steps() {
    local new_version=$1
    sleep 5
    # Perf Improvement
    sudo -u nginx php /opt/cyops-api/bin/console cache:warmup --no-debug

    if ! grep -q "EnvironmentFile=/etc/environment" /usr/lib/systemd/system/php-fpm.service; then
        sed -i '/Type=notify/a EnvironmentFile=/etc/environment' /usr/lib/systemd/system/php-fpm.service
        sed -i 's/PrivateTmp=true/PrivateTmp=no/g' /usr/lib/systemd/system/php-fpm.service
    fi
    
    if ! grep -q ";opcache.preload=/opt/cyops-api/config/preload.php" /opt/cyops/configs/php/10-opcache.ini; then
        sed -i 's#opcache.preload=/opt/cyops-api/config/preload.php#;opcache.preload=/opt/cyops-api/config/preload.php#g' /opt/cyops/configs/php/10-opcache.ini
    fi

    update_proxy
    
    # reload php-fpm.service for EnvironmentFile changes
    systemctl daemon-reload
    systemctl restart php-fpm.service

    sudo -u nginx php /opt/cyops-api/bin/console rabbitmq:setup-fabric
    systemctl enable fsr-api-consumer.service
    systemctl daemon-reload
    systemctl restart fsr-api-consumer.service

    # Set selinux user system_u
    semanage fcontext -a -s system_u -t var_log_t "/var/log/cyops/cyops-api"

    # touch file so that restorecon can set the right selinux permissions
    sudo -u nginx touch /var/log/cyops/cyops-api/last_system_publish.log
    sudo -u nginx touch /var/log/cyops/cyops-api/last_config_export.log
    sudo -u nginx touch /var/log/cyops/cyops-api/last_config_import.log
    sudo -u nginx touch /var/log/cyops/cyops-api/last_contenthubsync.log
    chown nginx:nginx -R /var/log/cyops/cyops-api/

    restorecon -rF /var/log/cyops/cyops-api
    restorecon -rF /opt/cyops-api
}

############
# Main/main/MAIN starts here
#########
context=$1
version=$2
previous_version=$3
init
# This script usually runs from the specfile or docker upgrade. Hence add -x here for debug.
set -x
validate_input "$context" "$version"
case $context in
upgrade)
    handle_upgrade_steps "$version" "$previous_version"
    ;;
common)
    handle_common_steps "$version"
    ;;
esac
exit 0
