<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Command;

use Exception;
use InvalidArgumentException;
use App\Entity\ModelMetadata\ClusterCommand;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Output\ConsoleOutputInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Filesystem\Filesystem;

class ClusterPublishCommand extends Command
{
    protected $container;

    /** @var Filesystem $filesystem */
    private $filesystem;

    public function __construct(ContainerInterface $container)
    {
        $this->container = $container;
        $this->filesystem = new Filesystem();
        parent::__construct();
    }

    protected function configure()
    {
        $this
            ->setName('app:cluster:publish')
            ->setAliases(['cybersponse:cluster:publish'])
            ->setDefinition([
                new InputOption(
                    'nodeId',
                    'nodeId',
                    InputOption::VALUE_OPTIONAL | InputOption::VALUE_IS_ARRAY,
                    'Arguments for self nodeId.'
                ),
                new InputOption(
                    'sequenceId',
                    'sId',
                    InputOption::VALUE_OPTIONAL | InputOption::VALUE_IS_ARRAY,
                    'Arguments for the command sequence id.'
                ),
                new InputOption(
                    'skip-cache-clear',
                    'c',
                    InputOption::VALUE_NONE,
                    'Arguments for Skipping cache clear'
                )
            ]);
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $io = new SymfonyStyle($input, $output);
        $errorIo = $io;
        if ($output instanceof ConsoleOutputInterface) {
            // If it's available, get stdErr output
            $errorOutput = $output->getErrorOutput();
            $errorIo = new SymfonyStyle($input, $errorOutput);
        }

        $nodeId = $input->getOption('nodeId');
        $sequenceId = $input->getOption('sequenceId');
        $defaultCommandData = [
            'runCondition' => true,
            'insertNewline' => false
        ];

        $dasUri = $this->container->getParameter('das.uri');
        $dasClient = $this->container->get("app.proxy.client.das");
        $headers = ["Content-type" => "application/json"];
        $retryCount = 0;
        $status = "passive";
        while ($retryCount < 2)
        {
            $dasRequests = new GuzzleRequest('get', $dasUri . '/cluster/?self_only=true', $headers);
            try {
                $dasResponse = $dasClient->send($dasRequests);
                $dasResponseData = $dasResponse->getBody()->getContents();
                $jsonDecodeResponse = json_decode($dasResponseData);
                $status = $jsonDecodeResponse->{"nodes"}[0]->{"status"};
                break;
            } catch (BadRequestHttpException | \Exception $error) {
                error_log("Call to DAS failed with below message:");
                $retryCount += 1;
                sleep(2);
            }
        }

        $commandsToRun = [
            'System Backup' => [
                'command' => 'app:backup:all'
            ],
            'Generating Models' => [
                'command' => 'app:metadata:generate:models'
            ],
            'Clearing Doctrine Metadata' => [
                'command' => 'doctrine:cache:clear-metadata',
            ],
            'Clearing Cache' => [
                'command' => 'cache:clear'
            ],
            'Cache Warmup' => [
                'command' => 'cache:warmup',
            ],
            'Cleaning Up Old Backups' => [
                'command' => 'app:cleanup:all',
            ],
            'Prepare Class Type Map Cache' => [
                'command' => 'app:type:class:map',
            ],
            'Clearing App Cache' => [
                'command' => 'app:cache:clear',
            ],
        ];
        $skipCacheClear = $input->getOption('skip-cache-clear') ?? false;
        if ($skipCacheClear === true) {
            unset($commandsToRun['Clearing App Cache']);
        }

        $errorCode = 0;

        if ($status != 'active') {
            unset($commandsToRun['Prepare Class Type Map Cache']);
            unset($commandsToRun['Clearing App Cache']);
        }

        try {
            //Run all commands
            $current = 0;
            foreach ($commandsToRun as $processingText => $commandData) {
                $counter = 5;
                $commandData = $commandData + $defaultCommandData;

                $io->writeln(sprintf('<fg=blue>%s) %s</fg=blue>', $current + 1, $processingText));
                if ($commandData['runCondition'] === true) {
                    if ($commandData['insertNewline'] === true) {
                        $io->writeln('');
                    }
                    $command = $commandData['command'];
                    while($counter > 1) {
                        list($returnVal, $returnString) = $this->executeCommand($command);
                        $io->writeln($returnString);

                        if ($returnVal !== 0) {
                            $io->writeln(' -> <fg=magenta>Retrying command: ' . $command . ' </fg=magenta>');
                            $counter--;
                            sleep(2);
                        } else {
                            $counter = 0;
                        }
                    }

                    if ($returnVal !== 0) {
                        throw new \Exception(sprintf('Command `%s` finished with code: %s', $command, $returnVal));
                    }
                    $io->write(' -> <fg=green>Finished</fg=green>', true);
                } else {
                    $io->write(' -> <fg=yellow>Skipped</fg=yellow>', true);
                }

                $current++;
            }
        } catch (InvalidArgumentException | \Exception $e) {
            //Perform rollback
            $errorIo->error(sprintf('Failed: %s, Restoring from backup...', $e->getMessage()));
            $errorCode = 1;
            $commands = ['app:restore:all', 'doctrine:decrypt:database --no-interaction', 'doctrine:encrypt:database --no-interaction'];
            foreach ($commands as $command) {
                list($returnVal, $returnString) = $this->executeCommand($command);
                if ($returnVal !== 0) {
                    //Todo What happens when the system is not restored properly?
                    throw new \Exception(sprintf('Command: %s ,Error Code: %s, Could not restore system: %s', $command, $returnVal, $returnString));
                }
            }
            $io->success('Successfully restored up the system. Please view the above errors and fix.');
        } finally {
            if ($nodeId && $sequenceId) {
                $entityManager = $this->container->get('doctrine.orm.entity_manager');
                $clusterCommandRepository = $entityManager->getRepository(ClusterCommand::class);
                try {
                    $clusterCommand = $clusterCommandRepository->findOneBy([
                        'nodeId' => $nodeId,
                        'sequenceId' => $sequenceId
                    ]);
                } catch (InvalidArgumentException | \Exception $e) {
                    $io->error($e->getMessage());
                }
                if ($clusterCommand && $errorCode == 0) {
                    $entityManager->remove($clusterCommand);
                    $entityManager->flush();
                } else if ($clusterCommand) {
                    $clusterCommand->setStatus('failed');
                    $entityManager->update($clusterCommand);
                    $entityManager->flush();
                }
            }
            $logFile = '/var/log/cyops/cyops-api/last_system_publish.log';
            $this->filesystem->touch($logFile);
            $io->success('Finished with system update!');
        }

        return $errorCode;
    }

    /**
     * @param $command
     * @return array
     */
    private function executeCommand($command, $indexOption = "")
    {
        $fullCommand = sprintf('/usr/bin/php /opt/cyops-api/bin/console %s -n', $command);
        $execOutput = null;
        exec($fullCommand, $execOutput, $returnVal);
        return [$returnVal, $execOutput];
    }
}
