<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Command;

use App\Constants\AppConstants;
use App\Entity\Core\Widget;
use App\Entity\Core\SolutionPack;
use Doctrine\ORM\EntityManagerInterface;
use App\Service\SolutionPackUtilityService;
use Symfony\Component\Filesystem\Filesystem;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\HttpKernel\Exception\HttpException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;


class ContentHubSyncCommand extends Command
{
    const BULK_UPSERT_SP_URL = '%s/api/3/bulkupsert/solutionpacks';
    private $container;
    private $utilityService;
    private $entityManager;
    /** @var Filesystem $filesystem */
    private $filesystem;

    public function __construct(
        ContainerInterface $container,
        SolutionPackUtilityService $utilityService,
        EntityManagerInterface $entityManager
    ) {
        $this->container = $container;
        $this->utilityService = $utilityService;
        $this->entityManager = $entityManager;
        $this->filesystem = new Filesystem();
        parent::__construct();
    }

    protected function configure(): void
    {
        $this
            ->setName('app:contenthub:sync')
            ->setAliases(['fortisoar:contenthub:sync'])
            ->setDescription('This command allows you to sync solution packs from fortisoar repo.')
            ->setDefinition([
                new InputOption(
                    'force',
                    'f',
                    InputOption::VALUE_NONE,
                    'Will force the re-sync regardless if sync is in progress.'
                ),
                new InputOption(
                    'skip_modifydate',
                    's',
                    InputOption::VALUE_NONE,
                    'Will force the re-sync regardless of current modify date.'
                )
            ]);
    }

    protected function execute(InputInterface $input, OutputInterface $output): int
    {
        $beforeTimeStamp = (new \DateTime('NOW', new \DateTimeZone('UTC')))->getTimestamp();
        $io = new SymfonyStyle($input, $output);
        $rootDir = $this->container->getParameter('kernel.project_dir');
        $progressFile = sprintf('%s/workspace/.contenthubsync', $rootDir);
        if ($input->getOption('force') === true && file_exists($progressFile)) {
            unlink($progressFile);
        }
        if (file_exists($progressFile)) {
            $io->success('ContentHub Sync already in progress');
            return Command::SUCCESS;
        }
        try {
            $this->filesystem->touch($progressFile);
            $progressFile = realpath($progressFile);
            $repoUrl = getenv("REPOSERVER") ? getenv("REPOSERVER") : getenv("product_yum_server");
            $allSolutions = $this->utilityService->getSolutionPackFromRepo();
            if (is_null($allSolutions)) {
                $io->error('Cannot fetch solution packs from the repo server. Check the connectivity to the repo server. If the issue persists, contact your administrator.');
            }
            $bulkPayload = [];
            $installedWidgets = $this->getAllWidgets();
            $installedConnectors = $this->getInstalledConnectors();
            $syncedSolutionpacks = $this->getAllSolutionpacks();
            $syncedSolutionpacksKeysArray = array_keys($syncedSolutionpacks);
            $this->filesystem->dumpFile($progressFile, "Fetched the repo data content hub metadata json.");

            foreach ($allSolutions as $key => $value) {
                $allSolutionsKeysArray[] = $value->name;
                //in case of solution pack uploaded by but also exists on repo.
                if (array_key_exists($value->name, $syncedSolutionpacks) && $syncedSolutionpacks[$value->name]['local'] && $value->type == "solutionpack") {
                    continue;
                }
                if ($input->getOption('skip_modifydate') === true || !array_key_exists($value->name, $syncedSolutionpacks)  || (array_key_exists($value->name, $syncedSolutionpacks) && $syncedSolutionpacks[$value->name]['publishedDate'] < $value->publishedDate)) {
                    $repoVersionNumber = str_replace('.', '', $value->version);
                    $availableVersionNumber =  !empty($syncedSolutionpacks) && array_key_exists($value->name, $syncedSolutionpacks) && $syncedSolutionpacks[$value->name]['installed'] ? (str_contains($syncedSolutionpacks[$value->name]['version'], "dev") ? null : str_replace('.', '', $syncedSolutionpacks[$value->name]['version'])) : null;
                    $availableLatestCompatibilityVersion = array_key_exists($value->name, $syncedSolutionpacks) && $syncedSolutionpacks[$value->name]['latestCompatibleVersion'] ? $syncedSolutionpacks[$value->name]['latestCompatibleVersion'] : null;

                    // If there is a newer version of the SP in the repo, update it and move on
                    if ($availableVersionNumber && $availableVersionNumber < $repoVersionNumber) {
                        $this->updateExistingSolutionPack($value->name, $value->version, $value->publishedDate, $value->scm, $availableLatestCompatibilityVersion);
                        continue;
                    } elseif ($availableVersionNumber == $repoVersionNumber || $availableVersionNumber > $repoVersionNumber) {
                        $value->latestAvailableVersion = null;
                        // For widgets and connectors from the repo, if the SP is already installed, we only want to check for a later
                        // version and make the update available, no need to overwrite other details
                        if (($value->type=='widget' || $value->type=='connector') && array_key_exists($value->name, $syncedSolutionpacks) && !$syncedSolutionpacks[$value->name]['local'] && $syncedSolutionpacks[$value->name]['installed']) {
                            continue;
                        }
                    }

                    if (is_array($syncedSolutionpacks) && isset($syncedSolutionpacks['type']) && $syncedSolutionpacks['type'] == 'solutionpack') {
                        $repoFsrMinCompatibilityNumber = str_replace('.', '', $value->fsrMinCompatibility);
                        $availableFsrMinCompatibilityNumber =  !empty($syncedSolutionpacks) && array_key_exists($value->name, $syncedSolutionpacks) && $syncedSolutionpacks[$value->name]['installed'] ? str_replace('.', '', $syncedSolutionpacks[$value->name][AppConstants::FSR_MIN_COMPATIBILITY]) : null;

                        if ($availableFsrMinCompatibilityNumber && $availableFsrMinCompatibilityNumber < $repoFsrMinCompatibilityNumber) {
                            $this->updateExistingSolutionPack($value->name, $value->version, $value->publishedDate, $value->scm, $value->fsrMinCompatibility);
                            continue;
                        } elseif ($availableFsrMinCompatibilityNumber == $repoFsrMinCompatibilityNumber || $availableFsrMinCompatibilityNumber > $repoFsrMinCompatibilityNumber) {
                            $value->latestCompatibleVersion = null;
                        }
                    }

                    $value->publisher = $value->publisher ?? 'Fortinet';
                    if (in_array(strtolower($value->publisher), ["cybersponse", "fortinet"])) {
                        $value->publisher = 'Fortinet';
                    } elseif (in_array(strtolower($value->publisher), ["cybersponse community", "community"])) {
                        $value->publisher = 'Community';
                    }

                    if ($value->type == 'connector') {
                        $value->recordTags = $value->tags;
                    }

                    $value->label = $value->label ?? ($value->display ?? $value->title);
                    $value->sourceControlStats = $value->scm;
                    if (strpos($value->infoPath, $repoUrl) === false && !is_null($value->infoPath)) {
                        $value->infoPath = sprintf('https://') . $repoUrl . $value->infoPath;
                        $client = new \GuzzleHttp\Client();
                        $responseObject = $client->request('GET', $value->infoPath . '/info.json', ['verify' => false]);
                        $infocontent = json_decode($responseObject->getBody()->getContents(), true);
                        if (array_key_exists($value->name, $syncedSolutionpacks) && array_key_exists('postInstallConfig', $syncedSolutionpacks[$value->name]['infoContent']) && $syncedSolutionpacks[$value->name]['infoContent']['postInstallConfig']) {
                            $infocontent['postInstallConfig'] = $syncedSolutionpacks[$value->name]['infoContent']['postInstallConfig'];
                        }
                        $value->infoContent = $infocontent;
                    }
                    if (strpos($value->iconLarge, $repoUrl) === false && !is_null($value->iconLarge)) {
                        if (strpos($value->iconLarge, 'data:image') === false) {
                            $value->iconLarge = sprintf('https://') . $repoUrl . $value->iconLarge;
                        }
                    }
                    if (array_key_exists($value->name . "-" . $value->version, $installedWidgets)) {
                        $value->installed = true;
                        $value->recordId = $installedWidgets[$value->name . "-" . $value->version]->getUuid();
                        if (!$value->infoContent['subTitle']) {
                            $value->infoContent['subTitle'] = $installedWidgets[$value->name . "-" . $value->version]->getSubTitle();
                        }
                        unset($installedWidgets[$value->name . "-" . $value->version]);
                    }
                    if (array_key_exists($value->name . "-" . $value->version, $installedConnectors)) {
                        $value->installed = true;
                        $value->recordId = $installedConnectors[$value->name . "-" . $value->version]->id;
                        if (property_exists($value, 'help_online') || property_exists($value, 'help')) {
                            $value->infoContent['help'] = $value->help_online ?? $value->help;
                        }
                        $value->configCount = count($installedConnectors[$value->name . "-" . $value->version]->configuration ?? []);
                        unset($installedConnectors[$value->name . "-" . $value->version]);
                    }
                    $category = $this->utilityService->prepareCategoryData($value->category);
                    $value->category = $category;
                    $value->local = false;
                    if (!empty($syncedSolutionpacks) && array_key_exists($value->name, $syncedSolutionpacks) && $syncedSolutionpacks[$value->name]['installed']) {
                        $value->buildNumber = $syncedSolutionpacks[$value->name]['buildNumber'];
                    }
                    $allSolutions[$key] = $value;
                    $bulkPayload[] = $value;
                } else {
                    $this->updateScm($value->name, $value->scm);
                    if (array_key_exists($value->name . "-" . $value->version, $installedWidgets)) {
                        unset($installedWidgets[$value->name . "-" . $value->version]);
                    }
                    if (array_key_exists($value->name . "-" . $value->version, $installedConnectors)) {
                        unset($installedConnectors[$value->name . "-" . $value->version]);
                    }
                }
            }
            $this->entityManager->flush();
            $this->filesystem->dumpFile($progressFile, "Updated the content-hub solution packs.");

            //This diff in array gives names of solutionpacks present in db but not on repo(including local)
            $spInDbButNotOnRepo = array_diff($syncedSolutionpacksKeysArray, $allSolutionsKeysArray);
            $uuidArrayOfSpToRemoveFromDb = [];
            foreach ($spInDbButNotOnRepo as $spName) {
                if (array_key_exists($spName, $syncedSolutionpacks) && !empty($syncedSolutionpacks[$spName]) && !$syncedSolutionpacks[$spName]['local'] && !$syncedSolutionpacks[$spName]['installed']) {
                    $uuidArrayOfSpToRemoveFromDb[] = $syncedSolutionpacks[$spName]['uuid'];
                }
            }
            if (!empty($uuidArrayOfSpToRemoveFromDb)) {
                $deleteFullURI = sprintf('%s/api/3/delete/solutionpacks', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
                $this->sendRequest($deleteFullURI, 'delete', ['ids' => $uuidArrayOfSpToRemoveFromDb, 'nonLocalNonRepoSpClean' => true]);
            }
            foreach ($installedConnectors as $value) {
                if ($value) {
                    $value->label = $value->label ?? $value->display;
                    if (!is_null($value->category) || $value->category != "") {
                        $value->category = $this->utilityService->prepareCategoryData($value->category);
                    }
                    $value->infoContent = (array)$value;
                    // If there is already an instealled solution pack for this name and version, then use its local value instead of defaulting to metadata
                    $value->local = (array_key_exists($value->name,$syncedSolutionpacks) && $syncedSolutionpacks[$value->name]['version']==$value->version && $syncedSolutionpacks[$value->name]['installed']) ? $syncedSolutionpacks[$value->name]['local'] : !$value->rpm_installed;
                    $value->type = 'connector';
                    $value->infoContent['help'] = $value->help_online;
                    $value->installed = true;
                    $value->recordId = $value->id;
                    $value->configCount = count($value->configuration ?? []);
                    $value->status = 'Installed';
                    $bulkPayload[] = $value;
                }
            }
            $this->filesystem->dumpFile($progressFile, "Updated the installed connectors solution packs.");
            foreach ($installedWidgets as $value) {
                if ($value) {
                    $serializer = $this->container->get('serializer');
                    $widget = $serializer->normalize($value, 'jsonld', ['ignore_authorization' => true, 'ignore_field_authorization' => true, 'ignore_relationship_authorization' => true]);
                    $widget['label'] = $widget['title'] ?? $widget['label'];
                    $widget['infoContent'] = ['metadata' => $widget['metadata'], 'subTitle' => $widget['subTitle']];
                    if (array_key_exists('help_online', $widget['metadata'])) {
                        $widget['infoContent']['help'] = $widget['metadata']['help_online'];
                    }
                    $widget['type'] = 'widget';
                    $widget['recordId'] = $widget['uuid'];
                    $widget['development'] = $widget['draft'];
                    $widget['status'] = 'Installed';
                    $bulkPayload[] = $widget;
                }
            }

            $this->sendUpdateSolutionPackRequestInChunks(["name", "type", "local"], $bulkPayload);

            $bulkPayload = [];
            $developmentConnectors = $this->getDevelopmentConnectors();
            foreach ($developmentConnectors as $key => $value) {
                $category = [];
                $value->label = $value->label ?? $value->display;
                if ($value->category != "" || !empty($value->category)) {
                    $category = $this->utilityService->prepareCategoryData($value->category);
                }
                $value->category = $category;
                $value->infoContent = (array)$value;
                if (property_exists($value, 'help_online') || property_exists($value, 'help')) {
                    $value->infoContent['help'] = $value->help_online ?? $value->help ?? null;
                }
                $value->local = true;
                $value->type = 'connector';
                $value->installed = false;
                $value->recordId = $value->id;
                $value->iconLarge = $value->icon_large;
                $value->configCount = is_array($value->configuration) ? count($value->configuration) : 0;
                $bulkPayload[] = $value;
            }

            $this->sendUpdateSolutionPackRequestInChunks(["name", "version", "type"], $bulkPayload);

            $this->filesystem->dumpFile($progressFile, "Updated the development connectors solution packs.");
            $io->success('Sync from FortiSOAR Content-Hub complete');
            $afterTimeStamp = (new \DateTime('NOW', new \DateTimeZone('UTC')))->getTimestamp();
            $io->writeln("Command took " . ($afterTimeStamp - $beforeTimeStamp) . " Second to execute");
        } catch (\Exception $e) {
            $io->error($e->getMessage());
            return Command::FAILURE;
        } finally {
            try {
                $this->filesystem->remove($progressFile);
            } catch (\Exception $e) {
                $io->error(sprintf("Failed to mark content-hub sync complete. Delete file %s before rerunning the command", $progressFile));
                return Command::FAILURE;
            }
        }
        return Command::SUCCESS;
    }

    protected function getAllWidgets()
    {
        $widgets = [];
        $widgetRepository = $this->entityManager->getRepository(Widget::class);
        $widgetExisting = $widgetRepository->findAll();
        foreach ($widgetExisting as $key => $value) {
            $widgets[$value->getName() . "-" . $value->getVersion()] = $value;
        }
        return $widgets;
    }

    protected function getInstalledConnectors()
    {
        $integrations = [];
        $uri = sprintf('%s/connectors/?format=json&page_size=1000', $this->container->getParameter('integrations.uri'));
        $client = $this->container->get("app.proxy.client.integrations");
        try {
            $integrationResponse = $client->request('GET', $uri, ['verify' => false]);
            $integrationResponse = json_decode($integrationResponse->getBody()->getContents());
            foreach ($integrationResponse->data as $value) {
                $integrations[$value->name . "-" . $value->version] = $value;
            }
            return $integrations;
        } catch (\Exception $e) {
            throw new HttpException(Response::HTTP_BAD_REQUEST, $e->getMessage(), $e);
        }
    }

    protected function getAllSolutionpacks()
    {
        $solutionpacks = [];
        $entityRepository = $this->entityManager->getRepository(SolutionPack::class);
        $qb = $entityRepository->createQueryBuilder('sp');
        $allSolutionPacks = $qb->where($qb->expr()
            ->notLike('sp.version', ':version_dev'))
            ->orderBy('sp.version', 'asc')
            ->setParameter(':version_dev', '%_dev')
            ->getQuery()
            ->toIterable();
        foreach ($allSolutionPacks as $value) {
            $solutionpacks[$value->getName()] = ['installed' => $value->getInstalled(), 'infoContent' => $value->getInfoContent(), 'version' => $value->getVersion(), 'name' => $value->getName(), 'local' => $value->getLocal(), 'buildNumber' => $value->getBuildNumber(), 'publishedDate' => $value->getPublishedDate() ? $value->getPublishedDate()->getTimestamp() : 0, 'latestCompatibleVersion' => $value->getLatestCompatibleVersion(), 'uuid' => $value->getUuid()];
        }
        return $solutionpacks;
    }

    private function updateExistingSolutionPack($name, $version, $publishDate, $scm, $availableLatestCompatibilityVersion)
    {
        $solutionRepo = $this->entityManager->getRepository(SolutionPack::class);
        $qb = $solutionRepo->createQueryBuilder('sp');
        $solutionPacks = $qb->where(
            $qb->expr()->andX($qb->expr()->eq('sp.name', ':connector_name'), $qb->expr()->notLike('sp.version', ':version_dev'))
        )->orderBy('sp.version', 'asc')->setParameter(':connector_name', $name)->setParameter(':version_dev', '%_dev')->getQuery()->toIterable();
        foreach ($solutionPacks as $solutionPack) {
            $solutionPack->setLatestAvailableVersion($version);
            $solutionPack->setLatestCompatibleVersion($availableLatestCompatibilityVersion);
            $solutionPack->setPublishedDate(\DateTime::createFromFormat('U', $publishDate));
            $solutionPack->setSourceControlStats(is_object($scm) ? (array)$scm : json_decode($scm, true));
            $this->entityManager->persist($solutionPack);
        }
    }

    private function updateScm($name, $scm)
    {
        $solutionRepo = $this->entityManager->getRepository(SolutionPack::class);
        $solutionPacks = $solutionRepo->findBy(["name" => $name], ['version' => 'desc']);
        foreach ($solutionPacks as $solutionPack) {
            $solutionPack->setSourceControlStats(is_object($scm) ? (array)$scm : json_decode($scm, true));
            $this->entityManager->persist($solutionPack);
        }
    }

    protected function getDevelopmentConnectors()
    {
        $integrations = [];
        $uri = sprintf('%s/connector/development/list/?page_size=1000', $this->container->getParameter('integrations.uri'));
        $client = $this->container->get("app.proxy.client.integrations");
        $integrationResponse = $client->request('POST', $uri, ['verify' => false]);
        $integrationResponse = json_decode($integrationResponse->getBody()->getContents());
        foreach ($integrationResponse->data as $value) {
            $integrations[$value->name . "-" . $value->version] = $value;
        }
        return $integrations;
    }

    protected function sendRequest($uri, $method, $data = null, $tokenType = 'JWT', $params = [])
    {
        $client = $this->container->get("app.proxy.client.crudhub");
        $headers = array(
            "Content-type" => "application/json",
        );
        $request = new GuzzleRequest($method, $uri, $headers, $data ? json_encode($data) : null);
        $params = array_merge($params, ['verify' => false]);
        try {
            $response = $client->send($request, $params);
            $statusCode = $response->getStatusCode();
            if ($statusCode < 400) {
                return $response->getBody()->getContents();
            } else {
                throw new \Exception($uri . '\n' . $response->getBody());
            }
        } catch (BadRequestHttpException | \Exception $exception) {
            $errCode = $exception->getCode();
            if ($errCode == 403) {
                $iri = str_replace($this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), '', $uri);
                $module = explode('/', $iri)[4];
                if ($module) {
                    throw new \Exception(sprintf('You do not have %s access on %s module', AppConstants::METHOD_MAP[$method], $module));
                } else {
                    $error = json_decode($exception->getResponse()->getBody(), true);
                    throw new \Exception($error['Error'] ?? $exception->getMessage());
                }
            } elseif ($errCode == 404) {
                $iri = str_replace($this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), '', $uri);
                $message = $data ? json_encode($data) : "{}";
                throw new \Exception(sprintf('The Requested Resource %s (%s) Does not exists ', $iri, $method));
            } elseif ($errCode == 409) {
                $message = json_decode($exception->getResponse()->getBody(), true);
                $message = $message['hydra:description'] ?? $exception->getMessage();
                throw new \Exception($message);
            }
            $message = $exception->getResponse() ? $exception->getResponse()->getBody() : $exception->getMessage();
            throw new \Exception(sprintf('%d: %s', $errCode, $message));
        }
    }


    /**
     * @param array $uniqueAttributes
     * @param $bulkPayload
     * @return void
     * @throws \Exception
     */
    private function sendUpdateSolutionPackRequestInChunks(array $uniqueAttributes, $bulkPayload): void
    {
        if (empty($bulkPayload)) return;

        $fullURI = sprintf(self::BULK_UPSERT_SP_URL, $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
        $chunkSize = AppConstants::DEFAULT_CONTENT_HUB_CHUNK_SIZE;

        if ($this->container->hasParameter(AppConstants::PARAM_CONTENT_HUB_REQ_CHUNK_SIZE)) {
            $chunkSize = (int) $this->container->getParameter(AppConstants::PARAM_CONTENT_HUB_REQ_CHUNK_SIZE);
        }

        $payloadChunks = array_chunk($bulkPayload, $chunkSize);
        foreach ($payloadChunks as $payload) {
            $body = [ "__unique" => $uniqueAttributes, "__replace" => true, "__data" => $payload ];
            $this->sendRequest($fullURI, 'post', $body);
        }
    }
}
