<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Command;

use Doctrine\ORM\EntityRepository;
use PhpAmqpLib\Message\AMQPMessage;
use App\Entity\ModelMetadata\ModelMetadata;
use Symfony\Component\Serializer\Serializer;
use Symfony\Component\Console\Command\Command;
use PhpAmqpLib\Connection\AMQPStreamConnection;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Serializer\Encoder\JsonEncoder;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\ObjectNormalizer;

class ElasticCreateCommand extends Command
{

    const PUBLISH_CREATE = 'publish_create';
    const PUBLISH_MODIFY = 'publish_modify';
    const PUBLISH_DELETE = 'publish_delete';
    const POPULATE_INDEX = 'populate_index';
    private $callback_queue;
    private $corelation;
    private $response;

    protected $container;

    public function __construct(ContainerInterface $container)
    {
        $this->container = $container;
        parent::__construct();
    }

    protected function configure()
    {
        $this
            ->setName('app:elastic:create')
            ->setAliases(['cybersponse:elastic:create'])
            ->setDefinition([
                new InputOption(
                    'index',
                    'i',
                    InputOption::VALUE_OPTIONAL | InputOption::VALUE_IS_ARRAY,
                    'Arguments for indexing modules in ES.'
                ),
                new InputOption(
                    'sync',
                    's',
                    InputOption::VALUE_OPTIONAL,
                    'Arguments for synchronous indexing of elasticsearch'
                ),
                new InputOption(
                    'timeout',
                    't',
                    InputOption::VALUE_OPTIONAL,
                    'Arguments for timeout in case of sync'
                ),
                new InputOption(
                    'update',
                    'u',
                    InputOption::VALUE_OPTIONAL | InputOption::VALUE_IS_ARRAY,
                    'Arguments for updating existing index in ES.'
                )
            ]);
    }

    public function remove_element($array, $value)
    {
        foreach ($value as $val) {
            if (($key = array_search($val, $array)) !== false) {
                unset($array[$key]);
            }
        }
        return $array;
    }

    protected function identify_module_to_populate($input)
    {
        $modelMetadatasWithKeys = $this->getAllMetadata(ModelMetadata::class);
        $indexName = [];
        $encoder = new JsonEncoder();
        $normalizer = new ObjectNormalizer();
        $defaultContext = [];
        $defaultContext['circular_reference_handler'] = function ($modelMetadatasWithKeys) {
            return $modelMetadatasWithKeys->getUuid();
        };

        $serializer = new Serializer([$normalizer], [$encoder]);

        $normalizedMetadata = $serializer->serialize($modelMetadatasWithKeys, 'json', $defaultContext);
        $normalizedMetadata = json_decode($normalizedMetadata, true);

        $object['modelToCreate'] = [];
        $indexNames = $input->getOption('index');
        $moduleExcludeList = $this->container->getParameter('module.exclude.list');
        $moduleExcludeList = $this->remove_element($moduleExcludeList, ["tenants", "routers", "agents", "workflows"]);
        if (is_array($indexNames) && array_key_exists("0", $indexNames)) {
            $indexName = json_decode($indexNames[0], true);
        }
        foreach ($normalizedMetadata as $key => $value) {
            if (count($indexNames)) {
                for ($i = 0; $i < count($indexName['type']); $i++) {
                    if ($indexName['type'][$i] == $value['type'] && !in_array($value['type'], $moduleExcludeList)) {
                        array_push($object['modelToCreate'], $value);
                    }
                }
            } else {
                if (!in_array($value['type'], $moduleExcludeList)) {
                    array_push($object['modelToCreate'], $value);
                }
            }
        }

        # This option is added for force update of workflow index on demand.
        if (!empty($indexName['type']) && in_array('workflows', $indexName['type'])) {
            $systemFixturesFile = file_get_contents('/opt/cyops-api/src/StaticFile/SystemFixtures.json');
            $systemFixture = json_decode($systemFixturesFile);
            array_push($object['modelToCreate'], $systemFixture->{'workflows'});
        }

        return $object['modelToCreate'];
    }

    protected function identify_module_to_update($input)
    {
        $indexNames = [];
        if ($input->getOption('update')[0]) {
            $indexNames = json_decode($input->getOption('update')[0], true);
        }
        return $indexNames;
    }

    protected function publish_message($input, $object, $publishType = self::PUBLISH_CREATE)
    {
        $mqCredObj = parse_ini_file('/opt/cyops/configs/rabbitmq/rabbitmq_users.conf', false, INI_SCANNER_NORMAL);
        $rabbitMqUsername = $mqCredObj['cyops.rabbitmq.user'] ? $mqCredObj['cyops.rabbitmq.user'] : 'guest';
        $rabbitMqPassword = $mqCredObj['cyops.rabbitmq.password'] ? $mqCredObj['cyops.rabbitmq.password'] : 'guest';
        $rabbitMqHost = $mqCredObj['cyops.rabbitmq.host'] ? $mqCredObj['cyops.rabbitmq.host'] : 'localhost';
        $rabbitMqPort = $mqCredObj['cyops.rabbitmq.port'] ? $mqCredObj['cyops.rabbitmq.port'] : '5672';
        $rabbitMqVHost = $mqCredObj['cyops.rabbitmq.vhost'] ? $mqCredObj['cyops.rabbitmq.vhost'] : '/';
        try {
            // Adding a connection timeout of 60 here since on cloud, the hostname resolution can take more than 10 seconds during initial config
            $connection = new AMQPStreamConnection($rabbitMqHost, $rabbitMqPort, $rabbitMqUsername, $rabbitMqPassword, $rabbitMqVHost,false,'AMQPLAIN',null,'en_US',60);
            $channel = $connection->channel();
            $channel->exchange_declare('cyops.crudhub.datanotify', 'topic', false, true, false);
        } catch (\Exception $e) {
            echo "Failure in running search index create command, please check prod.log for more information.";
            error_log($e->getMessage());
            exit;
        }
        if ($input->getOption('sync')) {
            $timeout = $input->getOption('timeout') ? (int)$input->getOption('timeout') : 7200;
            list($this->callback_queue,,) = $channel->queue_declare("", false, false, true, false);
            $channel->basic_consume($this->callback_queue, '', false, false, false, false, array($this, 'onResponse'));
            $this->response = null;
            $this->corelation = uniqid();
            $msg = new AMQPMessage(json_encode(array('esoperation' => $publishType, 'payload' => $object)), array(
                'correlation_id' => $this->corelation,
                'reply_to' => $this->callback_queue,
                'content_type' => 'application/json'
            ));
            $channel->basic_publish($msg, 'cyops.crudhub.datanotify', 'cyops.crudhub.systemnotify');
            while (!$this->response) {
                try {
                    $channel->wait(null, false, $timeout);
                } catch (\PhpAmqpLib\Exception\AMQPTimeoutException $e) {
                    echo "Timeout occurred please run following command : sudo -u nginx php /opt/cyops-api/bin/console app:elastic:create";
                    $channel->close();
                    $connection->close();
                    exit;
                } catch (\Exception $e) {
                    echo "Failure in running search index create command, please check prod.log for more information.";
                    error_log($e->getMessage());
                    $channel->close();
                    $connection->close();
                    exit;
                }
            }
        } else {
            $msg = new AMQPMessage(json_encode(array('esoperation' => $publishType, 'payload' => $object)), array('content_type' => 'application/json'));
            $channel->basic_publish($msg, 'cyops.crudhub.datanotify', 'cyops.crudhub.systemnotify');
        }
        $channel->close();
        $connection->close();
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $output->writeln([
            '===========================================================',
            'Passing Mapping To Falcon',
            '===========================================================',
            '',
        ]);
        $object['modelToUpdate'] = [];
        $object['modelToCreate'] = [];
        $indexOption = $input->getOption('index');
        $updateOption = $input->getOption('update');
        if ((!$indexOption && !$updateOption) || count($indexOption)) {
            $object['modelToCreate'] = $this->identify_module_to_populate($input);
        }

        if ($updateOption) {
            $object['modelToUpdate'] = $this->identify_module_to_update($input);
        }
        $meta = $input->getOption('index');
        $metaObject = $meta;
        if (count($object['modelToCreate'])) {
            foreach ($object['modelToCreate'] as $key => $value) {
                $objectToSend = [];
                if (isset($metaObject['meta']) && isset($metaObject['meta'][$value['type']]['modified'])) {
                    $value['meta'] = $metaObject['meta'][$value['type']];
                }
                $objectToSend['modelToCreate'] = [$value];
                $this->publish_message($input, $objectToSend);
            }
        }
        if ($object['modelToUpdate'] && count($object['modelToUpdate'])) {
            $this->publish_message($input, $object['modelToUpdate'], self::PUBLISH_MODIFY);
        }

        $output->writeln([
            '===========================================================',
            'Mapping Passed To Falcon',
            '===========================================================',
            '',
        ]);
        return Command::SUCCESS;
    }

    /**
     * @param $class
     * @return array
     */
    private function getAllMetadata($class)
    {
        $metadataDocumentManager = $this->container->get('doctrine.orm.entity_manager');

        /** @var EntityRepository $modelMetadataRepository */
        $modelMetadataRepository = $metadataDocumentManager->getRepository($class);

        /** @var ModelMetadata[] $modelMetadatas */
        $modelMetadatas = $modelMetadataRepository->findAll();
        $modelMetadatasWithKeys = [];

        foreach ($modelMetadatas as $modelMetadata) {
            $modelMetadatasWithKeys[$modelMetadata->getType()] = $modelMetadata;
        }
        return $modelMetadatasWithKeys;
    }

    public function onResponse($rep)
    {
        if ($rep->get('correlation_id') == $this->corelation) {
            $this->response = true;
        }
    }
}

