<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Command;

use DateTime;
use Exception;
use ZipArchive;
use Ramsey\Uuid\Uuid;
use App\Entity\Core\File;
use App\Entity\Core\PgFile;
use InvalidArgumentException;
use App\Entity\Core\ExportJob;
use RecursiveIteratorIterator;
use App\Constants\AppConstants;
use App\Service\SettingsHelper;
use RecursiveDirectoryIterator;
use App\Entity\Core\SolutionPack;
use App\Entity\Settings\SystemSettings;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\Finder\SplFileInfo;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Serializer\Serializer;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputOption;
use App\Service\ConfigExportImport\ExportService;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\Console\Input\InputInterface;
use App\EventSubscriber\MqMessagebroadcastSubscriber;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Serializer\Encoder\JsonEncoder;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\ObjectNormalizer;
use Symfony\Component\Filesystem\Exception\IOExceptionInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class ExportJobCommand
 */
class ExportJobCommand extends Command
{
    /** @var Filesystem $filesystem */
    protected $filesystem;

    /** @var SettingsHelper */
    private $settingsHelper;

    private $entityManager;

    /** @var ExportService $exportService */
    private $exportService;

    /** @var ContainerInterface */
    private $container;

    public function __construct(
        SettingsHelper $settingsHelper,
        EntityManagerInterface $entityManager,
        ExportService $exportService,
        ContainerInterface $container,
        private TokenStorageInterface $tokenStorage
    ) {
        $this->settingsHelper = $settingsHelper;
        $this->entityManager = $entityManager;
        $this->exportService = $exportService;
        $this->container = $container;
        $this->filesystem = new Filesystem();
        parent::__construct();
    }

    /**
     * @return array
     */
    protected function prepareChangeData(array $oldData, array $newData, string $entityId)
    {
        $deltaData = [];
        $deltaData['oldData'] = $oldData;
        $deltaData['newData'] = $newData;
        try {
            $user = $this->settingsHelper->getCurrentUser();
            if ($user) {
                $userUUID = $user->getUuid();
                $userName = $this->settingsHelper->extractUserName($user);
            } else {
                $userUUID = 'SYSTEM';
                $userName = 'SYSTEM';
            }
        } catch (InvalidArgumentException | Exception $e) {
            $userUUID = 'SYSTEM';
            $userName = 'SYSTEM';
        }
        $sourceIp = 'localhost';
        if (isset($_SERVER['HTTP_X_FORWARDED_FOR'])) {
            $sourceIp = $_SERVER['HTTP_X_FORWARDED_FOR'];
        } elseif (isset($_SERVER['REMOTE_ADDR'])) {
            $sourceIp = $_SERVER['REMOTE_ADDR'];
        }
        return [
            AppConstants::COMPONENT => 'crudhub',
            AppConstants::ENTITY_TYPE => 'ExportJob',
            AppConstants::ENTITY_UUID => $entityId,
            AppConstants::USER_UUID => $userUUID,
            AppConstants::USER => $userName,
            AppConstants::HTTP_METHOD => 'Update',
            AppConstants::TRANSACTION_DATE => (int)($_SERVER['REQUEST_TIME_FLOAT'] * 1000),
            AppConstants::DELTA_DATA => $deltaData,
            AppConstants::REMOTE_ADDR => $sourceIp
        ];
    }

    protected function configure()
    {
        $this
            ->setName('fortisoar:jobs:export')
            ->setDescription('Exports records and configuration into a zip file')
            ->setDefinition([
                new InputOption(
                    'job-uuid',
                    'j',
                    InputOption::VALUE_REQUIRED | InputOption::VALUE_IS_ARRAY,
                    'The uuid of the export job that will be started.'
                ),
                new InputOption(
                    'user-token',
                    't',
                    InputOption::VALUE_OPTIONAL,
                    'The user token for the export context. If empty, the appliance user will be used.'
                ),
                new InputOption(
                    'token-type',
                    'tt',
                    InputOption::VALUE_OPTIONAL,
                    'Type of token HMAC or JWT.'
                )
            ]);
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $io = new SymfonyStyle($input, $output);
        $solutionPack = [];
        $configExportData = null;
        $jobUuids = $input->getOption('job-uuid');
        if (empty($jobUuids)) {
            $io->error('Please specify a job uuid');
            return Command::FAILURE;
        }
        $jobUuid = $jobUuids[0];
        $rawToken = $input->getOption('user-token');
        $tokenType = $input->getOption('token-type');
        if ($rawToken && $tokenType) {
            $this->settingsHelper->setRefreshedToken($rawToken, $tokenType);
        }
        $repository = $this->entityManager->getRepository(ExportJob::class);
        /** @var ExportJob $exportJob */
        try {
            $exportJob =  $repository->findOneBy(['uuid' => $jobUuid]);
        } catch (InvalidArgumentException | \Exception $e) {
            $io->error($e->getMessage());
            return Command::FAILURE;
        }

        if ($exportJob && !is_null($exportJob->getSolutionPack())) {
            $solutionPack = $this->getSolutionPackDetails($exportJob->getSolutionPack()->getUuid());
            $exportJob->setType('SolutionPack Export');
        }

        $status = 'In Progress';
        $this->updateExportJob($exportJob, $status, null, null);

        $tmpDirectory = sys_get_temp_dir() . '/fsrexport-' . $jobUuid . '/';
        try {
            $this->filesystem->mkdir($tmpDirectory);
            // Check if iconLarge contains base64 image data and not a url or null
            if ($solutionPack && $solutionPack['iconLarge']) {
                $imageData = base64_decode(preg_replace('#^data:image/\w+;base64,#i', '', $solutionPack['iconLarge']));
                if ($imageData) {
                    $imageFilePath = $tmpDirectory . "data/images/fsr-icon-large.png";
                    $this->filesystem->dumpFile($imageFilePath, $imageData);
                    $solutionPack['iconLarge'] = "fsr-icon-large.png";
                }
            }
        } catch (IOExceptionInterface | \Exception $exception) {
            $message = sprintf('Error while creating temp directory: %s', $exception->getMessage());
            $this->updateExportJob($exportJob, 'Error', null, null, $message);
            $io->error($message);
        }

        try {
            $metadataFilePath = $tmpDirectory . 'data/info.json';
            $this->filesystem->dumpFile($metadataFilePath, json_encode(['contents' => []], JSON_PRETTY_PRINT));

            $configExportData = $this->exportService->exportConfiguration($exportJob, $this, $io, $tmpDirectory . 'data/', $tokenType, true, $exportJob);
        } catch (InvalidArgumentException | \Exception $exception) {
            $message = sprintf('Error while exporting file: %s', $exception->getMessage());
            $this->updateExportJob($exportJob, 'Error', null, null, $message);
            $io->error($message);
            $this->filesystem->remove($tmpDirectory);
        }

        if ($configExportData) {
            $isSolutionPack = is_null($exportJob->getSolutionPack()) ? false : true;
            $metadata = $this->getExportMetadata($isSolutionPack);
            if (is_array($metadata) && array_key_exists(AppConstants::FSR_MIN_COMPATIBILITY, $metadata)) {
                $metadata[AppConstants::FSR_MIN_COMPATIBILITY] = isset($solutionPack[AppConstants::FSR_MIN_COMPATIBILITY]) ? $solutionPack[AppConstants::FSR_MIN_COMPATIBILITY] : $metadata[AppConstants::FSR_MIN_COMPATIBILITY];
            }
            $subFileData = file_get_contents($tmpDirectory . 'data/info.json');
            $subJsonData = json_decode($subFileData, true);
            $metadata = array_merge($metadata, $subJsonData);
            $metadata = array_merge($solutionPack, $metadata);
            $metadataJson = json_encode($metadata, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT);
            $metadataFilePath = $tmpDirectory . 'data/info.json';
            $this->filesystem->dumpFile($metadataFilePath, $metadataJson);
            $filename = $this->cleanFileName(preg_replace('/\.zip$/', '', $exportJob->getFileName()));
            $zipFilename = $filename ? $filename . '.zip' : 'config.zip';
            $zipFilePath = $tmpDirectory . $zipFilename;
            $folderPath = $tmpDirectory . 'data/';
            try {
                $this->zipFolder($zipFilePath, $folderPath, $filename);
            } catch (InvalidArgumentException | \Exception $exception) {
                $message = sprintf('Error while building tar.gz file: %s', $exception->getMessage());
                $this->updateExportJob($exportJob, 'Error', null, null, $message);
                $io->error($message);
            }


            $pgFile = new PgFile();
            $fp = fopen($zipFilePath, 'rb');
            $fileSize = filesize($zipFilePath);
            $pgFile->setFile(base64_encode(fread($fp, $fileSize)));
            fclose($fp);
            $this->filesystem->remove($tmpDirectory);
            $pgFile->setUuid();
            $this->entityManager->persist($pgFile);
            $this->entityManager->flush();

            $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
            $mimetype = 'application/zip';
            $file = new File();
            $file->setFile($pgFile->getUuid());
            $file->setFilename($zipFilename);
            $file->setMimeType($mimetype);
            $file->setSize($fileSize);
            $file->setUploadDate($timestamp);
            $file->setUuid(Uuid::uuid4()->toString());
            $this->entityManager->persist($file);
            $this->entityManager->flush();

            $status = 'Export Complete';
            $this->updateExportJob($exportJob, $status, null, AppConstants::FILE_IRI . $file->getUuid());

            $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
            $exportTemplate = $exportJob->getTemplate();
            $exportTemplate->setLastExportDate($timestamp);
            $this->entityManager->persist($exportTemplate);
            $this->entityManager->flush();
            $io->success(sprintf('Export completed successfully.'));
            return Command::SUCCESS;
        }
        return Command::FAILURE;
    }

    private function getUserName($user)
    {
        if (method_exists($user, 'getFirstname')) {
            return $user->getFirstname() . ' ' . $user->getLastname();
        }
        return $user->getName();
    }

    private function getSolutionPackDetails($solutionPackUuid = null)
    {
        if ($solutionPackUuid) {
            $repository = $this->entityManager->getRepository(SolutionPack::class);
            $solutionPack =  $repository->findOneBy(['uuid' => $solutionPackUuid]);
            if ($solutionPack) {
                $categoryData = [];
                $categoryValues = $solutionPack->getCategory();
                if (!empty($categoryValues)) {
                    foreach ($categoryValues as $category) {
                        array_push($categoryData, $category->getItemValue());
                    }
                }

                $normalizedsolutionPack = [];
                $normalizedsolutionPack['name'] = $solutionPack->getName();
                $normalizedsolutionPack['version'] = $solutionPack->getVersion();
                $normalizedsolutionPack['fsrMinCompatibility'] = $solutionPack->getFsrMinCompatibility();
                $normalizedsolutionPack['type'] = $solutionPack->getType();
                $normalizedsolutionPack['local'] = $solutionPack->getLocal();
                if (!$solutionPack->getInstalled()) {
                    $normalizedsolutionPack['installed'] = $solutionPack->getInstalled();
                }
                $normalizedsolutionPack['label'] = $solutionPack->getLabel();
                $normalizedsolutionPack['dependencies'] = $solutionPack->getDependencies();
                $normalizedsolutionPack['prerequisite'] = $solutionPack->getPrerequisite();
                $normalizedsolutionPack['publisher'] = $solutionPack->getPublisher();
                $normalizedsolutionPack['description'] = $solutionPack->getDescription();
                $normalizedsolutionPack['help'] = $solutionPack->getInfoContent()['help'] ?? null;
                $normalizedsolutionPack['category'] = $categoryData;
                $normalizedsolutionPack['supportInfo'] = $solutionPack->getInfoContent()['supportInfo'] ?? null;
                $normalizedsolutionPack['iconLarge'] = $solutionPack->getIconLarge();
                $normalizedsolutionPack['postInstallConfig'] = $solutionPack->getInfoContent()['postInstallConfig'] ?? null;
                if ($solutionPack->getRecordTags()) {
                    foreach ($solutionPack->getRecordTags()->getValues() as $value) {
                        $normalizedsolutionPack['recordTags'][] = $value->getUuid();
                    }
                }
                return $normalizedsolutionPack;
            }
        }
        return [];
    }

    private function getExportMetadata(bool $isSolutionPack)
    {
        $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $systemRepository = $this->entityManager->getRepository(SystemSettings::class);
        $systemSetting = $systemRepository->findAll();
        $global_value = $systemSetting[0]->getGlobalValues();
        $hostname = isset($global_value['hostname']) ? $global_value['hostname'] : gethostname();
        $token = $this->tokenStorage->getToken();
        $user = $token->getUser();
        $userName = $this->getUserName($user);
        $metadata = [
            AppConstants::FSR_MIN_COMPATIBILITY => '7.2.0',
            'date' => $timestamp->format(DateTime::ATOM)
        ];
        if (!$isSolutionPack) {
            $metadata = array_merge($metadata, ['exported_from' => $hostname ? $hostname : 'Unknown', 'exported_by' => $userName]);
        }
        return $metadata;
    }

    public function updateExportJob(ExportJob $exportJob, $status = null, $progressPercent = null, $file = "", $errorMessage = null, $currentlyExporting = null)
    {
        $oldData = [
            '@id' => '/api/3/export_jobs/' . $exportJob->getUuid()
        ];
        $newData = [
            '@id' => '/api/3/export_jobs/' . $exportJob->getUuid()
        ];
        $oldData['id'] = $exportJob->getId();
        $newData['id'] = $exportJob->getId();
        if ($status !== null) {
            $oldStatus = $exportJob->getStatus();
            $exportJob->setStatus($status);
            $oldData['status'] = $oldStatus;
            $newData['status'] = $status;
        }
        if ($progressPercent !== null) {
            $oldProgressPercent = $exportJob->getProgressPercent();
            $exportJob->setProgressPercent($progressPercent);
            $oldData['progressPercent'] = $oldProgressPercent;
            $newData['progressPercent'] = $progressPercent;
        }
        if ($file !== "") {
            $oldFile = $exportJob->getFile();
            $exportJob->setFile($file);
            $oldData['file'] = $oldFile;
            $newData['file'] = $file;
        }
        if ($errorMessage !== null) {
            $oldErrorMessage = $exportJob->getErrorMessage();
            $exportJob->setErrorMessage($errorMessage);
            $oldData['errorMessage'] = $oldErrorMessage;
            $newData['errorMessage'] = $errorMessage;
        }
        if ($currentlyExporting !== null) {
            $old = $exportJob->getCurrentlyExporting();
            $exportJob->setCurrentlyExporting($currentlyExporting);
            $oldData['currentlyExporting'] = $old;
            $newData['currentlyExporting'] = $currentlyExporting;
        }
        $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $exportJob->setModifyDate($timestamp);
        $this->entityManager->persist($exportJob);
        $this->entityManager->flush();
        $body = $this->prepareChangeData($oldData, $newData, $exportJob->getUuid());
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->setContentType('application/json');
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->publish(json_encode([$body]), MqMessagebroadcastSubscriber::DEFALUT_KEY);
    }

    private function zipFolder($zipFilePath, $folderPath, $filename)
    {
        $zip = new ZipArchive();
        $zip->open($zipFilePath, ZipArchive::CREATE | ZipArchive::OVERWRITE);
        /** @var SplFileInfo[] $files */
        $files = new RecursiveIteratorIterator(
            new RecursiveDirectoryIterator($folderPath),
            RecursiveIteratorIterator::LEAVES_ONLY
        );

        foreach ($files as $file) {
            if (!$file->isDir()) {
                $filePath = $file->getRealPath();
                $relativePath = $filename . '/' . substr($filePath, strlen($folderPath));
                $zip->addFile($filePath, $relativePath);
            }
        }
        $zip->close();
    }

    protected function cleanFileName(string $name)
    {
        return preg_replace(AppConstants::FILE_NAME_CLEANUP_REGEX, '_', $name);
    }
}
