<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Command;

use App\Traits\Ownable;
use App\Traits\Taggable;
use App\Traits\Indexable;
use App\Traits\Queueable;
use App\Traits\Trackable;
use App\Traits\UserOwnable;
use App\Traits\PeerReplicable;
use App\Service\EntityGenerator;
use Doctrine\Inflector\InflectorFactory;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Mapping\ClassMetadataInfo;
use Symfony\Component\Routing\RequestContext;
use Symfony\Component\Console\Command\Command;
use App\Entity\ModelMetadata\AttributeMetadata;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Style\SymfonyStyle;
use App\Entity\ModelMetadata\StagingModelMetadata;
use Symfony\Component\Console\Input\InputInterface;
use App\Doctrine\ORM\Mapping\ResourceNamingStrategy;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Filesystem\Exception\IOException;

class GenerateModelsCommand extends Command
{

    /** @var StagingModelMetadata[] */
    protected $stagingModelMetadataMap;

    /** @var RouterInterface */
    protected $router;

    /** @var string */
    protected $customEntityDirectory;

    /** @var string */
    protected $customEntityNamespace;

    /** @var array */
    protected $typeToClassNameMap;

    protected $container;

    private $entityManager;

    private $inflector;

    /** @var ResourceNamingStrategy */
    protected $resourceNamingStrategy;

    /** @var ResourceCollectionInterface */
    protected $resourceCollection;

    protected function configure()
    {
        $this
            ->setName('app:metadata:generate:models')
            ->setAliases(['cybersponse:metadata:generate:models'])
            ->setDescription('Process current model metadata & perform code generation')
            ->setDefinition([
                new InputOption(
                    'type',
                    't',
                    InputOption::VALUE_REQUIRED | InputOption::VALUE_IS_ARRAY,
                    'The type of metadata, activites, alerts, etc.'
                )
            ]);
    }

    protected function getClassShortNameForModelMetadata($modelMetadata)
    {
        $singular = $modelMetadata->getDescriptions()['singular'];
        $escapedSingular = preg_replace('/[^a-z0-9\s]+/i', '', $singular);
        return $this->inflector->classify($escapedSingular);
    }

    public function __construct(ContainerInterface $container, EntityManagerInterface $entityManager)
    {
        $this->container = $container;
        $this->entityManager = $entityManager;
        $this->inflector = InflectorFactory::create()->build();
        parent::__construct();
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $io = new SymfonyStyle($input, $output);

        $progress = $io->createProgressBar();

        $this->resourceNamingStrategy = new ResourceNamingStrategy();

        /** @var EntityGenerator $entityGenerator */
        $entityGenerator = new EntityGenerator($this->container, $this->entityManager);
        $entityGenerator->setBackupExisting(false);
        $entityGenerator->setGenerateAnnotations(true);
        $entityGenerator->setGenerateStubMethods(true);
        $entityGenerator->setRegenerateEntityIfExists(true);
        $entityGenerator->setUpdateEntityIfExists(true);
        $entityGenerator->setNumSpaces(4);
        $entityGenerator->setAnnotationPrefix('Doctrine\ORM\Mapping\\');
        $entityGenerator->setFieldVisibility(EntityGenerator::FIELD_VISIBLE_PROTECTED);

        $metadataEntityManager = $this->container->get('doctrine.orm.entity_manager');

        $this->router = $this->container->get('router');
        $requestContext = new RequestContext();
        $requestContext->setMethod('GET');
        $this->router->setContext($requestContext);

        $sourcePath = sprintf('%s/src', $this->container->getParameter('kernel.project_dir'));

        $this->customEntityNamespace = $this->container->getParameter('custom_types_namespace');

        /** @var EntityRepository $modelMetaRepository */
        $stagingModelMetadataRepository = $metadataEntityManager->getRepository(StagingModelMetadata::class);

        /** @var StagingModelMetadata[] $stagingMetadatas */
        $stagingMetadatas = $stagingModelMetadataRepository->findAll();

        $this->stagingModelMetadataMap = [];
        $this->typeToClassNameMap = $this->container->getParameter('type_map');

        /** @var ManagerRegistry $managerRegistry */
        $managerRegistry = $this->container->get('doctrine');

        //Create fake model metadata from existing entities in the system:
        foreach ($this->typeToClassNameMap as $type => $className) {
            /** @var EntityManager $manager */
            $manager = $managerRegistry->getManagerForClass($className);
            $doctrineClassMetadataInfo = $manager->getClassMetadata($className);

            $stagingModelMetadata = $this->getModelMetadataFromDoctrineClassMetadataInfo($type, $doctrineClassMetadataInfo);
            $this->stagingModelMetadataMap[$type] = $stagingModelMetadata;
        }

        foreach ($stagingMetadatas as $stagingData) {
            $type = $stagingData->getType();
            $classShortName = $this->getClassShortNameForModelMetadata($stagingData);
            $this->typeToClassNameMap[$type] = sprintf('%s\%s', $this->customEntityNamespace, $classShortName);

            $attributeMetadatas = [];
            foreach ($stagingData->getAttributes() as $attributeMetadata) {
                $attributeMetadatas[$attributeMetadata->getName()] = $attributeMetadata;
            }
            foreach ($attributeMetadatas as $attributeMetadata) {
                $stagingData->addAttribute($attributeMetadata);
            }

            $this->stagingModelMetadataMap[$type] = $stagingData;
        }

        $types = $input->getOption('type');

        /** @var StagingModelMetadata[] $modelMetadatas */
        if (count($types) > 0) {
            $queryBuilder = $stagingModelMetadataRepository->createQueryBuilder('o');
            $stagingMetadatas = $queryBuilder->andWhere('o.type in (:types)')
                ->setParameter('types', $types)
                ->andWhere('o.system in (:system)')
                ->setParameter('system', false)
                ->getQuery()
                ->execute();
        } else {
            $stagingMetadatas = $stagingModelMetadataRepository->findAll();
        }

        $progress->setFormat(' %current%/%max% [%bar%] Working on %message%...');
        $progress->setBarWidth(50);
        $progress->start(count($stagingMetadatas));

        foreach ($stagingMetadatas as $stagingData) {

            $progress->setMessage($stagingData->getDescriptions()['singular']);

            //Build doctrine class metadata info
            $doctrineClassMetadataInfo = $this->buildDoctrineMetadata($stagingData);
            $parentClass = $this->getParentClassFromModelMetadata($stagingData);

            $entityGenerator->setClassToExtend($parentClass);
            $entityGenerator->setFieldVisibility(EntityGenerator::FIELD_VISIBLE_PROTECTED);
            $entityGenerator->setModelMetadata($stagingData);

            // Write entity class to a file
            try {
                $entityGenerator->writeEntityClass($doctrineClassMetadataInfo, $sourcePath, $stagingData);
            } catch (IOException | \Exception $exc) {
                $io->error(sprintf('Exception occurred while writing entity file for %s', $stagingData->getDescriptions()['singular']));
                return Command::FAILURE;
            }
            $progress->advance();
        }
        $progress->finish();

        $io->success(sprintf('Wrote all files to %s by namespace', $sourcePath));
        return Command::SUCCESS;
    }

    private function getAttributeType($stagingModelData, $columnName)
    {
        foreach ($stagingModelData->getAttributes() as $attribute) {
            if ($attribute->getName() == $columnName) {
                return $attribute->getFormType();
            }
        }
    }

    private function buildDoctrineMetadata(StagingModelMetadata $stagingModelData)
    {
        $attributeType = $stagingModelData->getType();
        $entityClass = $this->typeToClassNameMap[$attributeType];
        $parentClasses = [];
        $uniqeConst = [];

        $parentType = $stagingModelData->getParentType();
        if (array_key_exists($parentType, $this->typeToClassNameMap)) {
            $parentClasses[] = $this->typeToClassNameMap[$parentType];
        }

        $classMetadataInfo = new ClassMetadataInfo($entityClass, $this->resourceNamingStrategy);
        $uniqueConstraint = $stagingModelData->getUniqueConstraint();
        if ($uniqueConstraint && count($uniqueConstraint)) {
            $uniqueConstraint = $uniqueConstraint[0];
            $constraintName = array_keys($uniqueConstraint)[0];
        }

        if ($uniqueConstraint && count($uniqueConstraint)) {
            foreach ($uniqueConstraint as $uniqueConstraintName) {
                foreach ($uniqueConstraintName['columns'] as $column) {
                    if (in_array($this->getAttributeType($stagingModelData, $column), ['lookup', 'picklist'])) {
                        $uniqeConst[] = $this->inflector->tableize($column);
                    } else {
                        $uniqeConst[] = $column;
                    }
                }
                $uniqueConstraint[$constraintName]['columns'] = $uniqeConst;
            }
        }

        if ($stagingModelData->getTrackable()) {
            //Can be change to generic by exposing same in UI.
            $modifyDateIndexName = $stagingModelData->getType() . '_modifydate_idx';
            $indexList[$modifyDateIndexName]['columns'] = ["modifydate"];

            $createDateIndexName = $stagingModelData->getType() . '_createdate_idx';
            $indexList[$createDateIndexName]['columns'] = ["createdate"];
        }
        if ($stagingModelData->getIndexable()) {
            //Can be change to generic by exposing same in UI.
            $idIndexName = $stagingModelData->getType() . '_id_idx';
            $indexList[$idIndexName]['columns'] = ["id"];
        }
        if ($stagingModelData->getSoftDeleteable()) {
            $idIndexName = $stagingModelData->getType() . '_deletedat_idx';
            $indexList[$idIndexName]['columns'] = ["deletedat"];
            $indexList[$idIndexName]['options']['operator'] = "where";
            $indexList[$idIndexName]['options']['condition'] = "deletedat IS NULL";
        }

        if ($parentType ==='threat_intel_base'){
            $indexName = 'idx_value_tokens';
            $indexList[$indexName]['columns'] = ['value_tokens'];
            $indexList[$indexName]['flags'] = ['gin_search'];
        } 

        $classMetadataInfo->setPrimaryTable(['name' => sprintf('`%s`', $stagingModelData->getTableName()), 'uniqueConstraints' => $uniqueConstraint, 'indexes' => $indexList]);
        $classMetadataInfo->setCustomRepositoryClass('Doctrine\ORM\EntityRepository');
        $classMetadataInfo->setParentClasses($parentClasses);

        foreach ($stagingModelData->getAttributes() as $attribute) {
            $attributeType = $attribute->getType();

            if (array_key_exists($attributeType, $this->typeToClassNameMap)) {
                $relationshipMapping = [
                    'fieldName' => $attribute->getName(),
                    'isOwningSide' => $attribute->getOwnsRelationship(),
                    'targetEntity' => $this->typeToClassNameMap[$attributeType],
                    'cascade' => ['persist']
                ];

                /**
                 * The inverse side has to use the mappedBy attribute of the OneToOne, OneToMany, or ManyToMany mapping declaration.
                 * The mappedBy attribute contains the name of the association-field on the owning side.
                 * The owning side has to use the inversedBy attribute of the OneToOne, ManyToOne, or ManyToMany mapping declaration.
                 * The inversedBy attribute contains the name of the association-field on the inverse-side.
                 */

                $inversedField = $attribute->getInversedField(); //Determine relationship type automatically
                $relatedModelMetadata = $this->stagingModelMetadataMap[$attributeType];

                $relationshipType = 'manyTo'; //Default to manyTo

                if (!is_null($inversedField)) {
                    $relatedAttribs = $relatedModelMetadata->getAttributes()->getValues();
                    foreach ($relatedAttribs as $relatedAttribute) {
                        $relatedAttributes[$relatedAttribute->getName()] = $relatedAttribute;
                    }

                    if (array_key_exists($inversedField, $relatedAttributes)) {
                        $inversedFieldAttribute = $relatedAttributes[$inversedField];

                        $relationshipType = 'oneTo';

                        if ($inversedFieldAttribute->getCollection()) {
                            $relationshipType = 'manyTo';
                        }
                    }
                }

                if ($attribute->getCollection()) {
                    $relationshipType .= 'Many';
                } else {
                    $relationshipType .= 'One';
                }

                //TODO Might need to make this more functional
                if ($relationshipType == 'manyToMany') {
                    $relationshipMapping['fetch'] = ClassMetadataInfo::FETCH_EXTRA_LAZY;
                    $relationshipMapping['orphanRemoval'] = $attribute->getOrphanRemoval();
                    if ($attribute->getOrphanRemoval()) {
                        $relationshipMapping['cascade'] = ['persist', 'remove', 'detach'];
                    }

                    if (!empty($inversedField)) {
                        if ($attribute->getOwnsRelationship()) {
                            $relationshipMapping['inversedBy'] = $attribute->getInversedField();
                        } else {
                            $relationshipMapping['mappedBy'] = $attribute->getInversedField();
                        }
                    }
                    if ($attribute->getFormType() == "multiselectpicklist") {
                        $relationshipMapping['joinTable'] = [
                            'name' => $this->inflector->tableize($stagingModelData->getType()) . '_picklist_' . $this->inflector->tableize($attribute->getName())
                        ];
                    }
                } else if ($relationshipType == 'oneToMany') {
                    if (!empty($inversedField)) {
                        $relationshipMapping['mappedBy'] = $inversedField;
                    }
                    $relationshipMapping['orphanRemoval'] = $attribute->getOrphanRemoval();
                } else if ($relationshipType == 'manyToOne') {
                    if (!empty($inversedField)) {
                        $relationshipMapping['inversedBy'] = $inversedField;
                    }

                    $relationshipMapping['joinColumns'] = [
                        'joinColumn' => [
                            'name' => $this->inflector->tableize($attribute->getName()),
                            'onDelete' => 'SET NULL'
                        ]
                    ];
                } else if ($relationshipType == 'oneToOne' && !empty($inversedField)) {
                    $relationshipMapping['inversedBy'] = $inversedField;
                }

                $method = 'map' . ucfirst($relationshipType);
                $classMetadataInfo->$method($relationshipMapping);
            } else {
                $length = null;
                $validation = $attribute->getValidation();
                if ($validation !== null && array_key_exists('maxlength', $validation)) {
                    $length = $validation['maxlength'];
                    if ($length < 0) {
                        $length = null;
                    } else if ($length > 65535 && $attributeType == 'string') {
                        $attributeType = 'text';
                    }
                }
                $fieldMapping = [
                    'fieldName' => $attribute->getName(),
                    'type' => $attributeType,
                    'nullable' => !$attribute->getIdentifier(),
                    'unique' => $attribute->getUnique(),
                    'length' => $length
                ];

                $formType = $attribute->getFormType();

                //Json primitives to doctrine based on formtype
                if ($formType == 'textarea') {
                    $fieldMapping['type'] = 'text';
                }
                if (($formType == 'string' || $formType == 'richtext' ||  $formType == 'html') && (is_null($attribute->getLength()) || $attribute->getLength() > 255 ||  !$length)) {
                    $fieldMapping['type'] = 'text';
                }
                if ($attributeType == 'array' || $attributeType == 'object') {
                    $fieldMapping['type'] = 'jsonb';
                }
                if ($formType == 'date' || $formType == 'datetime' || $formType == 'livesync') {
                    $fieldMapping['type'] = 'datetime';
                }
                if ($formType == 'decimal') {
                    $fieldMapping['type'] = 'float';
                    $fieldMapping['precision'] = 10;
                    $fieldMapping['scale'] = 2;
                }

                $classMetadataInfo->mapField($fieldMapping);
            }
        }

        return $classMetadataInfo;
    }

    private function getParentClassFromModelMetadata(StagingModelMetadata $modelMetadata)
    {
        $parentType = $modelMetadata->getParentType();
        if (is_null($parentType)) {
            $sql = sprintf("SELECT DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME='%s' and column_name='uuid';", $modelMetadata->getTableName());
            $result = $this->entityManager->getConnection()->fetchAllAssociative($sql);
            // Tenant is require for backward compatability to be always string.
            if ($modelMetadata->getType() != 'tenants' && (!count($result) || $result[0]['data_type'] == 'uuid')) {
                $parentType = 'UuidBase';
            } else {
                $parentType = 'Base';
            }
        }

        return $this->typeToClassNameMap[$parentType];
    }


    /**
     * @param $type
     * @param $doctrineClassMetadataInfo
     * @return StagingModelMetadata
     */
    private function getModelMetadataFromDoctrineClassMetadataInfo($type, $doctrineClassMetadataInfo)
    {
        $traits = $doctrineClassMetadataInfo->getReflectionClass()->getTraitNames();

        $parentClass = $doctrineClassMetadataInfo->getReflectionClass()->getParentClass();

        $parentType = null;
        if (in_array($parentClass, $this->typeToClassNameMap)) {
            $parentType = array_search($parentClass, $this->typeToClassNameMap);
        }

        $stagingModelMetadata = new StagingModelMetadata();
        $stagingModelMetadata->setType($type);
        $stagingModelMetadata->setOwnable(array_key_exists(Ownable::class, $traits));
        $stagingModelMetadata->setTaggable(array_key_exists(Taggable::class, $traits));
        $stagingModelMetadata->setIndexable(array_key_exists(Indexable::class, $traits));
        $stagingModelMetadata->setUserOwnable(array_key_exists(UserOwnable::class, $traits));
        $stagingModelMetadata->setQueueable(array_key_exists(Queueable::class, $traits));
        $stagingModelMetadata->setTrackable(array_key_exists(Trackable::class, $traits));
        $stagingModelMetadata->setPeerReplicable(array_key_exists(PeerReplicable::class, $traits));
        $stagingModelMetadata->setTableName($doctrineClassMetadataInfo->getTableName());
        $stagingModelMetadata->setParentType($parentType);

        $attributeMetadatas = [];
        foreach ($doctrineClassMetadataInfo->getFieldNames() as $fieldName) {
            $attributeMetadata = new AttributeMetadata();
            $attributeMetadata->setCollection(false);
            $attributeMetadata->setName($fieldName);
            $attributeMetadatas[$fieldName] = $attributeMetadata;
        }

        foreach ($doctrineClassMetadataInfo->getAssociationMappings() as $assocationFieldName => $associationMapping) {
            $attributeMetadata = new AttributeMetadata();
            $attributeMetadata->setCollection(true);
            $attributeMetadata->setName($assocationFieldName);

            $inversedField = $associationMapping['isOwningSide'] ? $associationMapping['inversedBy'] : $associationMapping['mappedBy'];
            $attributeMetadata->setInversedField($inversedField);
            $attributeMetadata->setOwnsRelationship($associationMapping['isOwningSide']);

            $sourceEntity = $associationMapping['sourceEntity'];
            if (!is_null($sourceEntity) && ($attributeType = array_search($sourceEntity, $this->typeToClassNameMap)) !== false) {
                $attributeMetadata->setType($attributeType);
            }
            $attributeMetadatas[$assocationFieldName] = $attributeMetadata;
        }

        if ($attributeMetadatas) {
            foreach ($attributeMetadatas as $attrMetaData) {
                $stagingModelMetadata->addAttribute($attrMetaData);
            }
        }
        return $stagingModelMetadata;
    }
}
