<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Command;

use Exception;
use InvalidArgumentException;
use PhpAmqpLib\Message\AMQPMessage;
use App\Entity\ModelMetadata\ModelMetadata;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Serializer\Serializer;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\Console\Command\Command;
use PhpAmqpLib\Connection\AMQPStreamConnection;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Style\SymfonyStyle;
use App\Entity\ModelMetadata\StagingModelMetadata;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Serializer\Encoder\JsonEncoder;
use Symfony\Component\Console\Output\ConsoleOutputInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\ObjectNormalizer;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use App\Constants\AppConstants;
use App\Constants\ErrorConstants;

class SystemUpdateCommand extends Command
{
    /** @var EntityManager */
    private $entityManager;

    public function __construct(protected ContainerInterface $container)
    {
        $this->entityManager = $this->container->get('doctrine.orm.entity_manager');
        parent::__construct();
    }

    public const STATUS = 'status';
    public const PUBLISH_CREATE = 'publish_create';
    public const PUBLISH_MODIFY = 'publish_modify';
    public const PUBLISH_DELETE = 'publish_delete';
    public const ADDED_MODEL = 'created';
    public const MODIFIED_MODEL = 'modified';
    public const DELETED_MODEL = 'deleted';
    public const META_INFO = 'metainfo';
    public const COMPAREARRAY = ['ownable', 'userOwnable', 'queueable', 'taggable', 'trackable', 'peerReplicable', 'name', 'type'];
    /** @var Filesystem */
    protected $filesystem;

    protected function configure()
    {
        $this
            ->setName('app:system:update')
            ->setAliases(['cybersponse:system:update'])
            ->setDefinition([
                new InputOption(
                    'log-to-503',
                    'l',
                    InputOption::VALUE_NONE,
                    'Reports status to ./web/{clientName}/503.json file.'
                ),
                new InputOption(
                    'last-log',
                    'a',
                    InputOption::VALUE_NONE,
                    'Will remove the log file at the end of this command.'
                ),
                new InputOption(
                    'force',
                    'f',
                    InputOption::VALUE_NONE,
                    'Will force the regeneration regardless if there were no metadata updates.'
                ),
                new InputOption(
                    'index',
                    'i',
                    InputOption::VALUE_OPTIONAL | InputOption::VALUE_IS_ARRAY,
                    'Arguments for indexing modules in ES.'
                ),
                new InputOption(
                    'sync',
                    's',
                    InputOption::VALUE_OPTIONAL,
                    'Arguments for synchronous indexing of elasticsearch'
                ),
                new InputOption(
                    'skip-cache-clear',
                    'c',
                    InputOption::VALUE_NONE,
                    'Arguments for Skipping cache clear'
                ),
                new InputOption(
                    'skip-ha-broadcast',
                    'b',
                    InputOption::VALUE_NONE,
                    'Arguments for Skipping broadcasting to other ha nodes'
                ),
                new InputOption(
                    'skip-elasticsearch-indexing',
                    'k',
                    InputOption::VALUE_NONE,
                    'Arguments for Skipping elasticsearch indexing'
                )
            ]);
        $this->filesystem = new Filesystem();
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $io = new SymfonyStyle($input, $output);
        $errorIo = $io;
        $indexObject = [];
        if ($output instanceof ConsoleOutputInterface) {
            // If it's available, get stdErr output
            $errorOutput = $output->getErrorOutput();
            $errorIo = new SymfonyStyle($input, $errorOutput);
        }

        $globalOptions = '';

        if ($input->getOption('no-ansi') === true) {
            $globalOptions .= ' --no-ansi';
        } else {
            $globalOptions .= ' --ansi';
        }

        $isForceOptionEnabled = true;
        $isMinimalPublish = false;

        $modelMetadatasWithKeys = $this->getAllMetadata(ModelMetadata::class);
        $stagingModelMetadatasWithKeys = $this->getAllMetadata(StagingModelMetadata::class);

        $encoder = new JsonEncoder();
        $normalizer = new ObjectNormalizer();
        $defaultContext = [];
        $defaultContext['circular_reference_handler'] = fn ($modelMetadatasWithKeys) => $modelMetadatasWithKeys->getUuid();

        $serializer = new Serializer([$normalizer], [$encoder]);

        $normalizedMetadata = $serializer->serialize($modelMetadatasWithKeys, 'json', $defaultContext);
        $normalizedMetadata = json_decode($normalizedMetadata, true);

        $defaultContext['circular_reference_handler'] = fn ($stagingModelMetadatasWithKeys) => $stagingModelMetadatasWithKeys->getUuid();
        $normalizedStagingMetadata = $serializer->serialize($stagingModelMetadatasWithKeys, 'json', $defaultContext);
        $normalizedStagingMetadata = json_decode($normalizedStagingMetadata, true);
        $metadataUpdatesRequired = true == $input->getOption('force');

        if ($normalizedMetadata !== $normalizedStagingMetadata) {
            $metadataUpdatesRequired = true;
        }

        $indexObjectAll = $this->getIndexOptions($normalizedStagingMetadata, $normalizedMetadata);
        $indexObject = $indexObjectAll[self::ADDED_MODEL];
        $indexUpdateObject = $indexObjectAll[self::MODIFIED_MODEL];
        $indexMetaInfo = $indexObjectAll[self::META_INFO] ?? null;

        $defaultCommandData = [
            'runCondition' => true,
            'insertNewline' => false,
        ];

        $commandsToRun = [
            'Decrypt Database' => [
                'command' => 'doctrine:decrypt:database --no-interaction',
                'runCondition' => $metadataUpdatesRequired,
            ],
            'System Backup' => [
                'command' => 'app:backup:all',
            ],
            'Generating Models' => [
                'command' => 'app:metadata:generate:models',
            ],
            'Clearing Doctrine Metadata' => [
                'command' => 'doctrine:cache:clear-metadata',
            ],
            'Clearing Cache' => [
                'command' => 'cache:clear',
            ],
            'Cache Warmup' => [
                'command' => 'cache:warmup',
            ],
            'Schema Update' => [
                'command' => 'doctrine:schema:update --em=default --force',
                'runCondition' => $metadataUpdatesRequired,
            ],
            'Activating Latest ModelMetadata' => [
                'command' => 'app:metadata:activate',
                'runCondition' => $metadataUpdatesRequired,
            ],
            'Generating Modules' => [
                'command' => 'app:metadata:generate:modules',
                'runCondition' => $metadataUpdatesRequired,
            ],
            'Cleaning Up Old Backups' => [
                'command' => 'app:cleanup:all',
            ],
            'Encrypt Database' => [
                'command' => 'doctrine:encrypt:database --no-interaction',
                'runCondition' => $metadataUpdatesRequired,
            ],
            'Wait For Update Across Cluster' => [
                'command' => 'app:cluster:publish:wait',
            ],
            'Prepare Class Type Map Cache' => [
                'command' => 'app:type:class:map',
            ],
            'Clearing App Cache' => [
                'command' => 'app:cache:clear',
            ],
        ];
        $skipCacheClear = $input->getOption('skip-cache-clear') ?? false;
        if ($skipCacheClear === true) {
            unset($commandsToRun['Clearing App Cache']);
        }
        $skipHaBroadcast = $input->getOption('skip-ha-broadcast') ?? false;
        if ($skipHaBroadcast === true) {
            unset($commandsToRun['Wait For Update Across Cluster']);
        }
        $errorCode = 0;

        if (!$isForceOptionEnabled) {
            $isMinimalPublish = $this->checkForMinimalPublish($normalizedStagingMetadata, $normalizedMetadata);
            if ($isMinimalPublish) {
                unset($commandsToRun['Decrypt Database']);
                unset($commandsToRun['System Backup']);
                unset($commandsToRun['Generating Models']);
                unset($commandsToRun['Clearing Doctrine Metadata']);
                unset($commandsToRun['Cache Warmup']);
                unset($commandsToRun['Schema Update']);
                unset($commandsToRun['Generating Modules']);
                unset($commandsToRun['Cleaning Up Old Backups']);
                unset($commandsToRun['Encrypt Database']);
            }
        }

        try {
            //Run all commands
            $total = count($commandsToRun);
            $current = 0;
            $systemUpdate = new SystemUpdate();
            $systemUpdate
                ->setMessage('Starting system update.')
                ->setProgressPercent(0);

            $this->log($input, $systemUpdate);
            $available_free_space_bytes = disk_free_space($this->container->getParameter('kernel.project_dir'));
            $available_free_space_mb = $available_free_space_bytes / AppConstants::BYTES_TO_MB_DENOMINATOR;
            if ($available_free_space_mb < AppConstants::REQUIRED_MIN_SPACE_PUBLISH) {
                $errorMessage = sprintf(ErrorConstants::INSUFFICIENT_SPACE_FOR_PUBLISH, AppConstants::REQUIRED_MIN_SPACE_PUBLISH, $available_free_space_mb, '/opt');
                $errorIo->writeln($errorMessage);
                $this->logFinish($input);
                return Command::FAILURE;
            }
            # TODO: if return is not success, do not proceed
            $this->broadcastLockUnlock('cluster_publish_lock', 'cybersponse:publish:lock');
            foreach ($commandsToRun as $processingText => $commandData) {
                $commandData = $commandData + $defaultCommandData;

                $io->writeln(sprintf('<fg=blue>%s) %s</fg=blue>', $current + 1, $processingText));
                $systemUpdate
                    ->setMessage($processingText)
                    ->setProgressPercent((int) (($current / $total) * 100));
                $this->log($input, $systemUpdate);

                if ($commandData['runCondition'] === true) {
                    if ($commandData['insertNewline'] === true) {
                        $io->writeln('');
                    }
                    $command = $commandData['command'];
                    $beforetimestamp = (new \DateTime('NOW', new \DateTimeZone('UTC')))->getTimestamp();
                    [$returnVal, $returnString] = $this->executeCommand($command, $globalOptions);
                    $io->writeln($returnString);
                    $aftertimestamp = (new \DateTime('NOW', new \DateTimeZone('UTC')))->getTimestamp();
                    $io->writeln("Command took " . ($aftertimestamp - $beforetimestamp) . " Second to execute");
                    if ($returnVal !== 0) {
                        throw new \Exception(sprintf('Command `%s` finished with code: %s', $command, $returnVal));
                    }
                    $io->write(' -> <fg=green>Finished</fg=green>', true);
                } else {
                    $io->write(' -> <fg=yellow>Skipped</fg=yellow>', true);
                }

                $current++;
            }
            $this->log($input, $systemUpdate);
            $this->broadcastLockUnlock('cluster_publish_unlock', 'app:publish:unlock');
            $body = $this->prepareChangeData('Success');
            $this->logFinish($input);
            $this->notifyAuditingComponent([$body], $io);
        } catch (\Exception | InvalidArgumentException $e) {
            $this->broadcastLockUnlock('cluster_publish_unlock', 'app:publish:unlock');
            $error = [];
            $error['error'] = $e->getMessage();
            $body = $this->prepareChangeData('Failure', $error);
            $this->notifyAuditingComponent([$body], $io);
            //Perform rollback
            $errorIo->error(sprintf('Failed: %s, Restoring from backup...', $e->getMessage()));
            $errorCode = 1;
            $commands = ['app:restore:all', 'doctrine:decrypt:database --no-interaction', 'doctrine:encrypt:database --no-interaction', 'cache:warmup'];
            foreach ($commands as $command) {
                $beforetimestamp = (new \DateTime('NOW', new \DateTimeZone('UTC')))->getTimestamp();
                [$returnVal, $returnString] = $this->executeCommand($command, $globalOptions);
                $aftertimestamp = (new \DateTime('NOW', new \DateTimeZone('UTC')))->getTimestamp();
                $io->writeln("Command took " . ($aftertimestamp - $beforetimestamp) . " Second to execute");
                if ($returnVal !== 0) {
                    //Todo What happens when the system is not restored properly?
                    throw new \Exception(sprintf('Command: %s , Error Code: %s, Could not restore system: %s', $command, $returnVal, $returnString));
                }
                $io->writeln($returnString);
            }
            $io->success('Successfully restored up the system. Please view the above errors and fix.');
        } finally {
            $this->logFinish($input);
            if ($errorCode == 0) {
                $indexName = $input->getOption('index') ?? [];
                $indexOption = "";
                $updateOption = "";
                $typeModule = (object) null;
                if ($indexName && !empty($indexName)) {
                    $indexName = json_decode($indexName[0], true);
                    $indexObject = array_merge($indexObject, $indexName['type']);
                    $typeModule = ['type' => $indexObject];
                    $indexOption = " --index='" . json_encode($typeModule) . "'";
                } elseif ($indexObject && (is_countable($indexObject) ? count($indexObject) : 0)) {
                    $typeModule = ['type' => $indexObject, 'meta' => $indexMetaInfo];
                    $indexOption = " --index='" . json_encode($typeModule) . "'";
                }

                if ($input->getOption('sync')) {
                    $indexOption .= " --sync=" . $input->getOption('sync');
                }
                $typeModule = (object) null;
                if ($indexUpdateObject && (is_countable($indexUpdateObject) ? count($indexUpdateObject) : 0)) {
                    $typeModule = ['type' => $indexUpdateObject];
                    $updateOption = " --update='" . json_encode($typeModule) . "'";
                }
                if ($indexObject && (is_countable($indexObject) ? count($indexObject) : 0) || $indexUpdateObject && (is_countable($indexUpdateObject) ? count($indexUpdateObject) : 0)) {
                    $skipIndexing = $input->getOption('skip-elasticsearch-indexing') ?? false;
                    if ($skipIndexing === false) // $isMinimalPublish is removed from condition
                    {
                        $this->executeCommand('app:elastic:create', $globalOptions, $indexOption, $updateOption);
                    }
                }
            }
            $io->success('Finished with system update!');
        }

        return $errorCode;
    }

    private function broadcastLockUnlock($commandId, $command)
    {
        $dasUri = $this->container->getParameter('das.uri');
        $dasClient = $this->container->get("app.proxy.client.das");
        $objectData = ["commandId" => $commandId, "command" => "sudo -u nginx php /opt/cyops-api/bin/console " . $command, "includeFiredrillNodes" => false];
        $headers = ["Content-type" => "application/json"];
        $dasRequests = new GuzzleRequest('post', $dasUri . '/clustercommand/', $headers, json_encode($objectData));
        try {
            $dasResponse = $dasClient->send($dasRequests);
            $resCode = $dasResponse->getStatusCode();
            if ($resCode == 200) {
                return Command::SUCCESS;
            }
        } catch (BadRequestHttpException | \Exception $error) {
            error_log($error->getMessage());
        }
        return -1;
    }

    /**
     * @param $class
     * @return array
     */
    private function getAllMetadata($class)
    {
        $metadataDocumentManager = $this->container->get('doctrine.orm.entity_manager');

        /** @var EntityRepository $modelMetadataRepository */
        $modelMetadataRepository = $metadataDocumentManager->getRepository($class);

        /** @var ModelMetadata[] $modelMetadatas */
        $modelMetadatas = $modelMetadataRepository->findAll();
        $modelMetadatasWithKeys = [];

        foreach ($modelMetadatas as $modelMetadata) {
            $modelMetadatasWithKeys[$modelMetadata->getType()] = $modelMetadata;
        }
        return $modelMetadatasWithKeys;
    }

    protected function log(InputInterface $input, SystemUpdate $systemUpdate)
    {
        $logTo503 = true === $input->getOption('log-to-503');
        $logDir = $this->getLogDir();

        if ($logTo503) {
            $this->filesystem->dumpFile(sprintf('%s/503.json', $logDir), json_encode($systemUpdate));
        }
    }

    private function getLogDir()
    {
        $rootDir = $this->container->getParameter('kernel.project_dir');
        return sprintf('%s/public', $rootDir);
    }

    /**
     * @param $command
     * @param $globalOptions
     * @return array
     */
    private function executeCommand($command, $globalOptions, $indexOption = "", $updateOption = "")
    {
        if ($indexOption != "") {
            $globalOptions .= $indexOption;
        }
        if ($updateOption != "") {
            $globalOptions .= $updateOption;
        }
        $fullCommand = sprintf('/usr/bin/php /opt/cyops-api/bin/console %s %s', $command, $globalOptions);
        $execOutput = null;
        exec($fullCommand, $execOutput, $returnVal);
        return [$returnVal, $execOutput];
    }

    /**
     * @param $status
     * @param array $error
     * @return array
     */
    protected function prepareChangeData($status, $error = [])
    {
        $error['status'] = $status;
        $error['message'] = (object) null;
        return [
            AppConstants::COMPONENT => 'crudhub',
            AppConstants::ENTITY_TYPE => 'ModelMetadata',
            AppConstants::DELTA_DATA => $error,
            AppConstants::TRANSACTION_DATE => (int) ($_SERVER['REQUEST_TIME_FLOAT'] * 1000),
            AppConstants::ACTION => 'STATUS',
        ];
    }

    /**
     * @param $body
     * @return bool
     */
    protected function notifyAuditingComponent($body, $io)
    {
        $channel = null;
        $connection = null;
        try {
            $mqCredObj = parse_ini_file('/opt/cyops/configs/rabbitmq/rabbitmq_users.conf', false, INI_SCANNER_NORMAL);
            $rabbitMqUsername = $mqCredObj['cyops.rabbitmq.user'] ?: 'guest';
            $rabbitMqPassword = $mqCredObj['cyops.rabbitmq.password'] ?: 'guest';
            $rabbitMqHost = $mqCredObj['cyops.rabbitmq.host'] ?: 'localhost';
            $rabbitMqPort = $mqCredObj['cyops.rabbitmq.port'] ?: '5672';
            $rabbitMqVHost = $mqCredObj['cyops.rabbitmq.vhost'] ?: '/';

            $connection = new AMQPStreamConnection($rabbitMqHost, $rabbitMqPort, $rabbitMqUsername, $rabbitMqPassword, $rabbitMqVHost);
            $channel = $connection->channel();
            $channel->exchange_declare('cyops.crudhub.datanotify', 'topic', false, true, false);

            $msg = new AMQPMessage(json_encode($body), ['content_type' => 'application/json']);
            $channel->basic_publish($msg, 'cyops.crudhub.datanotify', 'cyops.crudhub.datanotify');
            $channel->basic_publish($msg, 'cyops.crudhub.datanotify', 'route.crudhub.instructions.remoterequest.modelmmdupdate');
            $channel->close();
            $connection->close();
            return true;
        } catch (BadRequestHttpException | \Exception $error) {
            $io->write(' -> <fg=yellow>Rabbitmq service is not reachable; hence, this publish might not be audited.</fg=yellow>', true);
            if ($channel && $connection) {
                $channel->close();
                $connection->close();
            }
            return false;
        }
    }

    /**
     * Finishes the log
     *
     */
    protected function logFinish(InputInterface $input)
    {
        $logTo503 = true === $input->getOption('log-to-503');
        $lastLog = true === $input->getOption('last-log');
        $logDir = $this->getLogDir();

        if ($logTo503 && $lastLog) {
            $this->filesystem->remove(sprintf('%s/503.json', $logDir));
        }
    }

    /** @return object
     * {
     * "created":[],
     * "deleted":[],
     * "modified":{
     *      "tasks":{
     *              "added":{"attributes":[{"newAttribute":{"name":"newAttribute","type":"string"}}]}
     *              },
     *      "indicators":{
     *              "modified":{"trackable":false,"taggable":false},
     *              "old":{"trackable":true,"taggable":true},
     *              "added":{"attributes":[{"newIndicatorField":{"name":"newIndicatorField","type":"string"}}]},
     *              "removed":{"attributes":[{"firstSeen":{"name":"firstSeen","type":"integer"}}]}
     *       }
     *  }
     *} */
    protected function getIndexOptions($stagingMetadata, $modelMetadata)
    {
        //use PHP in-built function to detect added/deleted/existing models
        $newModels = array_diff_key($stagingMetadata, $modelMetadata);
        $deletedModels = array_diff_key($modelMetadata, $stagingMetadata);
        $existingModels = array_intersect_key($stagingMetadata, $modelMetadata);
        $modelDiff = [];
        $modelDiff[self::ADDED_MODEL] = $newModels;
        $modelDiff[self::DELETED_MODEL] = $deletedModels;

        //loop through data to find diff using != operator on whole model.will be looping below on each model.
        foreach ($existingModels as $key => $value) {

            $stagingSoftDeleteable = $value['softDeleteable'];
            $modelMetaDataObject = $modelMetadata[$value['type']];
            $mmdSoftDeleteable = $modelMetaDataObject['softDeleteable'];

            if ($mmdSoftDeleteable && !$stagingSoftDeleteable) {
                $sql = sprintf("delete from %s where deletedat is not null", $value['tableName']);
                try {
                    $this->entityManager->getConnection()->prepare($sql)->executeQuery();
                } catch (InvalidArgumentException | \Exception $e) {
                    error_log($e->getMessage());
                }
            }

            $stagingValues = array_filter($stagingMetadata[$key], function ($varValue, $varKey) use ($stagingMetadata, $modelMetadata, $key) {
                if (\in_array($varKey, self::COMPAREARRAY) && $stagingMetadata[$key][$varKey] != $modelMetadata[$key][$varKey]) {
                    return true;
                } else {
                    return false;
                }
            }, ARRAY_FILTER_USE_BOTH);

            $stagingAttribute = [];
            $modelAttribute = [];
            $attrib = [];
            $attrib['added'] = [];
            $attrib['deleted'] = [];
            $attrib['modified'] = [];

            foreach ($stagingMetadata[$key]['attributes'] as $attribKey => $attribValue) {
                $stagingAttribute[$attribValue['name']] = ['name' => $attribValue['name'], 'type' => $attribValue['type'], 'formType' => $attribValue['formType'], 'encrypted' => $attribValue['encrypted']];
            }
            foreach ($modelMetadata[$key]['attributes'] as $attribKey => $attribValue) {
                $modelAttribute[$attribValue['name']] = ['name' => $attribValue['name'], 'type' => $attribValue['type'], 'formType' => $attribValue['formType'], 'encrypted' => $attribValue['encrypted']];
            }

            $diffVarHolder = array_diff_key($stagingAttribute, $modelAttribute);
            if ($diffVarHolder) {
                $attrib['added'] = $diffVarHolder;
            }
            $diffVarHolder = array_diff_key($modelAttribute, $stagingAttribute);
            if ($diffVarHolder) {
                $attrib['deleted'] = $diffVarHolder;
            }
            $existingModelsAttrib = array_intersect_key($modelAttribute, $stagingAttribute);
            foreach ($existingModelsAttrib as $existingModelsAttribkey => $existingModelsAttribvalue) {
                if ($stagingAttribute[$existingModelsAttribkey] !== $modelAttribute[$existingModelsAttribkey]) {
                    array_push($attrib['modified'], $stagingAttribute[$existingModelsAttribkey]);
                }
            }
            //check if we have any difference in mmd on exising model
            if (count((array) $stagingValues) > 0) {
                $modelDiff[self::MODIFIED_MODEL][$key] = [];
                $modelDiff[self::MODIFIED_MODEL][$key]['modified'] = $stagingValues;
                foreach ($stagingValues as $stagingKey => $satagingValue) {
                    $modelDiff[self::MODIFIED_MODEL][$key]['old'][$stagingKey] = $modelMetadata[$key][$stagingKey];
                }
            }
            if (count($attrib['modified']) > 0) {
                if (!isset($modelDiff[self::MODIFIED_MODEL][$key])) {
                    $modelDiff[self::MODIFIED_MODEL][$key] = [];
                }
                $modelDiff[self::MODIFIED_MODEL][$key]['modified']['attributes'] = $attrib['modified'];
                $modelDiff[self::MODIFIED_MODEL][$key]['old']['attributes'] = [];
                foreach ($attrib['modified'] as $attribKey => $attribValue) {
                    array_push($modelDiff[self::MODIFIED_MODEL][$key]['old']['attributes'], $modelAttribute[$attribValue['name']]);
                }
            }
            if ($attrib['added']) {
                $modelDiff[self::MODIFIED_MODEL][$key]['added']['attributes'] = $attrib['added'];
            }
            if ($attrib['deleted']) {
                $modelDiff[self::MODIFIED_MODEL][$key]['removed']['attributes'] = $attrib['deleted'];
            }
        }

        $indexNames = [];
        if ($modelDiff[self::ADDED_MODEL] && array_keys($modelDiff[self::ADDED_MODEL])) {
            $indexNames[self::ADDED_MODEL] = array_keys($modelDiff[self::ADDED_MODEL]);
        } else {
            $indexNames[self::ADDED_MODEL] = [];
        }
        if (isset($modelDiff[self::DELETED_MODEL]) && array_keys($modelDiff[self::DELETED_MODEL])) {
            $indexNames[self::DELETED_MODEL] = array_keys($modelDiff[self::DELETED_MODEL]);
        } else {
            $indexNames[self::DELETED_MODEL] = [];
        }
        if (isset($modelDiff[self::MODIFIED_MODEL])) {
            $indexNames[self::MODIFIED_MODEL] = $modelDiff[self::MODIFIED_MODEL];
            $getModifiedObject = [];
            // check of modified attrubutes if found send data to re index . only send --update the model if attributes are added .we can add model if attribute is dropped also
            foreach ($modelDiff[self::MODIFIED_MODEL] as $key => $value) {
                if (isset($value['modified']['attributes']) && (is_countable($value['modified']['attributes']) ? count($value['modified']['attributes']) : 0) > 0) {
                    array_push($indexNames[self::ADDED_MODEL], $key);
                    $getModifiedObject[$key] = $value;
                    unset($indexNames[self::MODIFIED_MODEL][$key]);
                }
            }
            $indexNames[self::META_INFO] = $getModifiedObject;
        } else {
            $indexNames[self::MODIFIED_MODEL] = [];
        }
        return $indexNames;
    }

    private function checkForMinimalPublish($stagingMetadatas, $modelMetadatas)
    {
        try {
            $arrayOfCascadeAttribs = AppConstants::PUBLISH_IGNORE_ATTRIBUTES;
            $newModels = array_diff_key($stagingMetadatas, $modelMetadatas);
            $deletedModels = array_diff_key($modelMetadatas, $stagingMetadatas);
            if ($newModels || $deletedModels) {
                return false;
            }
            foreach ($stagingMetadatas as $stagingMetadata) {
                $stagingMetadataType = $stagingMetadata['type'];
                foreach ($modelMetadatas as $modelMetadata) {
                    if ($stagingMetadataType == $modelMetadata['type']) {
                        $smmAttributes = $stagingMetadata['attributes'];
                        $mmAttributes = $modelMetadata['attributes'];
                        unset($stagingMetadata['attributes']);
                        unset($modelMetadata['attributes']);

                        sort($stagingMetadata);
                        sort($modelMetadata);

                        if ($stagingMetadata != $modelMetadata) {
                            return false;
                        } elseif (count($smmAttributes) > count($mmAttributes) || count($smmAttributes) < count($mmAttributes)) {
                            return false;
                        } elseif (count($smmAttributes) == count($mmAttributes)) {
                            if (!$this->checkAttributeDiffForMinimalPublish($smmAttributes, $mmAttributes, $arrayOfCascadeAttribs)) {
                                return false;
                            }
                        }
                    }
                }
            }
            return true;
        } catch (\Exception $error) {
            error_log($error->getMessage());
        }
    }

    private function checkAttributeDiffForMinimalPublish($smmAttributes, $mmAttributes, $arrayOfCascadeAttribs)
    {
        try {
            foreach ($smmAttributes as $smmAttribKey => $smmAttribValue) {
                $smmAttribKey = $smmAttribValue['name'];
                foreach ($mmAttributes as $mmAttribKey => $mmAttribValue) {
                    if ($smmAttribKey == $mmAttribValue['name']) {
                        foreach ($smmAttribValue as $smmKey => $smmValue) {
                            foreach ($mmAttribValue as $mmKey => $mmValue) {
                                if ($smmKey == $mmKey) {
                                    if (is_array($smmValue) && is_array($mmValue)) {
                                        sort($smmValue);
                                        sort($mmValue);
                                    }
                                    if ($smmValue != $mmValue && !in_array($smmKey, $arrayOfCascadeAttribs)) {
                                        return false;
                                    }
                                }
                            }
                        }
                    }
                }
            }
            return true;
        } catch (\Exception $error) {
            error_log($error->getMessage());
            return false;
        }
    }
}
