<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Doctrine\ORM;

use DateTime;
use DateTimeImmutable;
use App\Entity\Query\Query;
use Doctrine\ORM\Query\Expr;
use App\Service\MetadataHelper;
use App\Service\UtilityService;
use App\Service\OwnershipChecker;
use Doctrine\ORM\Query\Expr\From;
use App\Constants\AggregateOperators;
use App\Entity\ViewTemplate\Dashboard;
use App\Entity\ViewTemplate\Reporting;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\Persistence\ManagerRegistry;
use App\Constants\AttributeLevelOperators;
use App\Filter\AbstractFilterQueryBuilder;
use App\Service\AccessibleTeamIdsProvider;
use App\Entity\ViewTemplate\GenericTemplate;
use Symfony\Component\HttpFoundation\Response;
use ApiPlatform\Core\Api\IriConverterInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\HttpKernel\Exception\HttpException;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

class FilterQueryBuilder extends AbstractFilterQueryBuilder
{
    const AUTHORIZATION_TEAMS_PARAMETER = 'teams';
    const AUTHORIZATION_ROLES_PARAMETER = 'roles';
    const JSON_FIELD = 'jsonb';
    const GENERIC_TEMPLATE_CLASS = [GenericTemplate::class, Dashboard::class, Reporting::class];
    const QUERY_CLASS = ["App\Entity\Query\Query", "App\Entity\Query\SystemQuery", "App\Entity\Query\UserQuery"];
    const AUTHORIZATION_ACTORS_PARAMETER = 'actors';

    /** @var int */
    protected $parameterCount;
    /** @var array */
    protected $parameters;
    /** @var array */
    protected $selects;
    /** @var array */
    protected $aliases;
    /** @var array */
    protected $orderings;
    /** @var array */
    protected $groupings;
    /** @var array */
    protected $joins;
    /** @var array */
    protected $associationAliases;
    /** @var string */
    protected $fromAlias;
    /** @var ClassMetadata */
    protected $metadata;

    /** @var ManagerRegistry */
    protected $managerRegistry;

    protected $aggregateQuery = false;

    /** @var TokenStorageInterface */
    protected $tokenStorage;

    protected $ownershipChecker;

    protected $accessibleTeamIdsProvider;

    protected $iriConverter;
    private $metadataHelper;
    private $utilityService;
    private $requestStack;

    public function __construct(
        OwnershipChecker $ownershipChecker,
        AccessibleTeamIdsProvider $accessibleTeamIdsProvider,
        ManagerRegistry $managerRegistry,
        TokenStorageInterface $tokenStorage,
        IriConverterInterface $iriConverter,
        MetadataHelper $metadataHelper,
        UtilityService $utilityService,
        RequestStack $requestStack
    ) {
        $this->ownershipChecker = $ownershipChecker;
        $this->accessibleTeamIdsProvider = $accessibleTeamIdsProvider;
        $this->managerRegistry = $managerRegistry;
        $this->tokenStorage = $tokenStorage;
        $this->iriConverter = $iriConverter;
        $this->metadataHelper = $metadataHelper;
        $this->utilityService = $utilityService;
        $this->requestStack = $requestStack;
    }

    /**
     * @param array $query
     * @return self
     */
    public function fromQueryObject($query, $queryBuilder, $resourceClass, $fields = [], $dataGetRequest = true)
    {
        $request = $this->requestStack->getCurrentRequest();
        $this->parameterCount = 0;
        $this->parameters = [];
        $this->selects = $fields;
        $this->aliases = [];
        $this->orderings = [];
        $this->groupings = [];
        $this->joins = [];
        $this->associationAliases = [];
        /** @var From $fromClause */
        $fromClause = current($queryBuilder->getDQLPart('from'));
        $fromEntityClass = $fromClause->getFrom();
        $this->fromAlias = $fromClause->getAlias();
        $entityFieldMap = $this->metadataHelper->getEntityFields($resourceClass) ?? null;

        $this->metadata = $queryBuilder->getEntityManager()->getClassMetadata($fromEntityClass);
        // apply ownership filters
        if ((isset($GLOBALS["__associatedFieldRequest__"]) ? $GLOBALS["__associatedFieldRequest__"] : null) != true
            || $this->ownershipChecker->isClassConditionalOwnable($fromEntityClass)
        ) {
            $ownershipExpression = $this->getOwnershipExpression($this->fromAlias, $fromEntityClass);
            if ($ownershipExpression) {
                $queryBuilder->andWhere($ownershipExpression);
            }
        }

        if (in_array($fromEntityClass, self::GENERIC_TEMPLATE_CLASS)) {
            $userIdExpression = $this->getUserExpression($this->fromAlias);
            if ($userIdExpression) {
                $queryBuilder->orWhere($userIdExpression);
            }
            $ownershipExpression = $this->getRoleExpression($this->fromAlias);
            if ($ownershipExpression) {
                $queryBuilder->orWhere($ownershipExpression);
            }
        }

        if (in_array($fromEntityClass, self::QUERY_CLASS)) {
            $exploded_uri = [];
            if ($request) {
                $exploded_uri = explode('/', $request->getRequestUri());
            }
            if (empty($exploded_uri) || !(isset($exploded_uri[5]) && $exploded_uri[5]=='datasets')) {
                $queryAssigneeExpression = $this->getQueryAssigneeExpression($this->fromAlias);
                if ($queryAssigneeExpression) {
                    $queryBuilder->orWhere($queryAssigneeExpression);
                }
            }
        }

        // apply query filters
        if ($query) {
            $filterExpression = $this->getFilterExpression($query, $queryBuilder);
            if ($filterExpression) {
                $queryBuilder->andWhere($filterExpression);
            }
        }

        // apply parameters
        if (!empty($this->parameters)) {
            if (count($queryBuilder->getParameters()->toarray())) {
                $appendParam = [];
                $appendParam[':' . $queryBuilder->getParameters()->toarray()[0]->getName()] = $queryBuilder->getParameters()->toarray()[0]->getValue();
                $this->parameters = array_merge($this->parameters, $appendParam);
            }
            $queryBuilder->setParameters($this->parameters);
        }

        if ($query && array_key_exists(Query::KEYWORD_AGGREGATES, $query)) {
            foreach ($query[Query::KEYWORD_AGGREGATES] as $aggregate) {
                $aggregateExpression = $this->getAggregateExpression($aggregate);
                if ($aggregate['alias']) {
                    $this->aliases[] = $aggregate['alias'];
                }
                if (!is_null($aggregateExpression)) {
                    $this->selects[] = $aggregateExpression;
                }
            }
        }
        if (!is_null($entityFieldMap) && $query && !$this->aggregateQuery && (!array_key_exists(Query::KEYWORD_SORT, $query)
            || !is_iterable($query[Query::KEYWORD_SORT]))) {
            $query[Query::KEYWORD_SORT] = $entityFieldMap['defaultSort'];
        }

        if ($query && array_key_exists(Query::KEYWORD_SORT, $query) && is_iterable($query[Query::KEYWORD_SORT])) {
            foreach ($query[Query::KEYWORD_SORT] as $ordering) {
                $ordering = (array)$ordering;
                $relatedFields = $entityFieldMap['relatedFields'] ?? [];
                if (!in_array(explode('.', $ordering['field'])[0], $relatedFields)) {
                    $fieldAlias = $this->getFieldAlias($ordering['field']);
                    $this->orderings[] = [
                        'field' => $fieldAlias,
                        'direction' => $ordering['direction'],
                    ];
                }
            }
        }
        // $dataGetRequest this flag is added to avoid unwanted distinct and ordering for an internal query.
        // Where ordering of the data doesn't matter.
        if ($dataGetRequest && !$this->aggregateQuery && empty($this->groupings) && !empty($this->selects)) {
            $primaryIdentifier = $this->metadata->getIdentifier()[0];
            $this->selects[] = sprintf('o.%s', $primaryIdentifier);
            $queryBuilder->distinct();
        }

        if ($dataGetRequest && !$this->aggregateQuery && empty($this->orderings) 
            && empty($this->groupings) && empty($entityFieldMap['defaultSort'])) {
            // Apply default ordering
            // Todo this should be set by the metadata
            $identifier = $this->metadata->getIdentifier();
            $orderByField = current($identifier);
            $this->orderings[] = [
                'field' => sprintf('o.%s', $orderByField),
                'direction' => 'ASC',
            ];
        }

        if ($this->orderings) {
            foreach ($this->orderings as $ordering) {
                $queryBuilder->addOrderBy($ordering['field'], $ordering['direction']);
            }
        }

        if (!empty($this->groupings)) {
            foreach ($this->groupings as $grouping) {
                $queryBuilder->addGroupBy($grouping);
            }
        }

        if (count($this->selects) > 0) {
            $queryBuilder->select(implode(', ', $this->selects));
        }

        // Apply join clauses
        // TODO: How does this sort 10+ joins (t0.asdf, t1.asdf, t2.asdf, ..., t10.asdf)
        usort($this->joins, function ($join1, $join2) {
            return strcasecmp($join1['alias'], $join2['alias']);
        });

        foreach ($this->joins as $join) {
            // TODO: This is out of order

            if ($join['type'] && $this->metadata::TO_ONE) {
                $queryBuilder->leftJoin($join['join'], $join['alias']);
            } else {
                $queryBuilder->join($join['join'], $join['alias']);
            }
        }
        return $this;
    }

    /**
     * @param array $aggregateFilter
     * @return null|string
     */
    protected function getAggregateExpression($aggregateFilter)
    {
        $aggregateExpression = null;
        $aggregateOperator = $aggregateFilter['operator'];
        $this->convertFilterFieldToDotNotation($aggregateFilter);
        $fieldAlias = $this->getFieldAlias($aggregateFilter['field'], isset($aggregateFilter['fieldOperator']) ? $aggregateFilter['fieldOperator'] : null);

        switch ($aggregateOperator) {
            case AggregateOperators::SUM:
                $this->aggregateQuery = true;
                $aggregateExpression = sprintf('SUM(%s)', $fieldAlias);
                break;
            case AggregateOperators::COUNT_DISTINCT:
                $this->aggregateQuery = true;
                $aggregateExpression = $this->getExpression()->countDistinct($fieldAlias);
                break;
            case AggregateOperators::COUNT:
                $this->aggregateQuery = true;
                $aggregateExpression = $this->getExpression()->count($fieldAlias);
                break;
            case AggregateOperators::GROUP_BY:
                $this->groupings[] = $fieldAlias;
                $aggregateExpression = $fieldAlias;
                break;
            case AggregateOperators::SELECT:
            case AggregateOperators::FIELDS:
                $aggregateExpression = $fieldAlias;
                break;
            case AggregateOperators::MAX:
                $this->aggregateQuery = true;
                $aggregateExpression = $this->getExpression()->max($fieldAlias);
                break;
            case AggregateOperators::MIN:
                $this->aggregateQuery = true;
                $aggregateExpression = $this->getExpression()->min($fieldAlias);
                break;
            case AggregateOperators::AVG:
                $this->aggregateQuery = true;
                $aggregateExpression = $this->getExpression()->avg($fieldAlias);
                break;
            case AggregateOperators::MEDIAN:
                $this->aggregateQuery = true;
                $aggregateExpression = sprintf('percentile_cont(0.5) within group (order by %s)', $fieldAlias);
                break;
            case AggregateOperators::DISTINCT:
                break;
        }

        if (array_key_exists('alias', $aggregateFilter)) {
            $resultAlias = $aggregateFilter['alias'];
            $aggregateExpression = sprintf('%s AS %s', $aggregateExpression, $resultAlias);
        }

        return $aggregateExpression;
    }

    /**
     * @param string $fromAlias
     * @return Expr
     */
    protected function getOwnershipExpression($fromAlias, $fromEntityClass)
    {
        $teamExpression = null;
        $userExpression = null;
        if ($this->ownershipChecker->isClassOwnable($fromEntityClass) || $this->ownershipChecker->isClassConditionalOwnable($fromEntityClass)) {
            $accessibleTeams = $this->accessibleTeamIdsProvider->getDataForCurrentUser();
            $ownersAlias = $this->getAssociationAlias($this->getPropertyPath($fromAlias, 'owners'));
            $this->parameters[':' . self::AUTHORIZATION_TEAMS_PARAMETER] = $accessibleTeams;
            $teamExpression = $this->getExpression()->in($ownersAlias, ':' . self::AUTHORIZATION_TEAMS_PARAMETER) . ' or ' . $this->getExpression()->isNull($ownersAlias);
        }

        if ($this->ownershipChecker->isClassUserOwnable($fromEntityClass)) {
            $userUuid = $this->getCurrentUserUuid();
            $userOwnersAlias = $this->getAssociationAlias($this->getPropertyPath($fromAlias, 'userOwners'));
            $this->parameters[':' . self::AUTHORIZATION_ACTORS_PARAMETER] = [$userUuid];
            $userExpression = $this->getExpression()->in($userOwnersAlias, ':' . self::AUTHORIZATION_ACTORS_PARAMETER);
        }

        if ($teamExpression && $userExpression) {
            return '(' . $teamExpression . ') or (' . $userExpression . ')';
        } else if ($teamExpression) {
            return $teamExpression;
        } else if ($userExpression) {
            return $userExpression;
        }
    }

    /**
     * @param string $fromAlias
     * @return Expr
     */
    protected function getRoleExpression($fromAlias)
    {
        $accessibleRole = $this->getCurrentUserRoles();
        $roleAlias = $this->getAssociationAlias($this->getPropertyPath($fromAlias, 'roles'));
        $this->parameters[':' . self::AUTHORIZATION_ROLES_PARAMETER] = $accessibleRole;

        return $this->getExpression()->in($roleAlias, ':' . self::AUTHORIZATION_ROLES_PARAMETER); // . ' or ' . $this->getExpression()->eq($fromAlias.'.userid',$userUuid);
    }

    /**
     * @param string $fromAlias
     * @return Expr
     */
    protected function getUserExpression($fromAlias)
    {
        $userUuid = $this->getCurrentUserUuid();

        $this->parameters[':createuser'] = $userUuid;
        $this->parameters[':uuid'] = '82bc8209-ef8d-463d-bcd8-db3ac08882cb-463d';
        return $this->getExpression()->eq($fromAlias . '.createUser', ':createuser') . ' or ' . $this->getExpression()->eq($fromAlias . '.uuid', ':uuid');
    }

    /**
     * @param string $fromAlias
     * @return Expr
     */
    protected function getQueryAssigneeExpression($fromAlias)
    {
        $userUuid = $this->getCurrentUserAsIri();
        $fieldAlias = $this->getFieldAlias('assignee');
        $this->parameters[':assignee'] = $userUuid;
        $this->parameters[':blank'] = '';
        return $this->getExpression()->eq($fromAlias . '.assignee', ':assignee') . ' or ' . $this->getExpression()->isNull($fieldAlias) . ' or ' . $this->getExpression()->eq($fromAlias . '.assignee', ':blank');
    }

    /**
     * @return array string
     */
    protected function getCurrentUserRoles()
    {
        /** @var  $user Actor */
        $user = $this->tokenStorage->getToken()->getUser();
        return $user->getRoles();
    }

    /**
     * @return string
     */
    private function getCurrentUserUuid()
    {
        return $this->tokenStorage->getToken()->getUser()->getUuid();
    }

    /**
     * @return string
     */
    private function getCurrentUserAsIri()
    {
        $user = $this->tokenStorage->getToken()->getUser();
        return $this->iriConverter->getIriFromItem($user);
    }

    /**
     * @param array $filter
     * @return mixed|null
     * @throws \HttpException
     */
    public function getFilterExpression($filter = [], $queryBuilder = null)
    {
        $fromClause = current($queryBuilder->getDQLPart('from'));
        $fromEntityClass = $fromClause->getFrom();
        $this->fromAlias = $fromClause->getAlias();

        $this->metadata = $queryBuilder->getEntityManager()->getClassMetadata($fromEntityClass);

        $expression = null;

        if ($filter && array_key_exists(Query::KEYWORD_LOGIC, $filter) && array_key_exists(Query::KEYWORD_FILTERS, $filter)) {
            $logic = $filter[Query::KEYWORD_LOGIC];
            switch ($logic) {
                case 'AND':
                    $logicMethod = 'andX';
                    break;
                case 'OR':
                    $logicMethod = 'orX';
                    break;
                default:
                    $errormsg = (sprintf('Unknown logic type specified: "%s"', $logic));
                    throw new HttpException(Response::HTTP_BAD_REQUEST, $errormsg);
            }

            $filters = $filter[Query::KEYWORD_FILTERS];
            $expressions = [];

            foreach ($filters as $nestedFilter) {
                $nestedExpression = $this->getFilterExpression($nestedFilter, $queryBuilder);
                if ($nestedExpression) {
                    $expressions[] = $nestedExpression;
                }
            }

            if (count($expressions) > 1) {
                $expression = call_user_func_array([$this->getExpression(), $logicMethod], $expressions);
            } else if (count($expressions) === 1) {
                $expression = current($expressions);
            }
        } else if ($filter && array_key_exists('field', $filter)) {
            $expression = $this->getFieldExpression($filter);
        }

        return $expression;
    }

    /**
     * @param ClassMetadata $metadata
     * @param string $field
     * @return bool
     */
    protected function isAssociation($metadata, $field)
    {
        $fieldParts = explode('.', $field);
        $isAssociation = false;
        $currentField = array_shift($fieldParts);
        if ($currentField == $field && count($fieldParts) == 0) {
            $this->isField($metadata, $field);
        }
        if ($metadata->hasAssociation($currentField)) {
            if (count($fieldParts) > 0) {
                $nextField = implode('.', $fieldParts);
                $associationMapping = $metadata->getAssociationMapping($currentField);
                $targetEntity = $associationMapping['targetEntity'];
                $associatedMetadata = $this->getMetadataForClass($targetEntity);
                $isAssociation = $this->isAssociation($associatedMetadata, $nextField);
            } else {
                $isAssociation = true;
            }
        }

        return $isAssociation;
    }

    /**
     * @param string $field
     * @param string $value
     * @return array
     */
    protected function splitFieldFromJson($field, $value)
    {
        $queryParts = explode('.', $field);
        $currentField = array_shift($queryParts);
        $fieldParts = [$currentField];
        $metadata = $this->metadata;
        while ($metadata->hasAssociation($currentField)) {
            $metadata = $this->getMetadataForClass($metadata->associationMappings[$currentField]['targetEntity']);
            $currentField = array_shift($queryParts);
            if ($currentField) {
                $fieldParts[] = $currentField;
            }
        }

        $isJson = false;
        if ($metadata->hasField($currentField) && $metadata->fieldMappings[$currentField]['type'] == self::JSON_FIELD) {
            $isJson = true;
        }

        $jsonObject = $value;
        foreach (array_reverse($queryParts) as $part) {
            $jsonObject = [
                $part => $jsonObject,
            ];
        }

        $fieldAlias = call_user_func_array([$this, 'getPropertyPath'], $fieldParts);
        return [$fieldAlias, $jsonObject, $isJson];
    }

    /**
     * @param string $field
     * @return array
     */
    protected function getJsonQuery($field)
    {
        $queryParts = explode('.', $field);
        $currentField = array_shift($queryParts);
        $metadata = $this->metadata;
        while ($metadata->hasAssociation($currentField)) {
            $metadata = $this->getMetadataForClass($metadata->associationMappings[$currentField]['targetEntity']);
            $currentField = array_shift($queryParts);
        }

        return $queryParts;
    }

    /**
     * @param $class
     * @return ClassMetadata
     */
    protected function getMetadataForClass($class)
    {
        $manager = $this->managerRegistry->getManagerForClass($class);
        return $manager->getClassMetadata($class);
    }

    private function convertEpochtoDateTime($value)
    {
        $clone_value = $value;
        $value = DateTimeImmutable::createFromFormat('U.u', $value);
        if (!$value) {
            $value = \DateTime::createFromFormat('U', $clone_value);
        }
        return $value->format('Y-m-d H:i:s.u');
    }

    /**
     * @param array $filter
     * @return Expr\Comparison|null
     */
    protected function getFieldExpression(array $filter)
    {
        if (!array_key_exists('field', $filter)) {
            throw new HttpException(Response::HTTP_BAD_REQUEST, 'Filter must have a field.');
        }
        if (!array_key_exists('operator', $filter)) {
            throw new HttpException(Response::HTTP_BAD_REQUEST, 'Filter must have an operator.');
        }

        $this->convertFilterFieldToDotNotation($filter);
        $field = $filter['field'];
        list($field, $json, $isJson) = $this->splitFieldFromJson($field, $filter['value']);
        $fieldAlias = $this->getFieldAlias($field);
        $isAssociation = $this->isAssociation($this->metadata, $field);
        $operator = strtolower($filter['operator']);
        $fieldType = 'string';
        if (array_key_exists($field, $this->metadata->fieldMappings)) {
            $fieldType = $this->metadata->fieldMappings[$field]['type'];
        }

        $expression = null;

        if (array_key_exists('value', $filter)) {
            $value = $filter['value'];
            $type = array_key_exists('type', $filter) ? $filter['type'] : null;
            $dateFilterRegex = '/^{{\s*(.*)\((.*)\)\s*}}$/';
            if ($isAssociation && is_string($value) && preg_match('/^\/api\//', $value)) {
                #For fixing error in php7
                #https://stackoverflow.com/questions/4636166/only-variables-should-be-passed-by-reference
                $tmp = explode('/', $value);
                $value = end($tmp);
            } else if (is_string($value) && preg_match($dateFilterRegex, $value)) {
                $value = $this->utilityService->parseDateField($value)->format('Y-m-d H:i:s');
            }

            if (is_numeric($value) && $type == "datetime") {
                $value = $this->convertEpochtoDateTime($value);
            }

            switch ($operator) {
                case AttributeLevelOperators::EQUALS:
                    $value = !is_string($json) && !is_numeric($json) ? json_encode($json) : $value;
                    $expression = $this->getExpression()->eq($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::NOT_EQUALS:
                    $value = !is_string($json) && !is_numeric($json) ? json_encode($json) : $value;
                    $expression = $this->getExpression()->neq($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::LESS_THAN:
                    $expression = $this->getExpression()->lt($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::LESS_THAN_EQUALS:
                    $expression = $this->getExpression()->lte($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::GREATER_THAN:
                    $expression = $this->getExpression()->gt($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::GREATER_THAN_EQUALS:
                    $expression = $this->getExpression()->gte($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::IS_NULL:
                    $expressionBuilder = $this->getExpression();
                    $expression = filter_var($value, FILTER_VALIDATE_BOOLEAN)
                        ? $expressionBuilder->isNull($fieldAlias)
                        : $expressionBuilder->isNotNull($fieldAlias);
                    break;
                case AttributeLevelOperators::IN:
                case AttributeLevelOperators::NIN:
                    if (!is_array($value)) {
                        $value = explode('|', $value);
                    }

                    foreach ($value as $key => $item) {
                        if ($isAssociation && is_string($item) && preg_match('/^\/api\//', $item)) {
                            $item_explode = explode('/', $item);
                            $value[$key] = strtolower(end($item_explode));
                        } else if (is_string($item)) {
                            $value[$key] = strtolower($item);
                        }
                    }
                    $fullFieldAlias = $isAssociation ? $fieldAlias . '.uuid' : $fieldAlias;
                    if ($fieldType === 'string' || $fieldType === 'text') {
                        $functionValue = new Expr\Func('TEXT_LOWER', [$fullFieldAlias]);
                    } else {
                        $functionValue = $fullFieldAlias;
                    }
                    if ($operator == AttributeLevelOperators::NIN) {
                        $notInExpression = $this->getExpression()->notIn($functionValue, $this->getNextParameterName($value));
                        $isNullExpression = $this->getExpression()->isNull($functionValue);
                        $expression = $this->getExpression()->orX($notInExpression, $isNullExpression);
                    } else {
                        $expression = $this->getExpression()->in($functionValue, $this->getNextParameterName($value));
                    }
                    break;
                case AttributeLevelOperators::LIKE:
                    $value = strtolower($value);
                    if ($isJson) {
                        $expression = $this->getExpression()->jsonLike($fieldAlias, $this->getJsonQuery($filter['field']), $this->getNextParameterName($value));
                    } else {
                        $expression = $this->getExpression()->like('TEXT_LOWER(' . $fieldAlias . ')', $this->getNextParameterName($value));
                    }
                    break;
                case AttributeLevelOperators::NOT_LIKE:
                    $value = strtolower($value);
                    # ToDo Add Support for JSON Not like
                    if (!$isJson) {
                        $expression = $this->getExpression()->notLike('TEXT_LOWER(' . $fieldAlias . ')', $this->getNextParameterName($value));
                    }
                    break;
                case AttributeLevelOperators::CONTAINS:
                    $value = json_encode($json);
                    $expression = $this->getExpression()->jsonContains($fieldAlias, $this->getNextParameterName($value));
                    break;
                case AttributeLevelOperators::EXISTS:
                    $expression = $this->getExpression()->jsonExists($fieldAlias, $this->getJsonQuery($filter['field']), $this->getNextParameterName($value));
                    break;
                default:
                    $this->handleUnsupportedOperator($operator);
            }
        }

        return $expression;
    }

    /**
     * @param string $association
     * @param boolean $addJoins
     * @return string
     */
    protected function getAssociationAlias($association, $addJoins = true)
    {
        if (count(explode('.', $association)) < 2) {
            return $association;
        }

        if ($addJoins) {
            $this->addAssociationJoins($association);
        }

        if (!array_key_exists($association, $this->associationAliases)) {
            $this->associationAliases[$association] = 't' . count($this->associationAliases);
        }

        return $this->associationAliases[$association];
    }

    /**
     * @param string $association Example: association.field OR association1.association2.field
     */
    protected function addAssociationJoins($association)
    {
        if (!array_key_exists($association, $this->joins)) {
            $associations = explode('.', $association);
            $parent = current($associations);
            $metadata = $this->metadata;
            foreach ($associations as $index => $partialAssociation) {
                if ($index === 0) {
                    continue;
                }

                $fullAssociation = $this->getPropertyPath($parent, $partialAssociation);

                $parentAlias = $this->getAssociationAlias($parent, false);
                $alias = $this->getAssociationAlias($fullAssociation, false);

                $this->joins[$fullAssociation] = [
                    'join' => $this->getPropertyPath($parentAlias, $partialAssociation),
                    'alias' => $alias,
                    'type' => $metadata->associationMappings[$partialAssociation]['type'],
                ];

                // prep for next iteration
                $metadata = $this->getMetadataForClass($metadata->associationMappings[$partialAssociation]['targetEntity']);
                $parent = $fullAssociation;
            }
        }
    }

    /**
     * @param $value
     * @return string
     */
    protected function getNextParameterName($value)
    {
        $parameterName = ':param' . $this->parameterCount++;
        $this->parameters[$parameterName] = $value;
        return $parameterName;
    }

    /**
     * @param string $fieldPath
     * @return null|string
     */
    protected function getFieldAlias($fieldPath, $operation = null)
    {
        if (!$fieldPath) {
            return null;
        }

        if ($fieldPath === '*') {
            return $this->fromAlias;
        }

        if (in_array($fieldPath, $this->aliases)) {
            return $fieldPath;
        }

        $fieldPath = $this->getPropertyPath($this->fromAlias, $fieldPath);
        $aliasParts = explode('.', $fieldPath);

        if (count($aliasParts) > 2) {
            $associationParts = array_slice($aliasParts, 0, count($aliasParts) - 1);
            $field = end($aliasParts);
            $associationPath = call_user_func_array([$this, 'getPropertyPath'], $associationParts);
            $associationAlias = $this->getAssociationAlias($associationPath);
            $fieldAlias = $this->getPropertyPath($associationAlias, $field);
        } else {
            $field = end($aliasParts);
            if ($this->metadata->hasAssociation($field)) {
                $associationAlias = $this->getPropertyPath($this->fromAlias, $field);
                $fieldAlias = $this->getAssociationAlias($associationAlias);
            } else {
                $fieldAlias = $this->getPropertyPath($this->fromAlias, $field);
            }
        }
        //Computation Queries
        $fieldAlias = $this->checkForComputeQuery($fieldAlias, $operation);
        return $fieldAlias;
    }

    protected function checkForComputeQuery($fieldAlias, $fieldOperator = null)
    {
        $roiFieldExpr = explode(',', $fieldAlias);
        if (count($roiFieldExpr) > 1) {
            $op = '-';
            if ($fieldOperator == 'sum') {
                $op = '+';
            }
            $fieldAlias = str_replace(",", $op . 'o.', $fieldAlias);
        }
        return $fieldAlias;
    }

    /**
     * @return string
     */
    protected function getPropertyPath()
    {
        $propertyPathParts = func_get_args();
        return implode('.', $propertyPathParts);
    }

    /**
     * @return JsonExpressionBuilder
     */
    protected function getExpression()
    {
        return new JsonExpressionBuilder();
    }

    /**
     * @param $filter
     */
    protected function convertFilterFieldToDotNotation(&$filter)
    {
        $filter['field'] = preg_replace(sprintf('/%s/', AttributeLevelOperators::PROPERTY_OPERATOR), '.', $filter['field']);
    }

    /**
     * @param ClassMetadata $metadata
     * @param string $field
     * @return bool
     */
    private function isField(ClassMetadata $metadata, $field)
    {
        if (!($metadata->hasField($field) || $metadata->hasAssociation($field))) {
            throw new HttpException(Response::HTTP_BAD_REQUEST, 'Invalid query/body parameters [Field:' . $field . ' does not exist in module definition]');
        }
        return true;
    }
}
