<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Doctrine\ORM;


use App\Service\MetadataHelper;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\Persistence\ManagerRegistry;
use App\Constants\AttributeLevelOperators;
use App\Filter\AbstractFilterQueryBuilder;
use App\Entity\ModelMetadata\ModelMetadata;
use ApiPlatform\Core\Api\IriConverterInterface;
use App\Entity\ModelMetadata\AttribModelMetadata;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\PropertyAccess\PropertyAccessorInterface;

class SearchQueryBuilder extends AbstractFilterQueryBuilder
{

    const IGNORE_FORM_TYPE = ['manyToMany', 'oneToMany', 'oneToOne', 'lookup', 'picklist','date', 'datetime', 'decimal', 'checkbox', 'tags', 'multiselectpicklist','object', null];
    /** @var ClassMetadataFactory */
    protected $classMetadataFactory;
    /** @var IriConverterInterface */
    private $iriConverter;
    /** @var PropertyAccessorInterface */
    private $propertyAccessor;
    private $metadataHelper;
    private $entityManager;
    private $requestStack;

    /**
     * @param ManagerRegistry $managerRegistry
     * @param IriConverterInterface $iriConverter
     * @param PropertyAccessorInterface $propertyAccessor
     * @param null|array $properties Null to allow filtering on all properties with the exact strategy
     *                                              or a map of property name with strategy.
     */
    public function __construct(
        IriConverterInterface $iriConverter,
        PropertyAccessorInterface $propertyAccessor,
        MetadataHelper $metadataHelper,
        EntityManagerInterface $entityManager,
        RequestStack $requestStack
    ) {
        $this->iriConverter = $iriConverter;
        $this->propertyAccessor = $propertyAccessor;
        $this->metadataHelper = $metadataHelper;
        $this->entityManager = $entityManager;
        $this->requestStack = $requestStack;
    }

    public function fromQueryObject($query, $queryBuilder, $resourceClass)
    {
        $searchTerm = $this->extractProperties($query);
        if (empty($searchTerm)) {
            return;
        }
        $searchableFields = $this->getSearchableFields($resourceClass);
        if (empty($searchableFields)) {
            //Invalidate the query to ensure no results
            $queryBuilder->andWhere($queryBuilder->expr()->eq('333', '443'));
        }
        $searchTerm = '%' . strtolower($searchTerm) . '%';
        $searchExpressions = [];

        $doctrineClassMetadata = $queryBuilder->getEntityManager()->getClassMetadata($resourceClass);

        foreach ($searchableFields as $count => $field) {
            $searchParam = ':searchParam' . $count;
            $queryBuilder->setParameter($searchParam, $searchTerm);

            if ($doctrineClassMetadata->hasAssociation($field)) {
                continue;
            }

            $searchExpressions[] = $queryBuilder->expr()->like('TEXT_LOWER(o.' . $field . ')', $searchParam);
        }

        $searchExpression = call_user_func_array([$queryBuilder->expr(), 'orX'], $searchExpressions);
        $queryBuilder->andWhere($searchExpression);
    }

    /**
     * {@inheritdoc}
     */
    protected function extractProperties()
    {
        $request = $this->requestStack->getCurrentRequest();
        if (array_key_exists(AttributeLevelOperators::SEARCH_TERM, $_GET)) {
            return $request->query->get(AttributeLevelOperators::SEARCH_TERM);
        }
    }

    /**
     * @param ResourceInterface $resource
     * @return array
     */
    protected function getSearchableFields($resourceClass)
    {
        $type = $this->metadataHelper->getTypeFromClassName($resourceClass);
        $entityRepo = $this->entityManager->getRepository(ModelMetadata::class);
        $modelData = $entityRepo->findOneBy(['type' => $type]);
        $fields = [];
        if (!is_null($modelData)) {
            $attribEntityRepo = $this->entityManager->getRepository(AttribModelMetadata::class);
            $attribFields = $attribEntityRepo->findBy(['searchable' => true, 'sattrib' => $modelData->getUuid()]);
            foreach ($attribFields as $attribField) {
                if (!(in_array($attribField->getFormType(), self::IGNORE_FORM_TYPE))) {
                    $fields[] = $attribField->getName();
                }
            }
        } else {
            $systemFixturesFile = file_get_contents('/opt/cyops-api/src/StaticFile/SystemFixtures.json');
            $systemFixture = json_decode($systemFixturesFile);
            $stagingData = $systemFixture->$type;
            $attributes = $stagingData->attributes;
            foreach ($attributes as $attribute) {
                if (($attribute->searchable = true) && !(in_array($attribute->formType, self::IGNORE_FORM_TYPE))) {
                    $fields[] = $attribute->name;
                }
            }
        }
        return $fields;
    }
}
