<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Encoder\Hmac;

use Psr\Log\LoggerInterface;
use App\Constants\CacheConstants;
use App\Providers\SecretKeyProviderInterface;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class DasHmacSecretKeyProvider implements SecretKeyProviderInterface
{
    protected $containers;

    /** @var HmacClient */
    protected $dasClient;

    private $cache;

    /** @var bool */
    protected $isInitialized = false;

    public function __construct(ContainerInterface $containers, TagAwareCacheInterface $cache, protected LoggerInterface $logger)
    {
        $this->containers = $containers;
        $this->cache = $cache;
        $this->dasClient = $this->containers->get("app.proxy.client.das");
        $this->isInitialized = false;
    }

    const UUID = 'uuid';
    const SECRET_KEY = 'secretKey';
    const DAS_SECRET_KEY = 'pkey';
    const DAS_KEY_IDENTIFIER = 'pubkey';

    /** @var string */
    protected $uuid;

    /** @var string */
    protected $secretKey;

    /**
     * @param string $keyIdentifier
     * @return string
     * @throws \Exception
     */
    public function getSecretKey($keyIdentifier)
    {
        if (!$this->isInitialized) {
            $this->initializeCredentials($keyIdentifier);
        }

        return $this->secretKey;
    }

    /**
     * @return string
     * @throws \Exception
     */
    public function getUserId()
    {
        if (!$this->isInitialized) {
            throw new \Exception('Cannot load API user\'s UUID without having first retrieved the secret key.');
        }
        return $this->uuid;
    }

    /**
     * @param string $keyIdentifier
     * @return bool
     * @throws \Exception
     */
    protected function initializeCredentials($keyIdentifier)
    {
        // Only necessary once
        if ($this->isInitialized) {
            return true;
        }

        // Check cache first
        // and $this->cache->cacheContains($keyIdentifier)
        $cacheKeyName = base64_encode($keyIdentifier);
        $item = $this->cache->getItem($cacheKeyName);
        if (!$item->isHit()) {
            $publicKeyRequest = [
                self::DAS_KEY_IDENTIFIER => $keyIdentifier,
            ];
            $apikeysUri = sprintf('%s/%s/', $this->dasClient->getTargetUri(), 'apikeys');
            $encodedPublicKeyBody = json_encode($publicKeyRequest);
            $request = new GuzzleRequest('PUT', $apikeysUri, [], $encodedPublicKeyBody);

            try {
                $response = $this->dasClient->send($request);
                $responseBody = json_decode($response->getBody(), true);
                $this->secretKey = $responseBody[self::DAS_SECRET_KEY];
                $this->uuid = $responseBody[self::UUID];
                $credentials = [
                    self::UUID => $this->uuid,
                    self::SECRET_KEY => $this->secretKey,
                ];
                $item->set($credentials);
                $item->expiresAfter(600);
                $item->tag(CacheConstants::DAS_HMAC_SECRET_KEY_PROVIDER);
                $this->cache->save($item);
                $this->isInitialized = true;
            } catch (BadRequestHttpException | \Exception $clientException) {
                $this->logger->error($clientException->getMessage());
                throw new \Exception('Unable to load API credentials from cache or DAS.', 0, $clientException);
            }
        }
        $cachedCredentials = $item->get();
        $this->secretKey = $cachedCredentials[self::SECRET_KEY];
        $this->uuid = $cachedCredentials[self::UUID];
        $this->isInitialized = true;
        return true;
    }
}
