<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Entity\Actors;

use ApiPlatform\Core\Annotation\ApiResource;
use App\Entity\Authorization\Role;
use App\Entity\Authorization\Team;
use App\Entity\Base\BaseEntity;
use App\Repository\ActorRepository;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\Common\Collections\Collection;
use Doctrine\ORM\Mapping as ORM;
use Symfony\Component\Security\Core\User\UserInterface;
use Symfony\Component\Serializer\Annotation\Groups;

/**
 *  @ApiResource(
 *     attributes={"validation_groups"={"fsr_primary", "fsr_all"}, "fsr_type"="actors", "fsr_module"="people", "fsr_singular"="actors"},
 *     denormalizationContext={"groups"={"fsr_all", "password"}, "fetch_data"=false},
 *     normalizationContext={"max_depth"=2},
 *     collectionOperations={
 *          "get" = { "security_post_denormalize" = "is_granted('read.people', object)" },
 *          "post" = { "security_post_denormalize" = "is_granted('create.people', object)" }
 *     },
 *     itemOperations={
 *          "get" = { "security" = "is_granted('read.people', object)" },
 *          "put" = { "security" = "is_granted('update.people', object)" },
 *          "delete" = { "security" = "is_granted('delete.people', object)" }
 *     }
 * )
 * @ORM\Entity(repositoryClass=ActorRepository::class)
 * @ORM\InheritanceType("SINGLE_TABLE")
 * @ORM\DiscriminatorColumn(name="record_type", type="string", length=20)
 * @ORM\Table(name="actors")
 */
class Actor extends BaseEntity implements UserInterface
{

    /**
     * @Groups({"fsr_all", "roles"})
     * @ORM\ManyToMany(targetEntity=Role::class, cascade={"persist"})
     * @Doctrine\ORM\Mapping\JoinTable(name="actor_role",
     *   joinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="actor_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   },
     *   inverseJoinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="role_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   }
     * )
     */
    private $roles;

    /**
     * @Groups({"fsr_all", "teams"})
     * @ORM\ManyToMany(targetEntity=Team::class, mappedBy="actors", cascade={"persist"})
     */
    private $teams;

    /**
     * @Groups({"fsr_primary", "fsr_all", "userId"})
     * @Doctrine\ORM\Mapping\Column(name="user_id", type="string", length=36, unique=true)
     */
    private $userId;

    /**
     * @Groups({"fsr_primary", "fsr_all", "userType"})
     * @ORM\Column(name="user_type", type="string", length=36, nullable=true, unique=false)
     */
    protected $userType;

    /**
     * @var string
     * @Groups({"fsr_primary", "fsr_all", "avatar"})
     * @Doctrine\ORM\Mapping\Column(name="avatar", type="string", length=255, nullable=true, unique=true)
     */
    protected $avatar;

    public function __construct()
    {
        $this->roles = new ArrayCollection();
        $this->teams = new ArrayCollection();
    }

    /**
     * @return Collection|Role[]
     */
    public function getRoles($asArray = true, $isUuid = false)
    {
        if ($isUuid) {
            $rolesUuids = [];
            foreach ($this->roles->toArray() as $value) {
                $rolesUuids[] =  $value->getUuid();
            }
            return $rolesUuids;
        } else {
            return $asArray ? $this->roles->toArray() : $this->roles;
        }
    }

    public function addRole(Role $role): self
    {
        if (!$this->roles->contains($role)) {
            $this->roles[] = $role;
        }

        return $this;
    }

    public function removeRole(Role $role): self
    {
        if ($this->roles->contains($role)) {
            $this->roles->removeElement($role);
        }

        return $this;
    }

    /**
     * @return Collection|Team[]
     */
    public function getTeams(): Collection
    {
        return $this->teams;
    }

    public function addTeam(Team $team): self
    {
        if (!$this->teams->contains($team)) {
            $this->teams[] = $team;
            $team->addActor($this);
        }

        return $this;
    }

    public function removeTeam(Team $team): self
    {
        if ($this->teams->contains($team)) {
            $this->teams->removeElement($team);
            $team->removeActor($this);
        }

        return $this;
    }

    public function getUserId(): ?string
    {
        return $this->userId;
    }

    public function setUserId(string $userId): self
    {
        $this->userId = $userId;

        return $this;
    }


    /**
     * A visual identifier that represents this user.
     *
     * @see UserInterface
     */
    public function getUsername(): string
    {
        return (string) $this->userId;
    }

    /**
     * A visual identifier that represents this user.
     *
     * @see UserInterface
     */
    public function getUserIdentifier(): string
    {
        return (string) $this->userId;
    }

    /**
     * @see UserInterface
     */
    public function getPassword()
    {
        // not needed for apps that do not check user passwords
    }

    /**
     * @see UserInterface
     */
    public function getSalt()
    {
        // not needed for apps that do not check user passwords
    }

    /**
     * @see UserInterface
     */
    public function eraseCredentials()
    {
        // If you store any temporary, sensitive data on the user, clear it here
        // $this->plainPassword = null;
    }

    /**
     * @param string $userType
     */
    public function setUserType($userType)
    {
        $this->userType = $userType;
    }

    /**
     * @return string
     */
    public function getUserType()
    {
        return $this->userType;
    }

    /**
     * Get the value of avatar
     *
     */
    public function getAvatar()
    {
        return $this->avatar;
    }

    /**
     * Set the value of avatar
     *
     * @return  self
     */
    public function setAvatar($avatar)
    {
        $this->avatar = $avatar;

        return $this;
    }
}
