<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Entity\Authorization;

use ApiPlatform\Core\Annotation\ApiResource;
use App\Entity\Actors\Actor;
use App\Entity\Base\BaseEntity;
use App\Repository\TeamRepository;
use App\Traits\Importable;
use App\Validator\NonIntersectingCollectionFieldsConstraint;
use App\Validator\NotSelfReferentialConstraint;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\Common\Collections\Collection;
use Symfony\Component\Serializer\Annotation\Groups;
use Doctrine\ORM\Mapping as ORM;

/**
 *  @ApiResource(
 *     attributes={"validation_groups"={"fsr_primary", "fsr_all"}, "fsr_type"="teams", "fsr_module"="security", "fsr_singular"="Team"},
 *     denormalizationContext={"groups"={"fsr_all", "password"}, "fetch_data"=false},
 *     collectionOperations={
 *          "get" = { "security_post_denormalize" = "is_granted('read.security', object)" },
 *          "post" = { "security_post_denormalize" = "is_granted('create.security', object)" }
 *     },
 *     itemOperations={
 *          "get" = { "security" = "is_granted('read.security', object)" },
 *          "put" = { "security" = "is_granted('update.security', object)" },
 *          "delete" = { "security" = "is_granted('delete.security', object)" }
 *     }
 * )
 * @Doctrine\ORM\Mapping\Entity(repositoryClass=TeamRepository::class)
 * @Doctrine\ORM\Mapping\Table(name="teams")
 * @Doctrine\ORM\Mapping\EntityListeners({"App\EventSubscriber\TeamListener"})
 * @NotSelfReferentialConstraint(groups={"fsr_all"})
 * @NonIntersectingCollectionFieldsConstraint(groups={"fsr_all"})
 */
class Team extends BaseEntity
{
    use Importable;

    /**
     * @Groups({"fsr_primary", "fsr_all", "description"})
     * @Doctrine\ORM\Mapping\Column(type="text", nullable=true)
     */
    protected $description;

    /**
     * @Groups({"fsr_primary", "fsr_all", "name"})
     * @Doctrine\ORM\Mapping\Column(type="string", length=255, nullable=false, unique=true)
     */
    protected $name;

    /**
     * @Doctrine\ORM\Mapping\ManyToMany(targetEntity=Actor::class, inversedBy="teams")
     * @Doctrine\ORM\Mapping\JoinTable(name="team_actor",
     *   joinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="team_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   },
     *   inverseJoinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="actor_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   }
     * )
     * @Groups({"fsr_all", "actors"})
     */
    protected $actors;

    /**
     * @var \Doctrine\Common\Collections\Collection
     * @Doctrine\ORM\Mapping\ManyToMany(targetEntity="App\Entity\Authorization\Team", mappedBy="children", cascade={"persist"})
     * @Groups({"fsr_all", "parents"})
     */
    protected $parents;

    /**
     * @var \Doctrine\Common\Collections\Collection
     *
     * @Doctrine\ORM\Mapping\ManyToMany(targetEntity="App\Entity\Authorization\Team", inversedBy="siblings", cascade={"persist"})
     * @Doctrine\ORM\Mapping\JoinTable(name="team_sibling",
     *   joinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="sibling_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   },
     *   inverseJoinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="team_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   }
     * )
     * @Groups({"fsr_all", "siblings"})
     */
    protected $siblings;

    /**
     * @var \Doctrine\Common\Collections\Collection
     *
     * @Doctrine\ORM\Mapping\ManyToMany(targetEntity="App\Entity\Authorization\Team", inversedBy="parents", cascade={"persist"})
     * @Doctrine\ORM\Mapping\JoinTable(name="team_child",
     *   joinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="child_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   },
     *   inverseJoinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="team_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   }
     * )
     * @Groups({"fsr_all", "children"})
     */
    protected $children;

    public function __construct()
    {
        $this->actors = new ArrayCollection();
        $this->parents = new ArrayCollection();
        $this->children = new ArrayCollection();
        $this->siblings = new ArrayCollection();
    }

    public function getDescription(): ?string
    {
        return $this->description;
    }

    public function setDescription(?string $description): self
    {
        $this->description = $description;

        return $this;
    }

    public function getName(): ?string
    {
        return $this->name;
    }

    public function setName(?string $name): self
    {
        $this->name = $name;

        return $this;
    }

    /**
     * @return Collection|Actor[]
     */
    public function getActors(): Collection
    {
        return $this->actors;
    }

    public function addActor(Actor $actor): self
    {
        if (!$this->actors->contains($actor)) {
            $this->actors[] = $actor;
        }

        return $this;
    }

    public function removeActor(Actor $actor): self
    {
        if ($this->actors->contains($actor)) {
            $this->actors->removeElement($actor);
        }
        return $this;
    }

    /**
     * Add parent
     *
     * @param Team $parent
     * @return Team
     */
    public function addParent(Team $parent)
    {
        if (!$this->parents->contains($parent)) {
            $this->parents[] = $parent;
        }

        if (!in_array($this, $parent->getChildren()->toArray())) {
            $parent->addChild($this);
        }

        return $this;
    }

    /**
     * Remove parent
     * @param Team $parent
     */
    public function removeParent(Team $parent)
    {
        $this->parents->removeElement($parent);

        if (in_array($this, $parent->getChildren()->toArray())) {
            $parent->removeChild($this);
        }
    }

    /**
     * Get parents
     * @return \Doctrine\Common\Collections\Collection
     */
    public function getParents()
    {
        return $this->parents;
    }

    /**
     * Add sibling
     *
     * @param Team $sibling
     * @return Team
     */
    public function addSibling(Team $sibling)
    {
        if (!$this->siblings->contains($sibling)) {
            $this->siblings[] = $sibling;
        }

        if (!in_array($this, $sibling->getSiblings()->toArray())) {
            $sibling->addSibling($this);
        }

        return $this;
    }

    /**
     * Remove sibling
     * @param Team $sibling
     */
    public function removeSibling(Team $sibling)
    {
        $this->siblings->removeElement($sibling);

        if (in_array($this, $sibling->getSiblings()->toArray())) {
            $sibling->removeSibling($this);
        }
    }

    /**
     * Get siblings
     * @return \Doctrine\Common\Collections\Collection
     */
    public function getSiblings()
    {
        return $this->siblings;
    }

    /**
     * Add child
     * @param Team $child
     * @return Team
     */
    public function addChild(Team $child)
    {
        if (!$this->children->contains($child)) {
            $this->children[] = $child;
        }

        if (!in_array($this, $child->getParents()->toArray())) {
            $child->addParent($this);
        }

        return $this;
    }

    /**
     * Remove child
     * @param Team $child
     */
    public function removeChild(Team $child)
    {
        $this->children->removeElement($child);

        if (in_array($this, $child->getParents()->toArray())) {
            $child->removeParent($this);
        }
    }

    /**
     * Get children
     * @return \Doctrine\Common\Collections\Collection
     */
    public function getChildren()
    {
        return $this->children;
    }
}
