<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Entity\Core;

use App\Entity\Core\ExportJob;
use App\Entity\Core\ImportJob;
use App\Entity\Base\BaseEntity;
use Doctrine\ORM\Mapping as ORM;
use App\Entity\Core\ExportTemplate;
use Doctrine\Common\Collections\Collection;
use ApiPlatform\Core\Annotation\ApiResource;
use Doctrine\Common\Collections\ArrayCollection;
use Symfony\Component\Serializer\Annotation\Groups;
use Symfony\Component\Validator\Constraints\Length;
use Symfony\Component\Validator\Constraints\NotBlank;

/**
 *  @ApiResource(
 *     attributes={"validation_groups"={"fsr_primary", "fsr_all"}, "fsr_type"="solutionpacks", "fsr_module"="solutionpacks", "fsr_singular"="Solution Pack"},
 *     denormalizationContext={"groups"={"fsr_all", "password"}, "fetch_data"=false},
 *     normalizationContext={"circular_reference_limit"=2},
 *     collectionOperations={
 *          "get" = { "path"="solutionpacks", "security_post_denormalize" = "is_granted('read.solutionpacks', object)" },
 *          "post" = { "path"="solutionpacks", "security_post_denormalize" = "is_granted('create.solutionpacks', object)" }
 *     },
 *     itemOperations={
 *          "get" = { "path"="solutionpacks/{uuid}", "security" = "is_granted('read.solutionpacks', object)" },
 *          "put" = { "path"="solutionpacks/{uuid}", "security" = "is_granted('update.solutionpacks', object)" }
 *     }
 * )
 * 
 * @ORM\EntityListeners({"App\EventSubscriber\SolutionPackListener"})
 * @Doctrine\ORM\Mapping\Table(name="solutionpacks", uniqueConstraints={@Doctrine\ORM\Mapping\UniqueConstraint(name="solutionpack_unique", columns={"name", "version", "type"})})
 * @Doctrine\ORM\Mapping\Entity()
 */
class SolutionPack extends BaseEntity
{
    use \App\Traits\Trackable;
    use \App\Traits\Taggable;
    use \App\Traits\Importable;

    /**
     * @var string|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`name`", type="string", length=200, nullable=false, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "name"})
     * @Symfony\Component\Validator\Constraints\NotBlank(groups={"Default", "fsr_primary", "fsr_all"})
     * @Symfony\Component\Validator\Constraints\Length(groups={"Default", "fsr_primary", "fsr_all"}, max=200, maxMessage="name cannot be longer than 200 characters")
     */
    protected $name;

    /**
     * @var string|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`version`", type="string", length=30, nullable=false, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "version"})
     * @Symfony\Component\Validator\Constraints\NotBlank(groups={"Default", "fsr_primary", "fsr_all"})
     * @Symfony\Component\Validator\Constraints\Length(groups={"Default", "fsr_primary", "fsr_all"}, max=30, maxMessage="version cannot be longer than 30 characters")
     */
    protected $version;

    /**
     * @var string|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`type`", type="string", length=100, nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "type"})
     */
    protected $type;

    /**
     * @var bool|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`installed`", type="boolean", nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "installed"})
     */
    protected $installed = false;

    /**
     * @var mixed|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`latestavailableversion`", type="text", length=10, nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "latestAvailableVersion"})
     */
    protected $latestAvailableVersion;

    /**
     * @var mixed|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`fsrmincompatibility`", type="text", length=10, nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "fsrMinCompatibility"})
     */
    protected $fsrMinCompatibility;

    /**
     * @var mixed|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`latestcompatibleversion`", type="text", length=10, nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "latestCompatibleVersion"})
     */
    protected $latestCompatibleVersion;

    /**
     * @var mixed|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`dependencies`", type="jsonb", length=10485761, nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "dependencies"})
     */
    protected $dependencies;

    /**
     * @var mixed|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`featuredtags`", type="jsonb", nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "featuredTags"})
     */
    protected $featuredTags;

    /**
     * @var bool|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`featured`", type="boolean", nullable=true, unique=false, options={"default" = false})
     * @Groups({"fsr_primary", "fsr_all", "featured"})
     */
    protected $featured = false;

    /**
     * @var bool|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`development`", type="boolean", nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "development"})
     */
    protected $development = false;

    /**
     * @var bool|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`draft`", type="boolean", nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "draft"})
     */
    protected $draft = false;

    /**
     * @var bool|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`local`", type="boolean", nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "local"})
     */
    protected $local = true;

    /**
     * @var string|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`recordid`", type="string", length=100, nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "recordId"})
     */
    protected $recordId = null;

    /**
     * @var \DateTime|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`publisheddate`", type="datetime", length=50, nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "publishedDate"})
     */
    protected $publishedDate;

    /**
     * @var string|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`publisher`", type="string", length=100, nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "publisher"})
     */
    protected $publisher;

    /**
     * @var bool|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`certified`", type="boolean", nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "certified"})
     */
    protected $certified = false;

    /**
     * @Groups({"fsr_primary", "fsr_all", "options"})
     * @ORM\Column(type="jsonb", nullable=true)
     */
    private $options = [];

    /**
     * @Groups({"fsr_primary", "fsr_all", "draftoptions"})
     * @ORM\Column(type="jsonb", nullable=true)
     */
    private $draftOptions = [];

    /**
     * @Groups({"fsr_primary", "fsr_all", "installedoptions"})
     * @ORM\Column(type="jsonb", nullable=true)
     */
    private $installedOptions = [];

    /**
     * @Groups({"fsr_primary", "fsr_all", "infocontent"})
     * @ORM\Column(type="jsonb", nullable=true)
     */
    private $infoContent = [];

    /**
     * @var string|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`prerequisite`", type="text", length=10485761, nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "prerequisite"})
     */
    private $prerequisite;

    /**
     * @Groups({"fsr_primary", "fsr_all", "sourceControlStats"})
     * @ORM\Column(type="jsonb", nullable=true)
     */
    private $sourceControlStats = [];

    /**
     * @var string|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`infopath`", type="string", length=500, nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "infoPath"})
     */
    protected $infoPath;

    /**
     * @var string|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`iconlarge`", type="text", nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "iconLarge"})
     */

    protected $iconLarge;

    /**
     * @var string|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`description`", type="text", length=10485761, nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "description"})
     */

    protected $description = "";
    /**
     * @var string|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`label`", type="string", length=200, nullable=false, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "label"})
     * @Symfony\Component\Validator\Constraints\NotBlank(groups={"Default", "fsr_primary", "fsr_all"})
     * @Symfony\Component\Validator\Constraints\Length(groups={"Default", "fsr_primary", "fsr_all"}, max=100, maxMessage="label cannot be longer than 200 characters")
     */
    protected $label;

    /**
     * @var string|null
     *
     * @Doctrine\ORM\Mapping\Column(name="`status`", type="string", length=50, nullable=true, unique=false)
     * @Groups({"fsr_primary", "fsr_all", "status"})
     */
    protected $status;

    /**
     * @Groups({"fsr_primary", "fsr_all", "file"})
     * @ORM\Column(type="text", nullable=true)
     */
    private $file;


    /**
     * @Groups({"fsr_all","fsr_primary","template"})
     * @ORM\OneToOne(targetEntity="App\Entity\Core\ExportTemplate",cascade={"persist","remove"}, inversedBy="solutionPack")
     * @ORM\JoinColumn(name="template", referencedColumnName="uuid")
     */
    private $template;


    /**
     * @Groups({"fsr_all","fsr_primary", "exportJob"})
     * @ORM\OneToOne(targetEntity="App\Entity\Core\ExportJob", inversedBy="solutionPack", cascade={"persist", "remove"})
     * @ORM\JoinColumn(name="exportJob", referencedColumnName="uuid")
     */
    private $exportJob;

    /**
     * @Groups({"fsr_all","fsr_primary", "importJob"})
     * @ORM\OneToOne(targetEntity="App\Entity\Core\ImportJob", cascade={"persist","remove"},inversedBy="solutionPack")
     * @ORM\JoinColumn(name="importJob", referencedColumnName="uuid")
     */
    private $importJob;

    /**
     * @var \Doctrine\Common\Collections\Collection
     *
     * @Doctrine\ORM\Mapping\ManyToMany(targetEntity="App\Entity\Core\Picklist", cascade={"persist"}, fetch="EXTRA_LAZY")
     * @Doctrine\ORM\Mapping\JoinTable(name="solutionpack_picklist_solution_pack_category",
     *   joinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="solutionpack_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   },
     *   inverseJoinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="picklist_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   }
     * )
     * @Groups({ "fsr_all","fsr_primary", "category"})
     */
    protected $category;

    /**
     * @Groups({"fsr_all", "fsr_primary", "configCount"})
     * @ORM\Column(type="integer", nullable=true)
     */
    private $configCount = 0;

    /**
     * @Groups({"fsr_all", "fsr_primary", "buildNumber"})
     * @ORM\Column(type="integer", nullable=true)
     */
    private $buildNumber = 0;

    public function __construct()
    {
        $this->category = new \Doctrine\Common\Collections\ArrayCollection();
    }

    /**
     * Set name
     *
     * @param string  $name
     *
     * @return SolutionPack
     */
    public function setName($name = null)
    {
        $this->name = $name;

        return $this;
    }

    /**
     * Get name
     *
     * @return string 
     */
    public function getName()
    {
        return $this->name;
    }

    /**
     * Set version
     *
     * @param string  $version
     *
     * @return SolutionPack
     */
    public function setVersion($version = null)
    {
        $this->version = $version;

        return $this;
    }

    /**
     * Get version
     *
     * @return string 
     */
    public function getVersion()
    {
        return $this->version;
    }

    /**
     * Set type
     *
     * @param string  $type
     *
     * @return SolutionPack
     */
    public function setType($type = null)
    {
        $this->type = $type;

        return $this;
    }

    /**
     * Get type
     *
     * @return string
     */
    public function getType()
    {
        return $this->type;
    }

    /**
     * Set installed
     *
     * @param bool  $installed
     *
     * @return SolutionPack
     */
    public function setInstalled($installed = null)
    {
        $this->installed = $installed;

        return $this;
    }

    /**
     * Get installed
     *
     *  @return bool
     */
    public function getInstalled()
    {
        return $this->installed;
    }

    /**
     * Set latestAvailableVersion
     *
     * @param mixed  $latestAvailableVersion
     *
     * @return SolutionPack
     */
    public function setLatestAvailableVersion($latestAvailableVersion = null)
    {
        $this->latestAvailableVersion = $latestAvailableVersion;

        return $this;
    }

    /**
     * Get latestAvailableVersion
     *
     * @return mixed
     */
    public function getLatestAvailableVersion()
    {
        return $this->latestAvailableVersion;
    }

    /**
     * Set fsrMinCompatibility
     *
     * @param mixed  $fsrMinCompatibility
     *
     * @return SolutionPack
     */
    public function setFsrMinCompatibility($fsrMinCompatibility = null)
    {
        $this->fsrMinCompatibility = $fsrMinCompatibility;

        return $this;
    }

    /**
     * Get fsrMinCompatibility
     *
     * @return mixed
     */
    public function getFsrMinCompatibility()
    {
        return $this->fsrMinCompatibility;
    }



    /**
     * Set development
     *
     * @param bool  $development
     *
     * @return SolutionPack
     */
    public function setDevelopment($development = null)
    {
        $this->development = $development;

        return $this;
    }

    /**
     * Get development
     *
     * @return bool
     */
    public function getDevelopment()
    {
        return $this->development;
    }

    /**
     * Set featured
     *
     * @param bool  $featured
     *
     * @return SolutionPack
     */
    public function setFeatured($featured = null)
    {
        $this->featured = $featured;

        return $this;
    }

    /**
     * Get featured
     *
     * @return bool
     */
    public function getFeatured()
    {
        return $this->featured;
    }

    /**
     * Set draft
     *
     * @param bool  $draft
     *
     * @return SolutionPack
     */
    public function setDraft($draft = null)
    {
        $this->draft = $draft;

        return $this;
    }

    /**
     * Get draft
     *
     * @return bool
     */
    public function getDraft()
    {
        return $this->draft;
    }
    /**
     * Set publishedDate
     *
     * @param \DateTime  $publishedDate
     *
     * @return SolutionPack
     */
    public function setPublishedDate($publishedDate = null)
    {
        $this->publishedDate = $publishedDate;

        return $this;
    }

    /**
     * Get publishedDate
     *
     * @return \DateTime
     */
    public function getPublishedDate()
    {
        return $this->publishedDate;
    }

    /**
     * Set publisher
     *
     * @param string  $publisher
     *
     * @return SolutionPack
     */
    public function setPublisher($publisher = null)
    {
        $this->publisher = $publisher;

        return $this;
    }

    /**
     * Get publisher
     *
     * @return string
     */
    public function getPublisher()
    {
        return $this->publisher;
    }

    /**
     * Set certified
     *
     * @param bool  $certified
     *
     * @return SolutionPack
     */
    public function setCertified($certified = null)
    {
        $this->certified = $certified;

        return $this;
    }

    /**
     * Get certified
     *
     * @return bool
     */
    public function getCertified()
    {
        return $this->certified;
    }

    public function getOptions(): ?array
    {
        return $this->options;
    }

    public function setOptions(?array $options): self
    {
        $this->options = $options;

        return $this;
    }

    /**
     * Set infoPath
     *
     * @param string  $infoPath
     *
     * @return SolutionPack
     */
    public function setInfoPath($infoPath = null)
    {
        $this->infoPath = $infoPath;

        return $this;
    }

    /**
     * Get publisher
     *
     * @return string
     */
    public function getInfoPath()
    {
        return $this->infoPath;
    }

    /**
     * Get the value of iconLarge
     *
     * @return  string|null
     */
    public function getIconLarge()
    {
        return $this->iconLarge;
    }

    /**
     * Set the value of iconLarge
     *
     * @param  string|null  $iconLarge
     *
     * @return  self
     */
    public function setIconLarge($iconLarge)
    {
        $this->iconLarge = $iconLarge;

        return $this;
    }

    /**
     * Get the value of prerequisite
     *
     * @return  string|null
     */
    public function getPrerequisite()
    {
        return $this->prerequisite;
    }

    /**
     * Set the value of prerequisite
     *
     * @param  string|null  $prerequisite
     *
     * @return  self
     */
    public function setPrerequisite($prerequisite)
    {
        $this->prerequisite = $prerequisite;

        return $this;
    }

    /**
     * Get the value of label
     *
     * @return  string|null
     */
    public function getLabel()
    {
        return $this->label;
    }

    /**
     * Set the value of label
     *
     * @param  string|null  $label
     *
     * @return  self
     */
    public function setLabel($label)
    {
        $this->label = $label;

        return $this;
    }

    /**
     * Get the value of dependencies
     *
     * @return  mixed|null
     */
    public function getDependencies()
    {
        return $this->dependencies;
    }

    /**
     * Set the value of dependencies
     *
     * @param  mixed|null  $dependencies
     *
     * @return  self
     */
    public function setDependencies($dependencies = null)
    {
        $this->dependencies = $dependencies;

        return $this;
    }

    /**
     * Get the value of status
     *
     * @return  string|null
     */
    public function getStatus()
    {
        return $this->status;
    }

    /**
     * Set the value of status
     *
     * @param  string|null  $status
     *
     * @return  self
     */
    public function setStatus($status)
    {
        $this->status = $status;

        return $this;
    }

    public function getFile(): ?string
    {
        return $this->file;
    }

    public function setFile(?string $file): self
    {
        $this->file = $file;

        return $this;
    }

    /**
     * Get the value of infoContent
     */
    public function getInfoContent(): ?array
    {
        return $this->infoContent;
    }

    /**
     * Set the value of infoContent
     *
     * @return  self
     */
    public function setInfoContent(?array $infoContent): self
    {
        $this->infoContent = $infoContent;

        return $this;
    }

    /**
     * Get the value of description
     *
     * @return  string|null
     */
    public function getDescription()
    {
        return $this->description;
    }

    /**
     * Set the value of description
     *
     * @param  string|null  $description
     *
     * @return  self
     */
    public function setDescription($description)
    {
        $this->description = $description;

        return $this;
    }

    /**
     * Get the value of sourceControlStats
     */
    public function getsourceControlStats(): ?array
    {
        return $this->sourceControlStats;
    }

    /**
     * Set the value of sourceControlStats
     *
     * @return  self
     */
    public function setSourceControlStats(?array $sourceControlStats): self
    {
        $this->sourceControlStats = $sourceControlStats;

        return $this;
    }


    /**
     * Get the value of draftOptions
     */
    public function getDraftOptions()
    {
        return $this->draftOptions;
    }

    /**
     * Set the value of draftOptions
     *
     * @return  self
     */
    public function setDraftOptions($draftOptions)
    {
        $this->draftOptions = $draftOptions;

        return $this;
    }

    /**
     * Get the value of installedOptions
     */
    public function getInstalledOptions()
    {
        return $this->installedOptions;
    }

    /**
     * Set the value of installedOptions
     *
     * @return  self
     */
    public function setInstalledOptions($installedOptions)
    {
        $this->installedOptions = $installedOptions;

        return $this;
    }

    /**
     * @return ImportJob
     */
    public function getImportJob()
    {
        return $this->importJob;
    }

    public function setImportJob(ImportJob $importJob = null)
    {
        $this->importJob = $importJob;

        return $this;
    }
    /**
     * @return ExportTemplate
     */
    public function getTemplate()
    {
        return $this->template;
    }

    public function setTemplate(ExportTemplate $template = null)
    {
        $this->template = $template;

        return $this;
    }



    /**
     * @return null|ExportJob
     */
    public function getExportJob()
    {
        return $this->exportJob;
    }

    public function setExportJob(ExportJob $exportJob = null)
    {
        $this->exportJob = $exportJob;
        return $this;
    }

    /**
     * Get the value of local
     *
     * @return  bool|null
     */
    public function getLocal()
    {
        return $this->local;
    }

    /**
     * Set the value of local
     *
     * @param  bool|null  $local
     *
     * @return  self
     */
    public function setLocal($local)
    {
        $this->local = $local;

        return $this;
    }

    /**
     * Get the value of recordId
     *
     * @return  string|null
     */
    public function getRecordId()
    {
        return $this->recordId;
    }

    /**
     * Set the value of recordId
     *
     * @param  string|null  $recordId
     *
     * @return  self
     */
    public function setRecordId($recordId)
    {
        $this->recordId = $recordId;

        return $this;
    }

    /**
     * Add category
     *
     * @param \App\Entity\Core\Picklist  $category
     *
     * @return SolutionPack
     */
    public function addCategory($category)
    {
        if (!$this->category->contains($category)) {
            $this->category[] = $category;
        }
        return $this;
    }

    /**
     * Remove category
     *
     * @param \App\Entity\Core\Picklist  $category
     *
     * @return boolean TRUE if this collection contained the specified element, FALSE otherwise.
     */
    public function removeCategory($category)
    {
        return $this->category->removeElement($category);
    }

    /**
     * Get category
     *
     * @return \Doctrine\Common\Collections\Collection
     */
    public function getCategory()
    {
        return $this->category;
    }

    /**
     * Get the value of configCount
     */
    public function getConfigCount()
    {
        return $this->configCount;
    }

    /**
     * Set the value of configCount
     *
     * @return  self
     */
    public function setConfigCount($configCount)
    {
        $this->configCount = $configCount;

        return $this;
    }

    /**
     * Get the value of featuredTags
     *
     * @return  mixed|null
     */
    public function getFeaturedTags()
    {
        return $this->featuredTags;
    }

    /**
     * Set the value of featuredTags
     *
     * @param  mixed|null  $featuredTags
     *
     * @return  self
     */
    public function setFeaturedTags($featuredTags)
    {
        $this->featuredTags = $featuredTags;

        return $this;
    }

    /**
     * Get the value of buildNumber
     */
    public function getBuildNumber()
    {
        return $this->buildNumber;
    }

    /**
     * Set the value of buildNumber
     *
     * @return  self
     */
    public function setBuildNumber($buildNumber)
    {
        $this->buildNumber = $buildNumber;

        return $this;
    }

    /**
     * Get the value of latestCompatibleVersion
     *
     * @return  mixed|null
     */
    public function getLatestCompatibleVersion()
    {
        return $this->latestCompatibleVersion;
    }

    /**
     * Set the value of latestCompatibleVersion
     *
     * @param  mixed|null  $latestCompatibleVersion
     *
     * @return  self
     */
    public function setLatestCompatibleVersion($latestCompatibleVersion)
    {
        $this->latestCompatibleVersion = $latestCompatibleVersion;

        return $this;
    }
}
