<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Entity\Core;

use App\Entity\Core\File;
use App\Traits\Indexable;
use App\Traits\Trackable;
use App\Traits\Importable;
use App\Validator\IsJsonType;
use App\Entity\Base\BaseEntity;
use Doctrine\ORM\Mapping as ORM;
use App\Traits\ConditionalOwnable;
use App\Repository\WidgetRepository;
use App\EventSubscriber\WidgetListener;
use Doctrine\ORM\Mapping\UniqueConstraint;
use ApiPlatform\Core\Annotation\ApiResource;
use ApiPlatform\Core\Annotation\ApiSubresource;
use Symfony\Component\Serializer\Annotation\Groups;

/**
 *  @ApiResource(
 *     attributes={"validation_groups"={"fsr_primary", "fsr_all"}, "fsr_type"="widgets", "fsr_module"="widgets", "fsr_singular"="Widget"},
 *     denormalizationContext={"groups"={"fsr_all", "password"}, "fetch_data"=false},
 *     normalizationContext={"ignore_relationship_authorization"=true,"ignore_field_authorization"=true},
 *     collectionOperations={
 *          "get" = { "security_post_denormalize" = "is_granted('read.widgets', object)" },
 *          "post" = { "security_post_denormalize" = "is_granted('create.widgets', object)" }
 *     },
 *     itemOperations={
 *          "get" = { "security" = "is_granted('read.widgets', object)" },
 *          "put" = { "security" = "is_granted('update.widgets', object)" },
 *          "delete" = { "security" = "is_granted('delete.widgets', object)" }
 *     }
 * )
 * @ORM\Entity(repositoryClass=WidgetRepository::class)
 * @ORM\EntityListeners({"App\EventSubscriber\WidgetListener"})
 * @Doctrine\ORM\Mapping\Table(name="widgets", uniqueConstraints={@Doctrine\ORM\Mapping\UniqueConstraint(name="widgets_unique", columns={"name", "version"})}, indexes={@Doctrine\ORM\Mapping\Index(name="widgets_search_idx", columns={"modifydate"}), @Doctrine\ORM\Mapping\Index(name="widgets_id_search_idx", columns={"id"})})
 */
class Widget extends BaseEntity
{

    use Indexable;
    use Trackable;
    use Importable;
    use ConditionalOwnable;

    /**
     * @Groups({"fsr_primary", "fsr_all", "active"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $active;

    /**
     * @Groups({"fsr_primary", "fsr_all", "publishedDate"})
     * @ORM\Column(type="datetime", nullable=true)
     */
    private $publishedDate;

    /**
     * @Groups({"fsr_primary", "fsr_all", "subTitle"})
     * @ORM\Column(type="text", nullable=true)
     */
    private $subTitle;

    /**
     * @Groups({"fsr_primary", "fsr_all", "viewController"})
     * @ORM\Column(type="text", nullable=true)
     */
    private $viewController;

    /**
     * @Groups({"fsr_primary", "fsr_all", "editController"})
     * @ORM\Column(type="text", nullable=true)
     */
    private $editController;

    /**
     * @Groups({"fsr_primary", "fsr_all", "editHtml"})
     * @ORM\Column(type="text", nullable=true)
     */
    private $editHtml;

    /**
     * @Groups({"fsr_primary", "fsr_all", "viewHtml"})
     * @ORM\Column(type="text", nullable=true)
     */
    private $viewHtml;

    /**
     * @Groups({"fsr_primary", "fsr_all", "installed"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $installed;

    /**
     * @Groups({"fsr_primary", "fsr_all", "name"})
     * @ORM\Column(type="text", nullable=true)
     */
    private $name;

    /**
     * @Groups({"fsr_primary", "fsr_all", "label"})
     * @ORM\Column(type="text", nullable=true)
     */
    private $label;

    /**
     * @Groups({"fsr_primary", "fsr_all", "title"})
     * @ORM\Column(type="text", nullable=true)
     */
    private $title;

    /**
     * @Groups({"fsr_primary", "fsr_all", "version"})
     * @ORM\Column(type="text")
     */
    private $version;

    /**
     * @Groups({"fsr_primary", "fsr_all", "metadata"})
     * @IsJsonType("object")
     * @ORM\Column(type="jsonb", nullable=true)
     */
    private $metadata = [];

    /**
     * @Groups({"fsr_primary", "fsr_all", "draft"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $draft;

    /**
     * @var \Doctrine\Common\Collections\Collection
     * @Groups({"fsr_primary", "fsr_all", "file"})
     * @ApiSubresource(maxDepth=1)
     * @Doctrine\ORM\Mapping\ManyToMany(targetEntity="App\Entity\Core\File", cascade={"persist","remove","detach"}, orphanRemoval=true, fetch="EXTRA_LAZY")
     * @Doctrine\ORM\Mapping\JoinTable(name="widget_file",
     *   joinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="widget_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   },
     *   inverseJoinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="file_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   }
     * )
     */
    protected $files;

    /**
     * @Groups({"fsr_primary", "fsr_all", "enablePublish"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $enablePublish;

    /**
     * @var \Doctrine\Common\Collections\Collection
     *
     * @Doctrine\ORM\Mapping\ManyToMany(targetEntity="App\Entity\Core\Picklist", cascade={"persist"}, fetch="EXTRA_LAZY")
     * @Doctrine\ORM\Mapping\JoinTable(name="solutionpack_picklist_widget_category",
     *   joinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="widget_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   },
     *   inverseJoinColumns={
     *     @Doctrine\ORM\Mapping\JoinColumn(name="picklist_uuid", referencedColumnName="uuid", onDelete="CASCADE")
     *   }
     * )
     * @Groups({ "fsr_all","fsr_primary", "category"})
     */
    protected $category;

    /**
     * Constructor
     */
    public function __construct()
    {
        $this->category = new \Doctrine\Common\Collections\ArrayCollection();
        $this->files = new \Doctrine\Common\Collections\ArrayCollection();
    }


    public function getActive(): ?bool
    {
        return $this->active;
    }

    public function setActive(?bool $active): self
    {
        $this->active = $active;

        return $this;
    }

    public function getPublishedDate(): ?\DateTimeInterface
    {
        return $this->publishedDate;
    }

    public function setPublishedDate(?\DateTimeInterface $publishedDate): self
    {
        $this->publishedDate = $publishedDate;

        return $this;
    }

    public function getSubTitle(): ?string
    {
        return $this->subTitle;
    }

    public function setSubTitle(?string $subTitle): self
    {
        $this->subTitle = $subTitle;

        return $this;
    }

    public function getViewController(): ?string
    {
        return $this->viewController;
    }

    public function setViewController(?string $viewController): self
    {
        $this->viewController = $viewController;

        return $this;
    }

    public function getEditController(): ?string
    {
        return $this->editController;
    }

    public function setEditController(?string $editController): self
    {
        $this->editController = $editController;

        return $this;
    }

    public function getEditHtml(): ?string
    {
        return $this->editHtml;
    }

    public function setEditHtml(?string $editHtml): self
    {
        $this->editHtml = $editHtml;

        return $this;
    }

    public function getViewHtml(): ?string
    {
        return $this->viewHtml;
    }

    public function setViewHtml(?string $viewHtml): self
    {
        $this->viewHtml = $viewHtml;

        return $this;
    }

    public function getInstalled(): ?bool
    {
        return $this->installed;
    }

    public function setInstalled(?bool $installed): self
    {
        $this->installed = $installed;

        return $this;
    }

    public function getName(): ?string
    {
        return $this->name;
    }

    public function setName(?string $name): self
    {
        $this->name = $name;

        return $this;
    }

    public function getLabel(): ?string
    {
        return $this->label;
    }

    public function setlabel(?string $label): self
    {
        $this->label = $label;

        return $this;
    }

    public function getTitle(): ?string
    {
        return $this->title;
    }

    public function setTitle(?string $title): self
    {
        $this->title = $title;

        return $this;
    }

    public function getVersion(): ?string
    {
        return $this->version;
    }

    public function setVersion(?string $version): self
    {
        $this->version = $version;

        return $this;
    }

    public function getMetadata(): ?array
    {
        return $this->metadata;
    }

    public function setMetadata(?array $metadata): self
    {
        $this->metadata = $metadata;

        return $this;
    }

    public function getDraft()
    {
        return $this->draft;
    }

    public function setDraft($draft): self
    {
        $this->draft = $draft;

        return $this;
    }

    public function getEnablePublish()
    {
        return $this->enablePublish;
    }

    public function setEnablePublish($enablePublish): self
    {
        $this->enablePublish = $enablePublish;

        return $this;
    }

    /**
     * Add files
     *
     * @param \App\Entity\Core\File  $file
     *
     * @return Widget
     */
    public function addFile(\App\Entity\Core\File $file)
    {
        if (!$this->files->contains($file)) {
            $this->files[] = $file;
        }
        return $this;
    }

    /**
     * Remove files
     *
     * @param \App\Entity\Core\File  $file
     */
    public function removeFile(\App\Entity\Core\File $file)
    {
        $this->files->removeElement($file);
    }

    /**
     * Get files
     *
     * @return \Doctrine\Common\Collections\Collection 
     */
    public function getFiles()
    {
        return $this->files;
    }

    /**
     * Add category
     *
     * @param \App\Entity\Core\Picklist  $category
     *
     * @return Widget
     */
    public function addCategory($category)
    {
        if ($this->category && !$this->category->contains($category)) {
            $this->category[] = $category;
        }
        return $this;
    }

    /**
     * Remove category
     *
     * @param \App\Entity\Core\Picklist  $category
     *
     * @return boolean TRUE if this collection contained the specified element, FALSE otherwise.
     */
    public function removeCategory($category)
    {
        return $this->category->removeElement($category);
    }

    /**
     * Get category
     *
     * @return \Doctrine\Common\Collections\Collection
     */
    public function getCategory()
    {
        return $this->category;
    }
}
