<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Entity\ModelMetadata;

use App\Conditional\Query;
use Doctrine\ORM\Mapping as ORM;
use App\Validator\KeywordConstraint;
use App\Validator\ConditionSetConstraint;
use Doctrine\ORM\Mapping\UniqueConstraint;
use ApiPlatform\Core\Annotation\ApiResource;
use App\Entity\ModelMetadata\StagingModelMetadata;
use App\Traits\Importable;
use Symfony\Component\Serializer\Annotation\Groups;

/**
 *  @ApiResource(
 *     attributes={"validation_groups"={"fsr_primary", "fsr_all"}, "fsr_type"="attribute_metadatas", "fsr_module"="application", "fsr_singular"="attribute_metadata"},
 *     denormalizationContext={"groups"={"fsr_all", "password"}, "fetch_data"=false},
 *     collectionOperations={
 *          "get" = { "security_post_denormalize" = "is_granted('read.application', object)" },
 *          "post" = { "security_post_denormalize" = "is_granted('create.application', object)" }
 *     },
 *     itemOperations={
 *          "get" = { "security" = "is_granted('read.application', object)" }
 *     }
 * )
 * @ORM\Entity(repositoryClass="\Doctrine\ORM\EntityRepository")
 * @ConditionSetConstraint(fields={
 *      "ownsRelationship"={
 *          "notNull"={
 *             @Query(
 *                 result=true,
 *                 logic="AND",
 *                 filters={
 *                     {"field"="collection", "operator"="eq", "value"=true}
 *                 }
 *             )
 *          }
 *      },
 *      "visibility"={
 *          "isTrue"={
 *             @Query(
 *                 result=true,
 *                 logic="AND",
 *                 filters={
 *                     {"field"="validation[required]", "operator"="eq", "value"=true}
 *                 }
 *             )
 *          }
 *      }
 * })
 * @ORM\Table(name="attribute_metadatas", uniqueConstraints={@UniqueConstraint(name="staging_name_unique", columns={"name", "sattrib"})})
 */
class AttributeMetadata extends Metadatas
{

    use Importable;

    /**
     * @Groups({"fsr_primary", "fsr_all", "type"})
     * @ORM\Column(type="string", length=500, nullable=true)
     */
    private $type;

    /**
     * @Groups({"fsr_primary", "fsr_all", "name"})
     * @KeywordConstraint(groups={"Default", "fsr_primary", "fsr_all"})
     * @ORM\Column(type="string", length=500, nullable=true)
     */
    private $name;

    /**
     * @Groups({"fsr_primary", "fsr_all", "length"})
     * @ORM\Column(type="integer", nullable=true)
     */
    private $length;

    /**
     * @Groups({"fsr_primary", "fsr_all", "orderIndex"})
     * @ORM\Column(type="integer", nullable=true)
     */
    private $orderIndex;

    /**
     * @Groups({"fsr_primary", "fsr_all", "formType"})
     * @ORM\Column(type="string", length=500, nullable=true)
     */
    private $formType;

    /**
     * @Groups({"fsr_primary", "fsr_all", "dataSource"})
     * @ORM\Column(type="json", nullable=true)
     */
    private $dataSource;

    /**
     * @Groups({"fsr_primary", "fsr_all", "searchable"})
     * @ORM\Column(name="`searchable`", type="boolean", options={"default" = false})
     */
    private $searchable = false;

    /**
     * @Groups({"fsr_primary", "fsr_all", "peerReplicable"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $peerReplicable = true;

    /**
     * @Groups({"fsr_primary", "fsr_all", "system"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $system = false;

    /**
     * @Groups({"fsr_primary", "fsr_all", "encrypted"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $encrypted = false;

    /**
     * @Groups({"fsr_primary", "fsr_all", "gridColumn"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $gridColumn = false;

    /**
     * @Groups({"fsr_primary", "fsr_all", "collection"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $collection = false;

    /**
     * @Groups({"fsr_primary", "fsr_all", "inversedField"})
     * @ORM\Column(name="`inversefield`", type="string", length=200, nullable=true, unique=false)
     */
    private $inversedField;

    /**
     * @Groups({"fsr_primary", "fsr_all", "ownsRelationship"})
     * @ORM\Column(type="string", length=6, nullable=true)
     */
    private $ownsRelationship = false;

    /**
     * @Groups({"fsr_primary", "fsr_all", "validation"})
     * @ORM\Column(type="json", nullable=true)
     */
    private $validation;

    /**
     * @Groups({"fsr_primary", "fsr_all", "bulkAction"})
     * @ORM\Column(type="json", nullable=true)
     */
    private $bulkAction;

    /**
     * @Groups({"fsr_primary", "fsr_all", "dataSourceFilters"})
     * @ORM\Column(type="json", nullable=true)
     */
    private $dataSourceFilters;

    /**
     * @Groups({"fsr_primary", "fsr_all", "defaultValue"})
     * @ORM\Column(type="text", nullable=true)
     */
    private $defaultValue;

    /**
     * @Groups({"fsr_primary", "fsr_all", "tooltip"})
     * @ORM\Column(type="text", nullable=true)
     */
    private $tooltip;

    /**
     * @Groups({"fsr_primary", "fsr_all", "visibility"})
     * @ORM\Column(type="text", nullable=true, nullable=true)
     */
    private $visibility;

    /**
     * @Groups({"fsr_primary", "fsr_all", "htmlEscape"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $htmlEscape = false;

    /**
     * @Groups({"fsr_primary", "fsr_all", "orphanRemoval"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $orphanRemoval;

    /**
     * @Groups({"fsr_primary", "fsr_all", "readable"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $readable = true;

    /**
     * @Groups({"fsr_primary", "fsr_all", "writeable"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $writeable = true;

    /**
     * @Groups({"fsr_primary", "fsr_all", "identifier"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $identifier;

    /**
     * @Groups({"fsr_primary", "fsr_all", "unique"})
     * @ORM\Column(name="`unique`", nullable=true, type="boolean", nullable=true)
     */
    private $unique = false;

    /**
     * @Groups({"fsr_primary", "fsr_all", "recommend"})
     * @ORM\Column(type="boolean", nullable=true)
     */
    private $recommend = false;

    /**
     * @Groups({"fsr_primary", "fsr_all", "sattrib"})
     * @ORM\ManyToOne(targetEntity="StagingModelMetadata", inversedBy="attributes", fetch="EAGER", cascade={"persist"})
     * @Doctrine\ORM\Mapping\JoinColumns({
     *   @Doctrine\ORM\Mapping\JoinColumn(name="sattrib", referencedColumnName="uuid", onDelete="SET NULL")
     * })
     */
    private $sattrib;

    /**
     * @Groups({"fsr_primary", "fsr_all", "skipSerialization"})
     * @ORM\Column(name="skipserialization", type="boolean", options={"default":false})
     */
    private $skipSerialization = false;

    public function getType(): ?string
    {
        return $this->type;
    }

    public function setType($type): self
    {
        $this->type = $type;

        return $this;
    }

    public function getName(): ?string
    {
        return $this->name;
    }

    public function setName($name): self
    {
        $this->name = $name;

        return $this;
    }

    public function getLength(): ?int
    {
        return $this->length;
    }

    public function setLength($length): self
    {
        $this->length = $length;

        return $this;
    }

    public function getOrderIndex(): ?int
    {
        return $this->orderIndex;
    }

    public function setOrderIndex($orderIndex): self
    {
        $this->orderIndex = $orderIndex;

        return $this;
    }

    public function getFormType(): ?string
    {
        return $this->formType;
    }

    public function setFormType($formType): self
    {
        $this->formType = $formType;

        return $this;
    }

    public function getDataSource(): ?array
    {
        return $this->dataSource;
    }

    public function setDataSource(?array $dataSource): self
    {
        $this->dataSource = $dataSource;

        return $this;
    }

    public function getSearchable(): ?bool
    {
        return $this->searchable;
    }

    public function setSearchable($searchable): self
    {
        $this->searchable = $searchable;

        return $this;
    }

    public function getPeerReplicable(): ?bool
    {
        return $this->peerReplicable;
    }

    public function setPeerReplicable($peerReplicable): self
    {
        $this->peerReplicable = $peerReplicable;

        return $this;
    }

    public function getSystem(): ?bool
    {
        return $this->system;
    }

    public function setSystem($system): self
    {
        $this->system = $system;

        return $this;
    }

    public function getEncrypted(): ?bool
    {
        return $this->encrypted;
    }

    public function setEncrypted($encrypted): self
    {
        $this->encrypted = $encrypted;

        return $this;
    }

    public function getGridColumn(): ?bool
    {
        return $this->gridColumn;
    }

    public function setGridColumn($gridColumn): self
    {
        $this->gridColumn = $gridColumn;

        return $this;
    }

    public function getCollection(): ?bool
    {
        return $this->collection;
    }

    public function setCollection($collection): self
    {
        $this->collection = $collection;

        return $this;
    }

    public function getInversedField(): ?string
    {
        return $this->inversedField;
    }

    public function setInversedField($inversedField): self
    {
        $this->inversedField = $inversedField;

        return $this;
    }

    public function getOwnsRelationship()
    {
        return json_decode($this->ownsRelationship, true);
    }

    public function setOwnsRelationship($ownsRelationship)
    {
        $this->ownsRelationship = json_encode($ownsRelationship);

        return $this;
    }

    public function getValidation(): ?array
    {
        return $this->validation;
    }

    public function setValidation(?array $validation): self
    {
        $this->validation = $validation;

        return $this;
    }

    public function getBulkAction(): ?array
    {
        return $this->bulkAction;
    }

    public function setBulkAction(?array $bulkAction): self
    {
        $this->bulkAction = $bulkAction;

        return $this;
    }

    public function getDataSourceFilters(): ?array
    {
        return $this->dataSourceFilters;
    }

    public function setDataSourceFilters(?array $dataSourceFilters): self
    {
        $this->dataSourceFilters = $dataSourceFilters;

        return $this;
    }

    public function getDefaultValue(): ?string
    {
        return $this->defaultValue;
    }

    public function setDefaultValue($defaultValue): self
    {
        $this->defaultValue = $defaultValue;

        return $this;
    }

    public function getTooltip(): ?string
    {
        return $this->tooltip;
    }

    public function setTooltip($tooltip): self
    {
        $this->tooltip = $tooltip;

        return $this;
    }

    public function getVisibility()
    {
        return json_decode($this->visibility, true);
    }

    public function setVisibility($visibility)
    {
        $this->visibility = json_encode($visibility);
        return $this;
    }

    public function getReadable(): ?bool
    {
        return $this->readable;
    }

    public function setReadable($readable): self
    {
        $this->readable = $readable;

        return $this;
    }

    public function getWriteable(): ?bool
    {
        return $this->writeable;
    }

    public function setWriteable($writeable): self
    {
        $this->writeable = $writeable;

        return $this;
    }

    public function getIdentifier(): ?bool
    {
        return $this->identifier;
    }

    public function setIdentifier($identifier): self
    {
        $this->identifier = $identifier;

        return $this;
    }

    public function getUnique(): ?bool
    {
        return $this->unique;
    }

    public function setUnique($unique): self
    {
        $this->unique = $unique;

        return $this;
    }

    public function getSattrib(): ?StagingModelMetadata
    {
        return $this->sattrib;
    }

    public function setSattrib(?StagingModelMetadata $sattrib): self
    {
        $this->sattrib = $sattrib;

        return $this;
    }

    /**
     * Get the value of htmlEscape
     */
    public function getHtmlEscape()
    {
        return $this->htmlEscape;
    }

    /**
     * Set the value of htmlEscape
     *
     * @return  self
     */
    public function setHtmlEscape($htmlEscape)
    {
        $this->htmlEscape = $htmlEscape;

        return $this;
    }

    /**
     * Get the value of orphanRemoval
     */
    public function getOrphanRemoval()
    {
        return $this->orphanRemoval;
    }

    /**
     * Set the value of orphanRemoval
     *
     * @return  self
     */
    public function setOrphanRemoval($orphanRemoval)
    {
        $this->orphanRemoval = $orphanRemoval;

        return $this;
    }

    /**
     * Get the value of recommend
     */
    public function getRecommend()
    {
        return $this->recommend;
    }

    /**
     * Set the value of recommend
     *
     * @return  self
     */
    public function setRecommend($recommend)
    {
        $this->recommend = $recommend;

        return $this;
    }

    /**
     * Get the value of skipSerialization
     */
    public function getSkipSerialization()
    {
        return $this->skipSerialization;
    }

    /**
     * Set the value of skipSerialization
     *
     * @return  self
     */
    public function setSkipSerialization($skipSerialization)
    {
        $this->skipSerialization = $skipSerialization;

        return $this;
    }
}
