<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Entity\Settings;

use ApiPlatform\Core\Annotation\ApiResource;
use App\Repository\SystemSettingsRepository;
use Doctrine\ORM\Mapping as ORM;
use Doctrine\ORM\Mapping\UniqueConstraint;
use Symfony\Component\Serializer\Annotation\Groups;
use ApiPlatform\Core\Annotation\ApiProperty;

/**
 *  @ApiResource(
 *     attributes={"validation_groups"={"fsr_primary", "fsr_all"}, "fsr_type"="system_settings", "fsr_module"="application", "fsr_singular"="system_setting"},
 *     denormalizationContext={"groups"={"fsr_all", "password"}, "fetch_data"=false},
 *     collectionOperations={
 *          "get" = { "security_post_denormalize" = "is_granted('read.application', object)" },
 *          "post" = { "security_post_denormalize" = "is_granted('create.application', object)" }
 *     },
 *     itemOperations={
 *          "get" = { "security" = "is_granted('read.application', object)" },
 *          "put" = { "security" = "is_granted('update.application', object)" },
 *          "delete" = { "security" = "is_granted('delete.application', object)" },
 *     }
 * )
 * @ORM\Entity(repositoryClass="App\Repository\SystemSettingsRepository")
 * @ORM\Table(name="cyops.system_settings", schema="cyops", uniqueConstraints={@UniqueConstraint(name="restriction_unique", columns={"restriction"})})
 */
class SystemSettings extends Settings
{
    /**
     * @Groups({"fsr_primary", "fsr_all"})
     * @ORM\Column(type="string", length=1024, nullable=true, unique=true)
     */
    private $name;

    /**
     * @Groups({"fsr_primary", "fsr_all"})
     * @ORM\Column(type="jsonb", nullable=true)
     */
    private $privateValues;

    /**
     * @Groups({"fsr_primary", "fsr_all"})
     * @ORM\Column(type="jsonb", nullable=true)
     */
    private $globalValues;

    /**
     * @Groups({"fsr_primary", "fsr_all"})
     * @ORM\Column(type="string", length=255, nullable=true)
     */
    private $restriction;

    /**
     * @ORM\OneToMany(targetEntity="SystemSettings", mappedBy="parent", cascade={"persist"})
     * @Groups({"fsr_primary", "fsr_all", "sections"})
     */
    private $sections;

    /**
     * @ORM\ManyToOne(targetEntity="SystemSettings", inversedBy="sections", cascade={"persist"})
     * @ORM\JoinColumn(referencedColumnName="uuid", onDelete="CASCADE")
     * @Groups({"fsr_primary", "fsr_all", "parent"})
     * @ApiProperty(readableLink=false)
     */
    private $parent;

    public function __construct() {
        $this->sections = new \Doctrine\Common\Collections\ArrayCollection();
    }

    public function getPrivateValues(): ?array
    {
        return $this->privateValues;
    }

    public function setPrivateValues(?array $privateValues): self
    {
        $this->privateValues = $privateValues;

        return $this;
    }

    public function getRestriction(): ?string
    {
        return $this->restriction;
    }

    public function setRestriction(?string $restriction): self
    {
        $this->restriction = $restriction;

        return $this;
    }

    /**
     * Get the value of globalValues
     *
     * @return  array
     */ 
    public function getGlobalValues()
    {
        return $this->globalValues;
    }

    /**
     * Set the value of globalValues
     *
     * @param  array  $globalValues
     *
     * @return  self
     */ 
    public function setGlobalValues($globalValues)
    {
        $this->globalValues = $globalValues;

        return $this;
    }

    /**
     * Add Sections
     */
    public function addSection(SystemSettings $section)
    {
        if (!$this->sections->contains($section)) {
            $this->sections[] = $section;
            $section->setParent($this);
        }
        return $this;
    }

    /**
     * Remove Sections
     */
    public function removeSection(SystemSettings $section)
    {
        if ($this->sections->contains($section)) {
            $this->sections->removeElement($section);
            $section->setParent(null);
        }
        return $this;
    }

    /**
     * Get events
     *
     * @return \Doctrine\Common\Collections\Collection
     */
    public function getSections()
    {
        return $this->sections;
    }

    /**
     * Get the value of parent
     *
     * @return  App\Entity\Settings\SystemSettings
     */ 
    public function getParent()
    {
        return $this->parent;
    }

    /**
     * Set the value of parent
     *
     * @param  App\Entity\Settings\SystemSettings  $parent
     *
     * @return  self
     */ 
    public function setParent(SystemSettings $parent=null)
    {
        $this->parent = $parent;

        return $this;
    }

    /**
     * Get the value of name
     */ 
    public function getName()
    {
        return $this->name;
    }

    /**
     * Set the value of name
     *
     * @return  self
     */ 
    public function setName($name)
    {
        $this->name = $name;

        return $this;
    }
}
