<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use App\Entity\Actors\Actor;
use App\Entity\Custom\Agent;
use Psr\Log\LoggerInterface;
use App\Entity\Base\BaseEntity;
use App\Entity\Custom\Appliance;
use App\Entity\Authorization\Role;
use Doctrine\ORM\EntityManagerInterface;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class AgentSubscriber implements EventSubscriberInterface
{
    public const POSTMAN_CONFIG_DIRECTORY = '/opt/cyops-routing-agent/postman/configs/agents/';
    public const AGENT_ROLE_UUID = '3006d2f2-2f3a-4ae8-a36c-03ae75bf17fb';

    /** @var DasClient */
    protected $dasClient;

    public function __construct(
        protected EntityManagerInterface $entityManager,
        protected ContainerInterface $container,
        protected LoggerInterface $logger
    ) {
        $this->dasClient = $this->container->get("app.proxy.client.das");
    }

    public static function getSubscribedEvents()
    {
        // Adding this on POST WRITE PRIORITY in order so data is persisted in database and then handling of data can be done.
        return [
            KernelEvents::VIEW => [
                ['handleAgentCreate', 31],
                ['handleAgentDelete', 30]
            ]
        ];
    }

    public function handleAgentCreate(ViewEvent $event)
    {
        $request = $event->getRequest();
        $method = $request->getMethod();
        $object = $event->getControllerResult();
        $previousData = $request->attributes->get('previous_data');
        if (!$method || $method == Request::METHOD_GET || $method == Request::METHOD_DELETE || !$object instanceof BaseEntity) {
            return;
        }
        $entityClass = $object::class;
        if ($entityClass != Agent::class) {
            return;
        }

        if ($entityClass == Agent::class && is_null($previousData)) {
            $agentId = $object->getAgentId();
            $agentName = 'Agent' . '-' . $object->getName();
            $applianceRepo = $this->entityManager->getRepository(Appliance::class);
            $appliance = $applianceRepo->findBy(["name" => $agentName]);
            if (!$appliance) {
                $owners = $object->getOwners();
                $dasUri = $this->container->getParameter('das.uri');
                $headers = array(
                    "Content-type" => "application/json",
                );
                $applianceData = [
                    "domain" => $_SERVER['HTTP_HOST'],
                    "type" => 3,
                    "nfa" => [],
                    "system" => true,
                    "tenant" => $agentId . '_nXGkvPBdcR_' . $agentName,
                    "private_key_path" => self::POSTMAN_CONFIG_DIRECTORY . $agentId . '/APPLIANCE_PRIVATE_KEY',
                    "public_key_path" => self::POSTMAN_CONFIG_DIRECTORY . $agentId . '/APPLIANCE_PUBLIC_KEY',
                    "status" => 1,
                    "bind_name" => null,
                    "credentials" => ["loginid" => null]
                ];
                $dasRequests = new GuzzleRequest('post', $dasUri . '/users', $headers, json_encode($applianceData));
                try {
                    $dasResponse = $this->dasClient->send($dasRequests);
                    $resCode = $dasResponse->getStatusCode();
                    if ($resCode == 201) {
                        $response = json_decode($dasResponse->getBody());
                        $this->addApplianceDetails($agentName, $response->uuid, $owners);
                    }
                } catch (BadRequestHttpException | \Exception $error) {
                    $this->logger->error($error->getMessage());
                }
            }
        }
    }

    protected function addApplianceDetails($agentName, $userId, $teams)
    {
        $applianceData = new Appliance();
        $applianceData->setName($agentName);

        $roleRepo = $this->entityManager->getRepository(Role::class);
        $role = $roleRepo->findBy(["uuid" => self::AGENT_ROLE_UUID]);
        if ($role) {
            $role_obj = $role[0];
            $applianceData->addRole($role_obj);
        }
        $applianceData->setUserId($userId);
        foreach ($teams as $team) {
            $applianceData->addTeam($team);
        }
        $this->entityManager->persist($applianceData);
        $this->entityManager->flush();
    }
    public function handleAgentDelete(ViewEvent $event)
    {
        $request = $event->getRequest();
        $method = $request->getMethod();
        $previousData = $request->attributes->get('previous_data');

        if (!$method || $method != Request::METHOD_DELETE || !$previousData instanceof BaseEntity) {
            return;
        }
        $entityClass = $previousData::class;
        if ($entityClass != Agent::class) {
            return;
        }
        $agentName = 'Agent' . '-' . $previousData->getName();
        $applianceRepo = $this->entityManager->getRepository(Appliance::class);
        $appliances = $applianceRepo->findBy(["name" => $agentName]);
        $appliance = $appliances[0];
        $actorRepo = $this->entityManager->getRepository(Actor::class);
        $actor = $actorRepo->findBy(["uuid" => $appliance->getUuid()]);
        if ($actor) {
            $headers = array(
                "Content-type" => "application/json",
            );
            $dasUri = $this->container->getParameter('das.uri');
            $dasRequests = new GuzzleRequest('DELETE', $dasUri . '/users/'.$actor[0]->getUserId(), $headers);
            try {
                $dasResponse = $this->dasClient->send($dasRequests);
                $dasResponse->getStatusCode();
                $this->entityManager->remove($appliance);
                $this->entityManager->flush();
            } catch (BadRequestHttpException | \Exception $error) {
                $this->logger->error($error->getMessage());
            }
        }
    }

}
