<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use App\Traits\Ownable;
use App\Traits\UserOwnable;
use InvalidArgumentException;
use App\Entity\Base\BaseEntity;
use App\Service\MetadataHelper;
use Doctrine\ORM\EntityManager;
use Doctrine\Inflector\InflectorFactory;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\KernelEvents;
use ApiPlatform\Core\Api\IriConverterInterface;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\Serializer\SerializerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\PropertyAccess\PropertyAccessorInterface;

final class FileOwnershipCascadingSubscriber implements EventSubscriberInterface
{

    protected $container;
    private $inflector;

    public static function getSubscribedEvents()
    {
        return [
            KernelEvents::VIEW => ['FileOwnership', 31]
        ];
    }

    public function __construct(
        protected IriConverterInterface $iriConverter,
        protected SerializerInterface $serializer,
        protected EntityManagerInterface $entityManager,
        protected MetadataHelper $metadataHelper,
        protected RequestStack $requeststack,
        protected PropertyAccessorInterface $propertyAccessor
    ) {
        $this->inflector = InflectorFactory::create()->build();
    }

    public function FileOwnership(ViewEvent $event): void
    {
        $object = $event->getControllerResult();
        $request = $event->getRequest();
        $method = $request->getMethod();
        if (!$method || $method == Request::METHOD_GET || $method == Request::METHOD_DELETE || !$object instanceof BaseEntity) {
            return;
        }
        $request = $event->getRequest();
        $previousData = $request->attributes->get('previous_data');
        $previousOwner = [];
        $entityClass =  $object::class;
        $entityFieldMap = $this->metadataHelper->getEntityFields($entityClass) ?? null;
        if ($entityFieldMap && (is_countable($entityFieldMap['fileFields']) ? count($entityFieldMap['fileFields']) : 0)) { 
            // Added extra check because of missing type map, can be removed once all entity are mapped
            foreach ($entityFieldMap['fileFields'] as $key => $value) {
                $camelized = $this->inflector->camelize($value);
                $getMethod = 'get' . $camelized;
                if ($object->{$getMethod}()) {
                    try {
                        $fileObject = $this->iriConverter->getItemFromIri($object->{$getMethod}());
                        // $value = $this->normalize($fileObject, 'jsonld', $context);

                        if ($fileObject && (in_array(Ownable::class, class_uses($object::class)) || in_array(UserOwnable::class, class_uses($object::class)))) {
                            $previousOwner = [];
                            $previousUserOwner = [];

                            if ($previousData) {
                                $previousOwner = method_exists($previousData, 'getOwners') ? $previousData->getOwners() : [];
                                $previousUserOwner = method_exists($previousData, 'getUserOwners') ? $previousData->getUserOwners() : [];
                            }

                            $teams = method_exists($object, 'getOwners') ? $object->getOwners() : [];
                            $userOwners = method_exists($object, 'getUserOwners') ? $object->getUserOwners() : [];

                            // Removing Old Team from Current file objects
                            if ($previousOwner && (is_countable($previousOwner) ? count($previousOwner) : 0)) {
                                foreach ($previousOwner as $oldOwner) {
                                    $fileObject->removeOwner($oldOwner);
                                }
                            }

                            // Removing Old User Owners from Current file objects
                            if ($previousUserOwner && (is_countable($previousUserOwner) ? count($previousUserOwner) : 0)) {
                                foreach ($previousUserOwner as $oldOwner) {
                                    $fileObject->removeUserOwner($oldOwner);
                                }
                            }

                            // Adding current teams to file objects
                            if ($teams && (is_countable($teams) ? count($teams) : 0)) {
                                foreach ($teams as $team) {
                                    $fileObject->addOwner($team);
                                }
                            }

                            // Adding current user owners to file objects
                            if ($userOwners && (is_countable($userOwners) ? count($userOwners) : 0)) {
                                foreach ($userOwners as $owner) {
                                    $fileObject->addUserOwner($owner);
                                }
                            }

                            $this->entityManager->persist($fileObject);
                            $this->entityManager->flush();
                        }
                    } catch (InvalidArgumentException | \Exception $e) {
                        $value = null;
                    }
                }
            }
        }
    }
}
