<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use App\Service\TokenHelper;
use Psr\Log\LoggerInterface;
use App\Constants\CacheConstants;
use App\Service\UtilityService;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Event\LifecycleEventArgs;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class IndicatorsFeedsLinkingSubscriber
{
    private const REPUTATION_PRIORITY = [
        "ae98ebc6-beef-4882-9980-1d88fc6d87cd" => 1, // TBD
        "9a611980-1b5e-4ae9-8062-eb2c0c433cff" => 1, // No reputation available
        "b19b42aa-aee4-47df-9cda-894537dacb2a" => 1, // Good
        "50bfd06c-9aff-4f7d-b6d9-821339e31fe7" => 2, // Suspicious
        "7074e547-7785-4979-be32-c6d0c863e4bd" => 3  // Malicious
    ];

    private const REPUTATION_MAPPING_DEFAULT = [
        "setReputation" => false,
        "confidenceThreshold" => 90
    ];

    public function __construct(
        private EntityManagerInterface $entityManager,
        private ContainerInterface $containers,
        private TokenHelper $tokenHelper,
        protected TagAwareCacheInterface $cache,
        protected LoggerInterface $logger,
        protected UtilityService $utilityService
    ) {
        $this->cache = $cache;
    }

    public function prePersist(LifecycleEventArgs $args)
    {
        if ($this->utilityService->isTimSkuEnabled() && class_exists('\\App\\Entity\\Custom\\ThreatIntel') && class_exists('\\App\\Entity\\Custom\\Indicator')) {
            $entity = $args->getObject()::class;
            if ($entity === 'App\Entity\Custom\Indicator') {
                $this->linkFeeds($args->getObject());
            }
        }
    }

    private function linkFeeds($object): void
    {
        $reputationPreference = $this->getIndicatorPreference();
        if ($reputationPreference['setReputation'] === true) {
            $indicatorValue = method_exists($object, 'getValue') ? $object->getValue() : null;
            if (!is_null($indicatorValue)) {
                $indicatorToken = $this->tokenHelper->extractThreatFeedTokens($indicatorValue);
                $threatIntelRepository = $this->entityManager->getRepository('\\App\\Entity\\Custom\\ThreatIntel');
                $threatValuesQuery = $threatIntelRepository->createQueryBuilder('o')->andwhere("ARRAY_OVERLAP(o.valueTokens, :tokenValues) = TRUE")->setParameter('tokenValues', $indicatorToken)->getQuery();
                $threatValues = $threatValuesQuery->getResult();
                if (empty($threatValues)) {
                    return;
                }
                $maxConfidence = 0;
                $reputationToUse = $threatValues[0]->getReputation();
                if ($this->containers->hasParameter('indicator_reputation_priority')) {
                    $reputationPriority = $this->containers->getParameter('indicator_reputation_priority');
                } else {
                    $reputationPriority = self::REPUTATION_PRIORITY;
                }
                foreach ($threatValues as $threatValue) {
                    $threatConfidence = $threatValue->getConfidence() ?: 0;
                    if ($threatConfidence >= $reputationPreference['confidenceThreshold'] && $maxConfidence < $threatConfidence) {
                        $maxConfidence = $threatConfidence;
                        $reputationToUse = $threatValue->getReputation();
                    } elseif ($threatConfidence >= $reputationPreference['confidenceThreshold'] && $maxConfidence == $threatConfidence) {
                        if (
                            !is_null($threatValue->getReputation()) &&
                            array_key_exists($threatValue->getReputation()->getUuid(), $reputationPriority) &&
                            $reputationPriority[$threatValue->getReputation()->getUuid()] > $reputationPriority[$reputationToUse->getUuid()]
                        ) {
                            $reputationToUse = $threatValue->getReputation();
                        }
                    }
                    $now = new \DateTime('NOW', new \DateTimeZone('UTC'));
                    $threatValue->setLastSeen($now);
                    $threatValue->setExpiresOn(null);
                    $object->addThreatIntelFeed($threatValue);
                }
                if (
                    $maxConfidence >= $reputationPreference['confidenceThreshold'] &&
                    (is_null($object->getReputation())
                        || (array_key_exists($object->getReputation()->getUuid(), $reputationPriority) &&
                            $reputationPriority[$object->getReputation()->getUuid()] < $reputationPriority[$reputationToUse->getUuid()]))
                ) {
                    $object->setReputation($reputationToUse);
                }
            }
        }
    }

    private function getIndicatorPreference()
    {
        $reputationPreference = $this->cache->getItem(CacheConstants::INDICATOR_REPUTATION_PREFERENCE);
        if (!$reputationPreference->isHit()) {
            $wfClient = $this->containers->get("app.proxy.client.workflow");
            $headers = ["Content-type" => "application/json"];
            $workflowUri = $this->containers->getParameter('workflow.uri');
            $workflowRequests = new GuzzleRequest('get', $workflowUri . '/api/dynamic-variable/?format=json&name=' . CacheConstants::INDICATOR_REPUTATION_PREFERENCE, $headers);
            try {
                $workflowResponse = $wfClient->send($workflowRequests);
                $workflowResponseData = json_decode($workflowResponse->getBody()->getContents(), true);
                if ($workflowResponseData["hydra:totalItems"] > 0) {
                    $indicator_preference = json_decode($workflowResponseData["hydra:member"][0]["value"], true);
                } else {
                    // SP not installed. Macro not present
                    $indicator_preference = self::REPUTATION_MAPPING_DEFAULT;
                }
                $reputationPreference->set($indicator_preference);
                $reputationPreference->expiresAfter(3600);
                $this->cache->save($reputationPreference);
            } catch (BadRequestHttpException | \Exception $error) {
                $this->logger->error($error->getMessage());
                // do not set into cache. return default
                return self::REPUTATION_MAPPING_DEFAULT;
            }
        }
        return $reputationPreference->get();
    }
}
