<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use App\Entity\Base\BaseEntity;
use App\Service\MetadataHelper;
use Doctrine\Inflector\InflectorFactory;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\KernelEvents;
use ApiPlatform\Core\Api\IriConverterInterface;
use App\Entity\ModelMetadata\AttributeMetadata;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\PropertyAccess\PropertyAccessorInterface;
use Symfony\Component\PropertyAccess\Exception\NoSuchPropertyException;
use App\Constants\AppConstants;
use App\Service\UtilityService;

final class LinkedUnlinkSubscriber implements EventSubscriberInterface
{
    private $inflector;

    public static function getSubscribedEvents()
    {
        return [
            KernelEvents::VIEW => ['LinkUnLinkObjects', 35]
        ];
    }

    public function __construct(
        protected IriConverterInterface $iriConverter,
        protected RequestStack $requestStack,
        protected PropertyAccessorInterface $propertyAccessor,
        private EntityManagerInterface $entityManager,
        private MetadataHelper $metadataHelper,
        private UtilityService $utilityService
    ) {
        $this->inflector = InflectorFactory::create()->build();
    }

    protected function handleLinkItems($object, $linkItems, $request, &$attributes, array $queryParams = [])
    {
        $linkItems = $this->handleBadRelationData($linkItems, 'link', $object::class, $request, $attributes);
        foreach ($linkItems as $attribute => $itemToLinks) {
            try {
                $attributeValues = $this->propertyAccessor->getValue($object, $attribute);
            } catch (NoSuchPropertyException $e) {
                continue;
            }
            $camelized = $this->inflector->classify($attribute);
            $singulars = $this->inflector->singularize($camelized);
            $addMethod = 'add' . $singulars;
            if ($attribute == 'recordTags') {
                $attribute = 'tags';
            }
            $className = $this->metadataHelper->getResourceFromClassName($attribute);
            if (!is_null($attributeValues)) {
                // If sending value as string
                // eg alerts: "/api/3/alerts/123, /api/3/alerts/234"
                if (is_string($itemToLinks)) {
                    if (strlen(trim($itemToLinks)) === 0) {
                        continue;
                    }
                    $itemToLinks = array_map('trim', explode(',', $itemToLinks));
                }

                /* Validate if request is coming for custom modules
                Also validate if ignore missing relationship params is set over requests.
                If set, we will validate the the related record against database and check there validity.
                This will be done for all ManyToMany Fields.
                */
                $values = $itemToLinks; // Initialize the values with requested object requested for linking.
                if (
                    array_key_exists(AppConstants::IGNORE_MISSING_RELATIONSHIP, $queryParams) &&
                    $queryParams[AppConstants::IGNORE_MISSING_RELATIONSHIP] &&
                    str_starts_with($className, AppConstants::CUSTOM_MODULE_IDENTIFIER)
                ) {
                    $values = $this->utilityService->validateRelationFieldsValues($className, $itemToLinks);
                }
                foreach ($values as $value) {
                    // For records not part of custom modules these list of check need to be there.
                    if (is_string($value) && !str_contains($value, '/api/3')) {
                        $value = '/api/3/' . $attribute . '/' . $value;
                    } elseif (gettype($value) == 'object') {
                        $value = $value->{"@id"};
                    } elseif (gettype($value) == 'array' && array_key_exists('@id', $value)) {
                        $value = $value["@id"];
                    }
                    $itemToLinkObj = $this->iriConverter->getItemFromIri($value, ['fetch_data' => false]);
                    $object->{$addMethod}($itemToLinkObj);
                }
            }
        }
    }

    public function handleUnLinkItems($object, $unLinkItems, $request, &$attributes, array $queryParams = [])
    {
        $unLinkItems = $this->handleBadRelationData($unLinkItems, 'unlink', $object::class, $request, $attributes);
        foreach ($unLinkItems as $attribute => $itemToUnLinks) {
            try {
                $attributeValues = $this->propertyAccessor->getValue($object, $attribute);
            } catch (NoSuchPropertyException $e) {
                continue;
            }
            if (!is_null($attributeValues)) {
                $camelized = $this->inflector->classify($attribute);
                $singulars = $this->inflector->singularize($camelized);
                $removeMethod = 'remove' . $singulars;
                if ($attribute == 'recordTags') {
                    $attribute = 'tags';
                }
                $className = $this->metadataHelper->getResourceFromClassName($attribute);
                // If sending value as string
                // eg alerts: "/api/3/alerts/123, /api/3/alerts/234"
                if (is_string($itemToUnLinks)) {
                    if (strlen(trim($itemToUnLinks)) === 0) {
                        continue;
                    }
                    $itemToUnLinks = array_map('trim', explode(',', $itemToUnLinks));
                }
                /* Validate if request is coming for custom modules
                Also validate if ignore missing relationship params is set over requests.
                If set, we will validate the the related record against database and check there validity.
                This will be done for all ManyToMany Fields.
                */
                $values = $itemToUnLinks;  // Initialize the values with requested object requested for unlinking.
                if (
                    array_key_exists(AppConstants::IGNORE_MISSING_RELATIONSHIP, $queryParams) &&
                    $queryParams[AppConstants::IGNORE_MISSING_RELATIONSHIP] &&
                    str_starts_with($className, AppConstants::CUSTOM_MODULE_IDENTIFIER)
                ) {
                    $values = $this->utilityService->validateRelationFieldsValues($className, $itemToUnLinks);
                }
                foreach ($values as $value) {
                    if (is_string($value) && !str_contains($value, '/api/3')) {
                        $value = '/api/3/' . $attribute . '/' . $value;
                    } elseif (gettype($value) == 'object') {
                        $value = $value->{"@id"};
                    } elseif (gettype($value) == 'array' && array_key_exists('@id', $value)) {
                        $value = $value["@id"];
                    }
                    $itemToUnLinkObj = $this->iriConverter->getItemFromIri($value, ['fetch_data' => false]);
                    $object->{$removeMethod}($itemToUnLinkObj);
                }
            }
        }
    }

    public function LinkUnLinkObjects(ViewEvent $event): void
    {
        $object = $event->getControllerResult();
        $request = $event->getRequest();
        $request_method = $request->getMethod();
        if (!($object instanceof BaseEntity) || $request_method == Request::METHOD_GET || $request_method == Request::METHOD_DELETE) {
            return;
        }
        $this->manipulateObject($request, $object);
    }

    public function manipulateObject($request, $object)
    {
        // Validate if request is coming for custom modules
        // Also validate if ignore missing relationship params is set over requests.
        $queryParams = $request->query->all();
        $request_body = json_decode($request->getContent());
        if ($request_body && (isset($request_body->{"__link"}) || isset($request_body->{"__unlink"}))) {
            $linkItems = (array)$request_body->{'__link'} ?? [];
            $unlinkItems = property_exists($request_body, '__unlink') && isset($request_body->{"__unlink"}) ? (array)$request_body->{'__unlink'} : [];
            $attributes = null;
            if (count($linkItems)) {
                $this->handleLinkItems($object, $linkItems, $request, $attributes, $queryParams);
            }
            if (count($unlinkItems)) {
                $this->handleUnLinkItems($object, $unlinkItems, $request, $attributes, $queryParams);
            }
        }
    }

    private function handleBadRelationData($relationItems, $relationName, $className, $request, &$classAttributes)
    {
        $hasBadRelationData = false;
        foreach ($relationItems as $innerKey => $innerValue) {
            if ($innerKey === 'recordTags' && is_array($innerValue)) {
                foreach ($innerValue as $itemKey => $itemValue) {
                    $tag = !empty($itemValue) && is_string($itemValue) && str_contains($itemValue, '/api/3/tags/') ? end(explode('/', $itemValue)) : $itemValue;
                    if (empty(trim($tag)) || strlen($tag) <= AppConstants::MIN_TAG_LENGTH) {
                        unset($innerValue[$itemKey]);
                    }
                }
                $relationItems[$innerKey] = array_values($innerValue);
            }
            if (is_string($innerValue)) {
                $innerValue = $this->parseBadRelationData($innerValue);
                if (is_object($relationItems)) {
                    $relationItems->$innerKey = $innerValue;
                } else {
                    $relationItems[$innerKey] = $innerValue;
                }
                $hasBadRelationData = true;
            }
            # if relation UUIDs have been used instead of IRIs, convert them to IRIs
            # checking first element should suffice
            if ($innerKey != 'recordTags' && !empty($innerValue) && is_string($innerValue[0]) && !str_contains($innerValue[0], '/api/')) {
                $innerValue = $this->convertUUIDToIRI($innerValue, $innerKey, $className, $classAttributes);
                if (is_object($relationItems)) {
                    $relationItems->$innerKey = $innerValue;
                } else {
                    $relationItems[$innerKey] = $innerValue;
                }
                $hasBadRelationData = true;
            }
        }
        if ($hasBadRelationData === true) {
            $request->attributes->set('fsr_' . $relationName . '_data_modified', $relationItems);
        }
        return $relationItems;
    }

    private function parseBadRelationData($relationString)
    {
        $value = trim($relationString);
        if (strlen($value) > AppConstants::MIN_TAG_LENGTH) {
            return explode(",", $value);
        } else if (empty($value)) {
            return [];
        } else {
            return [];
        }
    }

    private function convertUUIDToIRI($uuidList, $key, $className, &$attributes)
    {
        $iriList = [];
        $relationType = null;
        //lazy load only if relations have UUID and send back to caller
        if (is_null($attributes)) {
            //lazy load only if relations have UUID and send back to caller
            $entityType = $this->metadataHelper->getTypeFromClassName($className);
            $modelMetaDataRepository = $this->entityManager->getRepository(AttributeMetadata::class);
            $attribute = $modelMetaDataRepository->createQueryBuilder('o')->andWhere('smd.type = :entityType')->andWhere('o.name = :name')->leftJoin('o.sattrib', 'smd')->setParameter('entityType', $entityType)->setParameter('name', $key)->getQuery()->getOneOrNullResult();
            if (is_null($attribute)) {
                # could be a system module
                $systemFixturesFile = file_get_contents('/opt/cyops-api/src/StaticFile/SystemFixtures.json');
                $systemFixture = json_decode($systemFixturesFile);
                $modelMetaDataForType = $systemFixture->{$entityType};
                $attributes = $modelMetaDataForType->{'attributes'};
                foreach ($attributes as $eachAttribute) {
                    if ($eachAttribute->getName() == $key) {
                        $relationType = $eachAttribute->getType();
                        break;
                    }
                }
            } else {
                $relationType = $attribute->getType();
            }
        }
        foreach ($uuidList as $uuid) {
            array_push($iriList, '/api/3/' . $relationType . '/' . $uuid);
        }
        return $iriList;
    }
}
