<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use stdClass;
use App\Entity\Custom\Agent;
use App\Utility\MsspUtility;
use Psr\Log\LoggerInterface;
use App\Entity\Core\Picklist;
use App\Entity\Custom\Router;
use App\Entity\Custom\Tenant;
use App\Traits\SoftDeleteable;
use App\Entity\Base\BaseEntity;
use App\Entity\Core\PicklistName;
use Doctrine\Inflector\InflectorFactory;
use Doctrine\ORM\EntityManagerInterface;
use App\Entity\ModelMetadata\ModelMetadata;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\KernelEvents;
use ApiPlatform\Core\Api\IriConverterInterface;
use App\Constants\AppConstants;
use App\Entity\ModelMetadata\AttribModelMetadata;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use App\Entity\ModelMetadata\StagingModelMetadata;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\Serializer\SerializerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\DependencyInjection\ParameterBag\ParameterBagInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

final class MqMessagebroadcastSubscriber implements EventSubscriberInterface
{

    public const AGENT_REPLICATION_ROUTINGKEY = 'route.crudhub.tenant.request.agentconfig';
    public const TENANT_REPLICATION_ROUTINGKEY = "route.crudhub.tenant.request.tenantconfig";
    public const ROUTER_REPLICATION_ROUTINGKEY = 'route.crudhub.tenant.request.routerconfig';
    public const DEFALUT_KEY = 'cyops.crudhub.datanotify';
    public const ELASTIC_SEARCH_OPERATION_KEY = 'esoperation';
    private $inflector;

    public static function getSubscribedEvents()
    {
        return [
            KernelEvents::VIEW => ['broadcastData', 15],
        ];
    }

    public function __construct(
        protected IriConverterInterface $iriConverter,
        protected RequestStack $requeststack,
        protected ContainerInterface $container,
        protected TokenStorageInterface $tokenStorage,
        protected SerializerInterface $serializer,
        protected MsspUtility $msspUtility,
        private ParameterBagInterface $parameters,
        private EntityManagerInterface $entityManager,
        private LoggerInterface $logger
    ) {
        $this->inflector = InflectorFactory::create()->build();
    }

    public function isSoftDeleteable($entity_class)
    {
        return in_array(SoftDeleteable::class, class_uses($entity_class));
    }

    public function broadcastData(ViewEvent $event)
    {
        $entity = null;
        $replicationOp = null;
        $requestOperation = null;
        $request = $event->getRequest();
        $requestMethod = $request->getMethod();
        $requestBody = new stdClass();
        if ($requestMethod == Request::METHOD_GET) {
            return;
        }
        $currentData = $event->getRequest()->attributes->get('data');
        $previousData = $request->attributes->get('previous_data');
        if (($currentData && !($currentData instanceof BaseEntity)) || ($previousData && !($previousData instanceof BaseEntity))) {
            return;
        }

        $normalized_current_data = json_decode($event->getControllerResult(), true);
        $normalized_previous_data = (object) null;
        $changeData = [];
        $data_to_broadcast = [];
        $isImporting = $request->attributes->get('fsr_subrequest_type') == 'import';

        if (is_null($previousData) && $currentData) {
            $entity = $currentData::class;
            $requestOperation = $isImporting ? 'Import' : 'Create';
            $replicationOp = 'POST';
            $data_to_broadcast[self::ELASTIC_SEARCH_OPERATION_KEY] = 'index';
        } elseif ($requestMethod == Request::METHOD_DELETE) {
            $replicationOp = 'DELETE';
            $data_to_broadcast[self::ELASTIC_SEARCH_OPERATION_KEY] = 'delete';
            $entity = get_class($previousData);
            $requestOperation = $isImporting ? 'Overwrite During Import' : 'Delete';
            $normalized_previous_data = $request->attributes->get('fsr_previous_data');
        } elseif ($previousData && $currentData) {
            $replicationOp = 'PUT';
            $entity = $currentData::class;
            $data_to_broadcast[self::ELASTIC_SEARCH_OPERATION_KEY] = 'modify';
            $requestOperation = $isImporting ? 'Update During Import' : 'Update';
            $normalized_previous_data = $request->attributes->get('fsr_previous_data');
            if (is_array($normalized_previous_data) && array_key_exists("@id", $normalized_previous_data)) {
                $entityType = explode("/", $normalized_previous_data['@id'])[3];
            } else {
                $entityType = null;
                $this->logger->debug("Normalized previous data is not array or not containing @id field");
                $this->logger->debug(gettype($normalized_previous_data));
            }
            $requestBody = json_decode($request->getContent());
            if (!is_null($request->attributes->get('fsr_link_data_modified'))) {
                $requestBody->{'__link'} = $request->attributes->get('fsr_link_data_modified');
            }
            if (!is_null($request->attributes->get('fsr_unlink_data_modified'))) {
                $requestBody->{'__unlink'} = $request->attributes->get('fsr_unlink_data_modified');
            }
            $changeData = $this->identifyChangeData($requestBody, $normalized_previous_data, $entityType, $normalized_current_data, $replicationOp);
        }

        // for lookup fields of non-system modules, audit and notifier expect iris
        if (!($currentData instanceof Picklist) && !($currentData instanceof PicklistName) && !($currentData instanceof StagingModelMetadata)) {
            $this->replaceLookupWithIri($normalized_current_data);
        }

        $softDelete = false;
        $isRestore = false;

        if (is_object($normalized_previous_data)) {
            $deletedAt = $normalized_previous_data->{'deletedAt'} ?? null;
        } else {
            $deletedAt = $normalized_previous_data['deletedAt'] ?? null;
        }

        $queryParams = $request->query->all();
        // softDelete is false when the query param "hardDelete" exists and is true (or) deletedAt is not null
        if ($this->isSoftDeleteable($entity) && $replicationOp == 'DELETE' && (!array_key_exists('$hardDelete', $queryParams) ||
            (array_key_exists('$hardDelete', $queryParams) && !$queryParams['$hardDelete'])) && is_null($deletedAt)) {
            $softDelete = true;
        } elseif ($this->isSoftDeleteable($entity) && $replicationOp == 'PUT' && $deletedAt && is_null($normalized_current_data['deletedAt'])) {
            $isRestore = true;
        }

        // Data to transmit to Audit, Notifier and Falcon Services
        $data_to_broadcast = array_merge($data_to_broadcast, $this->prepareChangeData(
            $normalized_current_data,
            $normalized_previous_data,
            $requestOperation,
            $changeData,
            $entity,
            $softDelete,
            $isRestore
        ));

        // Data to transmit to Postman service
        // This needs to be reimplemented in future
        // All services should work on standard message
        // Any alteration required in data should be done by component not the API layer service
        $diffObj = $this->msspUtility->prepareReplicationDiffData(
            $normalized_current_data,
            $replicationOp,
            $normalized_previous_data,
            $changeData,
            $isRestore
        );

        // All these needs to be standard over singile message
        if ($entity == Agent::class) {
            $data = $this->msspUtility->prepareReplicationData($diffObj, $replicationOp, null, null);
            $this->msspUtility->sendReplicationData($data, self::AGENT_REPLICATION_ROUTINGKEY);
        } elseif ($entity == Tenant::class) {
            $data = $this->msspUtility->prepareReplicationData($diffObj, $replicationOp, null, null);
            $this->msspUtility->sendReplicationData($data, self::TENANT_REPLICATION_ROUTINGKEY);
        } elseif ($entity == Router::class) {
            $data = $this->msspUtility->prepareReplicationData($diffObj, $replicationOp, null, null);
            $this->msspUtility->sendReplicationData($data, self::ROUTER_REPLICATION_ROUTINGKEY);
        }

        if ($event->getRequest()->attributes->get('fsr_subrequest_type') === null) {
            if (!($entity == Agent::class || $entity == Tenant::class || $entity == Router::class)) {
                $this->msspUtility->replicateData($entity, $requestBody, [$diffObj], $replicationOp);
            }
            $this->sendData([$data_to_broadcast]);
            return;
        }
        $audit_data = $request->attributes->get('fsr_audit_data') ?? [];
        $audit_data[] = $data_to_broadcast;
        $request->attributes->set('fsr_audit_data', $data_to_broadcast);
        $request->attributes->set('fsr_replication_data', $diffObj);
    }

    private function identifyChangeData($requestBody, $normalizedRecord, $entityType, $normalized_current_data, $replicationOp = null)
    {
        $attributes = null;
        foreach ($requestBody as $key => $value) {
            if ($key == '__link') {
                foreach ($requestBody->{'__link'} as $innerKey => $innerValue) {
                    if (is_string($innerValue)) {
                        $innerValue = explode(',', $innerValue);
                    } elseif (is_array($innerValue)) {
                        $innerVal = [];
                        foreach ($innerValue as $item) {
                            if (is_object($item)) {
                                $innerVal[] = $item->{'@id'};
                            } else {
                                $innerVal[] = $item;
                            }
                        }
                        $innerValue = $innerVal;
                    }
                    if (!empty($requestBody->$innerKey)) {
                        $requestBody->$innerKey = array_unique(array_merge($requestBody->$innerKey, $innerValue));
                    } else {
                        $requestBody->$innerKey = array_unique(array_merge($normalizedRecord[$innerKey], $innerValue));
                    }
                }
            } elseif ($key == '__unlink') {
                foreach ($requestBody->{'__unlink'} as $innerKey => $innerValue) {
                    if (is_string($innerValue)) {
                        $innerValue = explode(',', $innerValue);
                    } elseif (is_array($innerValue)) {
                        $innerVal = [];
                        foreach ($innerValue as $item) {
                            if (is_object($item)) {
                                $innerVal[] = $item->{'@id'};
                            } else {
                                $innerVal[] = $item;
                            }
                        }
                        $innerValue = $innerVal;
                    }
                    if (!empty($requestBody->$innerKey)) {
                        $requestBody->$innerKey = array_unique(array_diff($requestBody->$innerKey, $innerValue));
                    } else {
                        $requestBody->$innerKey = array_unique(array_diff($normalizedRecord[$innerKey], $innerValue));
                    }
                }
            }
            // we allow relations to be provided as a comma separated string also
            // that should be converted to an array for notifier, audit, postman to correctly identify them as a relation field
            elseif (array_key_exists($key, $normalizedRecord) && is_array($normalizedRecord[$key]) && $key != 'recordTags') {
                if (is_string($value)) {
                    $value = $this->parseBadRelationData($value);
                }
                if ($key != 'recordTags' && is_array($value) && !empty($value) && is_string($value[0]) && !str_contains($value[0], '/api/')) {
                    $value = $this->convertUUIDToIRI($value, $key, $entityType, $attributes);
                }
                $requestBody->$key = $value;
            }
        }

        if ($replicationOp == 'PUT') {
            $requestBody->{'modifyDate'} = array_key_exists('modifyDate', $normalized_current_data) ? $normalized_current_data['modifyDate'] : null;
            $requestBody->{'modifyUser'} = array_key_exists('modifyUser', $normalized_current_data) ? $normalized_current_data['modifyUser'] : null;
        }
        if (isset($requestBody->{"@id"})) {
            unset($requestBody->{"@id"});
        }
        return $requestBody;
    }

    private function parseBadRelationData($relationString)
    {
        $value = trim($relationString);
        if (strlen($value) > 0) {
            return explode(",", $value);
        } else {
            return [];
        }
    }

    private function replaceLookupWithIri(&$jsonData)
    {
        if (!is_null($jsonData) && is_array($jsonData)) {
            foreach ($jsonData as $key => $value) {
                if (is_array($value)) {
                    if (array_key_exists('@id', $value)) {
                        $jsonData[$key]  = $value['@id'];
                    } else {
                        // it could be a collection
                        $modifiedValues = [];
                        foreach ($value as $index => $item) {
                            if (is_array($item) && array_key_exists('@id', $item)) {
                                // this is an fsr entity
                                $this->replaceLookupWithIri($item);
                                array_push($modifiedValues, $item);
                            }
                        }
                        if (count($modifiedValues) > 0) {
                            $jsonData[$key] = $modifiedValues;
                        }
                    }
                }
            }
        }
    }

    private function convertUUIDToIRI($uuidList, $key, $entityType, &$attributes)
    {
        $iriList = [];
        $relationType = null;
        if (is_null($attributes)) {
            //lazy load only if relations have UUID and send back to caller
            $modelMetaDataRepository = $this->entityManager->getRepository(ModelMetadata::class);
            $modelMetaDataForType = $modelMetaDataRepository->findOneBy(['type' => $entityType]);
            if (is_null($modelMetaDataForType)) {
                # could be a system module
                $systemFixturesFile = file_get_contents('/opt/cyops-api/src/StaticFile/SystemFixtures.json');
                $systemFixture = json_decode($systemFixturesFile);
                $modelMetaDataForType = $systemFixture->{$entityType};
                $attributes = $modelMetaDataForType->{'attributes'};
            } else {
                $attributes = $modelMetaDataForType->getAttributes();
            }
        }
        if (is_null($attributes) || empty($attributes)) {
            return $uuidList;
        }
        if ($attributes[0]::class === AttribModelMetadata::class) {
            foreach ($attributes as $eachAttribute) {
                if ($eachAttribute->getName() == $key) {
                    $relationType = $eachAttribute->getType();
                    break;
                }
            }
        } else {
            foreach ($attributes as $eachAttribute) {
                if ($eachAttribute->{'name'} == $key) {
                    $relationType = $eachAttribute->{'type'};
                    break;
                }
            }
        }

        if ($relationType == 'object') {
            return $uuidList;
        }

        foreach ($uuidList as $uuid) {
            array_push($iriList, '/api/3/' . $relationType . '/' . $uuid);
        }
        return $iriList;
    }


    public function sendData(array $data, $key = self::DEFALUT_KEY)
    {
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->setContentType('application/json');
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->publish(json_encode($data), $key);
    }

    public function getUser()
    {
        return $this->tokenStorage->getToken()->getUser();
    }

    public function getUserName()
    {
        $user = $this->getUser();
        if (method_exists($user, 'getFirstname')) {
            return $user->getFirstname() . ' ' . $user->getLastname();
        }
        return $user->getName();
    }

    public static function extractEntityType($previousObject, $updatedObject, $identity)
    {
        if (is_array($updatedObject) && array_key_exists($identity, $updatedObject)) {
            return $updatedObject[$identity];
        } elseif (is_array($previousObject) && array_key_exists($identity, $previousObject)) {
            return $previousObject[$identity];
        }
        return null;
    }

    /**
     * @param $updatedObject
     * @param $previousObject
     * @param $method
     * @param array $changeData
     * @return array
     */
    public function prepareChangeData(
        $updatedObject,
        $previousObject,
        $method,
        $changeData = [],
        $entity = null,
        $softDelete = false,
        $isRestore = false,
        $title = null
    ) {
        $delta_data = [];
        // Gateway expect data to be coming inside data key
        if ($method == 'Delete') {
            $delta_data['softDelete'] = $softDelete;
        } elseif ($method == 'Update') {
            $delta_data['restore'] = $isRestore;
        }
        $delta_data['oldData'] = $previousObject;
        $delta_data['newData'] = $updatedObject;
        $delta_data['changeData'] = $changeData;
        $userName = $this->getUserName();
        $userId = $this->getUser()->getUuid();
        $entityType = self::extractEntityType($previousObject, $updatedObject, "@type");
        $entityId = $this->getEntityUuid($updatedObject, $previousObject);
        return [
            AppConstants::DELTA_DATA => $delta_data,
            AppConstants::HTTP_METHOD => $method,
            AppConstants::REMOTE_ADDR => $_SERVER['HTTP_X_FORWARDED_FOR'] ?? $_SERVER['REMOTE_ADDR'] ?? '127.0.0.1',
            AppConstants::WEB_SOCKET_SESSION_ID => $_SERVER['HTTP_WEBSOCKET_SESSIONID'] ?? null,
            AppConstants::COMPONENT => 'crudhub',
            AppConstants::ENTITY_TYPE => $entityType,
            AppConstants::ENTITY_UUID => $entityId,
            AppConstants::USER_UUID => $userId,
            AppConstants::TRANSACTION_DATE => round(microtime(true) * 1000),
            AppConstants::USER => $userName,
            AppConstants::PLAYBOOK_IRI => $_SERVER['HTTP_X_CYOPS_PLAYBOOKIRI'] ?? null,
            AppConstants::PLAYBOOK_NAME => $_SERVER['HTTP_X_CYOPS_PLAYBOOKNAME'] ?? null,
            AppConstants::TITLE => $title
        ];
    }

    private function getEntityUuid($updatedObject, $previousObject)
    {
        if (is_array($updatedObject)) {
            return $updatedObject['uuid'];
        }
        if (is_object($previousObject)) {
            return $previousObject->{'uuid'};
        } else {
            return $previousObject['uuid'];
        }
    }
}
