<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use App\Traits\Ownable;
use App\Traits\Trackable;
use App\Utility\MsspUtility;
use App\Entity\Custom\Tenant;
use App\Traits\PeerReplicable;
use App\Entity\Base\BaseEntity;
use App\Service\UtilityService;
use App\Constants\CacheConstants;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\KernelEvents;
use ApiPlatform\Core\Api\IriConverterInterface;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\HttpKernel\Exception\HttpException;
use ApiPlatform\Core\DataPersister\DataPersisterInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

class MsspEventSubscriber implements EventSubscriberInterface
{
    protected $requeststack;
    private const MASTER_ROLE_NAME = 'master';
    private const SELF_TENANT_IRI = 'self_tenant_iri';

    public function __construct(
        protected ContainerInterface $containers,
        protected TokenStorageInterface $tokenStorage,
        protected EntityManagerInterface $entityManager,
        protected DataPersisterInterface $dataPersister,
        protected TagAwareCacheInterface $cache,
        private UtilityService $utilityService,
        private MsspUtility $msspUtility,
        private IriConverterInterface $iriConverter
    ) {
    }

    public static function getSubscribedEvents()
    {
        return [
            KernelEvents::VIEW => ['handleObjectForMsspChanges', 34]
        ];
    }

    public function handleObjectForMsspChanges(ViewEvent $event)
    {
        $object = $event->getControllerResult();
        $request = $event->getRequest();
        $requestMethod = $request->getMethod();

        if (!($object instanceof BaseEntity) || $requestMethod == Request::METHOD_GET || $requestMethod == Request::METHOD_DELETE) {
            return;
        }
        $previousData = $request->attributes->get('previous_data');
        $fsrPreviousData = $request->attributes->get('fsr_previous_data');
        return $this->msspOperations($request, $object, $previousData, $fsrPreviousData, $event);
    }

    protected function msspOperations($request, $object, $previousData, $fsrPreviousData, $event)
    {
        $isDistributedInstance = $this->utilityService->getRequiredLicenseDetails(CacheConstants::IS_DISTRIBUTED_INSTANCE);
        if (!$isDistributedInstance) {
            return;
        }
        if (is_null($previousData)) {
            // POST Request
            $this->setTenantObject($object);
        } else {
            // PUT REQUEST
            $requestBody  = json_decode($request->getContent());
            $entityClass = $object::class;
            //Set Owners to Records when tenant is changed.

            if ($this->isPeerReplicable($entityClass) && is_null($object->getTenant())) {
                $object->setTenant($previousData->getTenant());
            }

            if (
                $this->isPeerReplicable($entityClass) && $this->isOwnable($entityClass) &&
                $previousData->getTenant()->getUuid() != $object->getTenant()->getUuid()
            ) {
                $tenantObject = $object->getTenant();
                foreach ($tenantObject->getOwners() as $owner) {
                    $object->addOwner($owner);
                }
            }

            $conflict = $this->isConflicts($object, $previousData, $requestBody);
            if ($conflict) {
                $response = new Response(json_encode($fsrPreviousData));
                $event->setControllerResult($previousData);
                $event->setResponse($response);
                $diffObj = $this->msspUtility->prepareReplicationDiffData(
                    $fsrPreviousData,
                    'PUT',
                    ['conflict' => true, 'tenant' => $fsrPreviousData['tenant']],
                    ['conflict' => true],
                    false
                );
                $request->attributes->set('fsr_replication_data', $diffObj);
                return $response;
            }
        }
    }

    private function isConflicts($object, $previousData, $requestBody = null)
    {
        if (isset($requestBody->{'__replicateData'})) {
            return false;
        }
        $entityClass = $object::class;
        if ($this->isPeerReplicable($entityClass) && $this->isTrackable($entityClass)) {
            $instanceRole = $this->utilityService->getRequiredLicenseDetails(CacheConstants::INSTANCE_ROLE);
            $modifyUserUuidOfRequestedObj = $this->tokenStorage->getToken()->getUser()->getUuid();
            $modifyUserUuidOfExistingObj = $previousData->getModifyUser() ? $previousData->getModifyUser()->getUuid() : null;
            $modifyDateOfExistingObj = is_null($previousData->getModifyDate()) ? 0 : $previousData->getModifyDate()->getTimestamp();
            $modifyDateOfRequestedObj = $requestBody->{'modifyDate'} ?? PHP_INT_MAX;
            if ($instanceRole != self::MASTER_ROLE_NAME && ($modifyUserUuidOfRequestedObj != $modifyUserUuidOfExistingObj) && ($modifyDateOfExistingObj >= $modifyDateOfRequestedObj)) {
                return true;
            }
            return false;
        }
    }

    public function isPeerReplicable($entityClass)
    {
        return in_array(PeerReplicable::class, class_uses($entityClass));
    }

    public function isTrackable($entityClass)
    {
        return in_array(Trackable::class, class_uses($entityClass));
    }

    public function isOwnable($entityClass)
    {
        return in_array(Ownable::class, class_uses($entityClass));
    }

    private function getSelfTenantObject()
    {
        $item = $this->cache->getItem(self::SELF_TENANT_IRI);
        if (!$item->isHit()) {
            $resourceClass = Tenant::class;
            $tenantRepo = $this->entityManager->getRepository($resourceClass);
            $selfTenantObj = $tenantRepo->findOneBy(["role" => "self"]);
            $item->set($this->iriConverter->getIriFromItem($selfTenantObj));
            $item->expiresAfter(86400);
            $item->tag(self::SELF_TENANT_IRI);
            $this->cache->save($item);
            return $selfTenantObj;
        }
        return $this->iriConverter->getItemFromIri($item->get(), ['fetch_data' => false]);
    }

    private function setTenantObject($object)
    {
        $entityClass = $object::class;
        // Check if Peer Replicable is enabled for class
        if ($this->isPeerReplicable($entityClass)) {
            // Set Self Tenant ID for Objects if not already came as part of requests
            if (!$object->getTenant() || is_null($object->getTenant())) {
                $selfTenantObject = $this->getSelfTenantObject();
                if (is_null($selfTenantObject)) {
                    throw new HttpException(Response::HTTP_BAD_REQUEST, "Self TenantId not defined for instance");
                }
                $object->setTenant($selfTenantObject);
            }
            //Set Owners to Records
            if ($this->isOwnable($entityClass)) {
                $tenantObject = $object->getTenant();
                foreach ($this->tokenStorage->getToken()->getUser()->getTeams() as $owner) {
                    $object->addOwner($owner);
                }
                foreach ($tenantObject->getOwners() as $owner) {
                    $object->addOwner($owner);
                }
            }
            $object->setConflict(false);
        }

        return $object;
    }
}
