<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use Error;
use App\Entity\Actors\Actor;
use Psr\Log\LoggerInterface;
use App\Entity\Custom\Person;
use App\Entity\Base\BaseEntity;
use App\Service\UtilityService;
use Doctrine\ORM\EntityManager;
use App\Entity\Settings\SystemSettings;
use Doctrine\ORM\EntityManagerInterface;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\HttpFoundation\Request;
use GuzzleHttp\Exception\BadResponseException;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use ApiPlatform\Core\EventListener\EventPriorities;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use App\Entity\Authorization\Role;
use App\Constants\ErrorConstants;

class PeopleSubscriber implements EventSubscriberInterface
{
    /** @var EntityManager */
    private $entityManager;
    /** @var ContainerInterface */
    private $container;
    /** @var UtilityService  */
    private $utilityService;

    // keeping same constant as in DAS 
    const FSR_USER_CREATE = "User Create";

    public function __construct(
        EntityManagerInterface $entityManager,
        ContainerInterface $container,
        UtilityService $utilityService,
        protected LoggerInterface $logger
    ) {
        $this->entityManager = $entityManager;
        $this->container = $container;
        $this->utilityService = $utilityService;
    }

    public static function getSubscribedEvents()
    {
        // Adding this on POST WRITE PRIORITY in order so data is persisted in database and then handling of data can be done.
        return [
            KernelEvents::VIEW => [
                ['handlePeopleCreate', EventPriorities::POST_WRITE],
                ['handlePeoplePreCreate', EventPriorities::PRE_WRITE]
            ]
        ];
    }

    // Send mail to user on create of People.
    public function handlePeopleCreate(ViewEvent $event)
    {
        $dasClient = $this->container->get("app.proxy.client.das");
        $integrationClient = $this->container->get("app.proxy.client.integrations");
        $request = $event->getRequest();
        $requestData = json_decode($request->getContent(), true);
        $method = $request->getMethod();
        $object = $event->getControllerResult();
        $previousData = $request->attributes->get('previous_data');

        if (!$method || $method == Request::METHOD_GET || $method == Request::METHOD_DELETE || !$object instanceof BaseEntity) {
            return;
        }
        $entityClass = get_class($object);
        if ($entityClass != Person::class) {
            return;
        }
        if (!is_null($previousData)) {
            return;
        }
        $dasUri = $this->container->getParameter('das.uri');
        $headers = array(
            "Content-type" => "application/json",
        );
        $keyFile = file_get_contents('/opt/cyops-api/src/keys/daspublic.key', true);
        $idKey = base64_encode($keyFile);
        $objectData = [];
        $objectData['idKey'] = $idKey;
        $objectData['uuid'] = $object->getUserId();
        $dasRequests = new GuzzleRequest('post', $dasUri . '/token?uuid=' . $object->getUserId(), $headers, json_encode($objectData));
        try {
            $dasResponse = $dasClient->send($dasRequests);
            $dasResponseData = $dasResponse->getBody()->getContents();
            $jsonDecodeResponse = json_decode($dasResponseData, true);
            $resCode = $dasResponse->getStatusCode();
        } catch (BadRequestHttpException | \Exception $error) {
            $this->logger->error($error->getMessage());
            $errVar = json_decode($error->getResponse()->getBody()->read(2048), true);
            $errMsg['message'] = $errVar['Error'];
            $errCode = $error->getCode();
            $this->logger->error($errMsg['message']);
            $this->logger->error($errCode);
            return;
        }
        $actorEmail = $object->getEmail();

        if ($resCode < 300) {
            $systemRepository = $this->entityManager->getRepository(SystemSettings::class);
            $systemSetting = $systemRepository->findAll();
            $global_value = $systemSetting[0]->getGlobalValues();
            $system_host = isset($global_value['hostname']) ? $global_value['hostname'] : getenv('HOSTNAME');

            $wfClient = $this->container->get("app.proxy.client.workflow");
            $headers = array(
                "Content-type" => "application/json",
            );
            $workflowUri = $this->container->getParameter('workflow.uri');
            $workflowRequests = new GuzzleRequest('get', $workflowUri . '/api/dynamic-variable/?format=json&name=Server_fqhn', $headers);
            try {
                $workflowResponse = $wfClient->send($workflowRequests);
                $workflowResponseData = json_decode($workflowResponse->getBody()->getContents(), true);
                if ($workflowResponseData["hydra:totalItems"] > 0) {
                    $system_host = $workflowResponseData["hydra:member"][0]["value"];
                }
            } catch (BadRequestHttpException | \Exception $error) {
                $this->logger->error($error->getMessage());
            }

            $dataToReplace = [];
            $dataToReplace['{{vars.userId}}'] = $requestData['username'] ?? $requestData['user']['loginid'];
            $dataToReplace['{{vars.token}}'] = $jsonDecodeResponse['token'];
            $dataToReplace['{{vars.request.baseUri}}'] = $request->getScheme() . "://" . $system_host;
            $dataToReplace['{{@Server_fqhn}}'] = $request->getScheme() . "://" . $system_host;
            $dataToReplace['{{@cyops_hostname}}'] = $request->getScheme() . "://" . $system_host;
            $mailDetails = [];
            $mailDetails['to'] = $actorEmail;
            $defaultMessage = 'newpassword';
            $mailResponse = $this->utilityService->sendMail(
                '12f23d8d-505c-4b4b-8911-6595c99b07f0',
                $dataToReplace,
                $mailDetails,
                $integrationClient,
                $defaultMessage
            );
            if ($mailResponse === true) {
                $sucessMsg['message'] = "Reset token sent to user";
                $this->logger->error($sucessMsg['message']);
            } else {
                $sucessMsg['message'] = "Reset token generated,mail not sent due to smtp connector failure";
                $this->logger->error($sucessMsg['message']);
            }
        }
    }

    public function handlePeoplePreCreate(ViewEvent $event)
    {
        $dasClient = $this->container->get("app.proxy.client.das");
        $request = $event->getRequest();
        $method = $request->getMethod();
        $object = $event->getControllerResult();

        if (!$method || $method == Request::METHOD_GET || $method == Request::METHOD_DELETE || $method == Request::METHOD_PUT || !$object instanceof BaseEntity) {
            return;
        }

        $entityClass = get_class($object);
        if ($entityClass != Person::class) {
            return;
        }

        $requestData = json_decode($request->getContent(), true);
        $rolesFromRequestData = $requestData['roles'];

        $this->validateUuid(Role::class, $rolesFromRequestData);

        if ($requestData['user']) {

            $dasUri = $this->container->getParameter('das.uri');
            $headers = array(
                "Content-type" => "application/json",
            );

            $accessType = $requestData['user']['accessType'] ?? 'Concurrent';

            if (!in_array($accessType, ['Concurrent', 'Named'])) {
                $accessType = 'Concurrent';
            }

            $body = [
                'credentials' => [
                    'loginid' => $requestData['user']['loginid']
                ],
                'status' => array_key_exists('status', $requestData['user']) ? $requestData['user']['status'] : 2,
                'type' => array_key_exists('user_type', $requestData['user']) ? $requestData['user']['user_type'] : 2,
                'domain' => $requestData['user']['domain'],
                'nfa' => array_key_exists('nfa', $requestData['user']) ? $requestData['user']['nfa'] : [],
                'bind_name' => $requestData['user']['bind_name'],
                'tenant' => $requestData['user']['tenant'],
                'accessType' => $accessType,
                'event_type' => self::FSR_USER_CREATE
            ];
            $dasRequests = new GuzzleRequest('post', $dasUri . '/users', $headers, json_encode($body));
            try {
                $dasResponse = $dasClient->send($dasRequests);
            } catch (BadResponseException | \Exception $exception) {
                $this->logger->error($exception->getMessage());
                if (method_exists($exception, 'getResponse')) {
                    throw new BadRequestHttpException($exception->getResponse()->getBody()->read(2048));
                }
                throw new BadRequestHttpException($exception->getMessage());
            }
            $dasResponseData = $dasResponse->getBody()->getContents();
            $jsonDecodeResponse = json_decode($dasResponseData, true);
            $actorRepo = $this->entityManager->getRepository(Actor::class);
            $user_id = $jsonDecodeResponse['uuid'];
            if ($actorRepo->findOneBy(["userId" => $user_id])) {
                $event->stopPropagation();
                throw new BadRequestHttpException(sprintf('The user with loginID %s already exists',  $requestData['user']['loginid']));
            }
            $object->setUserId($user_id);
        }
    }

    private function validateUuid($resource, $filterCriteria)
    {
        $repository = $this->entityManager->getRepository($resource);
        $result = $this->utilityService->getItemWithWhere($repository, ['uuid' => $filterCriteria]);
        if (count($result) < count($filterCriteria)) {
            $explodedResource = explode("\\", $resource);
            $className = end($explodedResource);
            $message = sprintf(ErrorConstants::INVALID_UUID_PROVIDED, strtolower($className));
            throw new BadRequestHttpException($message);
        }
    }
}
