<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

declare(strict_types=1);

namespace App\EventSubscriber;

use ApiPlatform\Core\Bridge\Symfony\Bundle\Test\Response;
use ApiPlatform\Core\EventListener\EventPriorities;
use ApiPlatform\Core\Metadata\Property\Factory\PropertyMetadataFactoryInterface;
use App\Constants\AttributeLevelOperators;
use App\Entity\Base\BaseEntity;
use App\Entity\Core\Picklist;
use App\Service\MetadataHelper;
use App\Service\NormalizerUtilityService;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * Final serializer that defines the response sent back as part of the API call. 
 * We normalize data in custom formats so that the event subscribers get all attributes they are looking for
 * But the final response should be consistent across modules and apis
 */
class ResponseSerializer implements EventSubscriberInterface
{
    private $metadataHelper;
    private $propertyMetadataFactory;
    /** @var NormalizerUtilityService */
    private $normalizerUtilityService;

    public function __construct(MetadataHelper $metadataHelper, PropertyMetadataFactoryInterface $propertyMetadataFactory, NormalizerUtilityService $normalizerUtilityService)
    {
        $this->metadataHelper = $metadataHelper;
        $this->propertyMetadataFactory = $propertyMetadataFactory;
        $this->normalizerUtilityService = $normalizerUtilityService;
    }

    public static function getSubscribedEvents()
    {
        return [
            KernelEvents::VIEW => ['createFinalJson', EventPriorities::PRE_RESPOND] // all other FSR subscribers to have higher priority so that this runs at the end
        ];
    }

    public function createFinalJson(ViewEvent $event): void
    {
        $controllerResult = $event->getControllerResult();
        $recordData = $event->getRequest()->attributes->get('data');
        $request = $event->getRequest();

        if ($controllerResult instanceof Response || !($request->attributes->getBoolean('_fsr_dynamic_context', false))) {
            return;
        }

        // skip serialization for perf intensive bulk inserts
        if ($this->normalizerUtilityService->getOption('statusOnly', [], false, true)){
            $event->setControllerResult(json_encode(array('status' => 'success')));
            return;
        }

        try {
            $decodedResponse = json_decode($controllerResult, true);
            $is_list = false;
            if (gettype(array_key_first($recordData)) == "string") {
                $item = $recordData[0];
                $is_list = true;
            } else {
                $item = $recordData;
            }

            $entity = get_class($item);
            if (!($item instanceof BaseEntity)) {
                return;
            }
            $entityFieldMap = $this->metadataHelper->getEntityFields($entity) ?? null;
            $key = AttributeLevelOperators::OPERATOR_TOKEN . 'relationships';
            $rawOption = $request->query->get($key);
            if (is_null($rawOption)) {
                $rawOption = $request->request->get($key, false);
            }
            $relationships = filter_var($rawOption, FILTER_VALIDATE_BOOLEAN);

            if ($is_list) {
                foreach ($decodedResponse as $eachRecord) {
                    $this->serializeRecord($eachRecord, $entity, $entityFieldMap, $relationships);
                }
            } else if ($decodedResponse) {
                $this->serializeRecord($decodedResponse, $entity, $entityFieldMap, $relationships);
            }
            $event->setControllerResult(json_encode($decodedResponse));
        } catch (\Throwable $e) {
            // TODO
        }
    }

    private function serializeRecord(&$decodedResponse, $entity, $entityFieldMap, $relationships)
    {
        // html escape
        foreach ($decodedResponse as $key => $value) {
            if (is_null($value)) {
                continue;
            }
            if ($entityFieldMap && count($entityFieldMap['htmlEscapeFields']) && in_array($key, $entityFieldMap['htmlEscapeFields'])) {
                $decodedResponse[$key] = htmlentities(htmlspecialchars_decode($value));
            }
            if (!is_array($value)) {
                continue;
            }
            // for non-relationship call, remove relation data, except tags and multiselect picklists
            if ($relationships === false) {
                $propertyMetadataLocal = $this->propertyMetadataFactory->create(
                    $entity,
                    $key,
                    []
                );
                $type = $propertyMetadataLocal->getType();
                if ($type && $type->isCollection()) {
                    $collectionType = $type->getCollectionValueType()->getClassName();
                    if (
                        $collectionType !== \App\Entity\Core\Tag::class && $collectionType !== Picklist::class
                        && is_subclass_of($collectionType, BaseEntity::class)
                    ) {
                        unset($decodedResponse[$key]);
                    }
                }
            }
        }
    }
}
