<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use App\Constants\CacheConstants;
use InvalidArgumentException;
use App\Entity\Base\BaseEntity;
use Doctrine\Inflector\InflectorFactory;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\KernelEvents;
use App\Entity\ModelMetadata\AttributeMetadata;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use App\Entity\ModelMetadata\StagingModelMetadata;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class StagingMetaDataSubscriber implements EventSubscriberInterface
{
    private $inflector;

    public function __construct(
        protected EntityManagerInterface $entityManager,
        protected TagAwareCacheInterface $cache
    ) {
        $this->inflector = InflectorFactory::create()->build();
    }
    public function handleStagingM2MUpdate(ViewEvent $event)
    {
        $newData = $event->getRequest()->attributes->get('data');
        $method = $event->getRequest()->getMethod();
        if (Request::METHOD_GET == $method || Request::METHOD_DELETE == $method || is_null($newData) || !($newData instanceof BaseEntity) || !($newData instanceof StagingModelMetadata)) {
            return;
        }

        $moduleName =  $newData->getDescriptions()['singular'];
        $moduleName = preg_replace('/[^a-z0-9\s]+/i', '', $moduleName);
        $escapedModuleName = $this->inflector->classify($moduleName);
        $escapedPlural = $this->inflector->pluralize($moduleName);
        if (!isset($newData->getDescriptions()['plural'])) {
            $description = [];
            $description['plural'] = $escapedPlural;
            $description['singular'] = $moduleName;
            $newData->setDescriptions($description);
        }
        if ($newData->getType() == null || $newData->getType() == '') {
            $newData->setType(\strtolower($this->inflector->tableize($escapedModuleName)));
        }
        if ($newData->getTableName() == null || $newData->getTableName() == '') {
            $newData->setTableName(\strtolower($this->inflector->tableize($escapedModuleName)));
        }

        $stagingRepo = $this->entityManager->getRepository(StagingModelMetadata::class);
        $attributeRepo = $this->entityManager->getRepository(AttributeMetadata::class);
        foreach ($newData->getAttributes() as $value) {
            if ($value->getName() == null || $value->getName() == '') {
                $apiKey = preg_replace('/[^a-z0-9\s]+/i', '', $value->getDescriptions()['singular']);
                $value->setName($this->inflector->classify($apiKey));
            }
            $relatedModel = !empty($value->getDataSource()) && array_key_exists('model', $value->getDataSource()) ? $value->getDataSource()['model'] : null;
            if (in_array($value->getFormType(), ['oneToMany']) && ($value->getInversedField() == null || $value->getInversedField() == '')) {
                $value->setInversedField($this->inflector->camelize($newData->getType()));
            }
            if ($value->getFormType() == 'manyToMany' && ($value->getInversedField() == null || $value->getInversedField() == '')) {
                if ($newData->getType() != $value->getType()) {
                    $stagingMetadata = $stagingRepo->findOneBy(array("type" => $relatedModel));
                    if (is_null($stagingMetadata)) {
                        continue;
                    }
                    $existingAttrib = $attributeRepo->findOneBy(array('sattrib' => $stagingMetadata->getUuid(), 'formType' => 'manyToMany', 'type' => $newData->getType()));
                    if ($existingAttrib) {
                        $value->setInversedField($existingAttrib->getName());
                        $value->setOwnsRelationship(!$existingAttrib->getOwnsRelationship());
                        continue;
                    }
                    // Few Content Pack MMD itself are having invalid value of inverseField, so better to let framework take call for safety of same
                    // We need more through review of MMD in Content Packs
                    if (is_null($value->getInversedField())) {
                        $value->setInversedField($this->inflector->camelize($newData->getType()));
                    }
                } else {
                    $value->setInversedField($value->getName());
                    $value->setOwnsRelationship(true);
                    continue;
                }
                if ($stagingMetadata) {
                    try {
                        $relatedField = new AttributeMetadata();
                        $relatedField->setDataSource(["model" => $newData->getType()]);
                        $relatedField->setDescriptions(["singular" => $newData->getDescriptions()['plural']]);
                        $relatedField->setDisplayName("{{" . $newData->getDescriptions()['plural'] . "}}");
                        $relatedField->setFormType("manyToMany");
                        $relatedField->setInversedField($value->getName());
                        $relatedField->setName($this->inflector->camelize($newData->getType()));
                        $relatedField->setType($newData->getType());
                        $relatedField->setCollection(true);
                        $value->setOwnsRelationship(true);
                        $relatedField->setOwnsRelationship(false);
                        $relatedField->setPeerReplicable(true);
                        $relatedField->setVisibility(true);
                        $relatedField->setReadable(true);
                        $relatedField->setHtmlEscape(false);
                        $relatedField->setOrderIndex(is_countable($stagingMetadata->getAttributes()) ? count($stagingMetadata->getAttributes()) : 0);
                        $stagingMetadata->addAttribute($relatedField);
                        $this->entityManager->persist($stagingMetadata);
                        $this->entityManager->flush();
                    } catch (\Exception | BadRequestHttpException $e) {
                        throw new BadRequestHttpException("Field name " . $newData->getDescriptions()['singular'] . " already exists on the module " . $stagingMetadata->getDescriptions()['singular'] . ". Remove the existing field.");
                    }
                }
            } elseif ($value->getFormType() == 'manyToMany' && $newData->getType() != $value->getType()) {
                $relatedModel = $value->getDataSource()['model'];
                $stagingMetadata = $stagingRepo->findOneBy(["type" => $relatedModel]);
                if ($stagingMetadata) {
                    $existingAttrib = $attributeRepo->findOneBy(['sattrib' => $stagingMetadata->getUuid(), 'formType' => 'manyToMany', 'type' => $newData->getType()]);
                    if ($existingAttrib) {
                        $value->setOwnsRelationship(!$existingAttrib->getOwnsRelationship());
                    }
                }
            } elseif ($value->getFormType() == 'manyToMany' && $newData->getType() == $value->getType()) {
                $value->setOwnsRelationship(true);
            }
        }
    }

    public function cascadeStagingDataToModelMetadata(ViewEvent $event)
    {
        $request = $event->getRequest();
        $requestMethod = $request->getMethod();
        $previousData = $request->attributes->get('previous_data');
        $currentData = $event->getRequest()->attributes->get('data');
        if ($requestMethod == Request::METHOD_GET || $requestMethod == Request::METHOD_DELETE || is_null($previousData) || !$currentData instanceof StagingModelMetadata) {
            return;
        }
        if ($currentData && $previousData) {
            $commandToExecute = "INSERT into model_metadatas select * from staging_model_metadatas
                ON CONFLICT (uuid)
                DO UPDATE set
                defaultsort=EXCLUDED.defaultsort,
                system=EXCLUDED.system,
                archivable=EXCLUDED.archivable,
                paused=EXCLUDED.paused,
                partitionby=EXCLUDED.partitionby,
                archivalcriteria=EXCLUDED.archivalcriteria,
                archivalfilters=EXCLUDED.archivalfilters,
                importedBy=EXCLUDED.importedBy";
            $metadataEntityDatabase = $this->entityManager->getConnection()->prepare($commandToExecute);
            $this->cache->invalidateTags([CacheConstants::ENTITY_REPLICATION_MAP]);
            try {
                $metadataEntityDatabase->executeQuery();
            } catch (InvalidArgumentException | \Exception $e) {
                throw new BadRequestHttpException('Failure: ' . $e->getMessage());
            }
        }
    }

    public static function getSubscribedEvents()
    {
        return [
            KernelEvents::VIEW => [
                ['handleStagingM2MUpdate', 67],
                ['cascadeStagingDataToModelMetadata', 14]
            ]
        ];
    }
}
