<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;

use App\Entity\Core\Widget;
use App\Service\UtilityService;
use App\Security\Voter\FsrVoter;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Doctrine\ORM\Event\PreUpdateEventArgs;
use Symfony\Component\HttpFoundation\RequestStack;

class WidgetListener
{
    const WIDGET_PATH = '/opt/cyops-ui/widgets/installed/';
    const WIDGET_DEVELOPMENT_PATH = '/opt/cyops-ui/widgets/development/';
    protected $container;
    protected $utilityService;
    private $requeststack;
    protected $fsrVoter;

    public function __construct(ContainerInterface $containers, UtilityService $utilityService, RequestStack $requeststack, FsrVoter $fsrVoter)
    {
        $this->container = $containers;
        $this->utilityService = $utilityService;
        $this->requeststack = $requeststack;
        $this->fsrVoter = $fsrVoter;
    }

    public function preUpdate(Widget $widget, PreUpdateEventArgs $previousWidgetData)
    {
        $request = $this->requeststack->getCurrentRequest();
        $requestData = json_decode($request->getContent(), true);
        $replace = array_key_exists('replace', $requestData) ? $requestData['replace']: false;
        $replaceVersions = $replace ? $requestData['replaceVersions'] : [];

        $widgetPath = self::WIDGET_PATH . $widget->getName() . '-' . $widget->getVersion();
        $widgetDevelopmentPath = self::WIDGET_DEVELOPMENT_PATH . $widget->getName() . '-' . $widget->getVersion();
        if ($previousWidgetData->hasChangedField('installed')) {
            // Publish Only
            // Old Draft - true - No need to check | New Draft - true - No need to check
            // Old Installed - false | New Installed - true
            // Copy from dev to prod
            if (!$previousWidgetData->getOldValue('installed') && $widget->getInstalled() && $widget->getDraft()) {
                $this->deleteExistingWidgets($widget->getName(), $widget->getVersion(), $replace, $replaceVersions);
                $this->utilityService->removeWidgetDir($widgetPath);
                $this->utilityService->copyWidgetFiles($widgetDevelopmentPath, $widgetPath, $widget->getName(), $widget->getVersion(), true);
                $this->utilityService->syncHANodes($widgetPath, 'copy');
            }

            // Publish and discard
            // Old Draft - true | New Draft - false
            // Old Installed - false | New Installed - true
            // Copy from dev to prod, and delete files from dev
            if (!$previousWidgetData->getOldValue('installed') && $widget->getInstalled() && !$widget->getDraft()) {
                $this->deleteExistingWidgets($widget->getName(), $widget->getVersion(), $replace, $replaceVersions);
                $this->utilityService->removeWidgetDir($widgetPath);
                $this->utilityService->copyWidgetFiles($widgetDevelopmentPath, $widgetPath, $widget->getName(), $widget->getVersion(), true);
                $this->utilityService->removeWidgetDir($widgetDevelopmentPath);
                $this->utilityService->syncHANodes($widgetDevelopmentPath, 'delete');
            }

            // Uninstall
            // Old Draft - true/false - No Need to check | New Draft - true - No need to check
            // Old Installed - true | New Installed - false
            // remove files from widget, if files not in dev folder, copy from prod
            if ($previousWidgetData->getOldValue('installed') && !$widget->getInstalled()) {
                if (!file_exists($widgetDevelopmentPath)) {
                    $this->utilityService->copyWidgetFiles($widgetPath, $widgetDevelopmentPath, $widget->getName(), $widget->getVersion(), false);
                    $this->utilityService->syncHANodes($widgetDevelopmentPath, 'copy');
                }
                $this->utilityService->removeWidgetDir($widgetPath);
                $this->utilityService->syncHANodes($widgetPath, 'delete');
            }
        } else if ($previousWidgetData->hasChangedField('draft')) {
            // Edit Installed, but no draft
            // Old Draft - false | New Draft - true
            // Old Installed - true - No need to check | New Installed - false - No need to check
            // Copy from prod to dev
            if (!$previousWidgetData->getOldValue('draft') && $widget->getDraft()) {
                $this->utilityService->copyWidgetFiles($widgetPath, $widgetDevelopmentPath, $widget->getName(), $widget->getVersion(), false);
                $this->utilityService->syncHANodes($widgetDevelopmentPath, 'copy');
            }

            // Widget already installed, publish and discard case
            else if ($widget->getInstalled() && !$widget->getDraft()) {
                $this->utilityService->copyWidgetFiles($widgetDevelopmentPath, $widgetPath, $widget->getName(), $widget->getVersion(), true);
                $this->utilityService->removeWidgetDir($widgetDevelopmentPath);
                $this->utilityService->syncHANodes($widgetPath, 'copy');
                $this->utilityService->syncHANodes($widgetDevelopmentPath, 'delete');
            }
        }

        // Widget already installed, publish only case 
        else if ($widget->getInstalled() && $widget->getDraft() && !$widget->getEnablePublish()) {
            $this->utilityService->removeWidgetDir($widgetPath);
            $this->utilityService->copyWidgetFiles($widgetDevelopmentPath, $widgetPath, $widget->getName(), $widget->getVersion(), true);
            $this->utilityService->syncHANodes($widgetPath, 'copy');
        }

        // Widget already installed, making changes in dvelopment
        else if ($widget->getEnablePublish()) {
            $this->utilityService->syncHANodes($widgetDevelopmentPath, 'copy');
        }
        // Edit Installed, with draft
        // Old Draft - true | New Draft - true
        // Old Installed - true | New Installed - false
        // No action needed 
        // Draft only - No Action
    }

    public function postUpdate(Widget $widget)
    {
        $this->sendWidgetCUDRequestMessage(["name" => $widget->getName(), "version" => $widget->getVersion()], 'Update');
    }

    public function postRemove(Widget $widget)
    {
        $widgetPath = self::WIDGET_PATH . $widget->getName() . '-' . $widget->getVersion();
        $widgetDevelopmentPath = self::WIDGET_DEVELOPMENT_PATH . $widget->getName() . '-' . $widget->getVersion();

        $this->utilityService->removeWidgetDir($widgetPath);
        $this->utilityService->removeWidgetDir($widgetDevelopmentPath);
        $this->utilityService->syncHANodes($widgetPath, 'delete');
        $this->utilityService->syncHANodes($widgetDevelopmentPath, 'delete');
        $normalizedWidget = [];
        $normalizedWidget['name'] = $widget->getName();
        $normalizedWidget['version'] = $widget->getVersion();
        $normalizedWidget['uuid'] = $widget->getUuid();
        $this->sendWidgetCUDRequestMessage($normalizedWidget, 'Delete');
    }

    protected function deleteExistingWidgets($widgetName, $widgetVersion, $replace, $replaceVersions)
    {
        $deleteWidgetVersions = false;
        /** @var EntityManager $entityManager */
        $entityManager = $this->container->get('doctrine')->getManagerForClass(Widget::class);
        $widgetRepository = $entityManager->getRepository(Widget::class);
        $serializer = $this->container->get('serializer');

        if ($replace) {
            if (!$this->fsrVoter->isGranted("delete.widgets")) {
                throw new AccessDeniedException();
            }
            if (empty($replaceVersions)) {
                $versionDeleteList = [];
                $widget = $widgetRepository->findBy(array('name' => $widgetName));
                $allWidgetList = $serializer->normalize($widget, 'jsonld');
                foreach ($allWidgetList as $w) {
                    if ($w['version'] !== $widgetVersion) {
                        array_push($versionDeleteList, $w['version']);
                    }
                }
                $replaceVersions = $versionDeleteList;
                $deleteWidgetVersions = true;
            } else {
                $deleteWidgetVersions = true;
            }
        } else {
            return;
        }

        if ($deleteWidgetVersions) {
            $queryBuilder = $widgetRepository->createQueryBuilder('w');
            $deleteWidgetCount = $serializer->normalize($queryBuilder
                ->delete()
                ->andWhere('w.name in (:widgetName)')
                ->setParameter('widgetName', $widgetName)
                ->andWhere('w.version in (:versions)')
                ->setParameter('versions', $replaceVersions)
                ->getQuery()
                ->getResult(), 'jsonld');
        }

        if ($deleteWidgetCount > 0) {
            foreach ($replaceVersions as $v) {
                $p = self::WIDGET_PATH . $widgetName . '-' . $v;
                $this->utilityService->removeWidgetDir($p);
                $this->utilityService->syncHANodes($p, 'delete');
            }
        }
    }

    private function sendWidgetCUDRequestMessage(array $data, $operation = "", $key = 'key.api.notify')
    {
        $message = [
            "data" => $data,
            "operation" => $operation,
            "entityType" => "Widget",
            "entitySingularName" => "widget"
        ];
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->setContentType('application/json');
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->publish(json_encode([$message]), $key);
    }
}
