<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

/*
 * This file is part of the API Platform project.
 *
 * (c) Kévin Dunglas <dunglas@gmail.com>
 *
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */

declare(strict_types=1);

namespace App\Providers;

use App\Entity\Query\Query;
use App\Traits\SoftDeleteable;
use Doctrine\ORM\QueryBuilder;
use App\Service\UtilityService;
use App\Service\OwnershipChecker;
use Doctrine\ORM\EntityManagerInterface;


use Doctrine\Persistence\ManagerRegistry;
use Doctrine\Persistence\Mapping\ClassMetadata;
use ApiPlatform\Core\Exception\RuntimeException;
use ApiPlatform\Core\Bridge\Doctrine\Orm\ItemDataProvider;
use ApiPlatform\Core\Identifier\IdentifierConverterInterface;
use ApiPlatform\Core\DataProvider\RestrictedDataProviderInterface;
use ApiPlatform\Core\Bridge\Doctrine\Common\Util\IdentifierManagerTrait;
use ApiPlatform\Core\Bridge\Doctrine\Orm\Extension\QueryCollectionExtensionInterface;
use ApiPlatform\Core\DataProvider\DenormalizedIdentifiersAwareItemDataProviderInterface;

/**
 * Collection data provider for the Doctrine ORM.
 *
 * @author Kévin Dunglas <dunglas@gmail.com>
 * @author Samuel ROZE <samuel.roze@gmail.com>
 * @final
 */
class CustomItemDataProvider extends ItemDataProvider implements DenormalizedIdentifiersAwareItemDataProviderInterface, RestrictedDataProviderInterface
{

    use IdentifierManagerTrait;

    private $managerRegistry;
    private $itemExtensions;
    private $utilityService;
    private $entityManager;
    protected $ownershipChecker;

    /**
     * @param QueryCollectionExtensionInterface[] $itemExtensions
     */
    public function __construct(
        ManagerRegistry $managerRegistry,
        UtilityService $utilityService,
        EntityManagerInterface $entityManager,
        OwnershipChecker $ownershipChecker,
        iterable $itemExtensions = []
    ) {
        $this->managerRegistry = $managerRegistry;
        $this->itemExtensions = $itemExtensions;
        $this->utilityService = $utilityService;
        $this->entityManager = $entityManager;
        $this->ownershipChecker = $ownershipChecker;
    }

    public function supports(string $resourceClass, string $operationName = null, array $context = []): bool
    {
        return $this->managerRegistry->getManagerForClass($resourceClass) instanceof EntityManagerInterface;
    }

    public function getItem(string $resourceClass, $id, string $operationName = null, array $context = [])
    {

        if ($this->isSoftDeleteable($resourceClass) && array_key_exists(Query::SHOW_DELETED, $_GET) && array_key_exists('softdeleteable', $this->entityManager->getFilters()->getEnabledFilters())) {
            $this->entityManager->getFilters()->disable('softdeleteable');
        }

        if ((\is_int($id) || \is_string($id)) && !($context[IdentifierConverterInterface::HAS_IDENTIFIER_CONVERTER] ?? false)) {
            $id = $this->normalizeIdentifiers($id, $this->entityManager, $resourceClass);
        }
        if (!\is_array($id)) {
            throw new \InvalidArgumentException(sprintf('$id must be array when "%s" key is set to true in the $context', IdentifierConverterInterface::HAS_IDENTIFIER_CONVERTER));
        }
        $identifiers = $id;

        $fetchData = $context['fetch_data'] ?? true;
        if (!$fetchData) {
            return $this->entityManager->getReference($resourceClass, $identifiers);
        }

        $repository = $this->entityManager->getRepository($resourceClass);
        if (!method_exists($repository, 'createQueryBuilder')) {
            throw new RuntimeException('The repository class must have a "createQueryBuilder" method.');
        }
        $queryBuilder = $repository->createQueryBuilder('o');
        $doctrineClassMetadata = $this->entityManager->getClassMetadata($resourceClass);

        $this->addWhereForIdentifiers($identifiers, $queryBuilder, $doctrineClassMetadata);


        $data = $queryBuilder->getQuery()->getOneOrNullResult();
        // validate Ownership
        if ($this->ownershipChecker->isClassOwnable($resourceClass) || $this->ownershipChecker->isClassConditionalOwnable($resourceClass) || $this->ownershipChecker->isClassUserOwnable($resourceClass)) {
            $this->utilityService->validateOwnerShip("request", $data);
        }
        return $data;
    }

    public function isSoftDeleteable($entity_class)
    {
        return in_array(SoftDeleteable::class, class_uses($entity_class));
    }

    /**
     * Add WHERE conditions to the query for one or more identifiers (simple or composite).
     */
    private function addWhereForIdentifiers(array $identifiers, QueryBuilder $queryBuilder, ClassMetadata $classMetadata)
    {
        $alias = $queryBuilder->getRootAliases()[0];
        foreach ($identifiers as $identifier => $value) {
            $placeholder = ':id_' . $identifier;
            $expression = $queryBuilder->expr()->eq(
                "{$alias}.{$identifier}",
                $placeholder
            );

            $queryBuilder->andWhere($expression);

            $queryBuilder->setParameter($placeholder, $value, $classMetadata->getTypeOfField($identifier));
        }
    }
}
