<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Providers;

use App\Entity\Query\Query;
use Psr\Log\LoggerInterface;
use InvalidArgumentException;
use App\Traits\PeerReplicable;
use App\Traits\SoftDeleteable;
use App\Constants\AppConstants;
use App\Entity\Workflow\Workflow;
use Psr\Container\ContainerInterface;
use App\Doctrine\ORM\FilterQueryBuilder;
use App\Entity\Workflow\WorkflowVersion;
use Doctrine\Persistence\ManagerRegistry;
use App\Entity\Workflow\WorkflowCollection;
use Doctrine\ORM\Mapping\ClassMetadataInfo;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Exception\HttpException;
use ApiPlatform\Core\Bridge\Doctrine\Orm\ItemDataProvider;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

class DeleteProvider extends ItemDataProvider
{

    private $managerRegistry;
    private $filterQueryBuilder;
    private $tokenStorage;
    private $container;


    public function __construct(
        ManagerRegistry $managerRegistry,
        FilterQueryBuilder $filterQueryBuilder,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        protected LoggerInterface $logger
    ) {
        $this->managerRegistry = $managerRegistry;
        $this->filterQueryBuilder = $filterQueryBuilder;
        $this->tokenStorage = $tokenStorage;
        $this->container = $container;
    }

    public function isPeerReplicable($entity_class)
    {
        return in_array(PeerReplicable::class, class_uses($entity_class));
    }

    public function isSoftDeleteable($entity_class)
    {
        return in_array(SoftDeleteable::class, class_uses($entity_class));
    }

    public function getUser()
    {
        return $this->tokenStorage->getToken()->getUser();
    }

    public function getUserName()
    {
        $user = $this->getUser();
        if (method_exists($user, 'getFirstname')) {
            return $user->getFirstname() . ' ' . $user->getLastname();
        }
        return $user->getName();
    }

    public function prepareChangeData(
        $previousObject,
        $method,
        $resourceShortName
    ) {
        $delta_data = [];
        $delta_data['oldData'] = $previousObject;
        $delta_data['changeData'] = $resourceShortName;
        $userName = $this->getUserName();
        $userId = $this->getUser()->getUuid();
        return [
            AppConstants::DELTA_DATA => $delta_data,
            AppConstants::HTTP_METHOD => $method,
            AppConstants::REMOTE_ADDR => $_SERVER['HTTP_X_FORWARDED_FOR'] ?? $_SERVER['REMOTE_ADDR'] ?? '127.0.0.1',
            AppConstants::WEB_SOCKET_SESSION_ID => isset($_SERVER['HTTP_WEBSOCKET_SESSIONID']) ? $_SERVER['HTTP_WEBSOCKET_SESSIONID'] : null,
            AppConstants::COMPONENT => 'crudhub',
            AppConstants::ENTITY_TYPE => $resourceShortName,
            AppConstants::USER_UUID => $userId,
            AppConstants::TRANSACTION_DATE => round(microtime(true) * 1000),
            AppConstants::USER => $userName
        ];
    }

    private function deletePlaybookVersionUsingWorkflow($manager, $queryBuilder)
    {
        $workflowVersionRepository = $manager->getRepository(WorkflowVersion::class);
        $qb = $workflowVersionRepository
            ->createQueryBuilder('d')
            ->delete();
        $qb->where($qb->expr()->in('d.workflow', $queryBuilder->getDQL()));
        $qb->setParameters(($queryBuilder->getParameters()));
        try {
            $qb->getQuery()->execute();
        } catch (\Exception $e) {
            $this->logger->error($e);
        }
    }

    private function deletePlaybookVersionUsingCollection($manager, $queryBuilder)
    {
        $workflowVersionRepository = $manager->getRepository(WorkflowVersion::class);
        $workflowRepository = $manager->getRepository(Workflow::class);

        $query = $workflowRepository
            ->createQueryBuilder('d');
        $query->where($query->expr()->in('d.collection', $queryBuilder->getDQL()));
        $query->setParameters(($queryBuilder->getParameters()));

        $qb = $workflowVersionRepository
            ->createQueryBuilder('wv')
            ->delete();
        $qb->where($qb->expr()->in('wv.workflow', $query->getDQL()));
        $qb->setParameters(($queryBuilder->getParameters()));
        try {
            $qb->getQuery()->execute();
        } catch (\Exception $e) {
            $this->logger->error($e);
        }
    }

    public function deleteQuery(string $entityClass, $query, $resourceShortName)
    {
        $returnField = ["o.uuid"];
        /** @var EntityManager $manager */
        $manager = $this->managerRegistry->getManagerForClass($entityClass);
        $repository = $manager->getRepository($entityClass);
        $queryBuilder = $repository
            ->createQueryBuilder('o');

        if (!(isset($query['logic']) && isset($query['filters']))) {
            $query['logic'] = "AND";
            $query['filters'] = [];
        }
        $this->filterQueryBuilder->fromQueryObject($query, $queryBuilder, $entityClass, $returnField, false);
        if ($this->isSoftDeleteable($entityClass) && array_key_exists(Query::SHOW_DELETED, $_GET) && array_key_exists('softdeleteable', $manager->getFilters()->getEnabledFilters())) {
            $manager->getFilters()->disable('softdeleteable');
        }
        $classMetaData = $manager->getClassMetadata($entityClass);

        foreach ($classMetaData->getAssociationNames() as $i => $association) {
            $mapping = $classMetaData->associationMappings[$association];

            if (ClassMetadataInfo::FETCH_EAGER === $mapping['fetch']) {
                $queryBuilder->leftJoin('o.' . $association, 'a' . $i);
                $queryBuilder->addSelect('a' . $i);
            }
        }

        // This is required to get the uuid of object to be deleted from falcon and can also be used for auditing.
        $broadcastObject = $queryBuilder->getQuery()->execute();
        $item_to_broadcast = $this->prepareChangeData(
            $broadcastObject,
            'Bulk Delete',
            $resourceShortName
        );
        $item_to_broadcast[AppConstants::ELASTIC_SEARCH_OPERATION_KEY] = 'bulk_delete';

        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->setContentType('application/json');
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->publish(json_encode([$item_to_broadcast]), AppConstants::DEFAULT_KEY);

        // Specific Handling for workflow version needs to be cleaned prior to deleting workflows and collections
        // Else it keeps giving error Key uuid is still referenced from table "workflow_versions"
        if ($entityClass == Workflow::class) {
            $this->deletePlaybookVersionUsingWorkflow($manager, $queryBuilder);
        } elseif ($entityClass == WorkflowCollection::class) {
            $this->deletePlaybookVersionUsingCollection($manager, $queryBuilder);
        }

        $qb = $repository
            ->createQueryBuilder('d')
            ->delete();
        $qb->where($qb->expr()->in('d.uuid', $queryBuilder->getDQL()));
        $qb->setParameters(($queryBuilder->getParameters()));
        try {
            $result = $qb->getQuery()->execute();
            if (in_array($entityClass, [Workflow::class, WorkflowCollection::class])) {
                $workflowGroupCascadeSql = 'delete from workflow_groups where uuid not in  (select workflowgroup_uuid from workflow_workflowgroup) and reusable=\'false\'';
                $manager->getConnection()->prepare($workflowGroupCascadeSql)->execute();

                $workflowStepCascadeSql = 'delete from workflow_steps where workflowgroup_uuid is null and uuid not in  (select workflowstep_uuid from workflow_workflowstep)';
                $manager->getConnection()->prepare($workflowStepCascadeSql)->execute();
            }
        } catch (\Throwable $th) {
            $errMsg = $th->getmessage();
            throw new HttpException(Response::HTTP_BAD_REQUEST, $errMsg);
        } catch (\Exception | InvalidArgumentException $e) {
            $errMsg = $e->getMessage();
            throw new HttpException(Response::HTTP_BAD_REQUEST, $errMsg);
        }
        return $result;
    }
}
