<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Security;

use App\Entity\Actors\Actor;
use App\Security\BaseAuthenticator;
use App\Providers\FingerprintProvider;
use Doctrine\ORM\EntityManagerInterface;
use App\Security\HmacPostAuthenticationToken;
use Symfony\Component\HttpFoundation\Request;
use App\Encoder\Hmac\DasHmacSecretKeyProvider;
use Symfony\Component\Security\Http\Authenticator\Passport\Passport;
use Symfony\Component\Security\Core\Authentication\Token\TokenInterface;
use Lexik\Bundle\JWTAuthenticationBundle\Services\JWTTokenManagerInterface;
use Symfony\Component\Security\Http\Authenticator\Passport\Badge\UserBadge;
use Symfony\Component\Security\Http\Authenticator\Token\PostAuthenticationToken;
use Symfony\Component\Security\Http\Authenticator\Passport\SelfValidatingPassport;
use Symfony\Component\Security\Core\Exception\CustomUserMessageAuthenticationException;
use Lexik\Bundle\JWTAuthenticationBundle\TokenExtractor\AuthorizationHeaderTokenExtractor;

class HmacAuthenticator extends BaseAuthenticator
{

    protected $jwtManager;
    /** @var DasHmacSecretKeyProvider */
    protected $secretKeyProvider;
    protected $userRepository;

    public function __construct(
        JWTTokenManagerInterface $jwtManager,
        DasHmacSecretKeyProvider $dasHmacSecretKeyProvider,
        EntityManagerInterface $entityManager
    ) {
        $this->jwtManager = $jwtManager;
        $this->secretKeyProvider = $dasHmacSecretKeyProvider;
        $this->userRepository = $entityManager->getRepository(Actor::class);
    }

    /**
     * Called on every request to decide if this authenticator should be
     * used for the request. Returning `false` will cause this authenticator
     * to be skipped.
     */
    public function supports(Request $request): ?bool
    {
        return $request->headers->has('Authorization') && strpos($request->headers->get('Authorization'), 'CS ') === 0;
    }

    public function authenticate(Request $request): Passport
    {
        $tokenExtractor = new AuthorizationHeaderTokenExtractor('CS', 'Authorization');

        $hmacToken = $tokenExtractor->extract($request);
        return $this->authenticateToken($hmacToken, $request);
    }

    public function authenticateToken($hmacToken, $request=null): Passport
    {
        if (null === $hmacToken) {
            // The token header was empty, authentication fails with HTTP Status
            // Code 401 "Unauthorized"
            throw new CustomUserMessageAuthenticationException('No API token provided');
        }
        FingerprintProvider::validateFingerprint($hmacToken, $this->secretKeyProvider, $request);
        $userId = $this->secretKeyProvider->getUserId();

        $passport = new SelfValidatingPassport(
            new UserBadge($userId, function ($userIdentifier) {
                return $this->userRepository->findOneBy(['userId' => $userIdentifier]);
            })
        );
        $passport->setAttribute('token', $hmacToken);

        return $passport;
    }

    /**
     * Shortcut to create a PostAuthenticationToken for you, if you don't really
     * care about which authenticated token you're using.
     */
    public function createToken(Passport $passport, string $firewallName): TokenInterface
    {
        return new HmacPostAuthenticationToken($passport->getUser(), $firewallName, $passport->getUser()->getRoles(true, true), $passport->getAttribute('token'));
    }
}
