<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Serializer;

use ApiPlatform\Core\Serializer\SerializerContextBuilderInterface;
use App\Constants\AttributeLevelOperators;
use App\Entity\Core\Picklist;
use App\Entity\Core\PicklistName;
use App\Entity\Custom\Comment;
use App\Entity\ModelMetadata\StagingModelMetadata;
use Symfony\Component\DependencyInjection\ParameterBag\ParameterBagInterface;
use Symfony\Component\HttpFoundation\Request;

final class DynamicContextBuilder implements SerializerContextBuilderInterface
{
    private $decorated;
    private $parameters;

    public function __construct(SerializerContextBuilderInterface $decorated, ParameterBagInterface $parameters)
    {
        $this->decorated = $decorated;
        $this->parameters = $parameters;
    }

    public function createFromRequest(Request $request, bool $normalization, ?array $extractedAttributes = null): array
    {
        $context = $this->decorated->createFromRequest($request, $normalization, $extractedAttributes);
        // skip for 1. GET; 2. denorm; 3. Query route
        if (
            $request->getMethod() == Request::METHOD_GET || false === $normalization ||
            (isset($extractedAttributes['_fsr_dynamic_context']) && $extractedAttributes['_fsr_dynamic_context'] === false)
        ) {
            return $context;
        }

        $fsrChangeFields = $request->attributes->get('fsr_change_fields');
        $newData = $request->attributes->get('data');
        $requestBody = json_decode($request->getContent());

        if (
            is_null($fsrChangeFields)
            || $newData instanceof Picklist
            || $newData instanceof StagingModelMetadata
            || $newData instanceof Comment
        ) {
            $fsrChangeFields = ["fsr_all", "fsr_primary", "password"];
        }
        $context["groups"] = $fsrChangeFields;

        // escape data before sending the response to UI
        $context["ignore_html_escape"] = true;

        // if the relationships flag is not from the request itself, cap the related count normalized
        if (!isset($context["fsr_max_relation_count"])) {
            $key = AttributeLevelOperators::OPERATOR_TOKEN . 'relationships';
            $rawOption = $request->query->get($key);
            if (is_null($rawOption)) {
                $rawOption = $request->request->get($key, false);
            }
            $context['relationships'] = $rawOption;
            $capRelations = !filter_var($rawOption, FILTER_VALIDATE_BOOLEAN);
            if ($capRelations) {
                // also check for conflict and sync flags
                //TODO: see if decode can be skipped
                $isConflict = false;
                try {
                    if (method_exists($newData, 'getConflict') && $newData->getConflict() === true) {
                        $isConflict = true;
                    }
                } catch (\Throwable | \Exception $e) {
                    $isConflict = false;
                }

                if (isset($requestBody->{"__sync"}) && $isConflict === false) {
                    $max_relation_count = 100;
                    if ($this->parameters->has('max_relation_count')) {
                        $max_relation_count = $this->parameters->get('max_relation_count', 100);
                    }
                    $context["fsr_max_relation_count"] = $max_relation_count;
                }
            }
        }

        if ($context['relationships'] && (isset($requestBody->{"__sync"}) && $requestBody->{"__sync"})) {
            $context["groups"][] = 'fsr_all';
            $max_relation_count = 100;
            if ($this->parameters->has('max_relation_count')) {
                $max_relation_count = $this->parameters->get('max_relation_count', 100);
            }
            $context["fsr_max_relation_count"] = $max_relation_count;
        }

        if (!$context['relationships'] && !($newData instanceof Picklist) && !($newData instanceof PicklistName) && !($newData instanceof StagingModelMetadata)) {
            $context["relationsAsJson"] = true;
        }

        // so that html escaping can be done before sending the response back
        $request->attributes->set('_fsr_dynamic_context', true);

        return $context;
    }
}
