<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

/*
 * This file is part of the API Platform project.
 *
 * (c) Kévin Dunglas <dunglas@gmail.com>
 *
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */

declare(strict_types=1);

namespace App\Serializer\Normalizer;

use Throwable;
use Psr\Log\LoggerInterface;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerAwareInterface;
use Symfony\Component\DependencyInjection\ParameterBag\ParameterBagInterface;
use Symfony\Component\Serializer\Normalizer\CacheableSupportsMethodInterface;

final class CompatibilityPartialCollectionViewNormalizer implements NormalizerInterface, NormalizerAwareInterface, CacheableSupportsMethodInterface
{
    /** @var PartialCollectionViewNormalizer */
    private $decorated;

    /** @var RequestStack */
    private $requestStack;

    /** @var bool */
    private $defaultLegacyEnabled;

    public function __construct(
        NormalizerInterface $decorated,
        RequestStack $requestStack,
        ParameterBagInterface $allParams,
        protected LoggerInterface $logger
    ) {
        $this->decorated = $decorated;
        $this->requestStack = $requestStack;
        $this->defaultLegacyEnabled = false;
        if ($allParams->has('legacy_collection_view')) {
            $this->defaultLegacyEnabled = $allParams->get('legacy_collection_view');
        }
    }

    public function normalize($object, $format = null, array $context = [])
    {
        $data = $this->decorated->normalize($object, $format, $context);
        if (array_key_exists('hydra:member', $data)) {
            try {
                $query = $this->requestStack->getCurrentRequest()->query;
                $enableLegacyParams = $query ? $query->get('$legacy_collection_view') : null;
                if (is_null($enableLegacyParams)) {
                    // check the default parameter
                    $enableLegacyParams = $this->defaultLegacyEnabled;
                } else {
                    $enableLegacyParams = filter_var($enableLegacyParams, FILTER_VALIDATE_BOOLEAN);
                }
                if ($enableLegacyParams === true) {
                    $data['hydra:itemsPerPage'] = $object->getItemsPerPage();
                    if (array_key_exists('hydra:view', $data)) {
                        $hydraView = $data['hydra:view'];
                        if (array_key_exists('hydra:first', $hydraView)) {
                            $data['hydra:firstPage'] = $hydraView['hydra:first'];
                            $data['hydra:lastPage'] = $hydraView['hydra:last'];
                            if (array_key_exists('hydra:next', $hydraView)) {
                                $data['hydra:nextPage'] = $hydraView['hydra:next'];
                            }
                            if (array_key_exists('hydra:previous', $hydraView)) {
                                $data['hydra:previousPage'] = $hydraView['hydra:previous'];
                            }
                        } else {
                            $data['hydra:firstPage'] = $hydraView['@id'];
                            $data['hydra:lastPage'] = $hydraView['@id'];
                        }
                    }
                }
            } catch (Throwable $e) {
                // legacy support on a best effort basis
                // TODO : add logging
                $this->logger->error("legacy support error:" . $e->getMessage());
            }
        }
        return $data;
    }

    public function supportsNormalization($data, $format = null)
    {
        return $this->decorated->supportsNormalization($data, $format);
    }

    public function hasCacheableSupportsMethod(): bool
    {
        return $this->decorated->hasCacheableSupportsMethod();
    }

    public function setNormalizer(NormalizerInterface $normalizer)
    {
        $this->decorated->setNormalizer($normalizer);
    }
}
