<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Serializer\Normalizer;

use InvalidArgumentException;
use UnexpectedValueException;
use App\Entity\Base\BaseEntity;
use App\Service\MetadataHelper;
use App\Security\Voter\FsrVoter;
use ApiPlatform\Core\Util\ClassInfoTrait;
use App\Entity\Workflow\WorkflowStepType;
use App\Service\NormalizerUtilityService;
use ApiPlatform\Core\Serializer\ContextTrait;
use ApiPlatform\Core\Api\IriConverterInterface;
use App\Providers\PermissionsDictionaryBuilder;
use Symfony\Component\HttpFoundation\RequestStack;
use ApiPlatform\Core\JsonLd\ContextBuilderInterface;
use ApiPlatform\Core\Serializer\AbstractItemNormalizer;
use ApiPlatform\Core\Api\ResourceClassResolverInterface;
use ApiPlatform\Core\JsonLd\Serializer\JsonLdContextTrait;
use Symfony\Component\PropertyAccess\PropertyAccessorInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Serializer\Normalizer\DenormalizerInterface;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Component\Serializer\NameConverter\NameConverterInterface;
use Symfony\Component\Serializer\Mapping\Factory\ClassMetadataFactoryInterface;
use ApiPlatform\Core\Metadata\Property\Factory\PropertyMetadataFactoryInterface;
use ApiPlatform\Core\Metadata\Resource\Factory\ResourceMetadataFactoryInterface;
use ApiPlatform\Metadata\Property\Factory\PropertyNameCollectionFactoryInterface;

class EfficientItemNormalizer extends AbstractItemNormalizer
{

    use ClassInfoTrait;
    use ContextTrait;
    use JsonLdContextTrait;

    private $decorated;

    private $requestStack;

    protected $iriConverter;

    private $metadataHelper;

    protected $normalizerUtilityService;
    protected $fsrVoter;
    protected $resourceClassResolver;
    protected $propertyMetadataFactory;
    private $contextBuilder;

    public const GROUPS = 'groups';
    private const FORMULAS_START_CHARACTERS = ['=', '-', '+', '@', "\t", "\r"];

    public function __construct(
        NormalizerInterface $decorated,
        RequestStack $requestStack,
        ResourceMetadataFactoryInterface $resourceMetadataFactory,
        IriConverterInterface $iriConverter,
        NormalizerUtilityService $normalizerUtilityService,
        FsrVoter $fsrVoter,
        PropertyNameCollectionFactoryInterface $propertyNameCollectionFactory,
        PropertyMetadataFactoryInterface $propertyMetadataFactory,
        ResourceClassResolverInterface $resourceClassResolver,
        ContextBuilderInterface $contextBuilder,
        MetadataHelper $metadataHelper,
        PropertyAccessorInterface $propertyAccessor = null,
        NameConverterInterface $nameConverter = null,
        ClassMetadataFactoryInterface $classMetadataFactory = null,
        array $defaultContext = [],
        iterable $dataTransformers = []
    ) {
        parent::__construct(
            $propertyNameCollectionFactory,
            $propertyMetadataFactory,
            $iriConverter,
            $resourceClassResolver,
            $propertyAccessor,
            $nameConverter,
            $classMetadataFactory,
            null,
            false,
            $defaultContext,
            $dataTransformers,
            $resourceMetadataFactory
        );
        if (!$decorated instanceof DenormalizerInterface) {
            throw new \InvalidArgumentException(sprintf('The decorated normalizer must implement the %s.', DenormalizerInterface::class));
        }
        $this->contextBuilder = $contextBuilder;
        $this->decorated = $decorated;
        $this->requestStack = $requestStack;
        $this->iriConverter = $iriConverter;
        $this->normalizerUtilityService = $normalizerUtilityService;
        $this->fsrVoter = $fsrVoter;
        $this->resourceClassResolver = $resourceClassResolver;
        $this->metadataHelper = $metadataHelper;
        $this->propertyMetadataFactory = $propertyMetadataFactory;
    }

    public function supportsNormalization($data, $format = null, array $context = []): bool
    {
        return ($data instanceof BaseEntity) && ($this->decorated->supportsNormalization($data, $format));
    }

    public function normalize($object, $format = null, array $context = [])
    {
        [$relationships, $export, $asJson, $showPassword, $hideUuid, $archived, $fsr_max_relation_count, $displayNameOnly, $csvNeutralization] = $this->normalizerUtilityService->getResponseIdentifier($context);
        $objectClass = $this->getObjectClass($object);
        if (!array_key_exists("object_class", $context)) {
            $context["object_class"] = [];
        }

        if (!array_key_exists("ignore_depth", $context)) {
            $context = $context + ["ignore_depth" => false];
        }
        $context["normalize_object_class"] = [];
        array_push($context["object_class"], $objectClass);
        $context["normalize_object_class"] = $objectClass;

        if (!array_key_exists('groups', $context) || count($context["object_class"]) > 1) {
            $context = $this->normalizerUtilityService->manipulateContextGroup($context, $relationships, $export, $asJson, $showPassword, $archived, $fsr_max_relation_count, $displayNameOnly, $csvNeutralization);
        }
        // conditions in which a related iri is returned:
        // when context doesn't have ignore_depth, export or json returns immediate iris, else one depth is allowed
        // when context has ignore depth, the same behavior but for one additional level of depth
        // in any case, a max depth of 3 is allowed except in case of WorkflowStepType where depth allowed is 4

        $exportDepth = 1;
        if ($object instanceof WorkflowStepType) {
            $exportDepth = 4;
        } elseif ($context["ignore_depth"] == true) {
            $exportDepth = 2;
        }
        $normalizeAsIri = false;
        $currentDepth = count($context["object_class"]);
        //return iri beyond max depth to prevent infinite normalization or circular reference
        if ($currentDepth > ($exportDepth + 1)) {
            $normalizeAsIri = true;
        }
        // flag
        $relationAsJson = false;
        if (isset($context['relationsAsJson']) && $context['relationsAsJson'] === true) {
            $relationAsJson = true;
        }

        if ($export || $asJson) {
            //for m2m, go upto max depth -1
            if ($currentDepth > $exportDepth) {
                $normalizeAsIri = true;
            } elseif ($currentDepth > 1) {
                // for lookups, iri irrespective of depth
                $propertyMetadataLocal = $this->propertyMetadataFactory->create(
                    $context['object_class'][$currentDepth - 2],
                    $context['api_attribute'],
                    $this->getFactoryOptions($context)
                );
                $type = $propertyMetadataLocal->getType();
                // only a check for collection suffices and resource class check is not needed since this normalizer will only be called
                // for instances of BaseEntity::class
                if ($type && !$type->isCollection()) {
                    $normalizeAsIri = true;
                }
            }
        } elseif ($relationAsJson && $currentDepth > $exportDepth) {
            if (!isset($context['api_attribute'])) {
                $normalizeAsIri = true;
            } else {
                $propertyMetadataLocal = $this->propertyMetadataFactory->create(
                    $context['object_class'][$currentDepth - 2],
                    $context['api_attribute'],
                    $this->getFactoryOptions($context)
                );
                $type = $propertyMetadataLocal->getType();
                if (is_null($type) || $type->isCollection()) {
                    $normalizeAsIri = true;
                }
            }
        }

        if ($normalizeAsIri) {
            try {
                return $this->iriConverter->getIriFromItem($object);
            } catch (\Exception $e) {
                // This will be a case of soft deleted
                // Record is still link but not accessible by entity Manager
                return null;
            }
        }

        $context['skip_null_values'] = false;
        if ($archived) {
            $context['skip_null_values'] = true;
        }
        $context['preserve_empty_objects'] = true;

        $outputClass = $this->getOutputClass($objectClass, $context);
        if (null !== $outputClass && !isset($context[self::IS_TRANSFORMED_TO_SAME_CLASS])) {
            $all_data = parent::normalize($object, $format, $context);
        } else {
            $resourceClass = $this->resourceClassResolver->getResourceClass($object, $context['resource_class'] ?? null);
            $context = $this->initContext($resourceClass, $context);
            try {
                $iri = $this->iriConverter->getIriFromItem($object);
            } catch (\Exception $e) {
                // This will be a case of soft deleted
                // Record is still link but not accessible by entity Manager
                return null;
            }
            $context['iri'] = $iri;
            $context['api_normalize'] = true;

            $metadata = $this->addJsonLdContext($this->contextBuilder, $resourceClass, $context);

            $data = parent::normalize($object, $format, $context);
            if (!\is_array($data)) {
                return $data;
            }

            $resourceMetadata = $this->resourceMetadataFactory->create($resourceClass);

            $metadata['@id'] = $iri;
            $metadata['@type'] = $resourceMetadata->getIri() ?: $resourceMetadata->getShortName();
            $all_data = $metadata + $data;
        }
        return $this->processFields($all_data, $context, $export, $hideUuid, $archived, $displayNameOnly, $csvNeutralization);
    }

    /**
     * Normalizes a collection of relations (to-many).
     * overridden to normalize a max number of related objects based on key 'fsr_max_relation_count'
     * @param iterable $attributeValue
     *
     * @throws UnexpectedValueException
     */
    protected function normalizeCollectionOfRelations($propertyMetadata, $attributeValue, string $resourceClass, ?string $format, array $context): array
    {
        $value = [];
        $limit = 0;
        if (array_key_exists('fsr_max_relation_count', $context) && $context['fsr_max_relation_count'] > 0) {
            $limit = $context['fsr_max_relation_count'] - 1;
        }

        foreach ($attributeValue as $index => $obj) {
            if (!\is_object($obj) && null !== $obj) {
                throw new UnexpectedValueException('Unexpected non-object element in to-many relation.');
            }

            $value[$index] = $this->normalizeRelation($propertyMetadata, $obj, $resourceClass, $format, $context);
            if ($limit !== 0 && $index >= $limit) {
                break;
            }
        }

        return $value;
    }

    public function supportsDenormalization($data, $type, $format = null, array $context = [])
    {
        // EfficientItemDeNormalizer has the functionality of denormalize
        return false;
    }

    public function denormalize($data, $class, $format = null, array $context = [])
    {
        return $this->decorated->denormalize($data, $class, $format, $context);
    }


    private function processFields($data, $context, $export, $hideUuid = false, $archived = false, $displayNameOnly = false, $csvNeutralization = false)
    {
        $entityFieldMap = null;
        $entity = $context["normalize_object_class"] ?? null;
        $module = $this->metadataHelper->getModuleFromClassName($entity);
        $context['ignore_authorization'] = $context['ignore_authorization'] ?? false;
        $context['ignore_field_authorization'] = $context['ignore_field_authorization'] ?? false;
        $context['ignore_relationship_authorization'] = $context['ignore_relationship_authorization'] ?? false;
        $entityFieldMap = $this->metadataHelper->getEntityFields($entity) ?? null;

        // Process HTML Escape and File Field here
        $returnData = [];
        if (!$context['ignore_authorization'] && !($this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module) ||
            $this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_EXECUTE . '.' . $module)) && count($context['object_class']) == 1) {
            throw new AccessDeniedException();
        }
        foreach ($data as $key => $value) {
            if ($key == "@id" && $export) {
                continue;
            }
            if ($key == "uuid" && $hideUuid) {
                continue;
            }
            if ($key == "@id" || $key == "@context" || $key == "@type") {
                $returnData[$key] = $value ?? null;
            } elseif (
                $entityFieldMap && array_key_exists('relatedFields', $entityFieldMap)
                && is_countable($entityFieldMap['relatedFields']) &&  count($entityFieldMap['relatedFields'])
                && in_array($key, $entityFieldMap['relatedFields']) && ($context['ignore_relationship_authorization']
                    || $this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module . '.' . $key))
            ) {
                $returnData[$key] = $value;
            } elseif ($context['ignore_field_authorization'] || $this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module . '.' . $key)) {
                $ignore_html_escape = $context['ignore_html_escape'] ?? false;
                if (!$ignore_html_escape && ($entityFieldMap && $value)) {
                    if (count($entityFieldMap['htmlEscapeFields']) && in_array($key, $entityFieldMap['htmlEscapeFields'])) {
                        $returnData[$key] = htmlentities(htmlspecialchars_decode($value));
                        continue;
                    }
                }
                if ($entityFieldMap && $value && count($entityFieldMap['fileFields']) && in_array($key, $entityFieldMap['fileFields'])) {
                    try {
                        $fileObject = $this->iriConverter->getItemFromIri($data[$key]);

                        $value = $this->normalize($fileObject, 'jsonld', $context + ['api_attribute' => $key]);
                    } catch (InvalidArgumentException | \Exception $e) {
                        $value = null;
                        continue;
                    }
                }
                if (is_array($context) && array_key_exists('csvNeutralization', $context) && $context['csvNeutralization'] && is_string($value) && \in_array(substr((string) $value, 0, 1), self::FORMULAS_START_CHARACTERS, true)) {
                    $value = "'" . $value;
                }
                $returnData[$key] = $value;
            }
        }

        if (
            ($archived || (is_array($context) && array_key_exists('displayNameOnly', $context) && $context['displayNameOnly'])) &&
            (is_array($entityFieldMap) && array_key_exists('displayName', $entityFieldMap))
        ) {
            $displayNameFields = $entityFieldMap['displayName'];
            $displayName = '';
            foreach ($displayNameFields as $displayNameField) {
                $displayName = $displayName . ' ' . $data[$displayNameField];
            }
            if ($archived || count($context["object_class"]) > 1) {
                $returnData['__displayName'] = trim($displayName);
            }
        }

        if (count($context["object_class"]) > 1 && array_key_exists('displayNameOnly', $context) && $context['displayNameOnly']) {
            return $returnData['__displayName'];
        }
        return $returnData;
    }

    public function hasCacheableSupportsMethod(): bool
    {
        return true;
    }

    protected function getFactoryOptions(array $context): array
    {
        $options = [];

        if (isset($context[self::GROUPS])) {
            /* @see https://github.com/symfony/symfony/blob/v4.2.6/src/Symfony/Component/PropertyInfo/Extractor/SerializerExtractor.php */
            $options['serializer_groups'] = (array) $context[self::GROUPS];
        }

        if (isset($context['collection_operation_name'])) {
            $options['collection_operation_name'] = $context['collection_operation_name'];
        }

        if (isset($context['item_operation_name'])) {
            $options['item_operation_name'] = $context['item_operation_name'];
        }

        return $options;
    }
}
