<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use stdClass;
use App\Kernel;
use App\Utility\MsspUtility;
use InvalidArgumentException;
use App\Constants\AppConstants;
use Psr\Container\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\JsonResponse;
use App\EventSubscriber\MqMessagebroadcastSubscriber;
use Symfony\Component\HttpKernel\HttpKernelInterface;
use Symfony\Component\HttpKernel\Exception\HttpException;

class BulkRequestService
{
    /**
     * @var HttpKernelInterface
     */
    protected $kernel;

    /** @var ContainerInterface */
    private $container;

    /** @var MSSPUtility */
    private $msspUtility;

    /** @var NormalizerUtilityService */
    private $normalizerUtility;

    public const AUDIT_PRODUCER =  "old_sound_rabbit_mq.cyops_crud_auditing_producer";

    public function __construct(HttpKernelInterface $kernel, ContainerInterface $container, MsspUtility $msspUtility, NormalizerUtilityService $normalizerUtility)
    {
        $this->kernel = $kernel;
        $this->container = $container;
        $this->msspUtility = $msspUtility;
        $this->normalizerUtility = $normalizerUtility;
    }

    public function prepareRecordSubRequest($resourceClass, $objectData, $subRequestPath, $headers, $queryParams, $method = 'POST', $reinitializeKernel = false, $subRequestType = 'bulk')
    {
        $successObj = [];
        $failedObj = [];
        $auditData = [];
        $replicationDataPost = [];
        $replicationDataPut = [];
        $replicationDataDelete = [];
        $workflowTriggers = [];
        $lastPublishIndex = -1;
        $failDetected = false;
        if (isset($objectData->{"data"}) && count($objectData->data) > 0) {
            $requestArray = $objectData->data;

            if (
                isset($objectData->data[0]->{'@type'}) &&
                in_array($objectData->data[0]->{'@type'}, ['Workflow', 'WorkflowCollection', 'WorkflowGroup'])
            ) {
                $reinitializeKernel = true;
            }
            foreach ($objectData->data as $k => $item) {
                if ($failDetected) {
                    $reinitializeKernel = true;
                    $failDetected = false;
                }
                $parentRequestBody = new stdClass();
                foreach ($objectData as $index => $arg) {
                    if ($index != 'data') {
                        $parentRequestBody->{$index} = $arg;
                    }
                }
                $this->updateReferenceToIriAndCheckForRelation($item);

                $response = $this->executeSubRequest(
                    $k,
                    $subRequestPath,
                    $method,
                    $headers,
                    $queryParams,
                    json_encode($item),
                    $auditData,
                    $replicationDataPost,
                    $replicationDataPut,
                    $replicationDataDelete,
                    $workflowTriggers,
                    $subRequestType,
                    $reinitializeKernel
                );
                if ($response->getStatusCode() == Response::HTTP_CREATED || $response->getStatusCode() == Response::HTTP_OK || $response->getStatusCode() == Response::HTTP_NO_CONTENT) {
                    $successObj[] = json_decode($response->getContent(), true);
                } else {
                    // this format is being used in postman to determine index of failed record
                    // in case of any modification that logic should also be modified.
                    // postman/dispatcher/callback/data/data_receiver.py#_find_indexes_of_failed_record
                    $failedObj[] = $method . " method for object at index #" . $k . " in the request payload failed with error: " . $response->getContent();
                    $failDetected = true;
                }

                // keep sending MQ and workflow data in batches so that the memory consumption does not become very high but bulk requests
                if (count($auditData) >= AppConstants::BATCH_SIZE_FOR_BULK) {
                    $parentRequestBody->data = array_slice($requestArray, $lastPublishIndex + 1, AppConstants::BATCH_SIZE_FOR_BULK);
                    $this->sendDataToServices($resourceClass, $parentRequestBody, $auditData, $replicationDataPost, $replicationDataPut, $replicationDataDelete, $workflowTriggers);
                    $lastPublishIndex = $k;
                }
            }
            if (!$parentRequestBody || !is_object($parentRequestBody)) {
                $parentRequestBody = new stdClass();
            }
            // send the remaining batch to audit and workflow
            $parentRequestBody->data = array_slice($requestArray, $lastPublishIndex + 1);
            $this->sendDataToServices($resourceClass, $parentRequestBody, $auditData, $replicationDataPost, $replicationDataPut, $replicationDataDelete, $workflowTriggers);
        } elseif (count($objectData->data) == 0) {
            return new JsonResponse(["message" => "No data provided in requests for bulk operations"], Response::HTTP_OK);
        } else {
            throw new HttpException(Response::HTTP_BAD_REQUEST, 'Data is not provided or is not in the list format');
        }

        return $this->processResults($successObj, $failedObj);
    }

    public function prepareUpdateDeleteSubRequest($resourceClass, $objectData, $subRequestPath, $headers, $queryParams, $method = 'DELETE', $reinitializeKernel = false, $subRequestType = 'bulk')
    {
        $successObj = [];
        $failedObj = [];
        $auditData = [];
        $replicationDataPost = [];
        $replicationDataPut = [];
        $replicationDataDelete = [];
        $workflowTriggers = [];
        $lastPublishIndex = -1;
        $failDetected = false;
        if ((is_array($objectData) && array_key_exists("data", $objectData)) || isset($objectData->data)) {
            if (isset($objectData->data[0]->{'@type'}) && ($objectData->data[0]->{'@type'} === 'Workflow' || $objectData->data[0]->{'@type'} === 'WorkflowCollection')) {
                $reinitializeKernel = true;
            }
            if ($method == 'DELETE') {
                if (is_array($objectData->data) && isset($objectData->data[0])) {
                    $urlEndpoint = explode('/', $objectData->data[0]->{'@id'});
                }
                if (isset($urlEndpoint[3]) && ($urlEndpoint[3] == 'workflows' || $urlEndpoint[3] == 'workflow_collections')) {
                    $reinitializeKernel = true;
                }
            }
            $requestArray = $objectData->data;
            foreach ($objectData->data as $k => $item) {
                if ($failDetected) {
                    $reinitializeKernel = true;
                    $failDetected = false;
                }
                $parentRequestBody = new stdClass();
                foreach ($objectData as $index => $arg) {
                    if ($index != 'data') {
                        $parentRequestBody->{$index} = $arg;
                    }
                }
                // PUT needs @id. if UUID only, generate @id from it
                if (!isset($item->{'@id'}) && isset($item->{'uuid'})) {
                    $item->{'@id'} = $subRequestPath . '/' . $item->{'uuid'};
                }
                $path = $item->{'@id'};
                $this->updateReferenceToIriAndCheckForRelation($item);
                $response = $this->executeSubRequest(
                    $k,
                    $path,
                    $method,
                    $headers,
                    $queryParams,
                    json_encode($item),
                    $auditData,
                    $replicationDataPost,
                    $replicationDataPut,
                    $replicationDataDelete,
                    $workflowTriggers,
                    $subRequestType,
                    $reinitializeKernel
                );
                if ($response->getStatusCode() == Response::HTTP_OK || $response->getStatusCode() == Response::HTTP_NO_CONTENT) {
                    $successObj[] = json_decode($response->getContent(), true);
                } else {
                    $failDetected = true;
                    $failedObj[] = $method . " method for object at index #" . $k . " in the request payload failed with error: " . $response->getContent();
                }
                // keep sending MQ and workflow data in batches so that the memory consumption does not become very high but bulk requests
                if (count($auditData) >= AppConstants::BATCH_SIZE_FOR_BULK) {
                    $parentRequestBody->data = array_slice($requestArray, $lastPublishIndex + 1, AppConstants::BATCH_SIZE_FOR_BULK);
                    $this->sendDataToServices($resourceClass, $parentRequestBody, $auditData, $replicationDataPost, $replicationDataPut, $replicationDataDelete, $workflowTriggers);
                    $lastPublishIndex = $k;
                }
            }
            if (!$parentRequestBody || !is_object($parentRequestBody)) {
                $parentRequestBody = new stdClass();
            }
            // send the remaining batch to audit and workflow
            $parentRequestBody->data = array_slice($requestArray, $lastPublishIndex + 1);
            $this->sendDataToServices($resourceClass, $parentRequestBody, $auditData, $replicationDataPost, $replicationDataPut, $replicationDataDelete, $workflowTriggers);
        } else {
            throw new HttpException(Response::HTTP_BAD_REQUEST, 'Invalid query parameters.');
        }

        $method == Request::METHOD_DELETE ? $responseCode = Response::HTTP_NO_CONTENT : $responseCode = Response::HTTP_OK;

        return $this->processResults($successObj, $failedObj, $responseCode);
    }

    private function processResults($successObj, $failedObj, $responseCode = Response::HTTP_CREATED)
    {
        $statusOnly = $this->normalizerUtility->getOption('statusOnly', array(), false, true);
        if (count($failedObj) == 0) {
            if ($statusOnly) {
                $data = ["success" => sizeof($successObj)];
            } else {
                $data = ['@type' => 'hydra:Collection', 'hydra:member' => $successObj, 'hydra:totalItems' => sizeof($successObj)];
            }
            return new JsonResponse($data, $responseCode);
        } else {
            $partialUpdatedObject = [];
            $partialUpdatedObject["success"] = $statusOnly ? sizeof($successObj) : $successObj;
            $partialUpdatedObject["failure"] = $statusOnly ? sizeof($successObj) : $failedObj;
            $responseCode = Response::HTTP_MULTI_STATUS;
            if (count($successObj) == 0) {
                $responseCode = Response::HTTP_INTERNAL_SERVER_ERROR;
            }
            // TODO: check if partial response is also returned as hydra collection
            return new JsonResponse($partialUpdatedObject, $responseCode);
        }
    }

    private function sendDataToServices($resourceClass, $objectData, &$auditData, &$replicationDataPost, &$replicationDataPut, &$replicationDataDelete, &$workflowTriggers)
    {
        $this->container->get(self::AUDIT_PRODUCER)->setContentType('application/json');
        $this->container->get(self::AUDIT_PRODUCER)->publish(json_encode($auditData), MqMessagebroadcastSubscriber::DEFALUT_KEY);


        if (!empty($replicationDataPost)) {
            $this->msspUtility->replicateData($resourceClass, $objectData, $replicationDataPost, 'POST');
        }
        if (!empty($replicationDataPut)) {
            $this->msspUtility->replicateData($resourceClass, $objectData, $replicationDataPut, 'PUT');
        }
        if (!empty($replicationDataDelete)) {
            $this->msspUtility->replicateData($resourceClass, $objectData, $replicationDataDelete, 'DELETE');
        }
        if (count($workflowTriggers)) {
            $workflowsData = ["data" => $workflowTriggers, "auth_info" => 'CS HMAC'];
            $this->container->get(self::AUDIT_PRODUCER)->setContentType('application/json');
            $this->container->get(self::AUDIT_PRODUCER)->publish(
                json_encode($workflowsData),
                AppConstants::WORKFLOW_EXECUTION_KEY
            );
        }

        $auditData = [];
        $replicationDataPost = [];
        $replicationDataPut = [];
        $replicationDataDelete = [];
        $workflowTriggers = [];
    }

    public function executeSubRequest(
        int $index,
        string $path,
        string $method,
        array $headers,
        array $queryParams,
        string $body,
        &$audit_data,
        &$replicationDataPost,
        &$replicationDataPut,
        &$replicationDataDelete,
        &$workflowTriggers,
        string $subRequestType = 'bulk',
        $reinitializeKernel = false
    ): Response {
        $subRequest = Request::create($path, $method, $queryParams, [], [], [], $body);
        $subRequest->headers->replace($headers);
        $subRequest->query->replace($queryParams);
        $subRequest->attributes->set('fsr_subrequest_type', $subRequestType);
        try {
            if ($reinitializeKernel === true) {
                $kernelSubRequest = new Kernel('prod', false);
                $response = $kernelSubRequest->handle($subRequest, HttpKernelInterface::MAIN_REQUEST, true);
            } else {
                $httpKernel = $this->container->get('http_kernel');
                $response = $httpKernel->handle($subRequest, HttpKernelInterface::MAIN_REQUEST, true);
            }

            if ($subRequest->attributes->has('fsr_audit_data')) {
                array_push($audit_data, $subRequest->attributes->get('fsr_audit_data'));
            }
            if ($subRequest->attributes->has('fsr_replication_data')) {
                $replicationData = $subRequest->attributes->get('fsr_replication_data');
                if ($replicationData->{'method'} == 'POST') {
                    array_push($replicationDataPost, $replicationData);
                } elseif ($replicationData->{'method'} == 'PUT') {
                    array_push($replicationDataPut, $replicationData);
                } elseif ($replicationData->{'method'} == 'DELETE') {
                    array_push($replicationDataDelete, $replicationData);
                }
            }
            if ($subRequest->attributes->has('fsr_workflow_triggers')) {
                array_push($workflowTriggers, ...$subRequest->attributes->get('fsr_workflow_triggers'));
            }
            if ($reinitializeKernel === true) {
                $kernelSubRequest->terminate($subRequest, $response);
                $kernelSubRequest->shutdown();
            } else {
                $httpKernel->terminate($subRequest, $response);
            }
            return $response;
        } catch (InvalidArgumentException | \Exception $e) {
            return new Response(sprintf('Batch element #%d failed, check the log files.', $index), 400);
        }
    }

    /**
     * @param $request_body
     * @return mixed
     */
    public function updateReferenceToIriAndCheckForRelation(&$request_body)
    {
        foreach ($request_body as $key => $value) {
            if (gettype($value) == "object" && isset($request_body->$key->{'@id'})) {
                $request_body->$key = $request_body->$key->{'@id'};
            }
        }
    }

    public function buildNormalizationContextFromQuery(array $queryParams): array
    {
        $context = [];
        foreach ($queryParams as $queryParam => $value) {
            if (str_starts_with($queryParam, "$")) {
                $key = substr($queryParam, 1);
                $context[$key] = $value;
            }
        }
        return $context;
    }
}
