<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use Exception;
use App\Entity\Core\PgFile;
use App\Entity\Actors\Actor;
use Psr\Log\LoggerInterface;
use App\Entity\Actors\Avatar;
use App\Constants\AppConstants;
use App\Service\SettingsHelper;
use App\Entity\Authorization\Role;
use App\Entity\Authorization\Team;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use GuzzleHttp\Psr7\Request as GuzzleRequest;

/**
 * Class ActorConfig
 * @package App\Service\ConfigExportImport
 */
class ActorConfig extends PortingConfig
{
    /** @var SettingsHelper */
    private $settingsHelper;

    private $authUserUri;

    private $authUserQueryUri;

    const CS_USERTYPE_USER = 2;
    const CS_USERTYPE_SAML = 6;
    const CS_USERTYPE_LDAP = 7;

    public function __construct(
        SettingsHelper $settingsHelper,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        protected LoggerInterface $logger
    ) {
        $this->settingsHelper = $settingsHelper;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
        $this->authUserUri = sprintf('%s/api/auth/users', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
        $this->authUserQueryUri = sprintf('%s/api/auth/query/users', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
    }
    /**
     * 
     * @param array $data List of actors
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        $repository = $this->loadRepository(Actor::class);
        $options = [];
        foreach ($data as $actor) {
            $existing = $repository->findOneBy(['uuid' => $actor['uuid']]);
            $individualOptions = [];
            $individualOptions['uuid'] = $actor['uuid'];
            $individualOptions['userType'] = $actor['@type'];
            $individualOptions['name'] = $actor['@type'] === 'Person' ? $actor['firstname'] . ' ' . $actor['lastname'] : $actor['name'];
            $individualOptions['existingActorUuid'] = null;
            if ($actor['user']) {
                $existingActorForUuid = $repository->findOneBy(['userId' => $actor['userId']]);
                if ($existingActorForUuid) {
                    $individualOptions['existingActorUuid'] = $existingActorForUuid->getUuid();
                }
                $individualOptions['username'] = $actor['user']['loginid'];
                try {
                    $authRequestBody = ["login_id" => [$individualOptions['username']]];
                    $response = $this->sendRequest($this->authUserQueryUri, 'POST', $authRequestBody, $tokenType);
                    $jsonResponse = json_decode($response, true);
                    $individualOptions['existingUsername'] = array_key_exists('usersresp', $jsonResponse) && empty($jsonResponse['usersresp']) ? false : $jsonResponse['usersresp'][0]['uuid'];
                    $individualOptions['csActive'] = array_key_exists('usersresp', $jsonResponse) && empty($jsonResponse['usersresp']) ? 0 : $jsonResponse['usersresp'][0]['status'];
                    $individualOptions['accessType'] = array_key_exists('usersresp', $jsonResponse) && empty($jsonResponse['usersresp']) ? 'Concurrent' : $jsonResponse['usersresp'][0]['access_type'];
                    if (!$existingActorForUuid && $individualOptions['existingUsername']) {
                        // If the username matches but the actor uuid does not, use the actor that matches the username
                        $existingActorForUsername = $repository->findOneBy(['userId' => $individualOptions['existingUsername']]);
                        if ($existingActorForUsername) {
                            $individualOptions['existingActorUuid'] = $existingActorForUsername->getUuid();
                        }
                    }
                } catch (BadRequestHttpException | \Exception $e) {
                    $individualOptions['existingUsername'] = false;
                }
            }
            $individualOptions['include'] = true;
            $individualOptions['existingUserId'] = $existing ? $existing->getUserId() : false;
            $individualOptions['exists'] = $existing || $individualOptions['existingUsername'] || $individualOptions['existingActorUuid'] ? true : false;
            $options[] = $individualOptions;
        }
        return $options;
    }

    protected function getNamedActiveUsersCount()
    {
        $usersDetails = [];
        $headers = [
            "Content-type" => "application/json",
        ];
        $headers = array("Content-type" => "application/json", "X-ADMIN-USER" => true);
        $dasClient = $this->container->get("app.proxy.client.das");
        $dasUri = $this->container->getParameter('das.uri');
        $dasRequest = new GuzzleRequest('get', $dasUri . '/license/?param=license_details', $headers);
        try {
            $response = $dasClient->send($dasRequest);
            $responseData = $response->getBody()->getContents();
            $jsonDecodeResponse = json_decode($responseData, true);
            $usersDetails['usersAllowed'] = $jsonDecodeResponse['details']['users'];
            $usersDetails['namedActiveUsers'] = $jsonDecodeResponse['user_stats']['fixed_active_users'];
            return $usersDetails;
        } catch (BadRequestHttpException | \Exception $error) {
            $this->logger->error($error->getMessage());
        }
    }

    protected function getSuperAdmin()
    {
        $superAdminDetails = [];
        $headers = ["Content-type" => "application/json"];
        $dasClient = $this->container->get("app.proxy.client.das");
        $dasUri = $this->container->getParameter('das.uri');
        $dasRequest = new GuzzleRequest('get', $dasUri . '/config?section=USER', $headers);
        try {
            $response = $dasClient->send($dasRequest);
            $responseData = $response->getBody()->getContents();
            $jsonDecodeResponse = json_decode($responseData, true);
            $superAdminDetails['superAdmin'] = $jsonDecodeResponse['hydra:member'][0];
            return $superAdminDetails;
        } catch (BadRequestHttpException | \Exception $error) {
            $this->logger->error($error->getMessage());
        }
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        $toAdd = [];
        $peopleToUpdate = [];
        $applianceToUpdate = [];
        $roleRepository = $this->loadRepository(Role::class);
        $teamRepository = $this->loadRepository(Team::class);
        $avatarRepository = $this->loadRepository(Avatar::class);
        foreach ($options as $actorOption) {
            if (!$actorOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($data as $actor) {
                if (!$actor) { // Todo: merge this condition with below if.
                    continue;
                }
                if ($actorOption['uuid'] == $actor['uuid']) {
                    $uuid = $actor['uuid'];
                    if ($actorOption['existingActorUuid']) {
                        // We cannot change the existing actor uuid for the given userId
                        $uuid = $actorOption['existingActorUuid'];
                    }
                    $moduleName = self::getModuleNameFromType($actor['@type']);
                    $actorIri = sprintf('/api/3/%s/%s', $moduleName, $uuid);
                    $existingRoles = [];
                    foreach ($actor['roles'] as $roleIri) {
                        $roleIriParts = explode('/', $roleIri);
                        $roleUuid = end($roleIriParts);
                        $existingRole = $roleRepository->findOneBy(['uuid' => $roleUuid]);
                        // Only include roles that exist
                        if ($existingRole) {
                            $existingRoles[] = $roleIri;
                        }
                    }
                    $actor['roles'] = $existingRoles;
                    $existingTeams = [];
                    foreach ($actor['teams'] as $teamIri) {
                        $teamIriParts = explode('/', $teamIri);
                        $teamUuid = end($teamIriParts);
                        $existingTeam = $teamRepository->findOneBy(['uuid' => $teamUuid]);
                        // Only include roles that exist
                        if ($existingTeam) {
                            $existingTeams[] = $teamIri;
                        }
                    }
                    $actor['teams'] = $existingTeams;
                    $actor['record_type'] = $moduleName;

                    if ($actor['avatar'] && $actor['avatar_file']) {
                        $avatar = $avatarRepository->findOneBy(['uuid' => $uuid]);
                        if (!$avatar) {
                            $avatar = new Avatar();
                            $avatar->setUuid($uuid);
                            $avatar->setId($uuid);
                        }

                        $avatar->setFilename($actor['avatar']['filename']);
                        $avatar->setMimeType($actor['avatar']['mimeType']);
                        $avatar->setSize($actor['avatar']['size']);

                        $entityManager = $this->container->get('doctrine')->getManagerForClass(PgFile::class);
                        $pgFile = null;
                        if ($avatar->getFile()) {
                            $pgFile = $entityManager->find(PgFile::class, $avatar->getFile());
                        }
                        if (!$pgFile) {
                            $pgFile = new PgFile();
                        }
                        $pgFile->setFile($actor['avatar_file']);
                        $entityManager->persist($pgFile);
                        $avatar->setFile($pgFile->getUuid());
                        $this->entityManager->persist($avatar);
                        $this->entityManager->flush();
                    }
                    if (is_array($actor['avatar'])) {
                        $actor['avatar'] = '/api/3/avatars/' . $actor['avatar']['uuid'];
                    }
                    if (!$actorOption['exists']) {
                        if ($actor['user']['user_type'] == self::CS_USERTYPE_SAML) {
                            $jsonData = $this->sendRequest(sprintf('%s/api/saml/config', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY)), 'get', null, $tokenType);
                            $result = json_decode($jsonData, true);
                            if (!$result['samlEnabled']) {
                                $actor['user']['user_type'] = self::CS_USERTYPE_USER;
                                $actor['userType'] = null;
                            }
                        } elseif ($actor['user']['user_type'] == self::CS_USERTYPE_LDAP) {
                            $authConfigUri = sprintf('%s/api/auth/config?section=LDAP', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
                            $jsonData = $this->sendRequest($authConfigUri, 'get', null, $tokenType);
                            $result = json_decode($jsonData, true);
                            foreach ($result['hydra:member'] as $ldapSetting) {
                                if ($ldapSetting['key'] == 'ldap_enabled' && !$ldapSetting['value']) {
                                    $actor['user']['user_type'] = self::CS_USERTYPE_USER;
                                    $actor['user']['bind_name'] = null;
                                    $actor['userType'] = null;
                                }
                            }
                        }
                        $actor['firstname'] = is_null($actor['firstname']) || $actor['firstname'] == ''  ? 'Change' : $actor['firstname'];
                        $actor['lastname'] = is_null($actor['lastname']) || $actor['lastname'] == ''  ? 'Me' : $actor['lastname'];
                        $toAdd[] = $actor;
                        break;
                    }
                    unset($actor['uuid']);
                    $actor['@id'] = $actorIri;
                    // If the user exists, we do not overwrite the user details
                    $actor['userId'] = $actorOption['existingUserId'] ? $actorOption['existingUserId'] : $actorOption['existingUsername'];
                    $actor['accessType'] = isset($actor['accessType']) ? $actor['accessType'] : 'Named';
                    if ($moduleName === 'people') {
                        $actor['existingStatus'] = $actorOption['csActive'] == 1 ? 1 : 0;
                        $actor['existingAccessType'] = $actorOption['accessType'];
                        $peopleToUpdate[] = $actor;
                    } else {
                        $applianceToUpdate[] = $actor;
                    }
                    break;
                }
            }
        }
        $results = [];
        $usersDetails = $this->getNamedActiveUsersCount();
        $superAdminDetails = $this->getSuperAdmin();
        $usersAllowed = $usersDetails['usersAllowed'] - $usersDetails['namedActiveUsers'];
        foreach ($toAdd as $item) {
            $item['accessType'] = isset($item['accessType']) ? $item['accessType'] : 'Named';
            $item['user']['access_type'] = $item['accessType'];
            $csactive = $item['csActive'];
            if ($item['accessType'] === 'Concurrent') {
                $csActive = $item['csActive'];
            } elseif (array_key_exists('value', $superAdminDetails['superAdmin']) && $item['userId'] == $superAdminDetails['superAdmin']['value']) {
                $csactive = 1;
            } elseif ($usersAllowed > 0 && $item['accessType'] === 'Named' && $item['csActive'] == 1) {
                $csActive = $item['csActive'];
                $usersAllowed--;
            } else {
                $csActive = 0;
            }
            $item['csActive'] = $csActive;
            $item['user']['status'] = $item['csActive'] == 1 ? 1 : 2;
            $fullURI = sprintf('%s/api/3/%s', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $actor['record_type']);
            $result = $this->sendRequest($fullURI, 'post', $item, $tokenType);
            $results[] = $result;
            // Finally, add user settings
            if ($item['@userSettings']) {
                $iri = sprintf('/api/3/%s/%s', $actor['record_type'], $item['uuid']);
                $userSettings = $this->getUserSettingsFromActorIRI($iri);
                $this->settingsHelper->saveUserSettings($userSettings, $item['@userSettings']);
            }
        }
        if (!empty($peopleToUpdate)) {
            $fullURI = sprintf('%s/api/3/update/people', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            foreach ($peopleToUpdate as $key => $people) {
                $csactive = $people['csActive'];
                if ($people['accessType'] === 'Concurrent') {
                    $csactive = $people['csActive'];
                } elseif (array_key_exists('value', $superAdminDetails['superAdmin']) && $people['userId'] == $superAdminDetails['superAdmin']['value']) {
                    $csactive = 1;
                } elseif ($people['accessType'] === 'Named' && $people['csActive'] == 1 && ($usersAllowed > 0 || ($people['existingAccessType'] == 'Named' && $people['existingStatus'] == 1))) {
                    $csactive = $people['csActive'];
                    $usersAllowed--;
                } else {
                    $csactive = 0;
                }
                $peopleToUpdate[$key]['csActive'] = $csactive;
            }

            $body = [
                "data" => $peopleToUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
            $dasRequestBody = ["update" => []];
            foreach ($peopleToUpdate as $item) {
                if ($item['@userSettings']) {
                    $userSettings = $this->getUserSettingsFromActorIRI($item['@id']);
                    $this->settingsHelper->saveUserSettings($userSettings, $item['@userSettings']);
                }
                $dasRequestBody['update'] = [
                    "uuid" => $item['userId'],
                    "access_type" => $item['accessType'],
                    "status" => $item['csActive'] == 1 ? 1 : 2
                ];
                $this->sendRequest($this->authUserUri, 'PUT', $dasRequestBody, $tokenType);
            }
        }

        if (!empty($applianceToUpdate)) {
            $fullURI = sprintf('%s/api/3/update/appliances', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $body = [
                "data" => $applianceToUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
        }

        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        $actors = $this->getResourceFromIRI($options, Actor::class, true, true);
        $authRequestBody = ["users" => []];
        foreach ($actors as $key => $value) {
            $authRequestBody['users'][] =  $value['userId'];
        }
        if (!empty($authRequestBody['users'])) {
            $response = $this->sendRequest($this->authUserQueryUri, 'POST', $authRequestBody, $tokenType);
            $userList = json_decode($response, true)['usersresp'];
        }

        foreach ($actors as $index => $actor) {
            $user = self::findBy($userList, 'uuid', $actor['userId'], true);
            foreach ($actor as $key => $value) {
                if (is_array($value) && !in_array($key, ['user', 'roles', 'teams'])) {
                    unset($actors[$index][$key]);
                }
            }
            $moduleName = self::getModuleNameFromType($actor['@type']);
            $actorIri = sprintf('/api/3/%s/%s', $moduleName, $actor['uuid']);
            $userSettings = $this->getUserSettingsFromActorIRI($actorIri);
            if ($userSettings) {
                $actors[$index]['@userSettings'] = $userSettings->getPublicValues();
            }
            if ($user) {
                $actors[$index]['user'] = $user;
            }

            if ($actor['avatar']) {
                $avatar = end($this->getResourceFromIRI([$actor['avatar']], Avatar::class));
                $avatarRepository = $this->entityManager->getRepository(Avatar::class);
                $iriParts = explode('/', $actor['avatar']);
                $avatarObject = $avatarRepository->findOneBy(['uuid' => end($iriParts)]);
                if ($avatarObject) {
                    $fileRepository = $this->entityManager->getRepository(PgFile::class);
                    $pgFile = $fileRepository->findOneBy(['uuid' => $avatarObject->getFile()]);
                    if ($pgFile) {
                        $actors[$index]['avatar_file'] = stream_get_contents($pgFile->getFile());
                        $actors[$index]['avatar'] = $avatar;
                    }
                }
            }
        }
        return $actors;
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        //Todo
        return [];
    }

    public function generateInfoContent(array $options)
    {
        // ToDo
        return [];
    }

    private function getUserSettingsFromActorIRI($actorIRI)
    {
        return $this->settingsHelper->getUserSettingsFromIRI($actorIRI);
    }

    private static function getModuleNameFromType($type)
    {
        return $type == 'Person' ? 'people' : 'appliances';
    }
}
