<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Constants\AppConstants;
use App\Entity\Core\Proxy;
use App\Entity\Settings\SystemSettings;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

/**
 * Class AppSettingsConfig
 * @package App\Service\ConfigExportImport
 */
class AppSettingsConfig extends PortingConfig
{

    private const PURGE_AUDIT_PLAYBOOK = 'b9144644-d69c-4b98-bf33-337c3b1a6cfa';
    private const PURGE_RUNNING_LOG_PLAYBOOK = '26f1077a-c334-43bb-a10a-77b2593a121f';

    /**
     * @param array $allData
     * @param string $key
     * @return void
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        $options = [];
        foreach ($data as $setting) {
            $individualOptions = [];
            $individualOptions['name'] =  $setting['type'];
            $individualOptions['include'] = true;

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        $results = [];
        $authConfigUri = sprintf('%s/api/auth/config', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
        $token = $this->getTokenFromStorage();
        $proxyUpdate = false;
        foreach ($options as $option) {
            if (!$option['include']) {
                continue;
            }
            $setting = $option['name'];
            $value = null;
            foreach ($data as $item) {
                if ($item['type'] == $setting) {
                    $value = $item['value'];
                }
            }
            if (!$value) {
                continue;
            }
            if ($setting == 'systemSettings') {
                $repository = $this->loadRepository(SystemSettings::class);
                if ($value && $value['publicValues'] && $value['publicValues']['auditlog_ttl_config']) {
                    $scheduleId = $this->savePurgeAuditLogSchedule($value['publicValues']['auditlog_ttl_config'], $tokenType);
                    $value['publicValues']['auditlog_ttl_config']['scheduleId'] = $scheduleId;
                }
                if ($value && $value['publicValues'] && $value['publicValues']['workflow_ttl_config']) {
                    $scheduleId = $this->savePurgeExecuteLogsSchedule($value['publicValues']['workflow_ttl_config'], $tokenType);
                    $value['publicValues']['workflow_ttl_config']['scheduleId'] = $scheduleId;
                }

                /** @var SystemSettings $existingSettings */
                $existingSettings = $repository->findAll()[0];
                if ($existingSettings->getGlobalValues() != null) {
                    $value['globalValues'] = $existingSettings->getGlobalValues();
                }

                // Branding does not exist in DB and branding in import - No Action Needed
                // Branding does not exist in DB and No branding in import - No Action Needed
                // Branding exists in DB and No branding in import
                $privateValues = $existingSettings->getPrivateValues();
                if ($privateValues !== null) {
                    if ($privateValues['authentication'] && $privateValues['authentication']['branding']) {
                        $value['privateValues']['authentication']['branding'] = $privateValues['authentication']['branding'];
                    }
                }

                $fullURI = sprintf('%s/api/3/system_settings/%s', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $existingSettings->getUuid());
                $existingSettingsAPI = json_decode($this->sendRequest($fullURI, 'get', null, $tokenType), true);
                $existingSections = $existingSettingsAPI['sections'];
                foreach ($value['sections'] as $section) {
                    // check if uuid already exists
                    $uuid = $section['uuid'];
                    $name = $section['name'];
                    $sectionExists = false;
                    foreach ($existingSections as &$existingSection) {
                        if ($uuid === $existingSection['uuid'] || $name === $existingSection['name']) {
                            // update
                            $existingSection['publicValues'] = $section['publicValues'];
                            $sectionExists = true;
                            break;
                        }
                    }
                    if (!$sectionExists) {
                        unset($section['@id']);
                        array_push($existingSections, $section);
                    }
                }
                $value['sections'] = $existingSections;

                $fullURI = sprintf('%s/api/3/system_settings/%s', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $existingSettings->getUuid());
                $results[] = (bool)$this->sendRequest($fullURI, 'put', $value, $tokenType);
            } elseif($setting == 'sso') {
                try {
                    if ($token) {
                        $jsonData = $this->sendRequest(sprintf('%s/api/saml/config', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY)), 'post', $value, $tokenType);
                    } else {
                        $jsonData = $this->sendGatewayRequest(sprintf('%s/saml/config', $this->container->getParameter('gateway.uri')), 'post', $value);
                    }
                    $result = json_decode($jsonData, true);
                    $results[] = true;
                } catch (BadRequestHttpException | \Exception $exception) {
                    $results[] = true;
                }
            } elseif($setting == 'syslog') {
                try {
                    if (count($value)) {
                        if ($token) {
                            $jsonData = $this->sendRequest(sprintf('%s/api/gateway/config/syslog', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY)), 'post', $value[0], $tokenType);
                        } else {
                            $jsonData = $this->sendGatewayRequest(sprintf('%s/config/syslog', $this->container->getParameter('gateway.uri')), 'post', $value[0], $tokenType);
                        }
                    }
                    $result = json_decode($jsonData, true);
                    $results[] = true;
                } catch (BadRequestHttpException | \Exception $exception) {
                    $results[] = true;
                }
            } elseif($setting == 'LDAP' || $setting == 'TOKEN' || $setting == 'HA' || $setting == 'RADIUS') {
                $sectionValues = [];
                foreach ($value as $item) {
                    $sectionValues[$item['key']] = $item['value'];
                }
                $body = [
                    "section" => $setting,
                    "value" => $sectionValues
                ];
                $jsonData = $this->sendRequest($authConfigUri, 'put', $body, $tokenType);
                $result = json_decode($jsonData, true);
                $results[] = $result['response'] == 'Success';
            } elseif($setting == 'proxy') {
                $proxyUpdate = true;
            }
        }
        if ($proxyUpdate) {
            $repository = $this->loadRepository(Proxy::class);
            $value['proxies'] = json_decode($value['proxies'], true);
            /** @var Proxy $proxySettings */
            $proxySettings = $repository->findAll()[0];
            if ($proxySettings) {
                $fullURI = sprintf('%s/api/3/proxy/%s', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $proxySettings->getUuid());
                $results[] = (bool)$this->sendRequest($fullURI, 'put', $value, $tokenType);
            } else {
                $fullURI = sprintf('%s/api/3/proxy', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
                $results[] = (bool)$this->sendRequest($fullURI, 'post', $value, $tokenType);
            }
        }
        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    private function savePurgeAuditLogSchedule($config, $tokenType)
    {
        $workflowUri = '/api/3/workflows/';
        $schedule = [
            'name' => 'Purge Audit Logs',
            'kwargs' => [
                'wf_iri' => $workflowUri . self::PURGE_AUDIT_PLAYBOOK
            ],
            'crontab' => [
                'minute' => '0',
                'hour' => '0',
                'day_of_week' => '*',
                'day_of_month' => '*',
                'month_of_year' => '*'
            ],
            'enabled' => $config['enabled'],
            'task' => 'workflow.tasks.periodic_task',
            'interval' => null
        ];

        return $this->saveSchedule($schedule, $tokenType);
    }

    private function savePurgeExecuteLogsSchedule($config, $tokenType)
    {
        $workflowUri = '/api/3/workflows/';
        $schedule = [
            'name' => 'Purge Executed Playbook Logs',
            'kwargs' => [
                'wf_iri' => $workflowUri . self::PURGE_RUNNING_LOG_PLAYBOOK
            ],
            'crontab' => [
                'minute' => '0',
                'hour' => '0',
                'day_of_week' => '*',
                'day_of_month' => '*',
                'month_of_year' => '*'
            ],
            'enabled' => $config['enabled'],
            'task' => 'workflow.tasks.periodic_task',
            'interval' => null
        ];

        return $this->saveSchedule($schedule, $tokenType);
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        $settings = [];
        $authConfigUri = sprintf('%s/api/auth/config', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
        foreach ($options as $setting) {
            $value = null;
            if ($setting == 'systemSettings') {
                $repository = $this->loadRepository(SystemSettings::class);
                /** @var SystemSettings $existingSettings */
                $existingSettings = $repository->findAll()[0];
                $context = $this->getDefaultNormalizationContext(SystemSettings::class, $export = false);
                $normalizedObject = $this->normalizer->normalize($existingSettings, 'jsonld', $context);
                $value = $normalizedObject;
            } elseif($setting == 'sso') {
                $jsonData = $this->sendGatewayRequest(sprintf('%s/saml/config', $this->container->getParameter('gateway.uri')), 'get', null);
                $data = json_decode($jsonData, true);
                $value = $data;
            } elseif($setting == 'syslog') {
                try {
                    $jsonData = $this->sendGatewayRequest(sprintf('%s/config/syslog', $this->container->getParameter('gateway.uri')), 'get', null);
                    $data = json_decode($jsonData, true);
                    $value = $data;
                } catch (BadRequestHttpException | \Exception $exception) {
                    $value = null;
                }
            } elseif($setting == 'LDAP' || $setting == 'TOKEN' || $setting == 'HA' || $setting == 'RADIUS') {
                $jsonData = $this->sendRequest(sprintf('%s?section=%s', $authConfigUri, $setting), 'get', null, $tokenType);
                $data = json_decode($jsonData, true);
                $value = $data['hydra:member'];
            } elseif($setting == 'proxy') {
                $context = $this->getDefaultNormalizationContext(Proxy::class);
                $repository = $this->loadRepository(Proxy::class);
                /** @var Proxy $proxySettings */
                $proxySettings = $repository->findAll()[0];
                if (!$proxySettings) {
                    // No proxy settings exist
                    continue;
                }
                $normalizedObject = $this->normalizer->normalize($proxySettings, AppConstants::NORMALIZER_METHOD, $context);
                $value = $normalizedObject;
            }
            $settings[] = [
                'type' => $setting,
                'value' => $value
            ];
        }

        return $settings;
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        foreach ($options as $key => $value) {
            array_push($exportOption, $key);
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        return [];
    }

    private function sendGatewayRequest($uri, $method, $data = null)
    {
        /** @var ProxyProvider $proxyProvider */
        $client = $this->container->get("app.proxy.client.saml");
        $headers = [
            "Content-type" => "application/json",
            "Authorization" => null
        ];
        $request = new GuzzleRequest($method, $uri, $headers, $data ? json_encode($data) : null);
        try {
            $response = $client->send($request, ['verify' => false]);
            $statusCode = $response->getStatusCode();
            if ($statusCode < 400) {
                //$jsonDecodeResponse = json_decode($response->getBody()->getContents(), true);
                return $response->getBody()->getContents();
            } else {
                throw new \Exception($response->getBody()->getContents());
            }
        } catch (BadRequestHttpException | \Exception $exception) {
            $errCode = $exception->getCode();
            $message = $exception->getResponse() ? $exception->getResponse()->getBody() : $exception->getMessage();
            throw new \Exception(sprintf('%d: %s', $errCode, $message));
        }
    }
}
