<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Security\Voter\FsrVoter;
use App\Entity\Core\SolutionPack;
use Doctrine\ORM\EntityManagerInterface;
use GuzzleHttp\Exception\RequestException;
use App\Service\SolutionPackUtilityService;
use App\Providers\PermissionsDictionaryBuilder;
use App\Service\ConfigExportImport\PlaybookConfig;
use Symfony\Component\Filesystem\Exception\IOException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class ConnectorConfig
 * @package App\Service\ConfigExportImport
 */
class ConnectorConfig extends PortingConfig
{
    /** @var fsrVoter */
    private $fsrVoter;

    /** @var PlaybookConfig */
    private $playbookConfig;
    /**@var ImportService */

    private $solutionPackUtilityService;

    protected $entityManager;

    public function __construct(
        FsrVoter $fsrVoter,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        PlaybookConfig $playbookConfig,
        SolutionPackUtilityService $solutionPackUtilityService,
        protected LoggerInterface $logger
    ) {
        $this->fsrVoter = $fsrVoter;
        $this->playbookConfig = $playbookConfig;
        $this->solutionPackUtilityService = $solutionPackUtilityService;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }

    private function updateConnectorObject($connector, $version)
    {
        $connInstallPath = sprintf('cyops-connector-%s-%s', $connector['name'], $version);
        $connector['version'] = $version;
        $connector['installer_path'] = $connInstallPath;
        $connector['rpm_name'] = $connInstallPath;

        return $connector;
    }

    /**
     *
     * @param array $data Playbook object containing collections & global vars
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        // at least one of these permissions
        if (
            (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . PermissionsDictionaryBuilder::MODULE_CONNECTORS) &&
                !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.' . PermissionsDictionaryBuilder::MODULE_CONNECTORS) &&
                !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.' . PermissionsDictionaryBuilder::MODULE_CONNECTORS)) ||
            empty($data)
        ) {
            return [];
        }
        $options = [];
        $existingConnectors = $this->getConnectors($tokenType);
        $existingConnectorNames = array_column($existingConnectors, 'name');
        $connectorsInRepo = $this->loadConnectorRpms();
        $importType = $this->importJob->getType();
        $options = [];
        foreach ($data as $connector) {
            $existing = in_array($connector['name'], $existingConnectorNames);
            $individualOptions = [];
            $individualOptions['name'] = $connector['name'];
            if (is_array($connector) && array_key_exists("label", $connector)) {
                $individualOptions['label'] = $connector['label'];
            }
            $individualOptions['configurationCount'] = is_array($connector) && array_key_exists("configurations", $connector) && is_array($connector['configurations']) ? count($connector['configurations']) : 0;

            // Special handling for installtion trigger from Solution Pack
            $updateIncludeInstall = true;
            if ($importType === 'SolutionPack Import' && $connector["install_mode"] === "rpm") {
                if (!array_key_exists("version", $connector) || is_null($connector['version'])) {
                    $solutionPackRecord = $this->solutionPackUtilityService->getLatestSolutionPack($connector['name']);
                    if ($solutionPackRecord == null) {
                        throw new \Exception($connector['name'] . ' solution pack record not found.');
                    }
                    // Set Connector to Latest Version, but installation would depend on includeInstall logic based on minVersion give orr not
                    $connVersion = is_null($solutionPackRecord->getLatestAvailableVersion()) ? $solutionPackRecord->getVersion() : $solutionPackRecord->getLatestAvailableVersion();
                    $connector = $this->updateConnectorObject($connector, $connVersion);
                } elseif($connector['version'] && !is_null($connector['version'])) {
                    $connVersion = $connector['version'];
                    $connector = $this->updateConnectorObject($connector, $connVersion);
                    $updateIncludeInstall = false;
                }
            }

            if (is_array($connector) && array_key_exists("version", $connector)) {
                $individualOptions['version'] = $connector['version'];
            }
            if (is_array($connector) && array_key_exists("rpm_name", $connector)) {
                $individualOptions['rpm_name'] = $connector['rpm_name'];
            }
            if (is_array($connector) && array_key_exists("configurations", $connector) && is_array($connector['configurations']) && count($connector['configurations']) > 0) {
                $individualOptions['includeConfigurations'] = true;
            }
            if (array_key_exists('install_mode', $connector) && array_key_exists('installer_path', $connector)) {
                $individualOptions['install_mode'] = $connector['install_mode'];
                $individualOptions['installer_path'] = $connector['installer_path'];
            }
            $individualOptions['exists'] = $existing;
            if ($existing) {
                foreach ($existingConnectors as $existingConnector) {
                    if ($existingConnector['name'] == $connector['name']) {
                        $individualOptions['existingId'] = $existingConnector['id'];
                        $existingVersion = str_replace('.', '', $existingConnector['version']);
                        $version = str_replace('.', '', $connector['version']);
                        if ($existingVersion >= $version) {
                            $individualOptions['includeInstall'] = false;
                            $individualOptions['locked'] = true;
                        } else {
                            $individualOptions['includeInstall'] = (bool)$connector['rpm_name'];
                            $individualOptions['existingVersion'] = $existingConnector['version'];
                        }
                        break;
                    }
                }
            } elseif((is_array($connector) && array_key_exists("rpm_name", $connector) && $connector['rpm_name']) || array_key_exists('install_mode', $connector)) {
                $individualOptions['includeInstall'] = true;
            }
            if ($individualOptions['includeInstall']) {
                $fullKey = sprintf('%s_%s', $connector['name'], array_key_exists("version", $connector) ? $connector['version'] : null);
                $individualOptions['rpmExists'] = array_key_exists($fullKey, $connectorsInRepo);
            }

            // Update includeInstall for Solution Pack
            if ($importType === 'SolutionPack Import' && $connector["install_mode"] === "rpm" && $updateIncludeInstall) {
                if (array_key_exists("minVersion", $connector) && $connector['minVersion'] && !is_null($connector['minVersion']) && $solutionPackRecord->getInstalled() && $solutionPackRecord->getVersion() >= $connector['minVersion']) {
                    $individualOptions['includeInstall'] = false;
                } elseif((array_key_exists("minVersion", $connector) && (!$connector['minVersion'] || is_null($connector['minVersion']))) && $solutionPackRecord->getInstalled()) {
                    $individualOptions['includeInstall'] = false;
                } else {
                    $individualOptions['includeInstall'] = true;
                }
            }

            $options[] = $individualOptions;
        }
        return $options;
    }

    private function getConnectors($tokenType = 'JWT')
    {
        $token = $this->getTokenFromStorage();
        $endpoint = '/connector_details/?ordering=label&page_size=10000&$export=true';
        if ($token) {
            $uri = sprintf('%s/api/integration%s', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $endpoint);
            $response = $this->sendRequest($uri, 'post', null, $tokenType);
            $jsonDecodeResponse = json_decode($response, true);
            return array_key_exists('data', $jsonDecodeResponse) ? $jsonDecodeResponse['data'] : $jsonDecodeResponse;
        } else {
            $uri = $this->container->getParameter('integrations.uri') . $endpoint;
            return $this->sendIntegrationRequest($uri, 'post');
        }
    }

    private function getDataImports($configurationUuid, $tokenType = 'JWT')
    {
        $token = $this->getTokenFromStorage();
        $endpoint = '/data-import/?configuration=' . $configurationUuid;
        if ($token) {
            $uri = sprintf('%s/api/integration%s', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $endpoint);
            $response = $this->sendRequest($uri, 'get', null, $tokenType);
            $jsonDecodeResponse = json_decode($response, true);
            return array_key_exists('data', $jsonDecodeResponse) ? $jsonDecodeResponse['data'] : $jsonDecodeResponse;
        } else {
            $uri = $this->container->getParameter('integrations.uri') . $endpoint;
            return $this->sendIntegrationRequest($uri, 'get');
        }
    }


    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        $token = $this->getTokenFromStorage();
        foreach ($options as $option) {
            if (!$option['include']) {
                continue;
            }
            foreach ($data as $connector) {
                if ($connector['rpm_name'] === $option['rpm_name']) {
                    $connectorId = array_key_exists('existingId', $option) ? $option['existingId'] : null;
                    // Install & wait if not exists
                    if ($option['includeInstall']) {
                        $result = $this->installConnectorRpm($connector, $tokenType);
                        if (!$result) {
                            // Connector failed to install
                            throw new \Exception(sprintf('Connector %s took longer than 4 minutes to install', $connector['rpm_name']));
                        }
                        $connectorId = $result['id'];
                    } elseif($option['exists']) {
                        $iri = sprintf('connectors/%s/%s/?format=json', $connector['name'], $connector['version']);
                        if ($token) {
                            $jsonData = $this->sendRequest(sprintf('%s/api/integration/%s', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $iri), 'post', $connector, $tokenType);
                        } else {
                            $jsonData = $this->sendIntegrationRequest(sprintf('%s/%s', $this->container->getParameter('integrations.uri'), $iri), 'post', $connector);
                        }
                        $result = json_decode($jsonData, true);
                    }
                    if (!$result || !$option['includeConfigurations']) {
                        continue;
                    }
                    // Now, update the roles for operations
                    if (array_key_exists('operation_roles', $connector)) {
                        $this->updateOperationRoles($connector['operation_roles'], $result['operations'], $tokenType);
                    }

                    // First, delete existing configs
                    $this->deleteExistingConfigurations($connector['configurations'], $result['configuration'], $tokenType);

                    // Next, ingestion playbooks need to go in
                    if (array_key_exists('dataImports', $connector) && !empty($connector['ingestionPlaybooks']['collections'])) {
                        $this->importIngestionPlaybooks($connector['ingestionPlaybooks'], $tokenType, $importedBy);
                    }

                    // Install configurations afterwards
                    $this->importConfigurations($connector['configurations'], $connectorId, $tokenType);

                    if (!array_key_exists('dataImports', $connector) || empty($connector['dataImports'])) {
                        continue;
                    }

                    $this->importDataImports($connector['dataImports'], $tokenType);
                }
            }
        }
        return true;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = [])
    {
        $connectorPath = $tmpDirectory . $key . '/';
        $defaultJsonPath = $connectorPath . 'data.json';
        $fileData = file_get_contents($defaultJsonPath);
        $data = json_decode($fileData, true);
        $token = $this->getTokenFromStorage();
        $importType = 'Import Wizard';
        if ($this->importJob && $this->importJob->getType() === 'SolutionPack Import') {
            $importType = 'SolutionPack Import';
        }

        foreach ($options as $option) {
            foreach ($data as $connector) {
                if ($importType === 'SolutionPack Import' && $connector["install_mode"] === "rpm" && $connector['name'] === $option['name']) {
                    $connector['version'] = $option['version'];
                    if (array_key_exists("installer_path", $option)) {
                        $connector['installer_path'] = $option['installer_path'];
                    }
                    $connector['rpm_name'] = $option['rpm_name'];
                }

                if ($option['name'] != $connector['name'] || $option['version'] != $connector['version']) {
                    continue;
                }

                if ((array_key_exists("install_mode", $connector) && array_key_exists("install_mode", $option) && $connector['install_mode'] == $option['install_mode']) && $connector['installer_path'] === $option['installer_path']) {
                    $connectorId = array_key_exists('existingId', $option) ? $option['existingId'] : null;
                    // Install & wait if not exists
                    if ($option['includeInstall'] && $connector['install_mode'] == 'rpm' && array_key_exists('rpmExists',$option) && $option['rpmExists']) {
                        $result = $this->installConnectorRpm($connector, $tokenType);
                        if (!$result) {
                            // Connector failed to install
                            throw new \Exception(sprintf('Connector %s took longer than 4 minutes to install', $connector['installer_path']));
                        }
                        $connectorId = $result['id'];
                    } elseif($option['includeInstall'] && $connector['install_mode'] == 'tgz' && $importType !== 'SolutionPack Import') {
                        $fileParts = explode('.', $connector['installer_path']);
                        $uri = sprintf('%s/import-connector/%s/?replace=true&format=json', $this->container->getParameter('integrations.uri'), $fileParts[0]);
                        $fileStream = fopen($connectorPath . $option['installer_path'], 'rb');
                        $token = $this->tokenStorage->getToken();
                        $authorization = $token ? sprintf("Bearer %s", $token->getCredentials()) : null;
                        $client = $this->container->get("app.proxy.client.integrations");
                        $params = [
                            'verify' => false,
                            'body' => $fileStream,
                            'headers' => [
                                'Authorization' => $authorization
                            ]
                        ];
                        try {
                            $integrationResponse = $client->request('POST', $uri, $params);
                        } catch (IOException | \Exception $e) {
                            $this->io->warning(sprintf('Skipping import for %s (from file %s) since the file is invalid or incompatible.', $option['name'], $option['installer_path']));
                            continue;
                        }
                        $result = json_decode($integrationResponse->getBody()->getContents(), true);
                        $connectorId = $result['id'];
                    } elseif($option['exists']) {
                        $iri = sprintf('connectors/%s/%s/?format=json', $connector['name'], $connector['version']);
                        if ($token) {
                            $jsonData = $this->sendRequest(sprintf('%s/api/integration/%s', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $iri), 'post', $connector, $tokenType);
                        } else {
                            $jsonData = $this->sendIntegrationRequest(sprintf('%s/%s', $this->container->getParameter('integrations.uri'), $iri), 'post', $connector);
                        }
                        $result = json_decode($jsonData, true);
                        $connectorId = $result['id'];
                    }
                    if ($result) {
                        $this->upsertImportedBy($option['name'], $importedBy);
                    }
                    if (!$result || (is_array($option) && array_key_exists("includeConfigurations", $option) && !$option['includeConfigurations'])) {
                        continue;
                    }
                    // Now, update the roles for operations
                    if (array_key_exists('operation_roles', $connector)) {
                        $this->updateOperationRoles($connector['operation_roles'], $result['operations'], $tokenType);
                    }
                    // First, delete existing configs
                    if (is_array($connector) && array_key_exists("configurations", $connector)) {
                        $this->deleteExistingConfigurations($connector['configurations'], $result['configuration'], $tokenType);
                    }

                    // Next, ingestion playbooks need to go in
                    if (array_key_exists('dataImports', $connector) && !empty($connector['ingestionPlaybooks']['collections'])) {
                        $this->importIngestionPlaybooks($connector['ingestionPlaybooks'], $tokenType);
                    }
                    // Install configurations afterwards
                    if (is_array($connector) && array_key_exists("configurations", $connector)) {
                        $this->importConfigurations($connector['configurations'], $connectorId, $tokenType);
                    }

                    if (!array_key_exists('dataImports', $connector) || empty($connector['dataImports'])) {
                        continue;
                    }
                    $this->importDataImports($connector['dataImports'], $tokenType);
                } else {
                    // Import configuration for installed custom connector
                    // For custom connector install_mode is not present in data.json
                    $connectorId = array_key_exists('existingId', $option) ? $option['existingId'] : null;
                    $this->importConfigurations($connector['configurations'], $connectorId, $tokenType);
                }

                break;
            }
        }
        return true;
    }

    private function installConnectorRpm($connector, $tokenType)
    {
        $token = $this->getTokenFromStorage();
        if ($token) {
            $this->sendRequest(sprintf('%s/api/integration/install-connector/?format=json', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY)), 'post', $connector, $tokenType);
        } else {
            $this->sendIntegrationRequest(sprintf('%s/install-connector/?format=json', $this->container->getParameter('integrations.uri')), 'post', $connector);
        }
        $status = 'In-Progress';
        $dep_installed = 'In-Progress';
        $counter = 0;
        while (($status == 'In-Progress' || $dep_installed == 'In-Progress') && $counter < 60) {
            sleep(4);
            $counter++;
            $iri = sprintf('connectors/%s/%s/?format=json', $connector['name'], $connector['version']);
            if ($token) {
                $jsonData = $this->sendRequest(sprintf('%s/api/integration/%s', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $iri), 'post', $connector, $tokenType);
            } else {
                $jsonData = $this->sendIntegrationRequest(sprintf('%s/%s', $this->container->getParameter('integrations.uri'), $iri), 'post', $connector);
            }
            $result = json_decode($jsonData, true);
            $status = $result['status'];
            $dep_installed = $result['requirements_installed'];
            if ($status == 'Failed') {
                // Connector failed to install
                throw new \Exception(sprintf('Connector %s failed to install. Please check connectors.log for more information.', $connector['rpm_name']));
            }
            if ($dep_installed == 'Failed') {
                // Connector failed to install
                $this->io->warning(sprintf('Connector %s installed, but dependencies failed to install. Please run dependency installation manually or check connectors.log for more information.', $connector['rpm_name']));
                return $result;
            }
        }

        return $counter == 60 ? false : $result;
    }

    private function updateOperationRoles($connectorOperationRoles, $operationRoles, $tokenType)
    {
        $token = $this->getTokenFromStorage();
        foreach ($operationRoles as $operation) {
            if (array_key_exists($operation['operation'], $connectorOperationRoles)) {
                $payload = [
                    'roles' => $connectorOperationRoles[$operation['operation']]
                ];
                $uri = sprintf('%s/api/integration/connectors/operations/%s/roles/?format=json', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $operation['id']);
                if ($token) {
                    $this->sendRequest($uri, 'post', $payload, $tokenType);
                } else {
                    $this->sendIntegrationRequest(str_replace('/api/integration/', '', $uri), 'post', $payload);
                }
            }
        }
    }

    private function deleteExistingConfigurations($configurations, $existingConfigurations, $tokenType)
    {
        $token = $this->getTokenFromStorage();
        if (isset($configurations)) {
            foreach ($configurations as $configuration) {
                foreach ($existingConfigurations as $existingConfiguration) {
                    if ($existingConfiguration['config_id'] == $configuration['config_id'] || $existingConfiguration['name'] == $configuration['name']) {
                        // Delete existing config
                        if ($token) {
                            $this->sendRequest(sprintf('%s/api/integration/configuration/%s/?format=json', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $existingConfiguration['id']), 'delete', null, $tokenType);
                        } else {
                            $this->sendIntegrationRequest(sprintf('%s/configuration/%s/?format=json', $this->container->getParameter('integrations.uri'), $existingConfiguration['id']), 'delete', null);
                        }
                        break;
                    }
                }
            }
        }
    }

    private function importDataImports($dataImports, $tokenType)
    {
        $token = $this->getTokenFromStorage();
        foreach ($dataImports as $dataImport) {
            try {
                if ($token) {
                    $scheduleId = array_key_exists('schedule', $dataImport) ?  $this->saveSchedule($dataImport['schedule'], $tokenType) : null;
                    $dataImport['metadata']['scheduleId'] = $scheduleId;
                    $this->sendRequest(sprintf('%s/api/integration/data-import/?format=json', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY)), 'post', $dataImport, $tokenType);
                } else {
                    $this->sendIntegrationRequest(sprintf('%s/data-import/?format=json', $this->container->getParameter('integrations.uri')), 'post', $dataImport);
                }
            } catch (BadRequestHttpException | \Exception $exception) {
                // Data import already exists, ignore
                continue;
            }
        }
    }

    private function importConfigurations($configurations, $connectorId, $tokenType)
    {
        $token = $this->getTokenFromStorage();
        if (isset($configurations)) {
            foreach ($configurations as $configuration) {
                $configuration['connector'] = $connectorId;
                if (array_key_exists('agent', $configuration)) {
                    unset($configuration['agent']);
                }
                // For Connector having no config eg nmap
                if (is_array($configuration['config']) && empty($configuration['config'])) {
                    unset($configuration['config']);
                }
                try {
                    if ($token) {
                        $this->sendRequest(sprintf('%s/api/integration/configuration/?format=json', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY)), 'post', $configuration, $tokenType);
                    } else {
                        $this->sendIntegrationRequest(sprintf('%s/configuration/?format=json', $this->container->getParameter('integrations.uri')), 'post', $configuration);
                    }
                } catch (BadRequestHttpException | \Exception $exception) {
                    // Config already exists, ignore
                    continue;
                }
            }
        }
    }

    private function importIngestionPlaybooks($ingestionPlaybooks, $tokenType, $importedBy = [])
    {
        if ($ingestionPlaybooks) {
            $allData = [
                'playbooks' => $ingestionPlaybooks
            ];
            $playbookOptions = $this->playbookConfig->getImportOptions($allData, 'playbooks', $tokenType);
            $this->playbookConfig->import($allData['playbooks'], $playbookOptions, $tokenType, $importedBy);
        }
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . PermissionsDictionaryBuilder::MODULE_CONNECTORS)) {
            return [];
        }
        $data = [];
        if (!empty($options)) {
            $connectors = $this->getConnectors($tokenType);
        }
        foreach ($options as $option) {
            foreach ($connectors as $connector) {
                // Load ingestion playbooks
                $rpmName = sprintf('cyops-connector-%s-%s', $connector['name'], $connector['version']);
                if ($option['value'] == $rpmName) {

                    $connectorObject = [
                        'name' => $connector['name'],
                        'label' => $connector['label'],
                        'category' => $connector['category'],
                        'description' => $connector['description'],
                        'publisher' => $connector['publisher'],
                        'version' => $connector['version'],
                        'operation_roles' => [],
                        'rpm_name' => null,
                        'configurations' => [],
                        'dataImports' => []
                    ];
                    foreach ($connector['operations'] as $operation) {
                        if (array_key_exists('roles', $operation) && !empty($operation['roles'])) {
                            $connectorObject['operation_roles'][$operation['operation']] = $operation['roles'];
                        }
                    }
                    if ($option['rpm']) {
                        $connectorObject['rpm_name'] = $rpmName;
                    }
                    if ($option['configurations']) {
                        $connectorObject['configurations'] = $connector['configuration'];
                        if (count($connector['configuration']) > 0) {
                            $configIds = array_column($connector['configuration'], 'config_id');
                            $options = [
                                'collections' => $configIds,
                                'globalVariables' => []
                            ];
                            $ingestionPlaybooks = $this->playbookConfig->export($options, $tokenType);
                            if ($ingestionPlaybooks) {
                                $connectorObject['ingestionPlaybooks'] = $ingestionPlaybooks;
                            }
                            foreach ($configIds as $configId) {
                                $configDataImports = $this->getDataImports($configId, $tokenType);
                                foreach ($configDataImports as $dataImport) {
                                    unset($dataImport['sample_data']);
                                    if ($dataImport['metadata']['scheduleId']) {
                                        $dataImport['schedule'] = $this->loadSchedule($dataImport['metadata']['scheduleId'], $tokenType);
                                    }
                                    $connectorObject['dataImports'][] = $dataImport;
                                }
                            }
                        }
                    }

                    $data[] = $connectorObject;
                    break;
                }
            }
        }
        return $data;
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        if (empty($options) || !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . PermissionsDictionaryBuilder::MODULE_CONNECTORS)) {
            return null;
        }
        $connectorDirectory = $tmpDirectory . $type;
        $connectorsInRepo = $this->loadConnectorRpms();
        $token = $this->getTokenFromStorage();
        $data = [];
        $metaData = [];
        $connectors = $this->getConnectors($tokenType);
        $exportJobType = $this->exportJob->getType();
        foreach ($options as $option) {
            foreach ($connectors as $connector) {
                // Load ingestion playbooks
                $rpmName = sprintf('cyops-connector-%s-%s', $connector['name'], $connector['version']);
                if ($option['value'] == $rpmName) {
                    $connectorObject = [
                        'name' => $connector['name'],
                        'label' => $connector['label'],
                        'category' => $connector['category'],
                        'description' => $connector['description'],
                        'publisher' => $connector['publisher'],
                        'version' => $connector['version'],
                        'operation_roles' => [],
                        'configurations' => [],
                        'dataImports' => []
                    ];
                    $metadataObject = [
                        'name' => $connectorObject['label'],
                        'apiName' => $connectorObject['name'],
                        'version' => $connectorObject['version']
                    ];
                    foreach ($connector['operations'] as $operation) {
                        if (array_key_exists('roles', $operation) && !empty($operation['roles'])) {
                            $connectorObject['operation_roles'][$operation['operation']] = $operation['roles'];
                        }
                    }
                    $fullKey = sprintf('%s_%s', $connector['name'], $connector['version']);
                    if (array_key_exists($fullKey, $connectorsInRepo) && $option['rpm']) {
                        $connectorObject['install_mode'] = 'rpm';
                        if ($exportJobType === 'SolutionPack Export') {
                            unset($connectorObject['version']);
                            unset($metadataObject['version']);
                        } else {
                            $connectorObject['rpm_name'] = $rpmName;
                            $connectorObject['installer_path'] = $rpmName;
                        }
                        # tgz export only if non-SP export
                    } elseif($option['rpm'] && ($exportJobType !== 'SolutionPack Export')) {
                        $uri = sprintf('%s/api/integration/connector/development/entity/%s/export/?format=json', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $connector['id']);
                        if ($token) {
                            $response = $this->sendRequest($uri, 'get', [], $tokenType);
                        } else {
                            $response = $this->sendIntegrationRequest(str_replace('/api/integration/', '', $uri), 'get', []);
                        }
                        $tgzFilename = $fullKey . '.tgz';
                        $this->writeBytesToFile(sprintf('%s/%s', $connectorDirectory, $tgzFilename), $response);
                        $connectorObject['install_mode'] = 'tgz';
                        $connectorObject['installer_path'] = $tgzFilename;
                    }

                    if ($option['configurations']) {
                        $connectorObject['configurations'] = $connector['configuration'];
                        if (is_countable($connector['configuration']) && count($connector['configuration']) > 0) {
                            $configIds = array_column($connector['configuration'], 'config_id');
                            $options = [
                                'collections' => $configIds,
                                'globalVariables' => []
                            ];
                            $ingestionPlaybooks = $this->playbookConfig->export($options, $tokenType);
                            if ($ingestionPlaybooks) {
                                $connectorObject['ingestionPlaybooks'] = $ingestionPlaybooks;
                            }
                            foreach ($configIds as $configId) {
                                $configDataImports = $this->getDataImports($configId, $tokenType);
                                foreach ($configDataImports as $dataImport) {
                                    unset($dataImport['sample_data']);
                                    if ($dataImport['metadata']['scheduleId']) {
                                        $dataImport['schedule'] = $this->loadSchedule($dataImport['metadata']['scheduleId'], $tokenType);
                                    }
                                    $connectorObject['dataImports'][] = $dataImport;
                                }
                            }
                        }
                    }

                    $data[] = $connectorObject;
                    $metaData[] = $metadataObject;
                    break;
                }
            }
        }
        $filename = $connectorDirectory . '/data.json';
        $this->writeToFile($filename, $data);
        $this->appendToFile($tmpDirectory . 'info.json', $type, $metaData);
        return $filename;
    }

    private function loadConnectorRpms()
    {
        try {
            $jsonUri = sprintf('https://%s/connectors/info/connectors-all.json', getenv("REPOSERVER") ? getenv("REPOSERVER") : getenv("product_yum_server"));
            $response = $this->sendRequest($jsonUri, 'GET', null, null, ['connect_timeout' => 15]);
            return json_decode($response, true);
        } catch (RequestException | \Exception $e) {
            $this->io->warning(sprintf('Cannot load connector RPMs from %s: %s', getenv("REPOSERVER") ? getenv("REPOSERVER") : getenv("product_yum_server"), $e->getMessage()));
            return [];
        }
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        foreach ($options as $value) {

            $value['value'] = sprintf('cyops-connector-%s-%s', $value['name'], $value['version']);
            $value['configurations'] = array_key_exists("includeConfigurations", $value) && $value['includeConfigurations'] &&  is_countable($value['configurationCount']) && count($value['configurationCount']) > 0 ? true : false;
            $value['include'] = true;
            $value['configCount'] = $value['configurationCount'];
            $value['rpm'] = true;
            array_push($exportOption, $value);
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        foreach ($options as $connector) {
            array_push($content, ["name" => $connector["label"]]);
        }
        return $content;
    }

    private function upsertImportedBy($connector, $importedBy)
    {
        /** @var SolutionPack */
        $solutionPack = $this->solutionPackUtilityService->getRecord($connector);
        if ($solutionPack  && $importedBy &&  $solutionPack->getName() != $importedBy[0]['apiName']) {
            $importArray = [];
            $appendImportBy = true;
            $existingImport = $solutionPack->getImportedBy();
            if ($existingImport) {
                foreach ($existingImport as $value) {
                    if ($value['name'] == $importedBy[0]['name']) {
                        $importArray[] = $importedBy[0];
                        $appendImportBy = false;
                    } else {
                        $importArray[] = $value;
                    }
                }
            } else {
                $importArray = $importedBy;
                $appendImportBy = false;
            }
            if ($appendImportBy) {
                $importArray[] = $importedBy[0];
            }
            $solutionPack->setImportedBy($importArray);
            $this->entityManager->persist($solutionPack);
            $this->entityManager->flush();
        }
    }
}
