<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Service\UtilityService;
use App\Security\Voter\FsrVoter;
use App\Entity\ViewTemplate\Dashboard;
use Doctrine\ORM\EntityManagerInterface;
use App\Providers\PermissionsDictionaryBuilder;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class DashboardConfig
 * @package App\Service\ConfigExportImport
 */
class DashboardConfig extends PortingConfig
{
    /** @var fsrVoter */
    private $fsrVoter;

    private $utilityService;

    public function __construct(
        FsrVoter $fsrVoter,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        UtilityService $utilityService,
        protected LoggerInterface $logger
    ) {
        $this->fsrVoter = $fsrVoter;
        $this->utilityService = $utilityService;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }

    /**
     * @param array $allData
     * @param string $key
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        if (
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.dashboard') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.dashboard') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.dashboard') ||
            !$data
        ) {
            return [];
        }
        $repository = $this->loadRepository(Dashboard::class);
        $options = [];
        foreach ($data as $dashboard) {
            $existingUuids = $repository->findOneBy(['uuid' => $dashboard['uuid']]);
            $existingDisplayNames = $repository->findOneBy(['displayName' => $dashboard['displayName']]);
            $foundRoles = $this->loadRolesFromReport($dashboard, $allData['roles']);

            $individualOptions = [];
            $individualOptions['uuid'] = $existingDisplayNames ? $existingDisplayNames->getUuid() : $dashboard['uuid'];
            $individualOptions['name'] = $dashboard['displayName'];
            $individualOptions['roles'] = $foundRoles;
            $individualOptions['include'] = true;
            $individualOptions['exists'] = $existingUuids || $existingDisplayNames ? true : false;

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        $dashboardsToAdd = [];
        $dashboardsToUpdate = [];
        foreach ($options as $dashboardOption) {
            if (!$dashboardOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($data as $dashboard) {
                if ($dashboardOption['uuid'] == $dashboard['uuid'] || $dashboardOption['name'] == $dashboard['displayName']) {
                    unset($dashboard['userId']);
                    $foundRoles = $this->loadRolesFromReport($dashboard, []);
                    $dashboard['roles'] = array_column($foundRoles, '@id');
                    $dashboard['filterArray'] = isset($dashboard['config']['filterArray']) ? $dashboard['config']['filterArray'] : $dashboard['filterArray'];
                    unset($dashboard['roleIds']);
                    $dashboard['importedBy'] = array_key_exists("importedBy", $dashboard) && !empty($dashboard['importedBy']) ? array_merge($dashboard['importedBy'], $importedBy) : $importedBy;
                    if (!$dashboardOption['exists']) {
                        $dashboardsToAdd[] = $dashboard;
                        break;
                    }
                    $dashboard['@id'] = sprintf('/api/3/dashboard/%s', $dashboardOption['uuid']);
                    unset($dashboard['uuid']);
                    $dashboardsToUpdate[] = $dashboard;
                    break;
                }
            }
        }
        $results = [];
        foreach ($dashboardsToAdd as $dashboard) {
            $fullURI = sprintf('%s/api/3/dashboard', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $result = $this->sendRequest($fullURI, 'post', $dashboard, $tokenType);
            $results[] = $result;
        }
        if (!empty($dashboardsToUpdate)) {
            $fullURI = sprintf('%s/api/3/update/dashboard', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $body = [
                "data" => $dashboardsToUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
        }

        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.dashboard')) {
            return [];
        }
        return $this->getResourceFromIRI($options, Dashboard::class, true);
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        foreach ($options as $value) {
            array_push($exportOption, $value['uuid']);
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        $filter = ['uuid' => $options];
        $entityRepo = $this->entityManager->getRepository(Dashboard::class);
        $dataObj = $this->utilityService->getItemWithWhere($entityRepo, $filter);
        foreach ($dataObj as $dashboard) {
            array_push($content, ['name' => $dashboard->getDisplayName()]);
        }
        return $content;
    }
    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $metaData = [];
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        foreach ($data as $item) {
            $filename = $tmpDirectory . $type . '/' . $this->cleanFileName($item['displayName']) . '.json';
            $this->writeToFile($filename, $item);
            $filenames[] = $filename;
            $metaData[] = ['name' => $item['displayName']];
        }
        $this->appendToFile($tmpDirectory . 'info.json', $type, $metaData);
        return $filenames;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $defaultJsonPath = $tmpDirectory . $key . '/data.json';
        if (file_exists($defaultJsonPath)) {
            $fileData = file_get_contents($defaultJsonPath);
            $jsonData = json_decode($fileData, true);
        } elseif (is_dir($tmpDirectory . $key)) {
            $jsonData = $this->loadDataFromDirectory($key, $tmpDirectory);
        }
        $roleData = [];
        if (is_dir($tmpDirectory . 'roles')) {
            $roleData = $this->loadDataFromDirectory('roles', $tmpDirectory);
        }
        $jsonData = isset($jsonData) ? $jsonData : [];
        return $this->getImportOptions([$key => $jsonData, 'roles' => $roleData], $key, $tokenType);
    }
}
