<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use Psr\Log\LoggerInterface;
use App\Command\SystemUpdate;
use InvalidArgumentException;
use App\Constants\AppConstants;
use App\Service\UtilityService;
use App\Constants\CacheConstants;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\Filesystem\Filesystem;
use App\Entity\ModelMetadata\StagingModelMetadata;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class ModuleConfig
 * @package App\Service\ConfigExportImport
 */
class ModuleConfig extends PortingConfig
{
    private $publishComplete = false;
    private $typeConflicts = [];
    private $filesystem;

    public function getTypeConflicts()
    {
        return $this->typeConflicts;
    }

    public function __construct(
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        private UtilityService $utilityService,
        protected LoggerInterface $logger
    ) {
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
        $this->utilityService = $utilityService;
        $this->filesystem = new Filesystem();
    }

    /**
     * @param array $allData
     * @param string $key
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        $moduleNames = array_column($data, 'type');
        $existingModules = $this->requestMmds($moduleNames, false);
        $repository = $this->loadRepository(StagingModelMetadata::class);
        $options = [];
        $uniqueConstraintAllowed = true;
        foreach ($data as $module) {
            /**
             * $module is like
             * {
             *   "@type": "StagingModelMetadata",
             *   "attributes": [...],
             *   "type": "alerts",
             *   "tableName": "alerts",
             *   ...
             * }
             */

            $existingModule = self::findBy($existingModules, 'type', $module['type'], true);
            $isModuleChanged = false;
            if (!$existingModule) {
                $existingModule = self::findBy($existingModules, 'tableName', $module['tableName'], true);
            }
            $moduleOptions = [];
            $moduleOptions['isModuleChanged'] = $isModuleChanged;
            $moduleOptions['include'] = true;
            $systemModuleTables = ['widgets', 'workflows', 'reporting', 'file'];
            $moduleOptions['isValid'] = $module['type'] && $module['tableName'] && $module['descriptions'] && $module['descriptions']['singular'] && $module['descriptions']['plural'] && $module['displayName'] && !in_array($module['tableName'], $systemModuleTables);
            $moduleOptions['type'] = $module['type'];
            $moduleOptions['name'] = $moduleOptions['isValid'] ? $module['descriptions']['plural'] : $module['type'];
            $moduleOptions['include'] = $moduleOptions['isValid'];
            $moduleOptions['exists'] = $existingModule ? true : false;
            $uniqueColumns = [];
            if (array_key_exists('uniqueConstraint', $module)) {
                $uniqueConstraintAllowed = $this->isUniqueConstraintAllowed($module, $existingModule);
                if ($uniqueConstraintAllowed && $module['uniqueConstraint'] && count($module['uniqueConstraint']) > 0) {
                    $uniqueColumns = array_map('strtolower', $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns']);
                } elseif ($existingModule && $existingModule['uniqueConstraint'] && count($existingModule['uniqueConstraint']) > 0) {
                    $uniqueColumns = array_map('strtolower', $existingModule['uniqueConstraint'][0][$existingModule['type'] . '_unique']['columns']);
                }
            }
            $moduleOptions['attributes'] = [];
            $newAttributeNames = [];
            foreach ($module['attributes'] as $attribute) {
                $existingAttribute = $existingModule ? self::findBy($existingModule['attributes'], 'name', $attribute['name'], true) : null;
                $attributeOptions = [];
                $attributeOptions['name'] = $attribute['name'];
                $attributeOptions['inUniqueConstraint'] = in_array(strtolower($attribute['name']), $uniqueColumns);
                $attributeOptions['title'] = $attribute['descriptions']['singular'];
                $attributeOptions['type'] = $attribute['formType'];
                $attributeOptions['exists'] = $existingAttribute ? true : false;
                if ($existingAttribute && $existingAttribute['system']) {
                    $attributeOptions['include'] = false;
                }
                if ($existingAttribute) {
                    $attributeOptions['@id'] = $existingAttribute['@id'];
                    $attributeOptions['system'] = $existingAttribute['system'];

                    if (is_array($existingAttribute['validation']) && (!array_key_exists('required', $existingAttribute['validation']) || $existingAttribute['validation']['required'] === false)) {
                        $attributeOptions['existingRequired'] = false;
                    }

                    if (is_array($existingAttribute['validation']) && !array_key_exists('maxlength', $existingAttribute['validation'])) {
                        $attributeOptions['existingMaxLength'] = 10485761;
                    } elseif (is_array($existingAttribute['validation']) && array_key_exists('maxlength', $existingAttribute['validation']) && $existingAttribute['validation']['maxlength'] > $attribute['validation']['maxlength']) {
                        $attributeOptions['existingMaxLength'] = $existingAttribute['validation']['maxlength'];
                    }

                    if (!$existingAttribute['system']) {
                        list($changes, $conflicts) = $this->compareModuleAttributes($attribute, $existingAttribute);
                        if (count($changes) > 0) {
                            $attributeOptions['changes'] = $changes;
                            $isModuleChanged = $isModuleChanged || (!empty($changes) && $changes) ? true : false;
                        }
                        if (count($conflicts) > 0) {
                            $attributeOptions['conflicts'] = $conflicts;
                            $moduleOptions['has_conflicts'] = true;
                            if ($this->importJob && $this->importJob->getType() === 'SolutionPack Import') {
                                $attributeOptions['include'] = false;
                            }
                        }
                    }
                }
                // If field is manyToMany or manyToOne and points to a non-existent module, skip it
                if (!array_key_exists('conflicts', $attributeOptions) && ($attribute['formType'] == 'manyToMany' || $attribute['formType'] == 'manyToOne' || $attribute['formType'] == 'oneToMany' || $attribute['formType'] == 'lookup')) {
                    $foundModule = self::findBy($data, 'type', $attribute['dataSource']['model'], true);
                    if (!$foundModule) {
                        // Doesn't exist in imported modules - check existing modules
                        $foundModule = $repository->findOneBy(['type' => $attribute['dataSource']['model']]);
                    }
                    if (!$foundModule) {
                        $attributeOptions['conflicts'] = [
                            [
                                'field' => 'type',
                                'existingValue' => null,
                                'newValue' => $attribute['dataSource']['model'],
                                'useExisting' => true,
                                'message' => 'Correlated module does not exist'
                            ]
                        ];
                        $attributeOptions['include'] = false;
                        $attributeOptions['inUniqueConstraint'] = false;
                        if (!empty($module['uniqueConstraint'][0][$module['type'] . '_unique']['columns']) && in_array($attribute['name'], $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'])) {
                            $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'] = array_diff($module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'], array($attribute['name']));
                        }
                    }
                }
                $newAttributeNames[] = strtolower($attribute['name']);
                $moduleOptions['attributes'][] = $attributeOptions;
            }

            if ($existingModule) {
                $changes = $this->compareModuleMetadata($module, $existingModule, $uniqueConstraintAllowed);
                $moduleOptions['changes'] = $changes;
                $isAtrributesCountChanged = count($module['attributes']) > 0 && (count($module['attributes']) < count($existingModule['attributes']) || count($module['attributes']) > count($existingModule['attributes']));
                $isModuleChanged = ($isAtrributesCountChanged || $isModuleChanged) || (!empty($changes) && $changes) ? true : false;
                $moduleOptions['isModuleChanged'] = $isModuleChanged;
                $moduleOptions['@id'] = $existingModule['@id'];
                // Check for attributes that may be removed if replaced
                foreach ($existingModule['attributes'] as $existingAttribute) {
                    if (array_search(strtolower($existingAttribute['name']), $newAttributeNames) === false) {
                        $foundAttribute = null;
                        foreach ($module['attributes'] as $attribute) {
                            $bothManyToMany = $existingAttribute['formType'] == 'manyToMany' && $attribute['formType'] == 'manyToMany';
                            $pointingToSameModel = isset($existingAttribute['dataSource']) && isset($attribute['dataSource']) && array_key_exists("model", $existingAttribute['dataSource']) && array_key_exists("model", $attribute['dataSource']) && $existingAttribute['dataSource']['model'] == $attribute['dataSource']['model'];
                            if ($bothManyToMany && $pointingToSameModel) {
                                $foundAttribute = $attribute;
                            }
                        }
                        if ($foundAttribute) {
                            // If another field is a many to many pointing to the same model, we consider them the same
                            continue;
                        }
                        $attributeOptions = [];
                        $attributeOptions['name'] = $existingAttribute['name'];
                        $attributeOptions['@id'] = $existingAttribute['@id'];
                        $attributeOptions['include'] = true;
                        $attributeOptions['title'] = $existingAttribute['descriptions']['singular'];
                        $attributeOptions['type'] = $existingAttribute['formType'];
                        $attributeOptions['skipped_in_import'] = true;
                        $attributeOptions['inUniqueConstraint'] = in_array(strtolower($existingAttribute['name']), $uniqueColumns);
                        $moduleOptions['attributes'][] = $attributeOptions;
                    }
                }
            }

            $options[] = $moduleOptions;
        }
        return $options;
    }

    private function isUniqueConstraintAllowed(array $module, $existingModule)
    {
        $traitsFixturesFile = file_get_contents('/opt/cyops-api/src/StaticFile/TraitsFixtures.json');
        $traitsFixtures = json_decode($traitsFixturesFile, true);

        foreach ($traitsFixtures as $key => $value) {
            if (array_key_exists($key, $module) && $module[$key]) {
                $module['attributes'] = array_merge($module['attributes'], $value);
            }
            if (is_array($existingModule) && array_key_exists($key, $existingModule) && $existingModule[$key]) {
                $existingModule['attributes'] = array_merge($existingModule['attributes'], $value);
            }
        }

        $uniquenessAllowed = true;
        try {
            $columns = [];
            $oldColumn = [];
            $this->logger->debug("Module Name:" . $module['type']);
            $uniqueness = array_key_exists('uniqueConstraint', $module) && is_array($module['uniqueConstraint']) && count($module['uniqueConstraint']) > 0;
            $uniquenessOld = $existingModule && array_key_exists('uniqueConstraint', $existingModule) && is_array($existingModule['uniqueConstraint']) && count($existingModule['uniqueConstraint']) > 0;

            if ($uniquenessOld) {
                $oldColumn = $existingModule['uniqueConstraint'][0][$existingModule['type'] . '_unique']['columns'];
            }

            if ($uniqueness) {
                $columns = $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'];
                if ($oldColumn && $oldColumn === $columns) {
                    $this->logger->debug("Old Column in Unique Constraint and new column are same, unique constraint is allowed");
                    return true;
                }
                $this->logger->debug("Procedings with validating if unique constraint can be allowed");

                if ($existingModule) {
                    $singular = $module['descriptions']['singular'];
                    $escapedSingular = preg_replace('/[^a-z0-9]+/i', '', $singular);
                    $classPath = $this->container->getParameter("type_map")[$escapedSingular] ?? '\\App\\Entity\\Custom\\' . $escapedSingular;
                    $queryBuilder = $this->entityManager->getRepository($classPath)->createQueryBuilder('a');
                    $queryBuilder->select('count(a) as the_count');

                    foreach ($columns as $column) {
                        $attribute = self::findBy($module['attributes'], 'name', $column, true);
                        $existingAttribute = self::findBy($existingModule['attributes'], 'name', $column, true);
                        if (!$attribute && !$existingAttribute) {
                            // The attribute does not exist
                            $this->logger->debug("the attribute does not exists, returning false");
                            return false;
                        }
                        if (!$existingAttribute) {
                            // We don't need to check non-existent attributes for uniqueness
                            $this->logger->debug("non-existent attributes found, returning true");
                            return true;
                        }
                        if ($attribute['dataSource'] && !empty($attribute['dataSource']) && $attribute['dataSource']['model']) {
                            $selectText = sprintf('IDENTITY(a.%s)', $column);
                        } else {
                            $selectText = sprintf('a.%s', $column);
                        }
                        $queryBuilder->addSelect($selectText);
                        $queryBuilder->andWhere($selectText . ' IS NOT NULL');
                        $queryBuilder->addGroupBy(sprintf('a.%s', $column));
                    }
                    $query = $queryBuilder
                        ->having('count(a) > 1')
                        ->getQuery();
                    $result = $query->getResult();
                    $uniquenessAllowed = count($result) == 0;
                }
            }
        } catch (InvalidArgumentException | \Exception $e) {
            $this->logger->error("Exception occurred while checking for unique constraint " . $e->getMessage());
            $uniquenessAllowed = false;
        }
        return $uniquenessAllowed;
    }

    /**
     * Determines changes and conflicts between attributes
     *
     * @param array $attribute
     * @param array $existingAttribute
     * @return array
     */
    private function compareModuleMetadata(array $module, array $existingModule, $uniquenessAllowed)
    {
        $changes = [];
        unset($module['attributes']);
        unset($existingModule['attributes']);
        $diffs = $this->arrayRecursiveDiff($module, $existingModule);
        $diffs['descriptions'] = array_diff($module['descriptions'], $existingModule['descriptions']);

        $existingModule['defaultSort'] = $existingModule['defaultSort'] ? $existingModule['defaultSort'] : [];
        $module['defaultSort'] = array_key_exists("defaultSort", $module) && $module['defaultSort'] ? $module['defaultSort'] : [];
        if (array_diff($module['defaultSort'], $existingModule['defaultSort']) || array_diff($existingModule['defaultSort'], $module['defaultSort'])) {
            $defaultSorts = array_column($module['defaultSort'], 'field');
            $existingDefaultSorts = array_column($existingModule['defaultSort'], 'field');
            $changes[] = [
                'field' => 'defaultSort',
                'message' => 'Changing default sort from (' . implode(', ', $existingDefaultSorts) . ') to (' . implode(', ', $defaultSorts) . ')'
            ];
        }

        if (array_key_exists('uniqueConstraint', $module)) {
            $uniqueness = is_array($module['uniqueConstraint']) && count($module['uniqueConstraint']) > 0;
            $existingUniqueness = array_key_exists('uniqueConstraint', $existingModule) && is_array($existingModule['uniqueConstraint']) && count($existingModule['uniqueConstraint']) > 0;
            $columns = [];
            if ($uniqueness) {
                $columns = $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'];
            }
            $existingColumns = [];
            if ($existingUniqueness) {
                $existingColumns = $existingModule['uniqueConstraint'][0][$existingModule['type'] . '_unique']['columns'];
            }
            if (count(array_diff($columns, $existingColumns)) > 0 || count(array_diff($existingColumns, $columns)) > 0) {
                if (!$uniquenessAllowed) {
                    $changes[] = [
                        'field' => 'uniqueConstraint',
                        'message' => 'Preventing a change to unique constraints on fields (' . implode(', ', $columns) . '). Either the existing data is not unique or one of the fields does not exist.',
                        'error' => true
                    ];
                } elseif ($uniqueness && $existingUniqueness) {
                    $changes[] = [
                        'field' => 'uniqueConstraint',
                        'message' => 'Changing unique constraints from fields (' . implode(', ', $existingColumns) . ') to fields (' . implode(', ', $columns) . ')'
                    ];
                } elseif ($uniqueness) {
                    $changes[] = [
                        'field' => 'uniqueConstraint',
                        'message' => 'Adding unique constraints on fields (' . implode(', ', $columns) . ')'
                    ];
                } elseif ($existingUniqueness) {
                    $changes[] = [
                        'field' => 'uniqueConstraint',
                        'message' => 'Removing unique constraints from fields (' . implode(', ', $existingColumns) . ')'
                    ];
                }
            }
        }

        unset($diffs['defaultSort']);
        unset($diffs['uniqueConstraint']);
        unset($diffs['@id']);
        unset($diffs['@type']);
        unset($diffs['sattrib']);
        unset($diffs['orderIndex']);
        unset($diffs['uuid']);
        foreach ($diffs as $key => $value) {
            $existingValue = array_key_exists($key, $existingModule) ? $existingModule[$key] : null;
            if (is_array($value)) {
                foreach ($value as $subKey => $subValue) {
                    $changes[] = [
                        'field' => $key . ' -> ' . $subKey,
                        'message' => ucfirst($key) . ' -> ' . $subKey . ' changed from ' . $existingValue[$subKey] . ' to ' . $subValue
                    ];
                }
                continue;
            }
            $changes[] = [
                'field' => $key,
                'message' => $key . ' changed from ' . $existingValue . ' to ' . $module[$key]
            ];
        }
        return $changes;
    }

    /**
     * Determines changes and conflicts between attributes
     *
     * @param array $attribute
     * @param array $existingAttribute
     * @return array
     */
    private function compareModuleAttributes(array $attribute, array $existingAttribute)
    {
        $conflicts = [];
        $changes = [];
        $supportedTypeConversions = [
            ['integer', 'string'],
            ['float', 'string'],
            ['manyToMany', 'manyToOne']
        ];
        $imcompatibleFormTypeConversion = ['picklist', 'multiselectpicklist'];
        // If the type conversion is not valid, issue a conflict
        $isDateTimeConversion = $existingAttribute['formType'] != $attribute['formType'] && ($existingAttribute['formType'] == 'datetime' || $attribute['formType'] == 'datetime' || $existingAttribute['formType'] == 'date' || $attribute['formType'] == 'date');
        $isPicklistConversion = in_array($existingAttribute['formType'], $imcompatibleFormTypeConversion) && in_array($attribute['formType'], $imcompatibleFormTypeConversion) && $existingAttribute['formType'] != $attribute['formType'];
        if ($isPicklistConversion || $isDateTimeConversion || $existingAttribute['type'] != $attribute['type'] && !in_array([$existingAttribute['type'], $attribute['type']], $supportedTypeConversions)) {
            $conflicts[] = [
                'field' => 'type',
                'existingValue' => $existingAttribute['type'],
                'newValue' => $attribute['type'],
                'useExisting' => true,
                'message' => 'Data type conversion from "' . $existingAttribute['type'] . '"('  . $existingAttribute['formType'] . ') to "' . $attribute['type'] . '" (' . $attribute['formType'] . ') not supported'
            ];
            $conflicts[] = [
                'field' => 'formType',
                'existingValue' => $existingAttribute['formType'],
                'newValue' => $attribute['formType'],
                'useExisting' => true
            ];
        }

        // Added below Line to remove array to string conversion warning.
        // Reason the same is coming is some attributes are containing JSONB or array information
        // Thus making the overall comparison happening for two dimensional array which gives warning.

        // $newAttributeDiffs = array_diff($attribute, $existingAttribute);
        $newAttributeDiffs = $this->arrayRecursiveDiff($attribute, $existingAttribute);
        unset($newAttributeDiffs['@id']);
        unset($newAttributeDiffs['@type']);
        unset($newAttributeDiffs['sattrib']);
        unset($newAttributeDiffs['orderIndex']);
        foreach ($newAttributeDiffs as $key => $value) {
            $existingValue = array_key_exists($key, $existingAttribute) ? $existingAttribute[$key] : null;
            $changes[$attribute['name']][] = [
                'field' => $key,
                'old' => $existingValue,
                'new' => $value
            ];
        }
        return [$changes, $conflicts];
    }

    private function arrayRecursiveDiff($sourceArray, $compareToArray)
    {
        $diff = array();
        foreach ($sourceArray as $mKey => $mValue) {
            if (array_key_exists($mKey, $compareToArray)) {
                if (is_array($mValue) && is_array($compareToArray[$mKey])) {
                    $aRecursiveDiff = $this->arrayRecursiveDiff($mValue, $compareToArray[$mKey]);
                    if (count($aRecursiveDiff)) {
                        $diff[$mKey] = $aRecursiveDiff;
                    }
                } else {
                    if ($mValue != $compareToArray[$mKey]) {
                        $diff[$mKey] = $mValue;
                    }
                }
            } else {
                $diff[$mKey] = $mValue;
            }
        }

        return $diff;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        $modulesToAdd = [];
        $modulesToUpdate = [];
        $moduleTypes = array_column($options, 'type');
        $existingModules = $this->requestMmds($moduleTypes, true);
        $isDistributedInstance = $this->utilityService->getRequiredLicenseDetails(CacheConstants::IS_DISTRIBUTED_INSTANCE);
        foreach ($options as $moduleOption) {
            $isModuleChanged = false;
            if ((array_key_exists('isValid', $moduleOption) && !$moduleOption['isValid']) || (array_key_exists('include', $moduleOption) && !$moduleOption['include'])) {
                // Do not merge skipped modules
                continue;
            }
            $module = null;
            foreach ($data as $item) {
                if ($item['type'] == $moduleOption['type']) {
                    $module = $item;
                    $isModuleChanged = array_key_exists('isModuleChanged', $moduleOption) && $moduleOption['isModuleChanged'] ?  true : false;
                    break;
                }
            }
            if (!$moduleOption['exists']) {
                // If it does not exist, just put it on the list to import
                unset($module['uuid']);
                unset($module['@id']);
                $existingModule = null;
                if (!$isDistributedInstance) {
                    if ($module['peerReplicable']) {
                        $module['peerReplicable'] = false;
                    }
                    if (array_key_exists('uniqueConstraint', $module) && !empty($module['uniqueConstraint'])) {
                        $uniqueConstraintColumns = $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'];
                        if (in_array('tenant', $uniqueConstraintColumns)) {
                            $uniqueConstraintColumns = array_diff($uniqueConstraintColumns, ["tenant"]);
                            $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'] = $uniqueConstraintColumns;
                        }
                    }
                }
            } else {
                $existingModule = self::findBy($existingModules, 'type', $moduleOption['type'], true);
                $module['@id'] = $moduleOption['@id'];

                // We don't allow changing these on import
                $module['peerReplicable'] = $existingModule['peerReplicable'];
                $module['trackable'] = $existingModule['trackable'];
                $module['ownable'] = $existingModule['ownable'];
                $module['userOwnable'] = $existingModule['userOwnable'];
                $module['importedBy'] = $existingModule['importedBy'];
                $module['archivable'] = $existingModule['archivable'];
                $module['paused'] = $existingModule['paused'];
                $module['softDeleteable'] = $existingModule['softDeleteable'];
                $module['partitionBy'] = $existingModule['partitionBy'];
                $module['archivalCriteria'] = $existingModule['archivalCriteria'];
                $module['archivalFilters'] = $existingModule['archivalFilters'];
                $module['replicationFilters'] = $existingModule['replicationFilters'];

                if (array_key_exists('uniqueConstraint', $module) && !$this->isUniqueConstraintAllowed($module, $existingModule)) {
                    $module['uniqueConstraint'] = $existingModule['uniqueConstraint'] ? $existingModule['uniqueConstraint'] : [];
                }
            }

            $moduleAttributes = [];

            $attributeList = [];
            foreach ($moduleOption['attributes'] as $attributeOption) {
                if (array_key_exists('exists', $attributeOption) && !$attributeOption['exists'] && array_key_exists('include', $attributeOption) && !$attributeOption['include']) {
                    // Doesn't exist and user chose not to include it
                    if (!empty($module['uniqueConstraint'][0][$module['type'] . '_unique']['columns']) && in_array($attributeOption['name'], $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns']) && count($module['uniqueConstraint'][0][$module['type'] . '_unique']['columns']) > 1) {
                        $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'] = array_diff($module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'], array($attributeOption['name']));
                    } elseif (!empty($module['uniqueConstraint'][0][$module['type'] . '_unique']['columns']) && in_array($attributeOption['name'], $module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'])) {
                        unset($module['uniqueConstraint'][0]);
                    }
                    continue;
                }
                $attribute = self::findBy($module['attributes'], 'name', $attributeOption['name'], true);
                if ($existingModule && (!$attribute || (array_key_exists('include', $attributeOption) && !$attributeOption['include']))) {
                    // !$attribute means the field is in existing mmd but not in imported one
                    // !$attributeOption['include'] means the imported one
                    $attribute = self::findBy($existingModule['attributes'], 'name', $attributeOption['name'], true);
                }
                $conflicts = array_key_exists('conflicts', $attributeOption) ? $attributeOption['conflicts'] : [];
                foreach ($conflicts as $conflict) {
                    if (!$conflict['useExisting']) {
                        continue;
                    }
                    $attribute[$conflict['field']] = $conflict['existingValue'];
                    if ($conflict['field'] == 'type') {
                        $this->typeConflicts[] = $attributeOption['name'];
                    }
                }
                unset($attribute['sattrib']);
                if ($moduleOption['exists'] && array_key_exists("@id", $attributeOption)) {
                    $attribute['@id'] = $attributeOption['@id'];
                    unset($attribute['uuid']);
                } else {
                    unset($attribute['@id']);
                }

                if (array_key_exists('existingRequired', $attributeOption) && $attributeOption['existingRequired'] === false) {
                    $attribute['validation']['required'] = false;
                }

                if (array_key_exists('existingMaxLength', $attributeOption) && $attributeOption['existingMaxLength']) {
                    $attribute['validation']['maxlength'] = $attributeOption['existingMaxLength'];
                }

                $attribute['importedBy'] = array_key_exists("importedBy", $module) && !empty($attribute['importedBy']) ? array_merge($attribute['importedBy'], $importedBy) : $importedBy;
                $moduleAttributes[] = $attribute;

                if (in_array(strtolower($attributeOption['name']), $attributeList)) {
                    throw new \Exception(sprintf("Duplicate field detected on module %s with attribute name %s", $moduleOption['type'], $attributeOption['name']));
                } else {
                    array_push($attributeList, \strtolower($attributeOption['name']));
                }
            }
            if (is_null($module)) {
                continue;
            }
            $module['attributes'] = $moduleAttributes;
            $module['importedBy'] = array_key_exists("importedBy", $module) && !empty($module['importedBy']) ? array_merge($module['importedBy'], $importedBy) : $importedBy;
            if ($moduleOption['exists'] && $isModuleChanged) {
                if (isset($module['uuid'])) {
                    unset($module['uuid']);
                }
                $modulesToUpdate[] = $module;
            } elseif (!$moduleOption['exists']) {
                $modulesToAdd[] = $module;
            }
        }
        // Once the lists of modules to add and update are there, we need to fix invalid references
        foreach ($modulesToAdd as $index => $module) {
            [$modulesToAdd[$index], $skippedAttributes] = $this->removeInvalidReferences($module, array_merge($modulesToAdd, $modulesToUpdate));
            if (!empty($module['uniqueConstraint'])) {
                $modulesToAdd[$index]['uniqueConstraint'][0][$module['type'] . '_unique']['columns'] = array_merge(array_diff($module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'], $skippedAttributes));
                if (empty($modulesToAdd[$index]['uniqueConstraint'][0][$module['type'] . '_unique']['columns'])) {
                    $modulesToAdd[$index]['uniqueConstraint'] = [];
                }
            }
        }
        foreach ($modulesToUpdate as $index => $module) {
            [$modulesToUpdate[$index], $skippedAttributes] = $this->removeInvalidReferences($module, array_merge($modulesToAdd, $modulesToUpdate));

            // Checking if the existing module's unique constraint conflicts with the module to import
            // This might still fail if the incoming JSON of records has duplicate records, but we cannot validate it.
            $existingModule = self::findBy($existingModules, 'type', $module['type'], true);
            if (!$this->isUniqueConstraintAllowed($module, $existingModule)) {
                $modulesToUpdate[$index]['uniqueConstraint'] = $existingModule['uniqueConstraint'] ? $existingModule['uniqueConstraint'] : [];
            } else {
                if (!empty($module['uniqueConstraint'])) {
                    $modulesToUpdate[$index]['uniqueConstraint'][0][$module['type'] . '_unique']['columns'] = array_merge(array_diff($module['uniqueConstraint'][0][$module['type'] . '_unique']['columns'], $skippedAttributes));
                    if (empty($modulesToUpdate[$index]['uniqueConstraint'][0][$module['type'] . '_unique']['columns'])) {
                        $modulesToUpdate[$index]['uniqueConstraint'] = [];
                    }
                }
            }
        }

        $results = [];
        // Import modules without attributes
        if (!empty($modulesToUpdate)) {
            $fullURI = sprintf('%s/api/3/update/staging_model_metadatas?$export=true&$relationships=true', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $body = [
                'data' => $modulesToUpdate
            ];
            $results[] = (bool)$this->sendRequest($fullURI, 'put', $body, $tokenType);
        }
        if (!empty($modulesToAdd)) {
            $fullURI = sprintf('%s/api/3/insert/staging_model_metadatas?$export=true&$relationships=true', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $body = [
                'data' => $modulesToAdd
            ];
            $results[] = (bool)$this->sendRequest($fullURI, 'post', $body, $tokenType);
        }
        if (!empty($modulesToUpdate) || !empty($modulesToAdd)) {
            $this->addMessage('Publishing changes');
            [$publishSuccess, $errors] = $this->publish();
            if (!$publishSuccess) {
                throw new \Exception(sprintf('Publish failed with exception: %s', implode(', ', $errors)));
            }
        }
        $this->publishComplete = true;

        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Parses the module and removes invalid relationships, inversed fields, and attributes
     *
     * @param array $module
     * @param array $data
     * @return array
     */
    private function removeInvalidReferences(array $module, array $modulesBeingImported)
    {
        $skippedAttributes = [];
        $moduleAttributesToKeep = [];
        foreach ($module['attributes'] as $attribute) {
            if ($attribute['dataSource'] && array_key_exists('model', $attribute['dataSource']) && ($attribute['formType'] == 'manyToMany' || $attribute['formType'] == 'oneToMany' || $attribute['formType'] == 'lookup')) {
                // Check inversedField
                // Get related module
                $relatedModule = self::findBy($modulesBeingImported, 'type', $attribute['dataSource']['model'], true);
                if (!$relatedModule) {
                    // Check existing modules
                    $existing_modules = $this->requestMmds([$attribute['dataSource']['model']], false);
                    $relatedModule = end($existing_modules);
                }
                if (!$relatedModule) {
                    //if a field is skipped adding it to the array
                    $skippedAttributes[] = $attribute['name'];
                    continue;
                }
                if (array_key_exists('inversedField', $attribute) && $attribute['inversedField']) {
                    $inversedAttribute = self::findBy($relatedModule['attributes'], 'name', $attribute['inversedField'], true);
                    // If it doesn't find the inversed attribute, clear it
                    if (!$inversedAttribute) {
                        $attribute['inversedField'] = null;
                        // Check backwards inversed field
                    } elseif (array_key_exists('inversedField', $inversedAttribute) && $inversedAttribute['inversedField'] && !self::findBy($module['attributes'], 'name', $inversedAttribute['inversedField'], true)) {
                        // If there is no backwards referencing field, remove this inversed field
                        $attribute['inversedField'] = null;
                    } elseif ($attribute['formType'] == 'lookup' && $attribute['dataSource']['model'] == $module['type']) {
                        // Lookups cannot have inversedFields
                        $attribute['inversedField'] = null;
                    }
                } elseif ($attribute['formType'] == 'manyToMany' && $attribute['dataSource']['model'] == $module['type']) {
                    // self referencing relationship - add an inversedField to help users out
                    $attribute['inversedField'] = $attribute['name'];
                }
            }
            $moduleAttributesToKeep[] = $attribute;
        }
        $module['attributes'] = $moduleAttributesToKeep;
        return [$module, $skippedAttributes];
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options list of MMDs to export
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        return $this->requestMmds($options);
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $metaData = [];
        $data = $this->requestMmds($options);
        if (!$data) {
            return $data;
        }
        $modulesDirectory = $tmpDirectory . $type . '/';
        foreach ($data as $item) {
            $filename = $modulesDirectory . $this->cleanFileName($item['type']) . '/mmd.json';
            if (array_key_exists('@id',$item)){
                unset($item['@id']);
            }
            if (array_key_exists('uuid',$item)){
                unset($item['uuid']);
            }
            $this->writeToFile($filename, $item);
            $metaData[$item['module']] = ['name' => $item['descriptions']['plural'], 'apiName' => $item['type']];
        }
        $this->appendToFile($tmpDirectory . 'info.json', $type, $metaData);
        return $filename;
    }

    /**
     * Requests MMDs with the option to use $export tag or not
     *
     * @param array $modules
     * @param boolean $export
     * @return array
     */
    private function requestMmds(array $modules, $ignoreTraitsAttributes = false)
    {
        /** @var ManagerRegistry $managerRegistry */
        $managerRegistry = $this->container->get('doctrine');
        $smmdManager = $managerRegistry->getManagerForClass(StagingModelMetadata::class);
        $smmdRepository = $smmdManager->getRepository(StagingModelMetadata::class);
        $context = $this->getDefaultNormalizationContext(StagingModelMetadata::class, false);
        $context['relationships'] = true;
        if ($ignoreTraitsAttributes && isset($context['ignored_attributes']) && !empty($context['ignored_attributes'])) {
            $importedByPos = array_search('importedBy', $context['ignored_attributes']);
            if ($importedByPos  !== false) {
                array_splice($context['ignored_attributes'], $importedByPos, 1);
            }
        }
        $exportModules = [];
        foreach ($modules as $module) {
            // backwards compatibility
            if (is_string($module)) {
                $value = $module;
            } else {
                $value = $module['value'];
            }
            $stagingMmd = $smmdRepository->findOneBy(['type' => $value]);
            if (is_null($stagingMmd)) {
                continue;
            }
            $normalizedObject = $this->normalizer->normalize($stagingMmd, 'jsonld', $context);
            $includeMetadata = true;
            if (is_array($module) && array_key_exists('includedAttributes', $module)) {
                $includedAttributes = [];
                foreach ($normalizedObject['attributes'] as $attribute) {
                    if (in_array($attribute['name'], $module['includedAttributes'])) {
                        unset($attribute['uuid']);
                        unset($attribute['@id']);
                        $includedAttributes[] = $attribute;
                    } else {
                        $includeMetadata = false;
                    }
                }
                $normalizedObject['attributes'] = $includedAttributes;
            }
            if (!$includeMetadata) {
                $normalizedObject = [
                    '@id' => $normalizedObject['@id'],
                    'uuid' => $normalizedObject['uuid'],
                    'descriptions' => $normalizedObject['descriptions'],
                    'type' => $normalizedObject['type'],
                    'attributes' => $normalizedObject['attributes'],
                    'tableName' => $normalizedObject['tableName'],
                    'displayName' => $normalizedObject['displayName'],
                ];
            }

            $exportModules[] = $normalizedObject;
        }
        return $exportModules;
    }

    /**
     * 
     *
     * @param array $config
     * @return array
     */
    public function publish()
    {
        $rootDir = $this->container->getParameter('kernel.project_dir');
        $file503 = sprintf('%s/web/503.json', $rootDir);

        $systemUpdate = new SystemUpdate();
        $systemUpdate->setMessage('Publish queued.');
        $systemUpdate->setProgressPercent(0);
        $this->filesystem->dumpFile($file503, json_encode($systemUpdate));

        $command = '/usr/bin/php /opt/cyops-api/bin/console app:system:update -la';

        chdir($this->container->getParameter('kernel.project_dir') . '/../');
        $logFile = '/var/log/cyops/cyops-api/last_system_publish.log';
        exec(sprintf('bash -c "%s > %s 2>&1"', $command, $logFile));

        $searchfor = ["An exception occurred", "ERROR:", "ORMException", "Compile Error:", "There is no column with name", "UnexpectedValueException"];
        $errors = [];
        $contents = file_get_contents($logFile);

        foreach ($searchfor as $searchString) {
            // escape special characters in the query
            $pattern = preg_quote($searchString, '/');
            // finalise the regular expression, matching the whole line
            $pattern = "/^.*$pattern.*$\\n.*$\\n/m";
            // search, and store all matching occurences in $matches
            if (preg_match_all($pattern, $contents, $matches)) {
                $errors = array_merge($errors, array_unique($matches[0]));
            }
        }
        if (empty($errors)) {
            return [true, null];
        } else {
            foreach ($errors as $key => $er) {
                $errors[$key] = preg_replace('#\\x1b[[][^A-Za-z]*[A-Za-z]#', '', $er);
            }
            return [false, $errors];
        }
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->getImportOptions([$key => $data], $key, $tokenType);
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = [])
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->import($data, $options, $tokenType, $importedBy);
    }

    public function loadDataFromDirectory(string $key, string $tmpDirectory)
    {
        $folderPath = $tmpDirectory . $key . '/';
        $folders = [];
        if (file_exists($folderPath)) {
            $folders = scandir($folderPath);
        }
        $data = [];
        foreach ($folders as $folder) {
            $mmdPath = $folderPath . $folder . '/mmd.json';
            if (file_exists($mmdPath)) {
                $fileData = file_get_contents($mmdPath);
                $jsonData = json_decode($fileData, true);
                $data[$folder] = $jsonData;
            }
        }
        return $data;
    }

    public function isPublishComplete()
    {
        return $this->publishComplete;
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        foreach ($options as $value) {
            if (is_array($value) && array_key_exists('includedAttributes', $value)) {
                $tempVar = [];
                $tempVar['name'] = $value['type'];
                $tempVar['includedAttributes'] = [];
                foreach ($value['attributes'] as $value) {
                    array_push($tempVar['includedAttributes'], $value['attributes']['name']);
                }
                array_push($exportOption, $tempVar);
            } else {
                array_push($exportOption, $value['type']);
            }
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        $mmdType = [];
        foreach ($options as $module) {
            if (is_array($module)) {
                $mmdType[] = array_key_exists('value', $module) ? $module['value'] : null;
            } elseif (is_string($module)) {
                $mmdType[] = $module;
            }
        }
        $filter = ['type' => $mmdType];
        $entityRepo = $this->entityManager->getRepository(StagingModelMetadata::class);
        $dataObj = $this->utilityService->getItemWithWhere($entityRepo, $filter);
        foreach ($dataObj as $item) {
            $content[$item->getType()] = ["name" => $item->getDescriptions()["plural"], "apiName" => $item->getType()];
        }
        return $content;
    }
}
