<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use Psr\Log\LoggerInterface;
use App\Entity\Core\Picklist;
use App\Constants\AppConstants;
use App\Service\UtilityService;
use App\Entity\Core\PicklistName;
use Psr\Container\ContainerInterface;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class PicklistNameConfig
 * @package App\Service\ConfigExportImport
 */
class PicklistNameConfig extends PortingConfig
{
    private $importOptionsGenerator;
    private $utilityService;

    public function __construct(
        ImportOptionsGenerator $importOptionsGenerator,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        UtilityService $utilityService,
        protected LoggerInterface $logger
    ) {
        $this->importOptionsGenerator = $importOptionsGenerator;
        $this->utilityService = $utilityService;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }

    /**
     * @param array $allData
     * @param string $key
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        $modules = array_key_exists($key, $allData) ? $allData['modules'] : [];
        $usedPicklistNames = [];
        foreach ($modules as $module) {
            foreach ($module['attributes'] as $attribute) {
                if ($attribute['formType'] == 'picklist' && array_key_exists('dataSource', $attribute) && array_key_exists('query', $attribute['dataSource'])) {
                    $picklistName = $attribute['dataSource']['query']['filters'][0]['value'];
                    $usedPicklistNames[$picklistName] = array_key_exists($picklistName, $usedPicklistNames) ? $usedPicklistNames[$picklistName] : [];
                    $usedPicklistNames[$picklistName][] = $module['type'];
                    $usedPicklistNames[$picklistName] = array_unique($usedPicklistNames[$picklistName]);
                }
            }
        }
        $repository = $this->loadRepository(PicklistName::class);
        $options = [];
        foreach ($data as $picklistName) {
            $existing = $repository->findOneBy(['uuid' => $picklistName['uuid']]);
            $existingName = $repository->findOneBy(['name' => $picklistName['name']]);
            $individualOptions = [];
            $individualOptions['uuid'] = $picklistName['uuid'];
            $individualOptions['existingUuid'] = $existingName ? $existingName->getUuid() : null;
            $individualOptions['name'] = $picklistName['name'];
            $individualOptions['count'] = count($picklistName['picklists']);
            $individualOptions['include'] = true;
            $individualOptions['exists'] = $existing || $existingName;
            $individualOptions['whenExists'] = 'keep';
            $individualOptions['requiredBy'] = array_key_exists($picklistName['name'], $usedPicklistNames) ? $usedPicklistNames[$picklistName['name']] : [];
            $individualOptions['required'] = count($individualOptions['requiredBy']) > 0;

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return mixed
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = [])
    {
        return parent::importFromFile($options, 'picklists', $tmpDirectory, $tokenType, $importedBy);
    }

    private function prepareExistingPicklistsMap($picklistRepository, $picklistNameOption)
    {
        $existingPicklists = [];
        $dbPicklists = $picklistRepository->findBy(["listName" => $picklistNameOption['existingUuid']]);
        foreach ($dbPicklists as $dbPicklist) {
            $existingPicklists[] = $this->normalizer->normalize($dbPicklist);
        }
        return $existingPicklists;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        $toAdd = [];
        $toUpdate = [];
        $picklistRepository = $this->loadRepository(Picklist::class);
        foreach ($options as $picklistNameOption) {
            if (!$picklistNameOption['include'] && !$picklistNameOption['required']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($data as $picklistName) {
                if ($picklistNameOption['uuid'] == $picklistName['uuid']) {
                    $existingPicklistsClone = [];
                    if ($picklistNameOption['exists'] && $picklistNameOption['whenExists'] == 'keep') {
                        $existingPicklists = $this->prepareExistingPicklistsMap($picklistRepository, $picklistNameOption);
                        $existingPicklistsClone = $existingPicklists;
                    }
                    $uuid = $picklistNameOption['existingUuid'] ? $picklistNameOption['existingUuid'] : $picklistName['uuid'];
                    $picklistName['@id'] = '/api/3/picklist_names/' . $uuid;
                    $newPicklists = [];
                    foreach ($picklistName['picklists'] as $picklist) {
                        if (isset($picklist['@id'])) {
                            unset($picklist['@id']);
                        }
                        foreach ($existingPicklistsClone as $index => $existingPicklistItem) {
                            if ($existingPicklistItem['uuid'] == $picklist['uuid'] || $existingPicklistItem['itemValue'] == $picklist['itemValue']) {
                                unset($existingPicklists[$index]);
                                break;
                            }
                        }

                        $existingPicklist = $picklistRepository->findOneBy(['uuid' => $picklist['uuid']]);
                        if (is_null($existingPicklist)) {
                            $existingPicklist = $picklistRepository->findOneBy(['itemValue' => $picklist['itemValue'], "listName" => $uuid]);
                        }
                        if ($existingPicklist) {
                            $picklist['@id'] = '/api/3/picklists/' . $existingPicklist->getUuid();
                            unset($picklist['uuid']);
                        }
                        $picklist['listName'] = $picklistName['@id'];
                        $newPicklists[] = $picklist;
                    }
                    if ($picklistNameOption['exists']) {
                        foreach ($existingPicklists as $picklist) {
                            $picklist['listName'] = $picklistName['@id'];
                            $picklist['@id'] = '/api/3/picklists/' . $picklist['uuid'];
                            $newPicklists[] = $picklist;
                        }
                    }
                    $picklistName['picklists'] = $newPicklists;
                    if (!$picklistNameOption['exists']) {
                        unset($picklistName['@id']);
                        $picklistName['importedBy'] = array_key_exists("importedBy", $picklistName) && $picklistName['importedBy'] ? array_merge($picklistName['importedBy'], $importedBy) : $importedBy;
                        $toAdd[] = $picklistName;
                        break;
                    }
                    unset($picklistName['uuid']);
                    $picklistNamerepository = $this->loadRepository(PicklistName::class);
                    $existingPicklistName = $picklistNamerepository->findOneBy(['uuid' => $uuid]);
                    $picklistName['importedBy'] = $existingPicklistName->getImportedBy() ? array_merge($existingPicklistName->getImportedBy(), $importedBy) : $importedBy;
                    $toUpdate[] = $picklistName;
                    break;
                }
            }
        }
        $results = [];
        foreach ($toAdd as $item) {
            $fullURI = sprintf('%s/api/3/picklist_names?$relationships=true', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $result = $this->sendRequest($fullURI, 'post', $item, $tokenType);
            $results[] = $result;
        }
        if (!empty($toUpdate)) {
            $fullURI = sprintf('%s/api/3/update/picklist_names?$relationships=true', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $body = [
                "data" => $toUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
        }

        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        return $this->getResourceFromIRI($options, PicklistName::class, true, false);
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $metaData = [];
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        foreach ($data as $item) {
            $filename = $tmpDirectory . 'picklists/' . $this->cleanFileName($item['name']) . '.json';
            $this->writeToFile($filename, $item);
            $filenames[] = $filename;
            $metaData[] = ['name' => $item['name']];
        }
        $this->appendToFile($tmpDirectory . 'info.json', $type, $metaData);
        return $filenames;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $jsonData = $this->loadDataFromDirectory('picklists', $tmpDirectory);
        $configTypes = $this->importOptionsGenerator->getConfigTypes();
        $moduleConfig = $this->container->get($configTypes['modules']);
        $moduleJsonData = $moduleConfig->loadDataFromDirectory('modules', $tmpDirectory);
        return $this->getImportOptions([$key => $jsonData, 'modules' => $moduleJsonData], $key, $tokenType);
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        foreach ($options as $value) {
            $value['existingUuid'] ?  array_push($exportOption, '/api/3/picklist_names/' . $value['existingUuid']) : array_push($exportOption, '/api/3/picklist_names/' . $value['uuid']);
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        $filter = ['uuid' => $options];
        $entityRepo = $this->entityManager->getRepository(PicklistName::class);
        $dataObj = $this->utilityService->getItemWithWhere($entityRepo, $filter);
        foreach ($dataObj as $picklist) {
            array_push($content, ['name' => $picklist->getName()]);
        }
        return $content;
    }
}
