<?php
/* Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use Ramsey\Uuid\Uuid;
use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Service\UtilityService;
use App\Security\Voter\FsrVoter;
use Doctrine\ORM\EntityManagerInterface;
use App\Entity\Workflow\WorkflowGroup;
use App\Providers\PermissionsDictionaryBuilder;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class PlaybookBlockConfig
 * @package App\Service\ConfigExportImport
 */
class PlaybookBlockConfig extends PortingConfig
{
    /** @var FsrVoter */
    private $fsrVoter;

    public function __construct(
        FsrVoter $fsrVoter,
        protected ImportOptionsGenerator $importOptionsGenerator,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        private UtilityService $utilityService,
        protected LoggerInterface $logger
    ) {
        $this->fsrVoter = $fsrVoter;
        $this->importOptionsGenerator = $importOptionsGenerator;
        $this->utilityService = $utilityService;

        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }

    /**
     * @param array $allData
     * @param string $key
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $repository = $this->loadRepository(WorkflowGroup::class);
        $data = $allData[$key];
        if (
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.workflows') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.workflows') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.workflows')
        ) {
            return [];
        }
        $options = [];
        if (array_key_exists('playbook_blocks', $data) && count($data['playbook_blocks']) > 0) {
            foreach ($data['playbook_blocks'] as $pbBlocks) {
                $existingByUuid = $repository->findOneBy(['uuid' => $pbBlocks['uuid']]);
                $existingByNameAndType = $repository->findOneBy(['name' => $pbBlocks['name'], 'type' => $pbBlocks['type']]);
                $individualOptions = [];
                $individualOptions['uuid'] = $pbBlocks['uuid'];
                $individualOptions['name'] = $pbBlocks['name'];
                $individualOptions['include'] = true;
                $individualOptions['mergeType'] = 'skip';
                // for solution packs, replace is the default behaviour
                if ($this->importJob && $this->importJob->getType() === 'SolutionPack Import') {
                    $individualOptions['mergeType'] = 'replace';
                }
                $individualOptions['exists'] = true;
                if (is_null($existingByUuid) && is_null($existingByNameAndType)) {
                    $individualOptions['mergeType'] = '';
                    $individualOptions['exists'] = false;
                }
                $individualOptions['exists_uuid'] = $existingByUuid ? true : false;
                $individualOptions['exists_name'] = $existingByNameAndType ? true : false;
                $individualOptions['existing_uuid'] = $existingByNameAndType ? $existingByNameAndType->getUuid() : $pbBlocks['uuid'];
                $options[] = $individualOptions;
            }
        }
        return $options;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->getImportOptions([$key => $data], $key, $tokenType);
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption['blocks'] = [];
        if (is_iterable($options)) {
            foreach ($options as $value) {
                if (is_array($value) && array_key_exists('uuid', $value) && !empty($value['uuid'])) {
                    array_push($exportOption['blocks'], $value['uuid']);
                }
            }
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        $workflowGroupRepository = $this->entityManager->getRepository(WorkflowGroup::class);
        $workflowBlocksUuids = [];
        if (array_key_exists('blocks', $options) && $options['blocks']) { //For sp create, we get array of uuids in 'block' key and for sp import we get array of arrays
            $workflowBlocksUuids = $options['blocks'];
        } else {
            foreach ($options as $option) {
                $workflowBlocksUuids[] = $option['uuid'];
            }
        }
        $filter = ['uuid' => $workflowBlocksUuids];
        $workflowBlocks = $this->utilityService->getItemWithWhere($workflowGroupRepository, $filter);
        if (!empty($workflowBlocks)) {
            foreach ($workflowBlocks as $workflowBlock) {
                $content[] = $workflowBlock->getName();
            }
        }
        return $content;
    }

    protected function loadDataFromDirectory(string $key, string $tmpDirectory)
    {
        $folderPath = $tmpDirectory . $key . '/';
        $items = [];
        $data = [];
        if (file_exists($folderPath)) {
            $items = scandir($folderPath);
        }
        $data['playbook_blocks'] = [];
        foreach ($items as $item) {
            $fileData = file_get_contents($folderPath . $item);
            $jsonData = json_decode($fileData, true);
            if (!is_null($jsonData)) {
                $data['playbook_blocks'][] = $jsonData;
            }
        }
        return $data;
    }

    private function sanitizePbBlocks(array &$playbookBlock)
    {
        $steps = [];
        foreach ($playbookBlock['workflowSteps'] as $step) {
            if (array_key_exists('group', $step)) {
                unset($step['group']);
            }
            $steps[] = $step;
        }
        $routes = [];
        foreach ($playbookBlock['workflowRoutes'] as $route) {
            if (array_key_exists('group', $route)) {
                unset($route['group']);
            }
            $routes[] = $route;
        }
        $playbookBlock['workflowSteps'] = $steps;
        $playbookBlock['workflowRoutes'] = $routes;
        return $playbookBlock;
    }

    /**
     * Imports playbook collections
     *
     * @param array $collections
     * @param array $collectionOptions
     * @return boolean
     */
    private function importPlaybookBlocks(array $pbBlocks, array $pbBlockOptions, $tokenType, $importedBy = [])
    {
        #$pbBlocksValues = $pbBlockOptions['values'];
        $deleteBlocks = [];
        $blockToInsert = [];
        $includeBlocks = [];
        foreach ($pbBlockOptions as $value) {
            if (!$value['include']) {
                continue;
            }
            if ($value['exists']) {
                if ($value['mergeType'] == 'skip') {
                    continue;
                }
                # If Blocks exists either by Uuid or Name&Type and value['mergeType'] == 'replace', then replace the block
                $deleteBlocks[] = $value['uuid'];
                $includeBlocks[] = $value['uuid'];
            } else {
                $includeBlocks[] = $value['uuid'];
            }
        }

        if (!empty($deleteBlocks)) {
            $deleteFullURI = sprintf('%s/api/3/delete/workflow_groups', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $this->sendRequest($deleteFullURI, 'delete', ['ids' => $deleteBlocks], $tokenType);
        }

        foreach ($pbBlocks as $pbBlock) {
            if (in_array($pbBlock['uuid'], $includeBlocks)) {
                $pbBlock = $this->sanitizePbBlocks($pbBlock);
                $pbBlock['reusable'] = true;
                $blockToInsert[] = $pbBlock;
            }
        }

        if (!empty($blockToInsert)) {
            $insertFullURI = sprintf('%s/api/3/insert/workflow_groups', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $response = $this->sendRequest($insertFullURI, 'post', ['data' => $blockToInsert], $tokenType);
        }
        return true;
    }


    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @param string $tokenType
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        if ($options) {
            return $this->importPlaybookBlocks($data['playbook_blocks'], $options, $tokenType, $importedBy);
        }
    }



    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = [])
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->import($data, $options, $tokenType, $importedBy);
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        if (array_key_exists('blocks', $options)) {
            $data['newPlaybookBlocks'] = [];
            if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.workflows')) {
                return [];
            }
            foreach ($options['blocks'] as $pb_block) {
                $iris = [];
                $iris[] = '/api/3/workflow_groups/' . $pb_block;
                $data['newPlaybookBlocks'] = array_merge($data['newPlaybookBlocks'], $this->getResourceFromIRI($iris, WorkflowGroup::class, true, true, true));
            }
        }
        if (array_key_exists('includeGlobalVariables', $options) && $options['includeGlobalVariables']) {
            $data['globalVariables'] = [];
            $usedMacros = $this->getMacrosFromPlaybookBlocks($data['newPlaybookBlocks']);
            $existingDynamicVariables = $this->getGlobalVariables($tokenType);
            foreach ($usedMacros as $dynamicVariableName) {
                foreach ($existingDynamicVariables as $existingDynamicVariable) {
                    if ($existingDynamicVariable['name'] == $dynamicVariableName) {
                        $data['globalVariables'][] = $existingDynamicVariable;
                        break;
                    }
                }
            }
        }
        return $data ?? null;
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $collectionMetaData = [];
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        $directory = $tmpDirectory . $type . '/';
        $filenames = [];
        foreach ($data['newPlaybookBlocks'] as $pb_block) {
            if ($pb_block['reusable']) {
                $pbBlockName = $this->cleanFileName($pb_block['name'], true);
                $pbBlockFilename = $directory . '/' . $pbBlockName . '.json';
                $this->writeToFile($pbBlockFilename, $pb_block);
                $collectionMetaData[] = ['name' => $pbBlockName];
                $filenames[] = $pbBlockFilename;
            }
        }
        if (array_key_exists('globalVariables', $data) && !empty($data['globalVariables'])) {
            $globalVariableDirectory = $tmpDirectory . 'playbooks' . '/';
            $globalVariableFilename = $globalVariableDirectory . 'globalVariables.json';
            if (file_exists($globalVariableFilename)) {
                $globalVariableFileData = file_get_contents($globalVariableFilename);
                $globalVariableFileJsonData = json_decode($globalVariableFileData, true);
            }
            $globalVariableFileJsonData = isset($globalVariableFileJsonData) ? $globalVariableFileJsonData : [];
            $globalVariables = array_merge($data['globalVariables'], $globalVariableFileJsonData);
            $globalVariableNameDataMap = [];
            foreach ($globalVariables as $globalVariable) {
                $globalVariableNameDataMap[$globalVariable['name']] = $globalVariable;
            }
            $this->writeToFile($globalVariableFilename, $globalVariableNameDataMap);
            $this->appendToFile($tmpDirectory . 'info.json', 'globalVariables', $globalVariableNameDataMap);
        }
        if (!empty($collectionMetaData)) {
            $this->appendToFile($tmpDirectory . 'info.json', $type, $collectionMetaData);
        }
        return $filenames;
    }

    private function getMacrosFromPlaybookBlocks($playbookBlocks)
    {
        $macros = [];
        foreach ($playbookBlocks as $playbookBlock) {
            if (array_key_exists('workflowSteps', $playbookBlock)) {
                foreach ($playbookBlock['workflowSteps'] as $workflowStep) {
                    if (array_key_exists('arguments', $workflowStep)) {
                        $argsJson = json_encode($workflowStep['arguments']);
                        $regex = '/(?>globalVars|macro)\.(\w+)/m';
                        $matches = [];

                        preg_match_all($regex, $argsJson, $matches, PREG_PATTERN_ORDER, 0);
                        $macros = array_merge($macros, $matches[1]);
                    }
                }
            }
        }
        return $macros;
    }
}
