<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use stdClass;
use Ramsey\Uuid\Uuid;
use App\Entity\Core\Tag;
use App\Entity\Core\Image;
use App\Entity\Core\PgFile;
use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Service\UtilityService;
use App\Security\Voter\FsrVoter;
use App\Entity\Workflow\Workflow;
use Doctrine\ORM\EntityManagerInterface;
use App\Entity\Workflow\WorkflowCollection;
use App\Providers\PermissionsDictionaryBuilder;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class PlaybookConfig
 * @package App\Service\ConfigExportImport
 */
class PlaybookConfig extends PortingConfig
{
    private const STEP_TYPES = [
        'MANUAL_INPUT' => 'fc04082a-d7dc-4299-96fb-6837b1baa0fe',
        'DECISION' => '12254cf5-5db7-4b1a-8cb1-3af081924b28'
    ];
    /** @var FsrVoter */
    private $fsrVoter;

    private $requiredMacros = [];

    private $utilityService;

    private const WORKFLOW_STEP_IRI = '/api/3/workflow_steps/';
    private const WORKFLOW_GROUP_IRI = '/api/3/workflow_groups/';
    private const SCHEDULE_KEY = 'includeSchedules';
    private const GLOBAL_VARIABLE_KEY = 'includeGlobalVariables';
    private const VERSIONS_KEY = 'includeVersions';

    public function __construct(
        FsrVoter $fsrVoter,
        protected ImportOptionsGenerator $importOptionsGenerator,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        UtilityService $utilityService,
        protected LoggerInterface $logger
    ) {
        $this->fsrVoter = $fsrVoter;
        $this->importOptionsGenerator = $importOptionsGenerator;
        $this->utilityService = $utilityService;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }

    /**
     * @param array $allData
     * @param string $key
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        if (
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.workflows') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.workflows') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.workflows')
        ) {
            return [];
        }
        $options = [];
        if (array_key_exists('collections', $data) && count($data['collections']) > 0) {
            $subOptions = $this->generatePlaybookCollectionOptions($data['collections']);
            $options['collections'] = [
                "include" => count($subOptions) > 0,
                "values" => $subOptions
            ];
        }
        if (array_key_exists('globalVariables', $data) && count($data['globalVariables']) > 0) {
            $subOptions = $this->generateGlobalVariableOptions($data['globalVariables'], $tokenType);
            $options['globalVariables'] = [
                "include" => count($subOptions) > 0,
                "values" => $subOptions
            ];
        }
        return $options;
    }

    /**
     * @param array $config
     * @return array
     */
    private function generateGlobalVariableOptions(array $dynamicVariables, $tokenType)
    {
        $existingDynamicVariables = $this->getGlobalVariables($tokenType);
        $existingDynamicVariableNames = array_column($existingDynamicVariables, 'name');
        $options = [];
        foreach ($dynamicVariables as $dynamicVariable) {
            $existing = in_array($dynamicVariable['name'], $existingDynamicVariableNames);
            $individualOptions = [];
            $individualOptions['name'] = $dynamicVariable['name'];
            $individualOptions['include'] = !$existing;
            $individualOptions['required'] = !$existing && in_array($dynamicVariable['name'], $this->requiredMacros);
            $individualOptions['exists'] = $existing;

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * @param array $config
     * @return array
     */
    private function generatePlaybookCollectionOptions(array $playbookCollections)
    {
        $repository = $this->loadRepository(WorkflowCollection::class);
        $options = [];
        foreach ($playbookCollections as $playbookCollection) {
            $existingByUuid = $repository->findOneBy(['uuid' => $playbookCollection['uuid']]);
            $existingByName = $repository->findOneBy(['name' => $playbookCollection['name']]);
            $individualOptions = [];
            $individualOptions['uuid'] = $playbookCollection['uuid'];
            $individualOptions['name'] = $playbookCollection['name'];
            $softDeleted = false;
            if (!is_null($existingByUuid)) {
                $softDeleted = !is_null($existingByUuid->getDeletedAt());
            } elseif (!is_null($existingByName)) {
                $softDeleted = !is_null($existingByName->getDeletedAt());
            }
            $individualOptions['softDeleted'] = $softDeleted;
            $individualOptions['include'] = true;
            $individualOptions['mergeType'] = 'merge_append';
            // for solution packs, replace is the default behaviour
            if ($this->importJob && $this->importJob->getType() === 'SolutionPack Import') {
                $individualOptions['mergeType'] = 'merge_replace';
            }
            $individualOptions['exists_uuid'] = $existingByUuid ? true : false;
            $individualOptions['existing_uuid'] = $existingByName ? $existingByName->getUuid() : $playbookCollection['uuid'];
            $individualOptions['exists_name'] = $existingByName ? true : false;
            $individualOptions['exists'] = $individualOptions['exists_uuid'] || $individualOptions['exists_name'];
            $results = $this->getMacrosAndScheduleCountFromPlaybooks($playbookCollection['workflows']);
            $individualOptions['macros'] = $results['macros'];
            $individualOptions['schedulesCount'] = $results['schedulesCount'];
            $individualOptions['includeSchedules'] = $individualOptions['schedulesCount'] > 0 ? true : false;
            $playbookCount = 0;
            if ($existingByName) {
                $playbookCount = count($existingByName->getWorkflows());
            } elseif ($existingByUuid) {
                $playbookCount = count($existingByUuid->getWorkflows());
            }
            $individualOptions['oldPlaybookCount'] = $playbookCount;
            $individualOptions['newPlaybookCount'] = count($playbookCollection['workflows']);

            $this->requiredMacros = array_unique(array_merge($this->requiredMacros, $individualOptions['macros']));

            $options[] = $individualOptions;
        }
        return $options;
    }

    private function getMacrosAndScheduleCountFromPlaybooks($playbooks)
    {
        $macros = [];
        $schedulesCount = 0;
        foreach ($playbooks as $playbook) {
            if (array_key_exists('steps', $playbook)) {
                foreach ($playbook['steps'] as $step) {
                    if (array_key_exists('arguments', $step)) {
                        $argsJson = json_encode($step['arguments']);
                        $regex = '/(?>globalVars|macro)\.(\w+)/m';
                        $matches = [];

                        preg_match_all($regex, $argsJson, $matches, PREG_PATTERN_ORDER, 0);
                        $macros = array_merge($macros, $matches[1]);
                    }
                }
            }
            if ($playbook['versions'] && count($playbook['versions'])) {
                foreach ($playbook['versions'] as $version) {
                    $version = json_decode($version['json'], true);
                    if (array_key_exists('steps', $version)) {
                        foreach ($version['steps'] as $step) {
                            if (array_key_exists('arguments', $step)) {
                                $argsJson = json_encode($step['arguments']);
                                $regex = '/(?>globalVars|macro)\.(\w+)/m';
                                $matches = [];
                                preg_match_all($regex, $argsJson, $matches, PREG_PATTERN_ORDER, 0);
                                $macros = array_merge($macros, $matches[1]);
                            }
                        }
                    }
                }
            }
            if (array_key_exists('schedules', $playbook)) {
                $schedulesCount += count($playbook['schedules']);
            }
        }
        return array('macros' => array_unique($macros), 'schedulesCount' => $schedulesCount);
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @param string $tokenType
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        $tagsResult = true;
        if (array_key_exists('exported_tags', $data)) {
            $tagsResult = $this->importTags($data['exported_tags'], $tokenType);
        }
        $collectionResult = true;
        if (array_key_exists('collections', $options) && $options['collections']['include']) {
            $collectionResult = $this->importPlaybookCollections($data['collections'], $options['collections']['values'], $tokenType, $importedBy);
        }
        $globalVariableResult = true;
        if (array_key_exists('globalVariables', $options) && $options['globalVariables']['include']) {
            $globalVariableResult = $this->importGlobalVariables($data['globalVariables'], $options['globalVariables']['values'], $tokenType);
        } elseif (array_key_exists('globalVariables', $options) && array_key_exists('collections', $options) && $options['collections']['include']) {
            // Import required global variables from playbooks
            $requiredMacros = [];
            foreach ($options['collections']['values'] as $collection) {
                $requiredMacros = array_merge($requiredMacros, $collection['macros']);
            }
            $uniqueRequiredMacros = array_unique($requiredMacros);
            $requiredMacrosToImport = [];
            foreach ($options['globalVariables']['values'] as $globalVariable) {
                if (in_array($globalVariable['name'], $uniqueRequiredMacros)) {
                    $requiredMacrosToImport[] = $globalVariable;
                }
            }
            $globalVariableResult = $this->importGlobalVariables($data['globalVariables'], $requiredMacrosToImport, $tokenType);
        }
        return $collectionResult && $globalVariableResult && $tagsResult;
    }

    private function importTags($tags, $tokenType)
    {
        $repository = $this->loadRepository(Tag::class);
        $tagObjects = [];
        foreach ($tags as $tag) {
            $existing = $repository->findOneBy(['uuid' => $tag]);
            if ($existing) {
                continue;
            }
            $tagObjects[] = [
                'uuid' => $tag
            ];
        }
        if (!empty($tagObjects)) {
            $fullURI = sprintf('%s/api/3/bulkupsert/tags', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $body = [
                "__unique" => ['uuid'],
                "__replace" => false,
                "__data" => $tagObjects
            ];
            return $this->sendRequest($fullURI, 'post', $body, $tokenType);
        }
        return true;
    }

    /**
     * Imports playbook collections
     *
     * @param array $collections
     * @param array $collectionOptions
     * @return boolean
     */
    private function importPlaybookCollections(array $collections, array $collectionOptions, $tokenType, $importedBy = [])
    {
        $toAdd = [];
        $toAppend = [];
        $deleteUuids = [];
        $playbooksToDelete = [];
        $schedulesToImport = [];
        $uuidMap = [];
        $playbookRepository = $this->loadRepository(Workflow::class);
        $collectionRepository = $this->loadRepository(WorkflowCollection::class);
        foreach ($collectionOptions as $playbookCollectionOption) {

            if (!$playbookCollectionOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($collections as $collection) {
                if ($playbookCollectionOption['uuid'] == $collection['uuid']) {
                    $newPlaybooks = [];
                    foreach ($collection['workflows'] as $playbook) {
                        $remove = false;
                        /** @var Workflow $existingPlaybook */
                        $existingPlaybookByUuid = $playbookRepository->findOneBy(['uuid' => $playbook['uuid']]);
                        if ($existingPlaybookByUuid && ($existingPlaybookByUuid->getCollection()->getUuid() != $collection['uuid'] || $playbookCollectionOption['mergeType'] === 'rename')) {
                            // If a uuid already exists in a different collection, or if we are renaming the current collection, generate a new one
                            $newUuid = Uuid::uuid4()->toString();
                            $uuidMap[$playbook['uuid']] = $newUuid;
                            $playbook['uuid'] = $newUuid;
                            // this is the only case where the stepids would also cause a conflict
                            $playbook = $this->fixStepUuids($playbook);
                        } elseif ($existingPlaybookByUuid && $existingPlaybookByUuid->getCollection()->getUuid() == $collection['uuid']) {
                            if ($playbookCollectionOption['mergeType'] === 'merge_replace') {
                                // If we are replacing existing playbooks, we need to remove any found playbooks by UUID in the same collection
                                $playbooksToDelete[] = $existingPlaybookByUuid->getUuid();
                            } elseif ($playbookCollectionOption['mergeType'] === 'merge_append') {
                                // If we are appending, we skip any existing playbooks
                                $remove = true;
                            }
                        }
                        if ($playbookCollectionOption['exists']) {
                            $existingPlaybookByName = $playbookRepository->findOneBy(['name' => $playbook['name'], 'collection' => $playbookCollectionOption['existing_uuid']]);
                            if ($existingPlaybookByName) {
                                if ($playbookCollectionOption['mergeType'] === 'merge_replace') {
                                    // If we are replacing existing playbooks, we need to remove any found playbooks by UUID in the same collection
                                    $playbooksToDelete[] = $existingPlaybookByName->getUuid();
                                } elseif ($playbookCollectionOption['mergeType'] === 'merge_append') {
                                    // If we are appending, we just need to add the playbook
                                    $remove = true;
                                }
                            }
                        }
                        unset($playbook['collection']);
                        $playbook['lastModifyDate'] = time();

                        if (array_key_exists('schedules', $playbook) && $playbookCollectionOption['includeSchedules']) {
                            foreach ($playbook['schedules'] as $schedule) {
                                $schedule['kwargs']['wf_iri'] = sprintf('/api/3/workflows/%s', $playbook['uuid']);
                                $schedulesToImport[] = $schedule;
                            }
                        }

                        if (!$remove) {
                            $playbook['importedBy'] = array_key_exists("importedBy", $playbook) && $playbook['importedBy'] ? array_merge($playbook['importedBy'], $importedBy) : $importedBy;
                            $newPlaybooks[] = $playbook;
                        }
                    }
                    $collection['workflows'] = $newPlaybooks;
                    $collection = $this->importCollectionImage($collection);
                    if (!$playbookCollectionOption['exists_uuid'] && !$playbookCollectionOption['exists_name']) {
                        $collection['importedBy'] = array_key_exists("importedBy", $collection) && $collection['importedBy'] ? array_merge($collection['importedBy'], $importedBy) : $importedBy;
                        $toAdd[] = $collection;
                        break;
                    }
                    if ($playbookCollectionOption['mergeType'] === 'rename') {
                        $number = 1;
                        while ($collectionRepository->findOneBy(['name' => $collection['name'] . ' (' . $number . ')'])) {
                            $number++;
                        }
                        $collection['name'] = $collection['name'] . ' (' . $number . ')';
                        if ($playbookCollectionOption['exists_uuid']) {
                            unset($collection['uuid']);
                        }
                        $collection['importedBy'] =  array_key_exists('importedBy', $collection) ? array_merge($collection['importedBy'], $importedBy) : $importedBy;
                        $toAdd[] = $collection;
                        break;
                    } elseif ($playbookCollectionOption['mergeType'] === 'replace') {
                        $deleteUuids[] = $playbookCollectionOption['existing_uuid'];
                        $toAdd[] = $collection;
                    } elseif (in_array($playbookCollectionOption['mergeType'], ['merge_replace', 'merge_append'])) {
                        // UI Name (Merge Replace Existing Playbook)
                        $collection['uuid'] = $playbookCollectionOption['existing_uuid'];
                        $collection['importedBy'] = array_key_exists('importedBy', $collection) ? array_merge($collection['importedBy'], $importedBy) : $importedBy;
                        $toAppend[] = $collection;
                    }
                    break;
                }
            }
            // Fix reference playbooks
            foreach ($toAdd as $collectionIndex => $collection) {
                foreach ($collection['workflows'] as $playbookIndex => $playbook) {
                    $toAdd[$collectionIndex]['workflows'][$playbookIndex] = $this->fixReferencePlaybookUuids($playbook, $uuidMap);
                }
            }
            foreach ($toAppend as $collectionIndex => $collection) {
                foreach ($collection['workflows'] as $playbookIndex => $playbook) {
                    $toAppend[$collectionIndex]['workflows'][$playbookIndex] = $this->fixReferencePlaybookUuids($playbook, $uuidMap);
                }
            }
        }
        $results = [];
        if (!empty($deleteUuids)) {
            $fullURI = sprintf('%s/api/3/delete-with-query/workflow_collections?$showDeleted=true', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $filterObj = new stdClass();
            $filterObj->{'field'} = "uuid";
            $filterObj->{'operator'} = "in";
            $filterObj->{'value'} = array_unique($deleteUuids);
            $filterObj->{'type'} = "primitive";
            $body = [$filterObj];

            $data = new stdClass();
            $data->{'logic'} = 'AND';
            $data->{'filters'} = $body;
            $this->sendRequest($fullURI, 'delete', $data, $tokenType);
        }
        if (!empty($playbooksToDelete)) {
            $filterObj = new stdClass();
            $filterObj->{'field'} = "uuid";
            $filterObj->{'operator'} = "in";
            $filterObj->{'value'} = array_unique($playbooksToDelete);
            $filterObj->{'type'} = "primitive";
            $body = [$filterObj];

            $data = new stdClass();
            $data->{'logic'} = 'AND';
            $data->{'filters'} = $body;

            $fullURI = sprintf('%s/api/3/delete-with-query/workflows?$showDeleted=true', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $this->sendRequest($fullURI, 'delete', $data, $tokenType);
        }
        foreach ($toAdd as $item) {
            $fullURI = sprintf('%s/api/3/workflow_collections', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $result = $this->sendRequest($fullURI, 'post', $item, $tokenType);
            $results[] = $result;
        }
        foreach ($toAppend as $collection) {
            foreach ($collection['workflows'] as $playbook) {
                $playbook['collection'] = sprintf('/api/3/workflow_collections/%s', $collection['uuid']);
                $fullURI = sprintf('%s/api/3/workflows', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
                $result = $this->sendRequest($fullURI, 'post', $playbook, $tokenType);
                $results[] = $result;
            }
            unset($collection['workflows']);
            $fullURI = sprintf('%s/api/3/workflow_collections/%s', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $collection['uuid']);
            $result = $this->sendRequest($fullURI, 'put', $collection, $tokenType);
            $results[] = $result;
        }
        foreach ($schedulesToImport as $schedule) {
            $this->saveSchedule($schedule, $tokenType);
        }
        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    private function importCollectionImage($collection)
    {
        $imageRepository = $this->entityManager->getRepository(Image::class);
        if ($collection['image'] && $collection['image_file']) {
            $image = $imageRepository->findOneBy(['uuid' => $collection['image']['uuid']]);
            if (!$image) {
                $image = new Image();
                $image->setUuid($collection['image']['uuid']);
                $image->setId($collection['image']['uuid']);
            }

            $image->setFilename($collection['image']['filename']);
            $image->setMimeType($collection['image']['mimeType']);
            $image->setSize($collection['image']['size']);

            $entityManager = $this->container->get('doctrine')->getManagerForClass(PgFile::class);
            $pgFile = null;
            if ($image->getFile()) {
                $pgFile = $entityManager->find(PgFile::class, $image->getFile());
            }
            if (!$pgFile) {
                $pgFile = new PgFile();
            }
            $pgFile->setFile($collection['image_file']);
            $entityManager->persist($pgFile);
            $image->setFile($pgFile->getUuid());
            $this->entityManager->persist($image);
            $this->entityManager->flush();
        }
        if (is_array($collection['image'])) {
            $collection['image'] = '/api/3/images/' . $collection['image']['uuid'];
        }
        return $collection;
    }

    private function fixReferencePlaybookUuids(array $playbook, array $uuidMap)
    {
        foreach ($playbook['steps'] as $index => $step) {
            if (array_key_exists('arguments', $step) && array_key_exists('workflowReference', $step['arguments'])) {
                $referenceUuid = str_replace('/api/3/workflows/', '', $step['arguments']['workflowReference']);
                if (array_key_exists($referenceUuid, $uuidMap)) {
                    $playbook['steps'][$index]['arguments']['workflowReference'] = sprintf('/api/3/workflows/%s', $uuidMap[$referenceUuid]);
                }
            }
        }
        return $playbook;
    }

    /**
     * Replaces the uuids for all playbook steps and points the routes to the new steps
     *
     * @param array $playbook
     * @return array
     */
    private function fixStepUuids(array $playbook)
    {
        $uuidMapping = [];
        $uuidGroupMapping = [];
        // First generate the new uuids and keep track using $uuidMapping
        $steps = [];
        foreach ($playbook['steps'] as $step) {
            $newUuid = Uuid::uuid4()->toString();
            $uuidMapping[$step['uuid']] = $newUuid;
            $step['uuid'] = $newUuid;
            $steps[] = $step;
        }
        $playbook['steps'] = $steps;

        $groups = [];
        foreach ($playbook['groups'] as $group) {
            $newUuid = Uuid::uuid4()->toString();
            $uuidGroupMapping[$group['uuid']] = $newUuid;
            $group['uuid'] = $newUuid;
            $groups[] = $group;
        }
        $playbook['groups'] = $groups;

        // Next, find all step references and replace them
        if ($playbook['triggerStep']) {
            $playbook['triggerStep'] = self::WORKFLOW_STEP_IRI . $uuidMapping[self::getUuidFromIri($playbook['triggerStep'])];
        }
        $routes = [];
        foreach ($playbook['routes'] as $route) {
            $route['uuid'] = Uuid::uuid4()->toString();
            $route['targetStep'] = self::WORKFLOW_STEP_IRI . $uuidMapping[self::getUuidFromIri($route['targetStep'])];
            $route['sourceStep'] = self::WORKFLOW_STEP_IRI . $uuidMapping[self::getUuidFromIri($route['sourceStep'])];
            $routes[] = $route;
        }
        $playbook['routes'] = $routes;
        $steps = [];
        foreach ($playbook['steps'] as $step) {
            if (array_key_exists('group', $step) && $step['group']) {
                $step['group'] = self::WORKFLOW_GROUP_IRI . $uuidGroupMapping[self::getUuidFromIri($step['group'])];
            }
            $stepTypeUuid = self::getUuidFromIri($step['stepType']);
            if ($stepTypeUuid == self::STEP_TYPES['MANUAL_INPUT']) {
                if (array_key_exists('timeout', $step['arguments']) && array_key_exists('step_iri', $step['arguments']['timeout'])) {
                    $step['arguments']['timeout']['step_iri'] = self::WORKFLOW_STEP_IRI . $uuidMapping[self::getUuidFromIri($step['arguments']['timeout']['step_iri'])];
                }
                $options = [];
                foreach ($step['arguments']['response_mapping']['options'] as $responseMappingOption) {
                    $responseMappingOption['step_iri'] = self::WORKFLOW_STEP_IRI . $uuidMapping[self::getUuidFromIri($responseMappingOption['step_iri'])];
                    $options[] = $responseMappingOption;
                }
                $step['arguments']['response_mapping']['options'] = $options;
            } elseif ($stepTypeUuid == self::STEP_TYPES['DECISION']) {
                $conditions = [];
                foreach ($step['arguments']['conditions'] as $condition) {
                    $condition['step_iri'] = self::WORKFLOW_STEP_IRI . $uuidMapping[self::getUuidFromIri($condition['step_iri'])];
                    $conditions[] = $condition;
                }
                $step['arguments']['conditions'] = $conditions;
            }
            $steps[] = $step;
        }
        $playbook['steps'] = $steps;
        return $playbook;
    }

    private function importGlobalVariables(array $globalVariables, array $globalVariableOptions, $tokenType)
    {
        $toAdd = [];
        foreach ($globalVariableOptions as $globalVariableOption) {
            if (!$globalVariableOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($globalVariables as $globalVariable) {
                if ($globalVariableOption['name'] == $globalVariable['name']) {
                    $toAdd[] = $globalVariable;
                    break;
                }
            }
        }
        $results = [];
        foreach ($toAdd as $item) {
            $token = $this->getTokenFromStorage();
            if ($token) {
                $workflowMacroEndpoint = '/api/wf/api/dynamic-variable/';
                $uri = $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY) . $workflowMacroEndpoint;
                $results[] = $this->sendRequest($uri, 'post', $item, $tokenType);
            } else {
                $workflowMacroEndpoint = '/api/dynamic-variable/';
                $uri = $this->container->getParameter('workflow.uri') . $workflowMacroEndpoint;
                $results[] = $this->sealabRequest($uri, 'post', $item, $tokenType);
            }
        }

        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.workflows')) {
            return [];
        }
        $usedMacros = [];
        if (array_key_exists('collections', $options)) {
            $uuids = $this->getUUIDsFromCollections($options['collections']);
            $data['collections'] = $this->getResourceFromIRI($uuids, WorkflowCollection::class, true);

            // Updating collection in case of include version selected for any collection
            $newCollections = [];
            $recordTags = [];
            foreach ($data['collections'] as $collection) {
                if (
                    $this->getValueOfKeyByCollectionUUID($options['collections'], $collection['uuid'], self::VERSIONS_KEY) ||
                    (array_key_exists('includeVersions', $options) && $options['includeVersions'])
                ) {
                    $iris = [];
                    foreach ($collection['workflows'] as $playbook) {
                        $iris[] = '/api/3/workflows/' . $playbook['uuid'];
                        foreach ($playbook['recordTags'] as $recordTag) {
                            $parts = explode('/', $recordTag);
                            $recordTags[] = end($parts);
                        }
                    }
                    $collection['workflows'] = $this->getResourceFromIRI($iris, Workflow::class, true, true, true);
                    $newCollections[] = $collection;
                    foreach ($collection['recordTags'] as $recordTag) {
                        $parts = explode('/', $recordTag);
                        $recordTags[] = end($parts);
                    }
                } else {
                    $newCollections[] = $collection;
                }
            }
            if (!empty($newCollections)) {
                $data['collections'] = $newCollections;
            }

            $recordTags = [];
            $newCollections = [];
            foreach ($data['collections'] as $collection) {
                foreach ($collection['recordTags'] as $recordTag) {
                    $parts = explode('/', $recordTag);
                    $recordTags[] = end($parts);
                }
                foreach ($collection['workflows'] as $index => $playbook) {
                    $collection['workflows'][$index]['isPrivate'] = false;
                    $collection['workflows'][$index]['owners'] = [];
                    foreach ($playbook['recordTags'] as $recordTag) {
                        $parts = explode('/', $recordTag);
                        $recordTags[] = end($parts);
                    }
                    foreach ($playbook['groups'] as $groupIndex => $group) {
                        if (array_key_exists('workflowSteps', $group)) {
                            unset($collection['workflows'][$index]['groups'][$groupIndex]['workflowSteps']);
                        }
                        if (array_key_exists('workflowRoutes', $group)) {
                            unset($collection['workflows'][$index]['groups'][$groupIndex]['workflowRoutes']);
                        }
                    }
                    foreach ($playbook['steps'] as $workflowStep) {
                        if (array_key_exists('stepType', $workflowStep) && $workflowStep['stepType'] === AppConstants::WORKFLOW_STEP_TYPES_IRI_MAP["WorkflowReference"] && array_key_exists('workflowReference', $workflowStep['arguments'])) {
                            $referencedWorkflowIris[] = $workflowStep['arguments']['workflowReference'];
                        }
                    }
                    if ($this->getValueOfKeyByCollectionUUID($options['collections'], $collection['uuid'], self::SCHEDULE_KEY)) {
                        $schedules = $this->loadSchedules($playbook['uuid'], $tokenType, 'playbook');
                        if (count($schedules) > 0) {
                            $collection['workflows'][$index]['schedules'] = $schedules;
                        }
                    }
                }
                if ((array_key_exists('includeGlobalVariables', $options) && $options['includeGlobalVariables']) ||
                    $this->getValueOfKeyByCollectionUUID($options['collections'], $collection['uuid'], self::GLOBAL_VARIABLE_KEY)
                ) {
                    $usedMacros = array_merge($usedMacros, $this->getMacrosAndScheduleCountFromPlaybooks($collection['workflows'])['macros']);
                }
                $newCollections[$collection['uuid']] = $collection;
            }
            $referencedWorkflows = $this->getResourceFromIRI($referencedWorkflowIris, Workflow::class, true, true, true);
            $referencedWorkflowCollectionIris = [];
            foreach ($referencedWorkflows as $referencedWorkflow) {
                $collectionUuid = end(explode('/', $referencedWorkflow['collection']));
                if (!array_key_exists($collectionUuid, $newCollections)) {
                    $referencedWorkflowCollectionIris[] = $referencedWorkflow['collection'];
                }
            }
            $referencedWorkflowCollections = $this->getResourceFromIRI($referencedWorkflowCollectionIris, WorkflowCollection::class, true);
            foreach ($referencedWorkflowCollections as $collectionIndex => $referencedWorkflowCollection) {
                foreach ($referencedWorkflowCollection['workflows'] as $workflowIndex => $workflow) {
                    $workflowIri = sprintf('/api/3/workflows/%s', $workflow['uuid']);
                    if (!in_array($workflowIri, $referencedWorkflowIris)) {
                        unset($referencedWorkflowCollections[$collectionIndex]['workflows'][$workflowIndex]);
                    }
                }
            }
            $data['collections'] = $referencedWorkflowCollections ? array_merge(array_values($newCollections), $referencedWorkflowCollections) : array_values($newCollections);
            sort($recordTags);
            $data['exported_tags'] = array_values(array_unique($recordTags));
        }
        if (!empty($usedMacros) || !empty($options['globalVariables'])) {
            $data['globalVariables'] = [];
            $existingDynamicVariables = $this->getGlobalVariables($tokenType);
            $macrosToImport = !empty($options['globalVariables']) && !empty($usedMacros) ? array_unique(array_merge($options['globalVariables'] ?? [], $usedMacros)) : (empty($options['globalVariables']) && !empty($usedMacros) ? array_unique($usedMacros) : array_unique($options['globalVariables']));
            foreach ($macrosToImport as $dynamicVariableName) {
                foreach ($existingDynamicVariables as $existingDynamicVariable) {
                    if ($existingDynamicVariable['name'] == $dynamicVariableName) {
                        unset($existingDynamicVariable['id']);
                        $data['globalVariables'][] = $existingDynamicVariable;
                        break;
                    }
                }
            }
        }

        return $data ?? null;
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $collectionMetaData = [];
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        $directory = $tmpDirectory . $type . '/';
        $filenames = [];
        foreach ($data['collections'] as $collection) {
            $collectionName = $this->cleanFileName($collection['name']);
            $collectionFolder = $directory . $collectionName . '/';
            foreach ($collection['workflows'] as $playbook) {
                $playbookName = $this->cleanFileName($playbook['name']);
                $playbookFilename = $collectionFolder . $playbookName . '.json';
                $this->writeToFile($playbookFilename, $playbook);
                $filenames[] = $playbookFilename;
            }
            unset($collection['workflows']);
            $collectionFilename = $collectionFolder . 'collection.metadata.json';
            if (array_key_exists('image', $collection) && $collection['image']) {
                $image = end($this->getResourceFromIRI([$collection['image']], Image::class));
                $imageRepository = $this->entityManager->getRepository(Image::class);
                $iriParts = explode('/', $collection['image']);
                $imageObject = $imageRepository->findOneBy(['uuid' => end($iriParts)]);
                if ($imageObject) {
                    $fileRepository = $this->entityManager->getRepository(PgFile::class);
                    $pgFile = $fileRepository->findOneBy(['uuid' => $imageObject->getFile()]);
                    if ($pgFile) {
                        $collection['image_file'] = stream_get_contents($pgFile->getFile());
                        unset($image['thumbnail']);
                        $collection['image'] = $image;
                    }
                }
            }
            $this->writeToFile($collectionFilename, $collection);
            $collectionMetaData[] = ['name' => $collectionName];
            $filenames[] = $collectionFilename;
        }
        $data['globalVariables'] = array_key_exists('globalVariables', $data) && !empty($data['globalVariables']) ? $data['globalVariables'] : [];
        if (!empty($data['globalVariables'])) {
            $globalVariableFilename = $directory . 'globalVariables.json';
            $this->writeToFile($globalVariableFilename, $data['globalVariables']);
            $filenames[] = $globalVariableFilename;
            $this->appendToFile($tmpDirectory . 'info.json', 'globalVariables', $data['globalVariables']);
        }
        if (!empty($data['exported_tags'])) {
            $exportedTagsFilename = $directory . 'tags.json';
            sort($data['exported_tags']);
            $this->writeToFile($exportedTagsFilename, $data['exported_tags']);
            $filenames[] = $exportedTagsFilename;
        }
        if (!empty($collectionMetaData)) {
            $this->appendToFile($tmpDirectory . 'info.json', $type, $collectionMetaData);
        }

        return $filenames;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->getImportOptions([$key => $data], $key, $tokenType);
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = [])
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->import($data, $options, $tokenType, $importedBy);
    }

    protected function loadDataFromDirectory(string $key, string $tmpDirectory)
    {
        $folderPath = $tmpDirectory . $key . '/';
        $items = [];
        $data = [];
        if (file_exists($folderPath)) {
            $items = scandir($folderPath);
        }
        $data['collections'] = [];
        foreach ($items as $item) {
            if ($item == 'globalVariables.json') {
                $fileData = file_get_contents($folderPath . $item);
                $jsonData = json_decode($fileData, true);
                $data['globalVariables'] = $jsonData;
            } elseif ($item == 'tags.json') {
                $fileData = file_get_contents($folderPath . $item);
                $jsonData = json_decode($fileData, true);
                $data['exported_tags'] = $jsonData;
            } elseif (file_exists($folderPath . $item . '/collection.metadata.json')) {
                // Folder
                $playbooks = scandir($folderPath . $item . '/');
                $collectionFolder = $folderPath . $item . '/';
                $collectionFileData = file_get_contents($collectionFolder . 'collection.metadata.json');
                $collection = json_decode($collectionFileData, true);
                $collection['workflows'] = [];
                foreach ($playbooks as $subFile) {
                    if ($subFile != 'collection.metadata.json' && substr($subFile, -5) == '.json') {
                        $pbFileData = file_get_contents($collectionFolder . $subFile);
                        $collection['workflows'][] = json_decode($pbFileData, true);
                    }
                }
                $data['collections'][] = $collection;
            }
        }
        return $data;
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption['collections'] = [];
        $exportOption['globalVariables'] = [];
        foreach ($options['collections']['values'] as $value) {
            array_push($exportOption['collections'], $value['existing_uuid']);
        }
        foreach ($options['globalVariables']['values'] as $value) {
            array_push($exportOption['globalVariables'], $value);
        }
        return $exportOption;
    }
    public function generateInfoContent(array $options)
    {
        $content = [];
        $collectionMetaData = [];
        if (array_key_exists('collections', $options)) {
            $filter = ['uuid' => $this->getUUIDsFromCollections($options['collections'])];
            $entityRepo = $this->entityManager->getRepository(WorkflowCollection::class);
            $dataObj = $this->utilityService->getItemWithWhere($entityRepo, $filter);
            foreach ($dataObj as $collection) {
                $collectionName = $this->cleanFileName($collection->getName());
                $collectionMetaData[] = ["name" => $collectionName];
            }
            $content["playbooks"] = $collectionMetaData;
        }
        $content["globalVariables"] = ['TBD. Depends on the variables used in the selected playbooks at the time of the export'];
        return $content;
    }

    private function getValueOfKeyByCollectionUUID(array $collections, $uuid, $key)
    {
        $index = array_search($uuid, array_column($collections, 'value'));
        return is_int($index) && array_key_exists($key, $collections[$index]) ? $collections[$index][$key] : null;
    }

    private function getUUIDsFromCollections(array $collections)
    {
        return !empty($collections) && is_string($collections[0]) ? $collections :
            array_map(function($collection){return $collection['value'];}, $collections);
    }
}
