<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use stdClass;
use Ramsey\Uuid\Uuid;
use App\Entity\Core\File;
use App\Entity\Core\PgFile;
use Psr\Log\LoggerInterface;
use App\Service\MetadataHelper;
use App\Service\UtilityService;
use App\Security\Voter\FsrVoter;
use App\Service\BulkRequestService;
use App\Doctrine\ORM\FilterQueryBuilder;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\Filesystem\Filesystem;
use App\Providers\PermissionsDictionaryBuilder;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class RecordSetConfig
 * @package App\Service\ConfigExportImport
 */
class RecordSetConfig extends PortingConfig
{
    private const SELF_TENANT = "b3a700f7-00be-4ef9-90c6-3c8fe6e1be63";
    private const SELF_AGENT = "973c17df-bb4b-41e5-b59c-a408666fdf27";
    private const DEFAULT_AGENT_STATUS = "135030fa-1b0b-453d-9fed-df9be5d4397f";
    private const DEFAULT_ROUTER_STATUS = "29395624-0300-46a5-90b3-a6af4c5154d9";

    /** @var MetadataHelper */
    private $metadataHelper;

    private $filterQueryBuilder;

    /** @var BulkRequestService */
    private $bulkRequestService;

    private $fsrVoter;

    /** @var UtilityService */
    private $utilityService;

    private $successes;

    private $failures;

    private $totalRecords;

    public function __construct(
        MetadataHelper $metadataHelper,
        FilterQueryBuilder $filterQueryBuilder,
        BulkRequestService $bulkRequestService,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        FsrVoter $fsrVoter,
        UtilityService $utilityService,
        protected LoggerInterface $logger
    ) {
        $this->metadataHelper = $metadataHelper;
        $this->filterQueryBuilder = $filterQueryBuilder;
        $this->bulkRequestService = $bulkRequestService;
        $this->fsrVoter = $fsrVoter;
        $this->utilityService = $utilityService;
        $this->importUser = '/api/3/people/3451141c-bac6-467c-8d72-85e0fab569ce';
        $this->successes = 0;
        $this->failures = 0;
        $this->totalRecords = 0;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }

    /**
     * @param array $allData
     * @param string $key
     * @return void
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        $options = [];
        $systemFixturesFile = file_get_contents('/opt/cyops-api/src/StaticFile/SystemFixtures.json');
        $systemFixture = json_decode($systemFixturesFile, true);
        foreach ($data as $key => $values) {
            $individualOptions = [];
            $individualOptions['type'] = $key;
            $metadata = $this->metadataHelper->getMetadataByType($key);
            if (!$metadata) {
                $newMetadata = $this->findBy($allData['modules'], 'type', $key);
                if ($newMetadata) {
                    $individualOptions['name'] = $newMetadata['descriptions']['plural'];
                } else {
                    $systemModuleMetadata = $this->findBy([$systemFixture], 'type', $key);
                    if ($systemModuleMetadata) {
                        $individualOptions['name'] = $key;
                        $individualOptions['moduleNotExists'] = false;
                    } else {
                        $individualOptions['name'] = $key;
                        $individualOptions['moduleNotExists'] = true;
                    }
                }
            } else {
                $individualOptions['name'] = $metadata->getDescriptions()['plural'];
            }
            $individualOptions['recordCount'] = count($values);
            $individualOptions['include'] = !$individualOptions['moduleNotExists'];
            $individualOptions['whenExists'] = 'replace';

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {

        $filePathData = $this->loadFilePathsFromDirectory($key, $tmpDirectory);
        $options = [];
        $systemFixturesFile = file_get_contents('/opt/cyops-api/src/StaticFile/SystemFixtures.json');
        $systemFixture = json_decode($systemFixturesFile, true);
        foreach ($filePathData as $moduleType => $filePaths) {
            $recordCount = 0;
            foreach ($filePaths as $index => $filePath) {
                if ($index === 'files') {
                    continue;
                }
                $fileData = file_get_contents($tmpDirectory . $filePath);
                $jsonData = json_decode($fileData, true);
                $recordCount += count($jsonData);
            }
            $individualOptions['type'] = $moduleType;
            $individualOptions['moduleNotExists'] = false;
            $metadata = $this->metadataHelper->getMetadataByType($moduleType);
            if (!$metadata) {
                $individualOptions['name'] = $moduleType;
                $newMetadata = file_exists(sprintf('%smodules/%s/mmd.json', $tmpDirectory, $moduleType));
                $systemFixtureModule = $systemFixture[$moduleType] ?? [];
                $systemModuleMetadata = $this->findBy([$systemFixtureModule], 'type', $moduleType);
                if ($newMetadata || $systemModuleMetadata) {
                    $individualOptions['moduleNotExists'] = false;
                } else {
                    $individualOptions['moduleNotExists'] = true;
                }
            } else {
                $individualOptions['name'] = $metadata->getDescriptions()['plural'];
            }
            $individualOptions['recordCount'] = $recordCount;
            $individualOptions['include'] = !$individualOptions['moduleNotExists'];
            $individualOptions['whenExists'] = 'replace';
            $options[] = $individualOptions;
        }


        return $options;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        // Record set import is not supported for .json files
        return true;
    }

    private function importData(array $option, array $records, array $typeConflicts, $tokenType = 'JWT', array $filesSkippedList = [])
    {
        $className = $this->metadataHelper->getResourceFromClassName($option['type']);
        if (!$className) {
            $this->addMessage(sprintf('Skipping record import for %s as the module was not found.', $option['type']));
            return [];
        }
        $relationshipFieldNames = $this->getRelationshipFieldNames($option['type']);
        [$filterItems, $module] = $this->utilityService->getUniqueResourceFields(null, $option['type']);
        $dataObject = new stdClass();
        $existingUuids = [];

        foreach ($records as $index => $object) {
            // Special case for agents
            if ($option['type'] === 'agents') {
                $records[$index]['active'] = false;
                $records[$index]['configurationHealth'] = self::DEFAULT_AGENT_STATUS;
            }
            // Special case for routers
            if ($option['type'] === 'routers') {
                $records[$index]['configstatus'] = self::DEFAULT_ROUTER_STATUS;
            }
            foreach ($object as $key => $value) {
                if (in_array($key, $relationshipFieldNames) || in_array($key, $typeConflicts)) {
                    unset($records[$index][$key]);
                } elseif ($key != 'uuid' && $key != '@type') {
                    if (($key == 'createUser' || $key == 'modifyUser') && $records[$index][$key]) {
                        $existingIris = $this->findExistingIRIs([$records[$index][$key]]);
                        $records[$index][$key] = count($existingIris) > 0 ? $existingIris[0] : $this->importUser;
                    }
                }
            }
            $entityRepo = $this->entityManager->getRepository($className);
            $uuidFilters = $this->utilityService->prepareFilterData($object, ['uuid'], $className);
            $existingObjectsByUuid = $this->utilityService->getItemWithWhere($entityRepo, $uuidFilters);

            $uniquenessFilters = $this->utilityService->prepareFilterData($object, $filterItems, $className);
            $existingObjectsByUniqueness = $uniquenessFilters ? $this->utilityService->getItemWithWhere($entityRepo, $uniquenessFilters) : [];

            foreach (array_merge($existingObjectsByUuid, $existingObjectsByUniqueness) as $existingItem) {
                $existingUuids[] = $existingItem->getUuid();
                if ($option['whenExists'] !== 'replace') {
                    // Skip item that matches the uniqueness constraint
                    unset($records[$index]);
                }
            }
        }

        $dataObject->{'data'} = $records;

        $headers = [];
        $token = $this->getTokenFromStorage();
        $tokenPrefix = $tokenType == 'JWT' ? 'Bearer ' : 'CS ';
        $headers['Authorization'] = $token ? $tokenPrefix . $token->getCredentials() : null;
        $headers['Content-Type'] = 'application/json';

        if ($option['whenExists'] === 'replace') {
            $deleteItems = [];
            foreach (array_unique($existingUuids) as $existingUuid) {
                $deleteItem = new stdClass();
                $deleteItem->{'uuid'} = $existingUuid;
                $deleteItems[] = $deleteItem;
            }
            if (!empty($deleteItems)) {
                try {
                    $classMeta = $this->entityManager->getClassMetadata($className);
                    $entityFieldMap = $this->metadataHelper->getEntityFields($className) ?? null;
                    $allFileField = "";
                    foreach ($entityFieldMap['fileFields'] as $fileField) {
                        $allFileField = $allFileField . strtolower($fileField) . '=NULL,';
                    }
                    $allFileField = rtrim($allFileField, ',');
                    $tableName = $classMeta->getTableName();
                    $setNullQuery = sprintf("update %s set %s where uuid in ('%s')", $tableName, $allFileField, implode("','", array_unique($existingUuids)));
                    $this->entityManager->getConnection()->prepare($setNullQuery)->executeQuery();
                } catch (\Exception $e) {
                    $this->logger->error("Error : " . $e->getMessage());
                }
            }
            $deleteDataObject = new stdClass();
            $deleteDataObject->{'data'} = $deleteItems;

            $this->bulkRequestService->prepareUpdateDeleteSubRequest(
                $className,
                $deleteDataObject,
                '/api/3/' . $option['type'],
                $headers,
                ['$hardDelete' => "true", '$showDeleted' => "true"],
                'DELETE',
                true,
                'import'
            );
        }

        //Removing irrevelent file reference from the database
        if (!empty($filesSkippedList)) {
            $entityFieldMap = $this->metadataHelper->getEntityFields($className) ?? null;
            $dataObject = $this->skipFileRecords($dataObject, $filesSkippedList, $entityFieldMap);
        }
        $result = $this->bulkRequestService->prepareRecordSubRequest(
            $className,
            $dataObject,
            '/api/3/' . $option['type'],
            $headers,
            [],
            'POST',
            true,
            'import'
        );
        $results = json_decode($result->getContent(), true);
        if (array_key_exists('success', $results)) {
            $successes = count($results['success']);
            $failures = count($results['failure']);
            foreach ($results['failure'] as $index => $failureMessage) {
                $this->addMessage(sprintf('> Failure message #%d: %s', $index, $failureMessage));
            }
        } else {
            $successes = array_key_exists('hydra:member', $results) ? count($results['hydra:member']) : 0;
            $failures = 0;
        }
        $this->successes += $successes;
        $this->failures += $failures;
        $this->addMessage(sprintf('> Successes: %s, Failures: %s', $successes, $failures));

        return array_unique($existingUuids);
    }
    private function skipFileRecords(stdClass $dataObject, array $filesSkippedList, array $entityFieldMap)
    {
        foreach ($dataObject->{'data'} as &$data) {
            foreach ($entityFieldMap['fileFields'] as $fileField) {
                if (!empty($data[$fileField])) {
                    $fileuuid = end(explode("/", $data[$fileField]));
                    if (in_array($fileuuid, $filesSkippedList)) {
                        $data[$fileField] = null;
                    }
                }
            }
        }
        return $dataObject;
    }
    private function importRelationships(array $option, array $records, array $existingUuids, array $typeConflicts, $tokenType = 'JWT')
    {
        $className = $this->metadataHelper->getResourceFromClassName($option['type']);
        if (!$className) {
            $this->addMessage(sprintf('Skipping record import for %s as the module was not found.', $option['type']));
            return [];
        }
        $relationshipFieldNames = $this->getRelationshipFieldNames($option['type']);

        $dataObject = new stdClass();
        foreach ($records as $index => $object) {
            if ($index === 'files') {
                continue;
            }
            if ($option['whenExists'] !== 'replace' && in_array($object['uuid'], $existingUuids)) {
                // skip this item
                unset($records[$index]);
                continue;
            }
            foreach ($object as $key => $value) {
                if ($key == 'uuid' || $key == '@type') {
                    continue;
                } elseif (!in_array($key, $relationshipFieldNames) || in_array($key, $typeConflicts)) {
                    unset($object[$key]);
                    continue;
                }
                if (is_string($value)) {
                    $items = $this->findExistingIRIs([$value]);
                    if (!empty($items)) {
                        $object[$key] = $items[0];
                    } else {
                        unset($object[$key]);
                    }
                } elseif (is_array($value) && count($value) > 0) {
                    $foundIris = $this->findExistingIRIs($value);
                    $object[$key] = $foundIris;
                }
            }
            $records[$index] = (object) $object;
        }

        $dataObject->{'data'} = $records;

        $headers = ["Content-type" => "application/json"];
        $token = $this->getTokenFromStorage();
        $tokenPrefix = $tokenType == 'JWT' ? 'Bearer ' : 'CS ';
        $headers['Authorization'] = $token ? $tokenPrefix . $token->getCredentials() : null;

        $result = $this->bulkRequestService->prepareUpdateDeleteSubRequest($className, $dataObject, '/api/3/' . $option['type'], $headers, [], 'PUT', true, 'import');

        $results = json_decode($result->getContent(), true);
        if (array_key_exists('success', $results)) {
            $successes = count($results['success']);
            $failures = array_key_exists('failure', $results) ? count($results['failure']) : 0;
            foreach ($results['failure'] as $index => $failureMessage) {
                $this->addMessage(sprintf('> Failure message #%d: %s', $index, $failureMessage));
            }
        } else {
            $successes = count($results['hydra:member']);
            $failures = 0;
        }
        $this->addMessage(sprintf('> Successes: %s, Failures: %s', $successes, $failures));
    }

    private function importFileRecords(array $fileRecords, bool $overwrite)
    {
        $restrictedMimeTypes = $this->utilityService->getRestrictedMimeTypeList();

        $fileRepository = $this->entityManager->getRepository(File::class);

        $filesSkippedList = [];
        foreach ($fileRecords as $fileRecord) {
            $existingFile = $fileRepository->findOneBy(['uuid' => $fileRecord['uuid']]);
            if (!empty($restrictedMimeTypes) && in_array($fileRecord['mimeType'], $restrictedMimeTypes) || $fileRecord['size'] == 0) {
                array_push($filesSkippedList, $fileRecord['uuid']);
                continue;
            }
            if ($existingFile && $overwrite) {
                $this->entityManager->remove($existingFile);
            } elseif ($existingFile) {
                // Skip if we are not overwriting
                continue;
            }
            $pgFile = new PgFile();
            $fp = fopen($fileRecord['_pgFile'], 'rb');
            $fileSize = filesize($fileRecord['_pgFile']);
            $pgFile->setFile(base64_encode(fread($fp, $fileSize)));
            fclose($fp);
            $pgFile->setUuid(Uuid::uuid4()->toString());
            $this->entityManager->persist($pgFile);
            $this->entityManager->flush();

            $timestamp = new \DateTime(sprintf('@%d', $fileRecord['uploadDate']), new \DateTimeZone('UTC'));
            $file = new File();
            $file->setFile($pgFile->getUuid());
            $file->setFilename($fileRecord['filename']);
            $file->setMimeType($fileRecord['mimeType']);
            $file->setThumbnail($fileRecord['thumbnail']);
            $file->setSize($fileSize);
            $file->setUploadDate($timestamp);
            $file->setUuid($fileRecord['uuid']);
            $file->setAssignee($fileRecord['assignee']);
            $file->setId($fileRecord['id']);
            $this->entityManager->persist($file);
            $this->entityManager->flush();
        }
        return $filesSkippedList;
    }

    private function getRelationshipFieldNames(string $type)
    {
        $modelMetadata = $this->metadataHelper->getMetadataByType($type);
        $relationshipFieldNames = ['owners', 'tenant', 'userOwners', 'fsrQueue'];
        if (is_null($modelMetadata)) {
            return [];
        }
        foreach ($modelMetadata->getAttributes() as $attribute) {
            if ($attribute->getFormType() == 'manyToMany' || $attribute->getFormType() == 'oneToMany' || $attribute->getFormType() == 'manyToOne' || $attribute->getFormType() == 'lookup') {
                $relationshipFieldNames[] = $attribute->getName();
            }
        }
        return $relationshipFieldNames;
    }

    /**
     * Narrows down a list of IRIs to existing ones
     *
     * @param array $iris
     * @return array
     */
    private function findExistingIRIs(array $iris)
    {
        if (empty($iris)) {
            return [];
        }
        $uuids = [];
        $iriParts = explode('/', $iris[0]);
        $type = $iriParts[3];
        foreach ($iris as $iri) {
            $uuids[] = self::getUuidFromIri($iri);
        }
        $className = $this->metadataHelper->getResourceFromClassName($type);
        $repository = $this->loadRepository($className);
        $items = $repository->findBy(['uuid' => $uuids]);
        $foundIris = [];
        foreach ($items as $item) {
            $foundIris[] = sprintf('/api/3/%s/%s', $type, $item->getUuid());
        }
        return $foundIris;
    }

    private function loadFilePathsFromDirectory(string $key, string $tmpDirectory)
    {
        $relativeFolderPath =  'records/';
        $folderPath = $tmpDirectory . $relativeFolderPath;
        $items = [];
        if (file_exists($folderPath)) {
            $items = scandir($folderPath);
        }
        $data = [];
        foreach ($items as $item) {
            if ($item == '.' || $item == '..' || $item == 'options.json' || !is_dir($folderPath . $item)) {
                continue;
            }
            $subFolder = $folderPath . $item . '/';
            $jsonFiles = [];
            if (file_exists($subFolder)) {
                $jsonFiles = scandir($subFolder);
            }
            foreach ($jsonFiles as $fileName) {
                $filePath = $subFolder . $fileName;
                if ($fileName == '.' || $fileName == '..' || (!str_ends_with($filePath, '.json') && $fileName != 'files')) {
                    continue;
                }
                $data[$item] = array_key_exists($item, $data) ? $data[$item] : [];
                if ($fileName != 'files') {
                    $data[$item][] = $relativeFolderPath . $item . '/' . $fileName;
                } else {
                    $subFiles = [];
                    if (file_exists($filePath)) {
                        $subFiles = scandir($filePath);
                    }
                    foreach ($subFiles as $subFileName) {
                        if ($subFileName == '.' || $subFileName == '..') {
                            continue;
                        }
                        $subFilePath = $filePath . '/' . $subFileName;
                        if (preg_match('/\.json$/', $subFileName)) {
                            $subFileData = file_get_contents($subFilePath);
                            $subJsonData = json_decode($subFileData, true);
                            $subJsonData['_pgFile'] = $filePath . '/' . $subJsonData['uuid'];
                            $data[$item]['files'] = array_key_exists('files', $data[$item]) ? $data[$item]['files'] : [];
                            $data[$item]['files'][] = $subJsonData;
                        }
                    }
                }
            }
        }
        return $data;
    }

    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = [])
    {
        $filesSkippedList = [];
        $folderPath = $tmpDirectory . 'records/';
        $moduleConfig = $this->container->get('fsr.core.service.module_config');
        if ($moduleConfig->isPublishComplete()) {
            // Since we just published, we need to spawn a new thread to import records, as MMDs may have changed
            $optionsJsonPath = $folderPath . 'options.json';
            $fullOptions = [
                'typeConflicts' => $moduleConfig->getTypeConflicts(),
                'options' => $options
            ];
            $filesystem = new Filesystem();
            $filesystem->dumpFile($optionsJsonPath, json_encode($fullOptions));
            $token = $this->tokenStorage->getToken();
            $tokenPart = sprintf('--user-token=%s --token-type=%s', $token->getCredentials(), $tokenType);
            $rootDir = $this->container->getParameter('kernel.project_dir');
            chdir($rootDir . '/../');
            $cmd = sprintf('/usr/bin/php /opt/cyops-api/bin/console app:import:recordsets --env=prod -d "%s" -o "%s" %s -u %s --job-uuid=%s >> /var/log/cyops/cyops-api/last_config_import.log 2>&1', $tmpDirectory, $optionsJsonPath, $tokenPart, $this->importUser, $this->importJob->getUuid());
            exec($cmd, $output, $errorCode);
            $this->entityManager->refresh($this->importJob);
            return !$errorCode;
        }

        $filePathData = $this->loadFilePathsFromDirectory($key, $tmpDirectory);
        $actualOptions = array_key_exists('options', $options) ? $options['options'] : $options;
        $typeConflicts = array_key_exists('typeConflicts', $options) ? $options['typeConflicts'] : [];
        // First, import the non-relationship data
        $existingUuids = [];
        foreach ($actualOptions as $option) {
            if (!$option['include']) {
                continue;
            }
            $filePaths = $filePathData[$option['type']];
            if (array_key_exists('files', $filePaths)) {
                $filesSkippedList = $this->importFileRecords($filePaths['files'], $option['whenExists'] === 'replace');
                unset($filePaths['files']);
            }
            foreach ($filePaths as $index => $filePath) {
                $this->addMessage(sprintf('Importing data for %s', $filePath));
                $fileData = file_get_contents($tmpDirectory . $filePath);
                $jsonData = json_decode($fileData, true);
                $this->totalRecords += count($jsonData);
                $foundExistingUuids = $this->importData($option, $jsonData, $typeConflicts, $tokenType, $filesSkippedList);
                $existingUuids = array_unique(array_merge($existingUuids, $foundExistingUuids));
            }
        }

        // Next, import the relationships
        foreach ($actualOptions as $option) {
            if (!$option['include']) {
                continue;
            }
            $filePaths = $filePathData[$option['type']];
            foreach ($filePaths as $index => $filePath) {
                if ($index === 'files') {
                    continue;
                }
                $this->addMessage(sprintf('Importing relationships for %s', $filePath));
                $fileData = file_get_contents($tmpDirectory . $filePath);
                $jsonData = json_decode($fileData, true);
                $this->importRelationships($option, $jsonData, $existingUuids, $typeConflicts, $tokenType);
            }
        }

        $this->addMessage(sprintf('Imported %d/%d successfully', $this->successes, $this->totalRecords));


        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        $recordSets = [];
        foreach ($options as $option) {
            $recordSets[$option['type']] = $this->exportPage($option, 10000, 1, 10000)['hydra:member'];
        }
        return $recordSets;
    }
    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        foreach ($options as $value) {
            array_push($exportOption, $value);
        }
        return $exportOption;
    }

    private function exportPage(array $option, $itemsPerPage, $page, $maxResults)
    {
        $relationships = array_key_exists('includeCorrelations', $option) ? $option['includeCorrelations'] : true;
        $ignoreDepth = false;
        $export = true;
        $className = $this->metadataHelper->getResourceFromClassName($option['type']);
        $repository = $this->entityManager->getRepository($className);
        $queryBuilder = $repository
            ->createQueryBuilder('o')
            ->setFirstResult(($page - 1) * $itemsPerPage)
            ->setMaxResults($maxResults);

        $this->filterQueryBuilder->fromQueryObject($option['query'], $queryBuilder, $className);
        $result = $queryBuilder->getQuery()->execute();
        if (count($result) == 0) {
            return [];
        }
        $serializer = $this->container->get('serializer');
        $context = $this->getDefaultNormalizationContext($className, $export, $relationships, $ignoreDepth);
        $context['ignored_attributes'] = ['id'];
        $data = $serializer->normalize($result, 'jsonld', $context);
        return $data['hydra:member'];
    }

    private function exportPgFile(string $fileIri, string $modulesDirectory, string $moduleType, Filesystem $filesystem)
    {
        $explodedIRI = explode('/', $fileIri);
        $fileUuid = end($explodedIRI);

        $repository = $this->entityManager->getRepository(File::class);
        /** @var AbstractFile $file */
        $file =  $repository->findOneBy(['uuid' => $fileUuid]);
        if (!$file) {
            return;
        }
        $serializer = $this->container->get('serializer');
        $context = $this->getDefaultNormalizationContext(File::class, true, false, false);
        $context['ignored_attributes'] = [];
        $data = $serializer->normalize($file, 'jsonld', $context);

        $pgFileRepo = $this->entityManager->getRepository(PgFile::class);
        /** @var AbstractFile $file */
        $pgFile = $pgFileRepo->findOneBy(['uuid' => $file->getFile()]);

        $freeSpace = disk_free_space(sys_get_temp_dir());
        if ($freeSpace - $file->getSize() < self::RESERVE_BUFFER_BYTES) {
            throw new \Exception(sprintf('Not enough free space in /tmp to import this file. Available: %.1fMb', $freeSpace / 1000000));
        }

        $jsonFilePath = sprintf('%s/%s/files/file_%s.json', $modulesDirectory, $moduleType, $fileUuid);
        $jsonData = json_encode($data, JSON_PRETTY_PRINT);
        $filesystem->dumpFile($jsonFilePath, $jsonData);
        $filePath = sprintf('%s/%s/files/%s', $modulesDirectory, $moduleType, $fileUuid);
        $filesystem->dumpFile($filePath, base64_decode(stream_get_contents($pgFile->getFile())));
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $modulesDirectory = $tmpDirectory . '/records';
        $filesystem = new Filesystem();
        $pageSize = 100;
        $metaData = [];

        foreach ($options as $option) {
            $className = $this->metadataHelper->getResourceFromClassName($option['type']);
            $module = $this->metadataHelper->getModuleFromClassName($className);
            if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $option['type']) && !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.' . $module)) {
                continue;
            }
            $page = 1;
            $option['query'] = $option['query'] ? $option['query'] : [];
            if (!(isset($option['query']['logic']) && isset($option['query']['filters']))) {
                $option['query']['logic'] = "AND";
                $option['query']['filters'] = [];
                $option['query']['limit'] = 100000;
            }
            if ($option['type'] === 'agents' || $option['type'] === 'tenants') {
                $option['query'] = [
                    'logic' => 'AND',
                    'limit' => $option['query']['limit'],
                    'filters' => [
                        $option['query'],
                        [
                            'field' => 'uuid',
                            'operator' => 'neq',
                            'value' => self::SELF_TENANT
                        ],
                        [
                            'field' => 'uuid',
                            'operator' => 'neq',
                            'value' => self::SELF_AGENT
                        ]
                    ]
                ];
            }
            $hasNextPage = true;
            while ($hasNextPage) {
                // If query limit is 117, get 100 the first page and 17 on the next page
                $maxResults = $page * $pageSize < $option['query']['limit'] ? $pageSize : $option['query']['limit'] - ($page - 1) * $pageSize;
                $maxResults = $maxResults < 0 ? 0 : $maxResults;
                $dataPage = $this->exportPage($option, $pageSize, $page, $maxResults);
                $modelMetadata = $this->metadataHelper->getMetadataByType($option['type']);
                $fileFieldNames = [];
                if ($modelMetadata) {
                    foreach ($modelMetadata->getAttributes() as $attribute) {
                        if ($attribute->getFormType() == 'file') {
                            $fileFieldNames[] = $attribute->getName();
                        }
                    }
                }
                foreach ($dataPage as $record) {
                    // Export files
                    foreach ($record as $key => $value) {
                        if ($value && in_array($key, $fileFieldNames)) {
                            $this->exportPgFile($value, $modulesDirectory, $option['type'], $filesystem);
                        }
                    }
                }
                if (count($dataPage) > 0) {
                    if (isset($metaData[$option['type']])) {
                        $metaData[$option['type']]['count'] +=  count($dataPage);
                    } else {
                        $metaData[$option['type']] = ['name' => $option['label'], 'apiName' => $option['type'], 'count' => count($dataPage)];
                    }
                    $filename = sprintf('%s/%s/%s%04d.json', $modulesDirectory, $option['type'], $option['type'], $page);
                    $jsonData = json_encode($dataPage, JSON_PRETTY_PRINT);
                    $freeSpace = disk_free_space(sys_get_temp_dir());
                    if ($freeSpace - strlen($jsonData) < self::RESERVE_BUFFER_BYTES) {
                        throw new \Exception(sprintf('Not enough free space in /tmp to import this file. Available: %.1fMb', $freeSpace / 1000000));
                    }
                    $filesystem->dumpFile($filename, $jsonData);
                    $page++;
                } else {
                    $hasNextPage = false;
                }
            }
        }
        if (!empty($metaData)) {
            $recordCount = [];
            foreach ($metaData as $value) {
                $recordCount[] = $value;
            }
            $this->appendToFile($tmpDirectory . 'info.json', $type, $recordCount);
        }
        return true;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        foreach ($options as $option) {
            $content[] = ['name' => $option['label'], 'apiName' => $option['type'], 'count' => 'TBD'];
        }
        return $content;
    }
}
