<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */


namespace App\Service\ConfigExportImport;

use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Service\UtilityService;
use App\Security\Voter\FsrVoter;
use App\Entity\ViewTemplate\Reporting;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\Filesystem\Filesystem;
use App\Providers\PermissionsDictionaryBuilder;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class ReportConfig
 * @package App\Service\ConfigExportImport
 */
class ReportConfig extends PortingConfig
{
    /** @var fsrVoter */
    private $fsrVoter;
    private $utilityService;
    private const SCHEDULE_KEY = 'includeSchedules';

    public function __construct(
        FsrVoter $fsrVoter,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        UtilityService $utilityService,
        protected LoggerInterface $logger
    ) {
        $this->fsrVoter = $fsrVoter;
        $this->utilityService = $utilityService;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }

    /**
     * @param array $allData
     * @param string $key
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        if (
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.reporting') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.reporting') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.reporting') ||
            !$data
        ) {
            return [];
        }
        $repository = $this->loadRepository(Reporting::class);
        $options = [];
        foreach ($data as $report) {
            $existing = $repository->findOneBy(['uuid' => $report['uuid']]);
            $foundRoles = $this->loadRolesFromReport($report, $allData['roles']);
            $individualOptions = [];
            $individualOptions['uuid'] = $report['uuid'];
            $individualOptions['name'] = $report['displayName'];
            $individualOptions['roles'] = $foundRoles;
            $individualOptions['include'] = true;
            $individualOptions['exists'] = $existing ? true : false;
            $individualOptions['schedulesCount'] = array_key_exists('schedules', $report) ? count($report['schedules']) : 0;
            $individualOptions['includeSchedules'] = $individualOptions['schedulesCount'] > 0 ? true : false;
            
            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        $reportsToAdd = [];
        $reportsToUpdate = [];
        $schedulesToImport = [];
        foreach ($options as $reportOption) {
            if (!$reportOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($data as $report) {
                if ($reportOption['uuid'] == $report['uuid']) {
                    unset($report['userId']);
                    $foundRoles = $this->loadRolesFromReport($report, []);
                    $report['roles'] = array_column($foundRoles, '@id');
                    $report['filterArray'] = isset($report['config']['filterArray']) ? $report['config']['filterArray'] : $report['filterArray'];

                    unset($report['roleIds']);
                    $report['importedBy'] = array_key_exists("importedBy", $report) && $report['importedBy'] ?  array_merge($report['importedBy'], $importedBy) : $importedBy;
                    if (array_key_exists('schedules', $report) && $reportOption['includeSchedules']) {
                        foreach ($report['schedules'] as $schedule) {
                            $schedule['kwargs']['referenceid'] = $report['uuid'];
                            $schedulesToImport[] = $schedule;
                        }
                    }
                    if (!$reportOption['exists']) {
                        $reportsToAdd[] = $report;
                        break;
                    }
                    $report['@id'] = sprintf('/api/3/reporting/%s', $report['uuid']);
                    unset($report['uuid']);
                    $reportsToUpdate[] = $report;
                    break;
                }
            }
        }
        $results = [];
        foreach ($reportsToAdd as $report) {
            $fullURI = sprintf('%s/api/3/reporting', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $result = $this->sendRequest($fullURI, 'post', $report, $tokenType);
            $results[] = $result;
        }
        if (!empty($reportsToUpdate)) {
            $fullURI = sprintf('%s/api/3/update/reporting', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $body = [
                "data" => $reportsToUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
        }

        foreach ($schedulesToImport as $schedule) {
            $this->saveSchedule($schedule, $tokenType);
        }

        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.reporting')) {
            return [];
        }
        $uuids = $this->getUUIDsFromOptions($options);
        $reports = $this->getResourceFromIRI($uuids, Reporting::class, true);
        foreach ($reports as $index => $report) {
            if ($this->getValueOfKeyByCollectionUUID($options, $report['uuid'], self::SCHEDULE_KEY)) {
                $schedules = $this->loadSchedules($report['uuid'], $tokenType, 'report');
                if (count($schedules) > 0) {
                    $reports[$index]['schedules'] = $schedules;
                }
            }
        }
        return $reports;
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        foreach ($options as $value) {
            array_push($exportOption, $value['uuid']);
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        $filter = ['uuid' => $this->getUUIDsFromOptions($options)];
        $entityRepo = $this->entityManager->getRepository(Reporting::class);
        $dataObj = $this->utilityService->getItemWithWhere($entityRepo, $filter);
        foreach ($dataObj as $report) {
            $content[] = ["name" => $report->getDisplayName()];
        }
        return $content;
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $metaData = [];
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        $filesystem = new Filesystem();
        foreach ($data as $item) {
            $filename = $tmpDirectory . $type . '/' . $this->cleanFileName($item['displayName']) . '.json';
            $this->writeToFile($filename, $item);
            $metaData[] = ['name' => $item['displayName']];
            $filenames[] = $filename;
        }
        $this->appendToFile($tmpDirectory . 'info.json', $type, $metaData);
        return $filenames;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $defaultJsonPath = $tmpDirectory . $key . '/data.json';
        if (file_exists($defaultJsonPath)) {
            $fileData = file_get_contents($defaultJsonPath);
            $jsonData = json_decode($fileData, true);
        } elseif (is_dir($tmpDirectory . $key)) {
            $jsonData = $this->loadDataFromDirectory($key, $tmpDirectory);
        }
        $roleData = [];
        if (is_dir($tmpDirectory . 'roles')) {
            $roleData = $this->loadDataFromDirectory('roles', $tmpDirectory);
        }
        $jsonData = isset($jsonData) ? $jsonData : [];
        return $this->getImportOptions([$key => $jsonData, 'roles' => $roleData], $key, $tokenType);
    }

    private function getValueOfKeyByCollectionUUID(array $collections, $uuid, $key) {
        $index = array_search($uuid, array_column($collections, 'value'));
        return is_int($index) && array_key_exists($key, $collections[$index]) ? $collections[$index][$key] : null;
    }

    private function getUUIDsFromOptions(array $options)
    {
        return !empty($options) && is_string($options[0]) ? $options :
            array_map(function ($report) { return $report['value']; }, $options);
    }
}
