<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Entity\Core\Module;
use App\Constants\AppConstants;
use App\Entity\Authorization\Role;

/**
 * Class RoleConfig
 * @package App\Service\ConfigExportImport
 */
class RoleConfig extends PortingConfig
{
    /**
     * @param array $allData
     * @param string $key
     * @return void
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        $repository = $this->loadRepository(Role::class);
        $options = [];
        foreach ($data as $role) {
            $iriParts = explode('/', $role['@id']);
            $uuid = end($iriParts);
            $existing = $repository->findOneBy(['uuid' => $uuid]);
            $existingByName = $repository->findOneBy(['name' => $role['name']]);
            $individualOptions = [];
            $individualOptions['uuid'] = $uuid;
            $individualOptions['existingUuid'] = $existingByName ? $existingByName->getUuid() : $uuid;
            $individualOptions['name'] = $role['name'];
            $individualOptions['description'] = $role['description'];
            $individualOptions['include'] = true;
            $individualOptions['exists'] = $existing || $existingByName ? true : false;

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        $toAdd = [];
        $toUpdate = [];
        $moduleRepository = $this->loadRepository(Module::class);
        foreach ($options as $roleOption) {
            if (!$roleOption['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($data as $role) {
                $iriParts = explode('/', $role['@id']);
                $uuid = end($iriParts);
                if ($roleOption['uuid'] == $uuid) {
                    $existingModulePermissions = [];
                    $moduleUuids = [];
                    foreach ($role['modulePermissions'] as $modulePermission) {
                        $module = $moduleRepository->findOneBy(['type' => $modulePermission['module']]);
                        // Only include permissions for existing modules
                        if ($module) {
                            $modulePermission['module'] = '/api/3/modules/' . $module->getUuid();
                            $moduleUuids[] = $module->getUuid();
                            unset($modulePermission['@id']);
                            unset($modulePermission['uuid']);
                            if (array_key_exists('fieldPermissions', $modulePermission)) {
                                $newFieldPermissions = [];
                                foreach ($modulePermission['fieldPermissions'] as $fieldPermission) {
                                    unset($fieldPermission['@id']);
                                    unset($fieldPermission['uuid']);
                                    $newFieldPermissions[] = $fieldPermission;
                                }
                                $modulePermission['fieldPermissions'] = $newFieldPermissions;
                            }
                            $existingModulePermissions[] = $modulePermission;
                        }
                    }
                    $role['importedBy'] = array_key_exists("importedBy", $role) && $role['importedBy'] ? array_merge($role['importedBy'], $importedBy) : $importedBy;
                    $role['modulePermissions'] = $existingModulePermissions;
                    if ($roleOption['exists']) {
                        $iri =  '/api/3/roles/' . $roleOption['existingUuid'];
                        /** @var Role $existing */
                        $roleOption = [
                            'value' => $iri
                        ];
                        $existingRole = $this->export([$roleOption], 'JWT', '@id')[0];
                        // Automatically merge existing permissions
                        foreach ($existingRole['modulePermissions'] as $modulePermission) {
                            $iriParts = explode('/', $modulePermission['module']);
                            $moduleUuid = end($iriParts);
                            if (!in_array($moduleUuid, $moduleUuids)) {
                                $role['modulePermissions'][] = $modulePermission;
                            }
                        }
                        $role['@id'] = $iri;
                        unset($role['uuid']);
                        // for solution packs, do not update the name and description if the role already exists
                        if ($this->importJob && $this->importJob->getType() === 'SolutionPack Import') {
                            if (isset($role['name'])) {
                                unset($role['name']);
                            }
                            if (isset($role['description'])) {
                                unset($role['description']);
                            }
                        }
                        $toUpdate[] = $role;
                    } else {
                        unset($role['@id']);
                        $role['uuid'] = $uuid;
                        $toAdd[] = $role;
                    }
                    break;
                }
            }
        }
        $results = [];
        foreach ($toAdd as $item) {
            $fullURI = sprintf('%s/api/3/roles', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $result = $this->sendRequest($fullURI, 'post', $item, $tokenType);
            $results[] = $result;
        }
        if (!empty($toUpdate)) {
            $fullURI = sprintf('%s/api/3/update/roles', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $body = [
                "data" => $toUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
        }

        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT', $moduleField = 'type')
    {
        $iris = array_column($options, 'value');
        $roles = $this->getResourceFromIRI($iris, Role::class, true, false, false, true);
        foreach ($roles as $index => $role) {
            $option = $this->findBy($options, 'value', $role['@id']);
            foreach ($role['modulePermissions'] as $modulePermissionIndex => $modulePermission) {
                if (array_key_exists('module', $modulePermission) && is_array($modulePermission['module'])) {
                    $moduleType = $modulePermission['module'][$moduleField];
                    $roles[$index]['modulePermissions'][$modulePermissionIndex]['module'] = $moduleType;
                    if (array_key_exists('modulePermissions', $option) && array_key_exists($moduleType, $option['modulePermissions']) && !$option['modulePermissions'][$moduleType]) {
                        unset($roles[$index]['modulePermissions'][$modulePermissionIndex]);
                    } else {
                        unset($roles[$index]['modulePermissions'][$modulePermissionIndex]['@id']);
                        unset($roles[$index]['modulePermissions'][$modulePermissionIndex]['uuid']);
                        foreach ($roles[$index]['modulePermissions'][$modulePermissionIndex]['fieldPermissions'] as $fieldPermissionIndex => $fieldPermission) {
                            unset($roles[$index]['modulePermissions'][$modulePermissionIndex]['fieldPermissions'][$fieldPermissionIndex]['@id']);
                            unset($roles[$index]['modulePermissions'][$modulePermissionIndex]['fieldPermissions'][$fieldPermissionIndex]['uuid']);
                        }
                    }
                }
            }
        }
        return $roles;
    }

    protected function getRoleModulePermissions($uuid)
    {
        $modulePermissions = [];
        $repository = $this->entityManager->getRepository(Role::class);
        $context = $this->getDefaultNormalizationContext(Role::class, false, true, true);
        $record = $repository->findOneBy(['uuid' => $uuid]);
        $normalizedObject = $this->normalizer->normalize($record, 'jsonld', $context);
        foreach ($normalizedObject['modulePermissions'] as $modulePermission) {
            $modulePermissions[$modulePermission['module']['type']] = false;
        }
        return $modulePermissions;
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        $fileExtractedPath = $this->tmpFilePath;
        $folderPath = $fileExtractedPath . 'roles';
        foreach ($options as $value) {
            $roleFilePath = sprintf('%s/%s-%s.json', $folderPath, $value['name'], $value['uuid']);
            $fileData = json_decode(file_get_contents($roleFilePath), true);
            $fileData['value'] = $fileData['@id'];
            $fileData['label'] = $fileData['name'];
            $fileData['include'] = true;
            $modulePermissions = $this->getRoleModulePermissions($value['uuid']);
            $fileData['modulePermissionDetails'] = [];
            if (array_key_exists('modulePermissions', $fileData)) {
                foreach ($fileData['modulePermissions'] as $index => $modulePermission) {
                    $moduleType = $modulePermission['module'];
                    unset($modulePermission['uuid']);
                    $fileData['modulePermissionDetails'][$moduleType] = $modulePermission;
                    $modulePermissions[$moduleType] = true;
                }
            }
            $fileData['modulePermissions'] = $modulePermissions;
            array_push($exportOption, $fileData);
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        foreach ($options as $role) {
            $content[] = ["name" => $role["label"]];
        }
        return $content;
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $metaData = [];
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        foreach ($data as $item) {
            $filename = $tmpDirectory . $type . '/' . $this->cleanFileName($item['name']) . '-' . $item['uuid'] . '.json';
            $this->writeToFile($filename, $item);
            $metaData[] = ['name' => $item['name']];
            $filenames[] = $filename;
        }
        $this->appendToFile($tmpDirectory . 'info.json', $type, $metaData);
        return $filenames;
    }
}
