<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Security\Voter\FsrVoter;
use Doctrine\ORM\EntityManagerInterface;
use App\Providers\PermissionsDictionaryBuilder;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class RuleConfig
 * @package App\Service\ConfigExportImport
 */
class RuleConfig extends PortingConfig
{
    /** @var FsrVoter */
    private $fsrVoter;

    public function __construct(
        FsrVoter $fsrVoter,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        protected LoggerInterface $logger
    ) {
        $this->fsrVoter = $fsrVoter;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }

    /**
     * @param array $allData
     * @param string $key
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        $result = [];
        if (
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_UPDATE . '.rules') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_CREATE . '.rules') ||
            !$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.rules')
        ) {
            $result = [];
        } elseif ($key == 'ruleChannels' && !empty($data)) {
            $result = $this->generateOptionsForType($data, 'channel', $tokenType);
        } elseif ($key == 'rules' && !empty($data)) {
            $result = $this->generateOptionsForType($data, 'rules', $tokenType);
        }
        return $result;
    }

    /**
     * @param array $config
     * @return array
     */
    private function generateOptionsForType(array $data, $entityType, $tokenType)
    {
        $existingItems = $this->loadFromSealab($entityType, $tokenType);
        $options = [];
        foreach ($data as $item) {
            $existingRecord = self::findBy($existingItems, 'name', $item['name']);
            $individualOptions = [];
            $individualOptions['name'] = $item['name'];
            if ($existingRecord) {
                $individualOptions['uuid'] = $existingRecord['uuid'];
            } else {
                $individualOptions['uuid'] = $item['uuid'];
            }
            if ($this->importJob && $this->importJob->getType() === 'SolutionPack Import') {
                $individualOptions['include'] = true;
            } else {
                $individualOptions['include'] = !$existingRecord;
            }
            $individualOptions['exists'] = (bool) $existingRecord;

            $options[] = $individualOptions;
        }
        return $options;
    }

    private function loadFromSealab($entityType, $tokenType = 'JWT')
    {
        $token = $this->getTokenFromStorage();
        $endpoint = sprintf('/api/%s/?offset=0&limit=2147483647&visible=true', $entityType);
        if ($token) {
            $uri = $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY) . '/api/rule' . $endpoint;
            $response = $this->sendRequest($uri, 'get', null, $tokenType);
            $jsonDecodeResponse = json_decode($response, true);
            return array_key_exists('hydra:member', $jsonDecodeResponse) ? $jsonDecodeResponse['hydra:member'] : $jsonDecodeResponse;
        } else {
            $uri = $this->container->getParameter('rule.uri') . $endpoint;
            return $this->sealabRequest($uri, 'get');
        }
    }


    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @param string $tokenType
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $key = '', $importedBy = [])
    {
        $toAdd = [];
        $toUpdate = [];
        foreach ($options as $option) {
            if (!$option['include']) {
                // Do not merge skipped items
                continue;
            }
            foreach ($data as $item) {
                if ($option['uuid'] == $item['uuid']) {
                    $item['importedBy'] = array_key_exists("importedBy", $item) && $item['importedBy'] ? array_merge($item['importedBy'], $importedBy) : $importedBy;
                    if ($option['exists']) {
                        $toUpdate[] = $item;
                    } else {
                        $toAdd[] = $item;
                    }
                    break;
                }
            }
        }
        $results = [];
        $endpointKey = $key == 'ruleChannels' ? 'channel' : 'rules';
        foreach ($toAdd as $item) {
            $token = $this->getTokenFromStorage();
            if ($token) {
                $uri = sprintf('%s/api/rule/api/%s/', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $endpointKey);
                $results[] = $this->sendRequest($uri, 'post', $item, $tokenType);
            } else {
                $uri = sprintf('%s/api/%s/', $this->container->getParameter('workflow.uri'), $endpointKey);
                $results[] = $this->sealabRequest($uri, 'post', $item, $tokenType);
            }
        }

        foreach ($toUpdate as $item) {
            $token = $this->getTokenFromStorage();
            if ($token) {
                $uri = sprintf('%s/api/rule/api/%s/%s/', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $endpointKey, $item['uuid']);
                $results[] = $this->sendRequest($uri, 'put', $item, $tokenType);
            } else {
                $uri = sprintf('%s/api/%s/%s/', $this->container->getParameter('workflow.uri'), $endpointKey, $item['uuid']);
                $results[] = $this->sealabRequest($uri, 'put', $item, $tokenType);
            }
        }

        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.rules')) {
            return [];
        }
        $data = [];
        if (!empty($options) && $options[0]['type'] == 'channel') {
            $existingChannels = $this->loadFromSealab('channel', $tokenType);
            foreach ($options as $channelOption) {
                foreach ($existingChannels as $existingChannel) {
                    if ($existingChannel['uuid'] == $channelOption['value']) {
                        $data[] = $existingChannel;
                        break;
                    }
                }
            }
        }
        if (!empty($options) && $options[0]['type'] == 'rule') {
            $existingRules = $this->loadFromSealab('rules', $tokenType);
            foreach ($options as $ruleOption) {
                foreach ($existingRules as $existingRule) {
                    if ($existingRule['uuid'] == $ruleOption['value']) {
                        $data[] = $existingRule;
                        break;
                    }
                }
            }
        }
        return $data;
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $metaData = [];
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        $directory = $tmpDirectory . $type . '/';
        foreach ($data as $item) {
            $name = $this->cleanFileName($item['name']);
            $filename = $directory . $name . '.json';
            $this->writeToFile($filename, $item);
            $filenames[] = $filename;
            $metaData[] = $item['name'];
        }
        $this->appendToFile($tmpDirectory . 'info.json', $type, $metaData);
        return $filenames;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->getImportOptions([$key => $data], $key, $tokenType);
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = [])
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->import($data, $options, $tokenType, $key, $importedBy);
    }

    protected function loadDataFromDirectory(string $key, string $tmpDirectory)
    {
        $folderPath = $tmpDirectory . $key . '/';
        $files = [];
        $data = [];
        if (file_exists($folderPath)) {
            $files = scandir($folderPath);
        }
        foreach ($files as $file) {
            if (substr($file, -5) == '.json') {
                $fileData = file_get_contents($folderPath . $file);
                $data[] = json_decode($fileData, true);
            }
        }
        return $data;
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        if (is_iterable($options)) {
            foreach ($options as $value) {
                $value['value'] = $value['uuid'];
                array_push($exportOption, $value);
            }
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        foreach ($options as $rule) {
            $content[] = $rule["label"] ?? $rule["name"];
        }

        return $content;
    }
}
