<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Constants\AppConstants;
use App\Entity\ModelMetadata\ModelMetadata;
use App\Entity\ViewTemplate\SystemViewTemplate;

/**
 * Class ViewConfig
 * @package App\Service\ConfigExportImport
 */
class ViewConfig extends PortingConfig
{
    /**
     * @param array $allData
     * @param string $key
     * @return array
     */
    public function getImportOptions(array $allData, string $typeKey, $tokenType)
    {
        $data = $allData[$typeKey];
        $options = [];
        $repository = $this->loadRepository(SystemViewTemplate::class);
        foreach ($data as $key => $value) {
            if (!$value) {
                continue;
            }
            if (count(array_filter(array_keys($value), 'is_string')) == 0) {
                // This is a sequential array, used by legacy module view templates
                foreach ($value as $view) {
                    $existing = $repository->findOneBy(['uuid' => $view['uuid']]);
                    $individualOptions = [];
                    $uuid = $view['uuid'];
                    if (!$uuid) {
                        $uuid = implode('-', explode('/', $view['id']));
                    }
                    $individualOptions['uuid'] = $uuid;
                    if ($view['config'] && array_key_exists('title', $view['config'])) {
                        $individualOptions['name'] = $view['config']['title'];
                    }
                    $individualOptions['module'] = $key;
                    $individualOptions['exists'] = $existing ? true : false;
                    $individualOptions['include'] = true;

                    $options[] = $individualOptions;
                }
                continue;
            }
            $existing = $repository->findOneBy(['uuid' => $value['uuid']]);
            $individualOptions = [];
            $uuid = $value['uuid'];
            $pieces = explode('-', $value['uuid']);
            if (count($pieces) === 3) {
                $individualOptions['module'] = $pieces[1];
            }
            if ($value['config'] && array_key_exists('title', $value['config'])) {
                $individualOptions['name'] = $value['config']['title'];
            }
            if ($value['uuid'] == 'app') {
                foreach ($value['config']['navigation'] as $key =>  $item) {
                    foreach ($existing->getConfig()['navigation'] as $existingItem) {
                        if ($existingItem['title'] == $item['title']) {
                            $value['config']['navigation'][$key]['exists'] = true;
                            $value['config']['navigation'][$key]['include'] = true;
                        }
                    }
                }
                foreach ($value['config']['navigation'] as $key =>  $item) {
                    if (!array_key_exists("exists", $value['config']['navigation'][$key]) || (array_key_exists("exists", $value['config']['navigation'][$key]) && $value['config']['navigation'][$key]['exists'] == false)) {
                        $value['config']['navigation'][$key]['exists'] = false;
                        $value['config']['navigation'][$key]['include'] = true;
                    }
                }
            }
            if (is_array($pieces) && array_key_exists("2", $pieces) && $pieces[2] == 'settings') {
                $individualOptions['_hidden'] = true;
            }
            $individualOptions['uuid'] = $uuid;
            $individualOptions['exists'] = $existing ? true : false;
            $individualOptions['include'] = true;
            $individualOptions['config'] = $value['config'];

            $options[] = $individualOptions;
        }
        return $options;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        $toAdd = [];
        $toUpdate = [];
        foreach ($options as $option) {
            if (is_array($option) && array_key_exists("_hidden", $option) && $option['_hidden'] && strpos($option['uuid'], '-settings')) {
                $detailUuid = str_replace('settings', 'detail', $option['uuid']);
                $detailView = self::findBy($options, 'uuid', $detailUuid);
                $option['include'] = $detailView['include'];
            }
            if (!$option['include']) {
                // Do not merge skipped items
                continue;
            }
            $option['appendNavigation'] = array_key_exists('mergeType', $option) && $option['mergeType'] == 'replace' ? false : true;
            $foundView = null;
            foreach ($data as $value) {
                if (!$value) {
                    continue;
                }
                if (count(array_filter(array_keys($value), 'is_string')) == 0) {
                    // This is a sequential array, used by module view templates
                    foreach ($value as $view) {
                        if ($option['uuid'] == $view['uuid']) {
                            $foundView = $view;
                            break;
                        }
                    }
                    continue;
                }
                if ("app" == $value['uuid'] && $option['appendNavigation']) {
                    $fullURI = sprintf('%s/api/views/1/app', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
                    $oldNav = json_decode($this->sendRequest($fullURI, 'GET', null, $tokenType), true);
                    $foundView = $this->mergeNAppendNav($option['config']['navigation'], $oldNav);
                } elseif ($option['uuid'] == $value['uuid']) {
                    $value['importedBy'] = array_key_exists("importedBy", $value) && $value['importedBy'] ? array_merge($value['importedBy'], $importedBy) : $importedBy;
                    $foundView = $value;
                    break;
                }
            }
            if ($option['exists'] && $foundView) {
                $toUpdate[] = $foundView;
            } elseif ($foundView) {
                $toAdd[] = $foundView;
            }
        }
        $results = [];
        if (!empty($toAdd)) {
            $fullURI = sprintf('%s/api/3/', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $result = $this->sendRequest($fullURI, 'post', $toAdd, $tokenType);
            $results[] = $result;
        }
        if (!empty($toUpdate)) {
            $fullURI = sprintf('%s/api/3/update/system_view_templates', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
            $body = [
                "data" => $toUpdate
            ];
            $result = $this->sendRequest($fullURI, 'put', $body, $tokenType);
            $results[] = $result;
        }

        if (in_array(false, $results)) {
            return false;
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        return $this->getResourceFromIRI($options, SystemViewTemplate::class);
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $modifiedOptions = [];
        $appendNavigation = [];
        $metaData = [];
        $key = null;
        foreach ($options as $value) {
            if (is_string($value)) {
                $modifiedOptions[] = $value;
            }
            if (is_array($value) && !empty($value)) {
                $modifiedOptions[] = $value['value'];
                if (!empty($value['appendNavigation'])) {
                    $appendNavigation = $value['appendNavigation'];
                }
            }
        }
        $data = $this->export($modifiedOptions, $tokenType);
        if (!$data) {
            return $data;
        }
        foreach ($data as $item) {
            $uuidParts = explode('-', $item['uuid']);
            if ($uuidParts[0] === 'modules') {
                if ($uuidParts[2] === 'settings') {
                    $typeName = 'settings.json';
                } else {
                    $typeName = $uuidParts[2] . '-layout.json';
                }
                $filename = $tmpDirectory . 'modules/' . $this->cleanFileName($uuidParts[1]) . '/' . $typeName;
                $metaData[$uuidParts[1]][] = $uuidParts[2];
            } else {
                if ($item['uuid'] == 'app') {
                    $item['uuid'] = 'navigation';
                    $metaData[$item['uuid']] = [];
                    $appendList = [];
                    $navigation = [];
                    foreach ($item['config']['navigation'] as $value) {
                        if (in_array($value['title'], $appendNavigation)) {
                            $value = $this->sortArrayByKey($value); // First and second level sorting
                            foreach ($value as $key => $arrayElement) {
                                if (is_array($arrayElement) && !empty($arrayElement)) {
                                    $value[$key] = $this->sortArrayByKey($value[$key]); // Third level sorting
                                }

                                foreach ($arrayElement as $arrayKey => $arrayValue) {
                                    if (is_array($arrayValue) && !empty($arrayValue)) {
                                        $arrayElement[$arrayKey] = $this->sortArrayByKey($arrayElement[$arrayKey]); // Fourth level sorting
                                    }
                                }
                                $value[$key] = $arrayElement;
                            }
                            $appendList[] = $value;
                            array_push($navigation, $value['title']);
                        }
                    }
                    $metaData[$item['uuid']] = $navigation;
                    $item['config']['navigation'] =  $appendList;
                }
                $filename = $tmpDirectory . 'views/' . $item['uuid'] . '.json';
            }
            $this->writeToFile($filename, $item);
            $filenames[] = $filename;
        }
        if ($type == 'viewTemplates') {
            $viewMetaData = [];
            foreach ($metaData as $key => $value) {
                $mmd = $this->getMetadataByType($key);
                $viewMetaData[$key] = ['name' => $mmd->getDescriptions()['plural'], 'apiName' => $key, 'views' => $value];
            }
            $metaData = $viewMetaData;
        }
        $this->appendToFile($tmpDirectory . 'info.json', $type, $metaData);
        return $filenames;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->getImportOptions([$key => $data], $key, $tokenType);
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = [])
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->import($data, $options, $tokenType, $importedBy);
    }

    protected function loadDataFromDirectory(string $key, string $tmpDirectory)
    {
        // Module views
        if ($key == 'viewTemplates') {
            $folderPath = $tmpDirectory . 'modules/';
            $folders = [];
            if (file_exists($folderPath)) {
                $folders = scandir($folderPath);
            }
            $data = [];
            foreach ($folders as $folder) {
                $viewPaths = [
                    $folderPath . $folder . '/detail-layout.json',
                    $folderPath . $folder . '/list-layout.json',
                    $folderPath . $folder . '/form-layout.json',
                    $folderPath . $folder . '/settings.json'
                ];
                foreach ($viewPaths as $viewPath) {
                    if (file_exists($viewPath)) {
                        $fileData = file_get_contents($viewPath);
                        $jsonData = json_decode($fileData, true);
                        $data[] = $jsonData;
                    }
                }
            }
        } else {
            // Other views
            $folderPath = $tmpDirectory . $key . '/';
            $folders = [];
            if (file_exists($folderPath)) {
                $folders = scandir($folderPath);
            }
            $data = [];
            foreach ($folders as $file) {
                $viewPath = $folderPath . $file;
                if (file_exists($viewPath)) {
                    $fileData = file_get_contents($viewPath);
                    $jsonData = json_decode($fileData, true);
                    if (isset($jsonData['uuid']) && $jsonData['uuid'] == 'navigation') {
                        $jsonData['uuid'] = 'app';
                    }
                    $data[] = $jsonData;
                }
            }
        }

        return $data;
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        foreach ($options as $value) {
            if ($value['uuid'] == 'app') {
                $value['appendNavigation'] = [];
                foreach ($value['config']['navigation'] as $item) {
                    $value['appendNavigation'][] = $item['title'];
                }
                array_push($exportOption, ['value' => $value['uuid'], 'appendNavigation' => $value['appendNavigation']]);
            } else {
                array_push($exportOption, $value['uuid']);
            }
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        foreach ($options as $item) {
            if (is_array($item)) {
                if ($item["value"] == "app") {
                    $content["navigation"] = $item["appendNavigation"];
                }
            } else {
                $moduleParts = explode('-', $item);
                if (!array_key_exists($moduleParts[1], $content)) {
                    $module = $this->getMetadataByType($moduleParts[1]);
                    $content[$moduleParts[1]] = [
                        "name" => $module->getDescriptions()['plural'],
                        "apiName" => $moduleParts[1],
                        "views" => []
                    ];
                }
                array_push($content[$moduleParts[1]]["views"], $moduleParts[2]);
            }
        }

        return $content;
    }

    public function getMetadataByType($type)
    {
        $repository = $this->entityManager->getRepository(ModelMetadata::class);
        return $repository->findOneBy(['type' => $type]);
    }

    protected function mergeNAppendSubNav($oldSubNavItems, $newSubNavItems)
    {
        $updatedSubNavItems = $oldSubNavItems;
        foreach ($newSubNavItems as $newItem) {
            $itemExists = false;
            foreach ($oldSubNavItems as $key => $oldItem) {
                if ($oldItem['title'] == $newItem['title']) {
                    $itemExists = true;
                    $updatedSubNavItems[$key] = $newItem;
                    break;
                }
            }
            if (!$itemExists) {
                array_push($updatedSubNavItems, $newItem);
            }
        }

        return $updatedSubNavItems;
    }

    protected function mergeNAppendNav($newNav, $oldNav)
    {
        $updatedNav = $oldNav;
        foreach ($newNav as $key => $newNavItem) {
            if ((array_key_exists("include", $newNavItem) && $newNavItem['include'] == true) || (array_key_exists("include", $newNavItem) && $newNavItem['include'] == "true")) {
                $itemExists = false;
                unset($newNavItem['include']);
                unset($newNavItem['exists']);
                foreach ($oldNav['config']['navigation'] as $key => $oldNavItem) {
                    $updatedNavItems = [];
                    if ($oldNavItem['title'] == $newNavItem['title']) {
                        $itemExists = true;
                        $oldNavItems = $oldNavItem['items'] ?? [];
                        $newNavItems = $newNavItem['items'] ?? [];

                        $updatedNav['config']['navigation'][$key] = $newNavItem; //merge values like icon from the new nav
                        // $updatedNav['config']['navigation'][$key]['items'] =  $oldNavItems; //keep the items from old nav, as we would be matching each sub item in seoerate for loop
                        if (empty($oldNavItems) && !empty($newNavItems)) {
                            $updatedNavItems = $newNavItems;
                        }

                        if (!empty($oldNavItems) && !empty($newNavItems)) {
                            $updatedNavItems = $this->mergeNAppendSubNav($oldNavItems, $newNavItems);
                        }
                        if (!empty($updatedNavItems)) {
                            $updatedNav['config']['navigation'][$key]['items'] = $updatedNavItems; //?? incase both old and new item is emplty
                        }
                    }
                }
                if (!$itemExists) {
                    array_push($updatedNav['config']['navigation'], $newNavItem);
                }
            }
        }
        return $updatedNav;
    }
}
