<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Entity\Core\Widget;
use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Service\UtilityService;
use App\Security\Voter\FsrVoter;
use Doctrine\ORM\EntityManagerInterface;
use GuzzleHttp\Exception\RequestException;
use App\Service\SolutionPackUtilityService;
use App\Service\ConfigExportImport\PortingConfig;
use Symfony\Component\Filesystem\Exception\IOException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Class WidgetConfig
 */
class WidgetConfig extends PortingConfig
{
    private $fsrVoter;
    private $solutionPackUtilityService;
    private $utilityService;

    public function __construct(
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        FsrVoter $fsrVoter,
        UtilityService $utilityService,
        SolutionPackUtilityService $solutionPackUtilityService,
        protected LoggerInterface $logger
    ) {
        $this->fsrVoter = $fsrVoter;
        $this->utilityService = $utilityService;
        $this->solutionPackUtilityService = $solutionPackUtilityService;
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }

    private function getIndividualOptionsConfig($widget, $individualOptions)
    {
        $individualOptions['name'] = $widget['name'];
        $individualOptions['version'] = $widget['version'];
        $individualOptions['title'] = $widget['title'];
        if(array_key_exists("install_mode",$widget)){
            $individualOptions['install_mode'] = $widget['install_mode'];
        }
        if(array_key_exists("installer_path",$widget)){
            $individualOptions['installer_path'] = $widget['installer_path'];
        }
        return $individualOptions;
    }

    private function includeInstallSP($widget, $solutionPackRecord)
    {
        if ($widget['minVersion'] && !is_null($widget['minVersion']) && $solutionPackRecord->getInstalled() && $solutionPackRecord->getVersion() >= $widget['minVersion']) {
            $includeInstall = false;
        } elseif((!$widget['minVersion'] || is_null($widget['minVersion'])) && $solutionPackRecord->getInstalled() && is_null($solutionPackRecord->getLatestAvailableVersion())) {
            // $solutionPackRecord->getLatestAvailableVersion() is null if the widget installed is latest. Hence avoiding installation of widget.
            $includeInstall = false;
        } else {
            $includeInstall = true;
        }
        return $includeInstall;
    }

    private function getIndividualOptionsSP($widget, $individualOptions)
    {
        $individualOptions['name'] = $widget['name'];
        $individualOptions['title'] = $widget['title'];
        $individualOptions['install_mode'] = $widget['install_mode'];

        if ($widget['install_mode'] == 'rpm') {
            if (!$widget['version'] || is_null($widget['version'])) {
                $solutionPackRecord = $this->solutionPackUtilityService->getLatestSolutionPack($widget['name']);
                if ($solutionPackRecord == null) {
                    throw new \Exception($widget['name'] . ' widget record not found.');
                }
                // Set Widget to Latest Version, but installation would depend on includeInstall logic based on minVersion give orr not
                $widgetVersion = is_null($solutionPackRecord->getLatestAvailableVersion()) ? $solutionPackRecord->getVersion() : $solutionPackRecord->getLatestAvailableVersion();
                $individualOptions['include'] = $this->includeInstallSP($widget, $solutionPackRecord);
            } elseif($widget['version'] && !is_null($widget['version'])) {
                $widgetVersion = $widget['version'];
            }
            $widgetInstallPath = sprintf('%s-%s', $widget['name'], $widgetVersion);
            $individualOptions['version'] = $widgetVersion;
            $individualOptions['installer_path'] = $widgetInstallPath;
            $individualOptions['rpm_name'] = $widgetInstallPath;
        }

        return $individualOptions;
    }

    /**
     *
     * @param array $data Playbook object containing collections & global vars
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        if ($tokenType && !($this->fsrVoter->isGranted('update.widgets') || $this->fsrVoter->isGranted('create.widgets') || $this->fsrVoter->isGranted('read.widgets'))) {
            return [];
        }
        $options = [];
        $repository = $this->loadRepository(Widget::class);
        $existingWidgets = $repository->findAll();
        $existingWidgetNames = array_map(function ($item) {
            return [$item->getName(), $item->getVersion()];
        }, $existingWidgets);
        $options = [];
        $importType = $this->importJob->getType();
        foreach ($data as $widget) {
            $individualOptions = [];
            if (array_key_exists('install_mode', $widget)) {
                $importOption['install_mode'] = $widget['install_mode'];
                $importOption['installer_path'] = $widget['installer_path'];
            } else {
                $importOption['install_mode'] = 'rpm';
            }

            $individualOptions = $importType == 'SolutionPack Import' ? $this->getIndividualOptionsSP($widget, $individualOptions) : $this->getIndividualOptionsConfig($widget, $individualOptions);
            $existing = in_array([$individualOptions['name'], $individualOptions['version']], $existingWidgetNames);
            $individualOptions['exists'] = $existing;
            if (!array_key_exists('include', $individualOptions) || is_null($individualOptions['include'])) {
                $individualOptions['include'] = !$existing;
            }

            $options[] = $individualOptions;
        }
        return $options;
    }


    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @return mixed
     */
    public function import(array $data, array $options, $tokenType, $importedBy = [])
    {
        foreach ($options as $option) {
            $this->upsertImportedBy($option['name'], $importedBy);
            if ($option['exists'] || !$option['include']) {
                // Skip existing widgets
                continue;
            }
            // Install & wait if not exists
            $params = [
                'name' => $option['name'],
                'version' => $option['version'],
                'replace' => true,
                'importedBy' => $importedBy
            ];
            $this->sendRequest(sprintf('%s/api/3/widgets/install', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY)), 'post', $params, $tokenType);
        }
        return true;
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = [])
    {
        $widgetsPath = $tmpDirectory . $key . '/';
        $widgetsInRepo = $this->loadRepoWidgets();
        $importType = $this->importJob->getType();
        foreach ($options as $option) {
            if ($option['exists'] || !$option['include']) {
                // Skip existing widgets
                continue;
            }
            // Install & wait if not exists
            if ($option['install_mode'] == 'rpm') {
                $this->upsertImportedBy($option['name'], $importedBy);
                if (!array_key_exists($option['installer_path'], $widgetsInRepo)) {
                    continue;
                }
                $params = [
                    'name' => $option['name'],
                    'version' => $option['version'],
                    'replace' => true,
                    'importedBy' => $importedBy
                ];
                $this->sendRequest(sprintf('%s/api/3/widgets/install', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY)), 'post', $params, $tokenType);
            } elseif($option['install_mode'] == 'tgz' && $importType != 'SolutionPack Import') {
                $uri = sprintf('%s/api/3/widgets/import', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY));
                $fileStream = fopen($widgetsPath . $option['installer_path'], 'r');
                $token = $this->tokenStorage->getToken();
                $authorization = $token ? sprintf("Bearer %s", $token->getCredentials()) : null;
                $client = new \GuzzleHttp\Client();
                $params = [
                    'verify' => false,
                    'multipart' => [
                        [
                            'name' => 'replace',
                            'contents' => true,
                            'headers' => ['Content-Type' => 'application/json']
                        ],
                        [
                            'name' => 'file',
                            'contents' => $fileStream,
                            'filename' => $option['installer_path'],
                            'headers' => ['Content-Type' => 'application/gzip']
                        ]
                    ],
                    'headers' => [
                        'Authorization' => $authorization
                    ]
                ];
                try {
                    $client->request('POST', $uri, $params);
                } catch (IOException | \Exception $e) {
                    $this->io->warning(sprintf('Skipping import for %s (from file %s) since the file is invalid or incompatible.', $option['name'], $option['installer_path']));
                    continue;
                }
            }
        }
        return true;
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        if (!$this->fsrVoter->isGranted('read.widgets')) {
            return [];
        }
        $data = [];
        if (!empty($options)) {
            $repository = $this->loadRepository(Widget::class);
            $widgets = $repository->findAll();
        }
        foreach ($options as $option) {
            foreach ($widgets as $widget) {
                if ($option == $widget->getName()) {
                    $widgetObject = [
                        'name' => $widget->getName(),
                        'version' => $widget->getVersion(),
                        'title' => $widget->getTitle()
                    ];

                    $data[] = $widgetObject;
                    break;
                }
            }
        }
        return $data;
    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     * @return void
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        if (!$this->fsrVoter->isGranted('read.widgets') || count($options) == 0) {
            return null;
        }
        $widgetDirectory = $tmpDirectory . $type;
        $widgetObjects = [];
        $metaData = [];
        $repository = $this->loadRepository(Widget::class);
        $widgets = $repository->findAll();
        $widgetsInRepo = $this->loadRepoWidgets();
        $exportJobType = $this->exportJob->getType();
        foreach ($options as $option) {
            foreach ($widgets as $widget) {
                if ($option == $widget->getName()) {
                    // Verify the existence of the widget in the repo
                    $rpmName = sprintf('%s-%s', $widget->getName(), $widget->getVersion());
                    $widgetObject = [
                        'name' => $widget->getName(),
                        'version' => $widget->getVersion(),
                        'title' => $widget->getTitle()
                    ];
                    if (array_key_exists($rpmName, $widgetsInRepo)) {
                        $widgetObject['install_mode'] = 'rpm';
                        $widgetObject['installer_path'] = $rpmName;
                    }
                    # tgz export only if non-SP export
                    elseif($exportJobType != 'SolutionPack Export') {
                        // Download tgz
                        $params = [
                            'development' => false
                        ];
                        $response = $this->sendRequest(sprintf('%s/api/3/widgets/export/%s', $this->container->getParameter(AppConstants::CRUDHUB_PARAMETER_KEY), $widget->getUuid()), 'post', $params, $tokenType);
                        $tgzFilename = $rpmName . '.tgz';
                        $this->writeBytesToFile(sprintf('%s/%s', $widgetDirectory, $tgzFilename), $response);
                        $widgetObject['install_mode'] = 'tgz';
                        $widgetObject['installer_path'] = $tgzFilename;
                    }

                    // Version Agnostic Widget for Solution Pack
                    if ($exportJobType == 'SolutionPack Export') {
                        unset($widgetObject['version']);
                        unset($widgetObject['installer_path']);
                    }

                    $widgetObjects[] = $widgetObject;
                    $metaData[] = ['apiName' => $widgetObject['name'], 'name' => $widgetObject['title']];
                    break;
                }
            }
        }
        $filename = $widgetDirectory . '/data.json';
        $this->writeToFile($filename, $widgetObjects);
        $this->appendToFile($tmpDirectory . 'info.json', $type, $metaData);
        return $filename;
    }

    private function loadRepoWidgets()
    {
        try {
            $widgetJsonUri = sprintf('https://%s/fsr-widgets/widgets.json', getenv("REPOSERVER") ? getenv("REPOSERVER") : getenv("product_yum_server"));
            $response = $this->sendRequest($widgetJsonUri, 'GET', null, null, ['connect_timeout' => 15]);
            return json_decode($response, true);
        } catch (RequestException | \Exception $e) {
            $this->io->warning(sprintf('Cannot load widget RPMs from %s: %s', getenv("REPOSERVER") ? getenv("REPOSERVER") : getenv("product_yum_server"), $e->getMessage()));
            return [];
        }
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        foreach ($options as $value) {
            array_push($exportOption, $value['name']);
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        $filter = ['name' => $options];
        $entityRepo = $this->entityManager->getRepository(Widget::class);
        $dataObj = $this->utilityService->getItemWithWhere($entityRepo, $filter);
        foreach ($dataObj as $widget) {
            $content[] = ["apiName" => $widget->getName(), 'version' => $widget->getVersion(), 'name' => $widget->getTitle()];
        }
        return $content;
    }

    //ToDo move to utility service.
    private function upsertImportedBy($widget, $importedBy)
    {
        /** @var SolutionPack */
        $solutionPack = $this->solutionPackUtilityService->getRecord($widget);
        if ($solutionPack  && $importedBy &&  $solutionPack->getName() != $importedBy[0]['apiName']) {
            $importArray = [];
            $appendImportBy = true;
            $existingImport = $solutionPack->getImportedBy();
            if ($existingImport) {
                foreach ($existingImport as $value) {
                    if ($value['name'] == $importedBy[0]['name']) {
                        $importArray[] = $importedBy[0];
                        $appendImportBy = false;
                    } else {
                        $importArray[] = $value;
                    }
                }
            } else {
                $importArray = $importedBy;
                $appendImportBy = false;
            }
            if ($appendImportBy) {
                $importArray[] = $importedBy[0];
            }
            $solutionPack->setImportedBy($importArray);
            $this->entityManager->persist($solutionPack);
            $this->entityManager->flush();
        }
    }
}
