<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use Doctrine\DBAL\Types\Type;
use Doctrine\Inflector\Inflector;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\Inflector\InflectorFactory;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\Common\Collections\Collection;
use Doctrine\ORM\Mapping\ClassMetadataInfo;
use App\Entity\ModelMetadata\StagingModelMetadata;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Doctrine\ORM\Tools\EntityGenerator as DoctrineEntityGenerator;

class EntityGenerator extends DoctrineEntityGenerator
{

    const RELATION_TYPE = ['manyToMany', 'oneToMany'];
    /** @var ModelMetadata */
    protected $modelMetadata;

    /** @var AttributeMetadata[] */
    protected $attributeMetadataMap;

    /** @var TraitGeneratorInterface[] */
    protected $traitGenerators;

    /** @var Inflector|null */
    protected $inflector;

    /** @var Entity Manager|null */
    protected $entityManager;

    private $nonTrackingModuleList = [];
    private $nonIndexableModuleList = [];

    /**
     * @var string
     */
    protected static $addMethodAndRelationshipTemplate =
    '/**
 * <description>
 *
 * @param <variableType>$<variableName>
 * @return <entity>
 */
public function <methodName>(<methodTypeHint>$<variableName>)
{
<spaces>if (!$this-><fieldName>->contains($<variableName>)) {
<spaces><spaces>$this-><fieldName>[] = $<variableName>;
<spaces><spaces>$<variableName>-><relationshipMethodName>($this);
<spaces>}
<spaces>return $this;
}';

    const COMMENT_REPLY_TEMPLATE = '
    /**
     * @Groups({"childCommentCount"})
     */
    public function getChildCommentCount()
    {
        return $this->replies->count();
    }
    ';

    /**
     * @var string
     */
    protected static $removeMethodAndRelationshipTemplate =
    '/**
 * <description>
 *
 * @param <variableType>$<variableName>
 */
public function <methodName>(<methodTypeHint>$<variableName>)
{
<spaces>if ($this-><fieldName>->contains($<variableName>)) {
<spaces><spaces>$this-><fieldName>->removeElement($<variableName>);
<spaces><spaces>$<variableName>-><relationshipMethodName>(<relationshipMethodArgument>);
<spaces>}
}';

    /**
     * @var string
     */
    protected static $addMethodTemplate =
    '/**
 * <description>
 *
 * @param <variableType> $<variableName>
 *
 * @return <entity>
 */
public function <methodName>(<methodTypeHint>$<variableName>)
{
<spaces>if (!$this-><fieldName>->contains($<variableName>)) {
<spaces><spaces>$this-><fieldName>[] = $<variableName>;
<spaces>}
<spaces>return $this;
}';

    /** @var string */
    protected static $setMethodTemplate =
    '/**
 * <description>
 *
 * @param <variableType> $<variableName>
 *
 * @return <entity>
 */
public function <methodName>(<methodTypeHint>$<variableName><variableDefault>)
{
<spaces>$this-><fieldName> =<stringify> $<variableName>;

<spaces>return $this;
}';

    protected $container;

    public function __construct(ContainerInterface $container, EntityManagerInterface $entityManager)
    {
        $this->container = $container;
        $this->entityManager = $entityManager;
        $allModule = $this->entityManager->getRepository(StagingModelMetadata::class)->findAll();
        $this->inflector = InflectorFactory::create()->build();
        parent::__construct();
        $this->typeAlias = array_merge([
            'jsonb' => 'mixed',
        ], $this->typeAlias);
        foreach ($allModule as $module) {
            if ($module->getTrackable() == false) {
                $this->nonTrackingModuleList[] = $module->getType();
            }
            if ($module->getIndexable() == false) {
                $this->nonIndexableModuleList[] = $module->getType();
            }
        }
    }

    /**
     * @param ModelMetadata $modelMetadata
     */
    public function setModelMetadata($modelMetadata)
    {
        $this->modelMetadata = $modelMetadata;
        $this->attributeMetadataMap = [];

        foreach ($modelMetadata->getAttributes() as $attribute) {
            $this->attributeMetadataMap[$attribute->getName()] = $attribute;
        }
    }

    /**
     * @param TraitGeneratorInterface[] $traitGenerators
     */
    public function setTraitGenerators($traitGenerators)
    {
        $this->traitGenerators = $traitGenerators;
    }

    /**
     * @param array $fieldMapping
     * @param ClassMetadataInfo $metadata
     * @return string
     */
    protected function generateFieldMappingPropertyDocBlock(array $fieldMapping, ClassMetadataInfo $metadata, $attribute = null)
    {
        $fieldMapping['columnName'] = sprintf('`%s`', strtolower($fieldMapping['fieldName']));

        $linesString = parent::generateFieldMappingPropertyDocBlock($fieldMapping, $metadata);

        return $this->modifyMappingPropertyDocBlock($fieldMapping, $metadata, $linesString, $attribute);
    }

    /**
     * @param array $fieldMapping
     * @param ClassMetadataInfo $metadata
     * @param $linesString
     * @return string
     */
    protected function modifyMappingPropertyDocBlock(array $fieldMapping, ClassMetadataInfo $metadata, $linesString, $attribute = null)
    {
        $lines = explode("\n", $linesString);

        $lastLine = array_pop($lines);
        $attributeName = $attribute->getName();

        $groupsText = $this->spaces . ' * ' . '@Groups({';

        if (in_array($attribute->getFormType(), self::RELATION_TYPE) && $attribute->getFormType() != "multiselectpicklist") {
            $groupsText = $groupsText . '"fsr_all", "' . $attributeName . '"';
        } elseif ($attribute->getFormType() == 'password') {
            $groupsText = $groupsText . '"password"';
        } else {
            $groupsText = $groupsText . '"fsr_primary", "fsr_all", "' . $attributeName . '"';
        }
        if ($attribute->getEncrypted()) {
            $groupsText = $groupsText . ', "fsr_encrypted"';
        }
        $lines[] = $groupsText . '})';
        if ($attribute->getWriteable() == false) {
            $lines[] = $this->spaces . ' * ' .  '@ApiProperty(writable=false)';
        }

        if (method_exists($attribute, 'getValidation') && $attribute->getValidation()) {
            if ($attribute->getValidation()['required'] === true) {
                $lines[] = $this->spaces . ' * ' . '@Symfony\Component\Validator\Constraints\NotBlank(groups={"Default", "fsr_primary", "fsr_all"})';
            }
            if (isset($attribute->getValidation()['maxlength']) && $attribute->getValidation()['maxlength'] > 0 && $attribute->getValidation()['maxlength'] < 10485761) {
                $maxValue = $attribute->getValidation()['maxlength'];
                $maxMessage = sprintf("%s cannot be longer than %s characters", $attribute->getName(), $maxValue);
                $lines[] = $this->spaces . ' * ' . '@Symfony\Component\Validator\Constraints\Length(groups={"Default", "fsr_primary", "fsr_all"}, max=' . $maxValue . ', maxMessage="' . $maxMessage . '")';
            }
        }

        if ($attribute->getEncrypted() === true) {
            $lines[] = $this->spaces . ' * ' . '@Encrypted';
        }

        $lines[] = $lastLine;

        return implode("\n", $lines);
    }

    protected function generateEntityConstructor(ClassMetadataInfo $metadata)
    {
        $collections = [];

        if ($this->hasMethod('__construct', $metadata)) {
            $collections[] = 'parent::__construct();';
        }

        foreach ($metadata->associationMappings as $mapping) {
            if ($mapping['type'] & ClassMetadataInfo::TO_MANY) {
                $collections[] = '$this->' . $mapping['fieldName'] . ' = new \Doctrine\Common\Collections\ArrayCollection();';
            }
        }

        if ($collections) {
            return $this->prefixCodeWithSpaces(str_replace("<collections>", implode("\n" . $this->spaces, $collections), self::$constructorMethodTemplate));
        }

        return '';
    }

    /**
     * @param ClassMetadataInfo $metadata
     *
     * @return string
     */
    protected function generateEntityFieldMappingProperties(ClassMetadataInfo $metadata, $stagingData = null)
    {
        $lines = [];
        $attributes = [];
        foreach ($stagingData->getAttributes() as $attribute) {
            $attributes[$attribute->getName()] = $attribute;
        }

        foreach ($metadata->fieldMappings as $fieldMapping) {
            if (
                isset($fieldMapping['declaredField'], $metadata->embeddedClasses[$fieldMapping['declaredField']]) ||
                $this->hasProperty($fieldMapping['fieldName'], $metadata) ||
                $metadata->isInheritedField($fieldMapping['fieldName'])
            ) {
                continue;
            }

            $defaultValue = '';
            if (isset($fieldMapping['options']['default'])) {
                if ($fieldMapping['type'] === 'boolean' && $fieldMapping['options']['default'] === '1') {
                    $defaultValue = ' = true';
                } else {
                    $defaultValue = ' = ' . var_export($fieldMapping['options']['default'], true);
                }
            }

            $lines[] = $this->generateFieldMappingPropertyDocBlock($fieldMapping, $metadata, $attributes[$fieldMapping['fieldName']]);
            $lines[] = $this->spaces . $this->fieldVisibility . ' $' . $fieldMapping['fieldName'] . $defaultValue . ";\n";
        }

        if ($stagingData->getType() == 'comments') {
            $lines[] = $this->spaces . self::COMMENT_REPLY_TEMPLATE;
        }

        return implode("\n", $lines);
    }

    /**
     * @param array             $associationMapping
     * @param ClassMetadataInfo $metadata
     *
     * @return string
     */
    protected function generateAssociationMappingPropertyDocBlock(array $associationMapping, ClassMetadataInfo $metadata, $attribute = null)
    {
        $lines = [];
        $lines[] = $this->spaces . '/**';

        if ($associationMapping['type'] & ClassMetadataInfo::TO_MANY) {
            $lines[] = $this->spaces . ' * @var \Doctrine\Common\Collections\Collection';
        } else {
            $lines[] = $this->spaces . ' * @var \\' . ltrim($associationMapping['targetEntity'], '\\');
        }

        if ($this->generateAnnotations) {
            $lines[] = $this->spaces . ' *';

            if (isset($associationMapping['id']) && $associationMapping['id']) {
                $lines[] = $this->spaces . ' * @' . $this->annotationsPrefix . 'Id';

                if ($generatorType = $this->getIdGeneratorTypeString($metadata->generatorType)) {
                    $lines[] = $this->spaces . ' * @' . $this->annotationsPrefix . 'GeneratedValue(strategy="' . $generatorType . '")';
                }
            }

            $type = null;
            switch ($associationMapping['type']) {
                case ClassMetadataInfo::ONE_TO_ONE:
                    $type = 'OneToOne';
                    break;
                case ClassMetadataInfo::MANY_TO_ONE:
                    $type = 'ManyToOne';
                    break;
                case ClassMetadataInfo::ONE_TO_MANY:
                    $type = 'OneToMany';
                    break;
                case ClassMetadataInfo::MANY_TO_MANY:
                    $type = 'ManyToMany';
                    break;
            }
            $typeOptions = [];

            if (isset($associationMapping['targetEntity'])) {
                $typeOptions[] = 'targetEntity="' . $associationMapping['targetEntity'] . '"';
            }

            if (isset($associationMapping['inversedBy'])) {
                $typeOptions[] = 'inversedBy="' . $associationMapping['inversedBy'] . '"';
            }

            if (isset($associationMapping['mappedBy'])) {
                $typeOptions[] = 'mappedBy="' . $associationMapping['mappedBy'] . '"';
            }

            if ($associationMapping['cascade']) {
                $cascades = [];

                if ($associationMapping['isCascadePersist']) {
                    $cascades[] = '"persist"';
                }

                if ($associationMapping['isCascadeRemove']) {
                    $cascades[] = '"remove"';
                }

                if ($associationMapping['isCascadeDetach']) {
                    $cascades[] = '"detach"';
                }

                if ($associationMapping['isCascadeMerge']) {
                    $cascades[] = '"merge"';
                }

                if ($associationMapping['isCascadeRefresh']) {
                    $cascades[] = '"refresh"';
                }

                if (count($cascades) === 5) {
                    $cascades = ['"fsr_all"'];
                }

                $typeOptions[] = 'cascade={' . implode(',', $cascades) . '}';
            }

            if (isset($associationMapping['orphanRemoval']) && $associationMapping['orphanRemoval']) {
                $typeOptions[] = 'orphanRemoval=true';
            }

            if (isset($associationMapping['fetch']) && $associationMapping['fetch'] !== ClassMetadataInfo::FETCH_LAZY) {
                $fetchMap = [
                    ClassMetadataInfo::FETCH_EXTRA_LAZY => 'EXTRA_LAZY',
                    ClassMetadataInfo::FETCH_EAGER => 'EAGER',
                ];

                $typeOptions[] = 'fetch="' . $fetchMap[$associationMapping['fetch']] . '"';
            }

            $lines[] = $this->spaces . ' * @' . $this->annotationsPrefix . '' . $type . '(' . implode(', ', $typeOptions) . ')';

            if (isset($associationMapping['joinColumns']) && $associationMapping['joinColumns']) {
                $lines[] = $this->spaces . ' * @' . $this->annotationsPrefix . 'JoinColumns({';

                $joinColumnsLines = [];

                foreach ($associationMapping['joinColumns'] as $joinColumn) {
                    if ($joinColumnAnnot = $this->generateJoinColumnAnnotation($joinColumn)) {
                        $joinColumnsLines[] = $this->spaces . ' *   ' . $joinColumnAnnot;
                    }
                }

                $lines[] = implode(",\n", $joinColumnsLines);
                $lines[] = $this->spaces . ' * })';
            }

            if (isset($associationMapping['joinTable']) && $associationMapping['joinTable']) {
                $joinTable = [];
                $joinTable[] = 'name="' . $associationMapping['joinTable']['name'] . '"';

                if (isset($associationMapping['joinTable']['schema'])) {
                    $joinTable[] = 'schema="' . $associationMapping['joinTable']['schema'] . '"';
                }

                $lines[] = $this->spaces . ' * @' . $this->annotationsPrefix . 'JoinTable(' . implode(', ', $joinTable) . ',';
                $lines[] = $this->spaces . ' *   joinColumns={';

                $joinColumnsLines = [];

                foreach ($associationMapping['joinTable']['joinColumns'] as $joinColumn) {
                    $joinColumnsLines[] = $this->spaces . ' *     ' . $this->generateJoinColumnAnnotation($joinColumn);
                }

                $lines[] = implode("," . PHP_EOL, $joinColumnsLines);
                $lines[] = $this->spaces . ' *   },';
                $lines[] = $this->spaces . ' *   inverseJoinColumns={';

                $inverseJoinColumnsLines = [];

                foreach ($associationMapping['joinTable']['inverseJoinColumns'] as $joinColumn) {
                    $inverseJoinColumnsLines[] = $this->spaces . ' *     ' . $this->generateJoinColumnAnnotation($joinColumn);
                }

                $lines[] = implode("," . PHP_EOL, $inverseJoinColumnsLines);
                $lines[] = $this->spaces . ' *   }';
                $lines[] = $this->spaces . ' * )';
            }

            if ($attribute->getCollection() == true && !isset($associationMapping['orderBy']) && $attribute->getFormType() != "multiselectpicklist") {
                if (!in_array($attribute->getDataSource()['model'], $this->nonTrackingModuleList)) {
                    $associationMapping['orderBy'] = array('modifyDate' => 'DESC');
                } elseif (!in_array($attribute->getDataSource()['model'], $this->nonIndexableModuleList)) {
                    $associationMapping['orderBy'] = array('id' => 'DESC');
                }
            }
            if (isset($associationMapping['orderBy'])) {
                $lines[] = $this->spaces . ' * @' . $this->annotationsPrefix . 'OrderBy({';

                foreach ($associationMapping['orderBy'] as $name => $direction) {
                    $lines[] = $this->spaces . ' *     "' . $name . '"="' . $direction . '",';
                }

                $lines[count($lines) - 1] = substr($lines[count($lines) - 1], 0, strlen($lines[count($lines) - 1]) - 1);
                $lines[] = $this->spaces . ' * })';
            }
        }

        if (in_array($attribute->getFormType(), self::RELATION_TYPE) && $attribute->getSkipSerialization()) {
            $lines[] = $this->spaces . ' * ' . '@Groups({"' . $attribute->getName() . '"})';
        } elseif (in_array($attribute->getFormType(), self::RELATION_TYPE)  && $attribute->getFormType() != "multiselectpicklist") {
            $lines[] = $this->spaces . ' * ' . '@Groups({"fsr_all", "' . $attribute->getName() . '"})';
        } else {
            $lines[] = $this->spaces . ' * ' . '@Groups({"fsr_primary", "fsr_all", "' . $attribute->getName() . '"})';
        }

        if (method_exists($attribute, 'getValidation') && $attribute->getValidation()) {
            if ($attribute->getValidation()['required'] === true) {
                $lines[] = $this->spaces . ' * ' . '@Symfony\Component\Validator\Constraints\NotBlank(groups={"Default", "fsr_primary", "fsr_all"})';
            }
        }

        $lines[] = $this->spaces . ' */';

        return implode("\n", $lines);
    }

    protected function generateEntityAssociationMappingProperties(ClassMetadataInfo $metadata, $stagingData = null)
    {
        $lines = [];
        $attributes = [];
        foreach ($stagingData->getAttributes() as $attribute) {
            $attributes[$attribute->getName()] = $attribute;
        }

        foreach ($metadata->associationMappings as $associationMapping) {
            if ($this->hasProperty($associationMapping['fieldName'], $metadata)) {
                continue;
            }

            $lines[] = $this->generateAssociationMappingPropertyDocBlock($associationMapping, $metadata, $attributes[$associationMapping['fieldName']]);
            $lines[] = $this->spaces . $this->fieldVisibility . ' $' . $associationMapping['fieldName']
                . ($associationMapping['type'] == 'manyToMany' ? ' = array()' : null) . ";\n";
        }

        return implode("\n", $lines);
    }

    /**
     * @param ClassMetadataInfo $metadata
     *
     * @return string
     */
    protected function generateEntityBody(ClassMetadataInfo $metadata, $stagingData = null)
    {
        $traitsBlock = $this->generateTraitBlock($stagingData);
        $fieldMappingProperties = $this->generateEntityFieldMappingProperties($metadata, $stagingData);
        $embeddedProperties = $this->generateEntityEmbeddedProperties($metadata);
        $associationMappingProperties = $this->generateEntityAssociationMappingProperties($metadata, $stagingData);
        $stubMethods = $this->generateEntityStubMethods ? $this->generateEntityStubMethods($metadata, $stagingData) : null;
        $lifecycleCallbackMethods = $this->generateEntityLifecycleCallbackMethods($metadata);

        $code = [];

        if ($fieldMappingProperties) {
            $code[] = $fieldMappingProperties;
        }

        if ($embeddedProperties) {
            $code[] = $embeddedProperties;
        }

        if ($associationMappingProperties) {
            $code[] = $associationMappingProperties;
        }

        $code[] = $this->generateEntityConstructor($metadata);

        if ($stubMethods) {
            $code[] = $stubMethods;
        }

        if ($lifecycleCallbackMethods) {
            $code[] = $lifecycleCallbackMethods;
        }

        $body = implode("\n", $code);

        return sprintf("%s\n\n%s", $traitsBlock, $body);
    }

    protected function generateTraitBlock($stagingData)
    {
        $traits = [];

        $trait_map = $this->container->getParameter('trait_map');

        if ($stagingData->getTrackable()) {
            $traits[] = $trait_map['trackable'];
        }
        if ($stagingData->getPeerReplicable()) {
            $traits[] = $trait_map['peerReplicable'];
        }
        if ($stagingData->getOwnable()) {
            $traits[] = $trait_map['ownable'];
        }
        if ($stagingData->getIndexable()) {
            $traits[] = $trait_map['indexable'];
        }
        if ($stagingData->getTaggable()) {
            $traits[] = $trait_map['taggable'];
        }
        if ($stagingData->getUserOwnable()) {
            $traits[] = $trait_map['userOwnable'];
        }
        if ($stagingData->getQueueable()) {
            $traits[] = $trait_map['queueable'];
        }
        if ($stagingData->getSoftDeleteable()) {
            $traits[] = $trait_map['softDeleteable'];
        }

        $traitsLines = array_map(function ($trait) {
            return sprintf('%suse \\%s;', $this->spaces, $trait);
        }, $traits);
        $traitsBlock = implode("\n", $traitsLines);

        return $traitsBlock;
    }

    /**
     * @param $type
     * @param $fieldName
     * @return string
     */
    private function getMethodName($type, $fieldName)
    {
        $methodName = $type . $this->inflector->classify($fieldName);
        if (in_array($type, array("add", "remove"))) {
            return $this->inflector->singularize($methodName);
        }
        return $methodName;
    }

    /**
     * @param array $fieldMapping
     *
     * @return string|null
     */
    private function nullableFieldExpression(array $fieldMapping)
    {
        if (isset($fieldMapping['nullable']) && $fieldMapping['nullable'] === true) {
            return 'null';
        }

        return null;
    }

    /**
     * @return string
     */
    protected function generateEntityStubMethods(ClassMetadataInfo $metadata, $stagingData = null)
    {
        $methods = [];
        $encryptedFields = [];

        foreach ($stagingData->getAttributes() as $attribute) {
            if ($attribute->getEncrypted()) {
                $encryptedFields[] = $attribute->getName();
            }
        }

        foreach ($metadata->fieldMappings as $fieldMapping) {
            if (isset($fieldMapping['declaredField'], $metadata->embeddedClasses[$fieldMapping['declaredField']])) {
                continue;
            }

            $encrypted = false;
            if (in_array($fieldMapping['fieldName'], $encryptedFields)) {
                $encrypted = true;
            }

            $nullableField = $this->nullableFieldExpression($fieldMapping);

            if (
                (!$metadata->isEmbeddedClass || !$this->embeddablesImmutable)
                && (!isset($fieldMapping['id']) || !$fieldMapping['id'] || $metadata->generatorType === ClassMetadataInfo::GENERATOR_TYPE_NONE)
                && $code = $this->generateEntityStubMethod($metadata, 'set', $fieldMapping['fieldName'], $fieldMapping['type'], $nullableField, $encrypted)
            ) {
                $methods[] = $code;
            }

            if ($code = $this->generateEntityStubMethod($metadata, 'get', $fieldMapping['fieldName'], $fieldMapping['type'], $nullableField)) {
                $methods[] = $code;
            }
        }

        foreach ($metadata->embeddedClasses as $fieldName => $embeddedClass) {
            if (isset($embeddedClass['declaredField'])) {
                continue;
            }

            if (!$metadata->isEmbeddedClass || !$this->embeddablesImmutable) {
                if ($code = $this->generateEntityStubMethod($metadata, 'set', $fieldName, $embeddedClass['class'])) {
                    $methods[] = $code;
                }
            }

            if ($code = $this->generateEntityStubMethod($metadata, 'get', $fieldName, $embeddedClass['class'])) {
                $methods[] = $code;
            }
        }

        foreach ($metadata->associationMappings as $associationMapping) {
            if ($associationMapping['type'] & ClassMetadataInfo::TO_ONE) {
                $nullable = $this->isAssociationIsNullable($associationMapping) ? 'null' : null;
                if ($code = $this->generateEntityStubMethod($metadata, 'set', $associationMapping['fieldName'], $associationMapping['targetEntity'], $nullable)) {
                    $methods[] = $code;
                }

                if ($code = $this->generateEntityStubMethod($metadata, 'get', $associationMapping['fieldName'], $associationMapping['targetEntity'], $nullable)) {
                    $methods[] = $code;
                }
            } elseif ($associationMapping['type'] & ClassMetadataInfo::TO_MANY) {
                if ($code = $this->generateEntityStubMethod($metadata, 'add', $associationMapping['fieldName'], $associationMapping['targetEntity'])) {
                    $methods[] = $code;
                }

                if ($code = $this->generateEntityStubMethod($metadata, 'remove', $associationMapping['fieldName'], $associationMapping['targetEntity'])) {
                    $methods[] = $code;
                }

                if ($code = $this->generateEntityStubMethod($metadata, 'get', $associationMapping['fieldName'], Collection::class)) {
                    $methods[] = $code;
                }
            }
        }

        return implode("\n\n", $methods);
    }

    protected function generateEntityStubMethod(ClassMetadataInfo $metadata, $type, $fieldName, $typeHint = null, $defaultValue = null, $encrypted = false)
    {
        $methodName = $this->getMethodName($type, $fieldName);

        $variableName = $this->inflector->camelize($fieldName);
        if (in_array($type, array("add", "remove"))) {
            $variableName = $this->inflector->singularize($variableName);
        }

        if ($this->hasMethod($methodName, $metadata)) {
            return '';
        }
        $this->staticReflection[$metadata->name]['methods'][] = $methodName;

        $var = sprintf('%sMethodTemplate', $type);

        $inverseMethodName = null;
        $relationshipMethodArgument = '$this';

        if ($metadata->isAssociationInverseSide($fieldName) || $this->isFieldSelfInversing($metadata, $fieldName)) {
            $associationMapping = $metadata->getAssociationMapping($fieldName);
            $inversedField = is_null($associationMapping['mappedBy']) ? $associationMapping['inversedBy'] : $associationMapping['mappedBy'];
            if ($associationMapping['type'] == ClassMetadata::MANY_TO_MANY && in_array($type, array('add', 'remove'))) {
                $inverseMethodName = $this->getMethodName($type, $inversedField);

                $var = sprintf('%sMethodAndRelationshipTemplate', $type);
            }
            if ($associationMapping['type'] == ClassMetadata::ONE_TO_MANY && in_array($type, array('add', 'remove'))) {
                $inverseMethodName = $this->getMethodName("set", $inversedField);

                $var = sprintf('%sMethodAndRelationshipTemplate', $type);
                if ($type == 'remove') {
                    $relationshipMethodArgument = 'null';
                }
            }
        }

        $template = self::$$var;

        $methodTypeHint = null;
        $types = Type::getTypesMap();
        $variableType = $typeHint ? $this->getType($typeHint) . ' ' : null;

        if ($typeHint && !isset($types[$typeHint])) {
            $variableType = '\\' . ltrim($variableType, '\\');
            $methodTypeHint = '\\' . $typeHint . ' ';
        }

        $stringify = null;
        if ($encrypted) {
            $stringify = ' (string)';
        }

        $replacements = array(
            '<description>' => ucfirst($type) . ' ' . $fieldName,
            '<methodTypeHint>' => $methodTypeHint,
            '<variableType>' => $variableType,
            '<variableName>' => $variableName,
            '<methodName>' => $methodName,
            '<fieldName>' => $fieldName,
            '<variableDefault>' => ($defaultValue !== null) ? (' = ' . $defaultValue) : '',
            '<entity>' => $this->getClassName($metadata),
            '<relationshipMethodArgument>' => $relationshipMethodArgument,
            '<stringify>' => $stringify,
        );

        if (!is_null($inverseMethodName)) {
            $replacements['<relationshipMethodName>'] = $inverseMethodName;
        }

        $method = str_replace(
            array_keys($replacements),
            array_values($replacements),
            $template
        );

        return $this->prefixCodeWithSpaces($method);
    }

    /**
     * Returns true if the given field name is self inversing
     *
     * @param ClassMetadataInfo $metadata
     * @param $fieldName
     *
     * @return bool
     */
    private function isFieldSelfInversing($metadata, $fieldName)
    {
        if (!$metadata->hasAssociation($fieldName)) {
            return false;
        }
        $associationMapping = $metadata->getAssociationMapping($fieldName);
        return $associationMapping['targetEntity'] === $associationMapping['sourceEntity'];
    }

    private function generateEntityListenerAnnotation(ClassMetadataInfo $metadata): string
    {
        if (0 === \count($metadata->entityListeners)) {
            return '';
        }

        $processedClasses = [];
        foreach ($metadata->entityListeners as $event => $eventListeners) {
            foreach ($eventListeners as $eventListener) {
                $processedClasses[] = '"' . $eventListener['class'] . '"';
            }
        }

        return \sprintf(
            '%s%s({%s})',
            '@' . $this->annotationsPrefix,
            'EntityListeners',
            \implode(',', \array_unique($processedClasses))
        );
    }

    /**
     * @param string $constraintName
     * @psalm-param array<string, array<string, mixed>> $constraints
     *
     * @return string
     */
    protected function generateTableConstraints($constraintName, array $constraints)
    {
        $annotations = [];
        $operator = null;
        $condition = null;
        $flags = null;
        foreach ($constraints as $name => $constraint) {
            $columns = [];
            $flags = null;
            foreach ($constraint['columns'] as $column) {
                $columns[] = '"' . $column . '"';
            }
            if (array_key_exists('options', $constraint) && array_key_exists('operator', $constraint['options']) && array_key_exists('condition', $constraint['options'])) {
                $operator = $constraint['options']['operator'];
                $condition = $constraint['options']['condition'];
            }

            if (array_key_exists('flags', $constraint)) {
                $flags = $constraint['flags'];
            }

            if ($operator && $condition) {
                $annotations[] = '@' . $this->annotationsPrefix . $constraintName . '(name="' . $name . '", columns={' . implode(', ', $columns) . '}, options={"' . $operator . '": "(' . $condition . ')"})';
            } else {
                $annotations[] = '@' . $this->annotationsPrefix . $constraintName . '(name="' . $name . '", columns={' . implode(', ', $columns) . '})';
            }

            if ($flags) {
                $annotations[] = '@' . $this->annotationsPrefix . $constraintName . '(name="' . $name . '", columns={' . implode(', ', $columns) . '}, flags={"' . implode(',', $flags) . '"})';
            }
        }
        return implode(', ', $annotations);
    }

    /**
     * @param ClassMetadataInfo $metadata
     *
     * @return string
     */
    protected function generateEntityDocBlock(ClassMetadataInfo $metadata, $stagingData = null)
    {
        $lines = [];
        $lines[] = '/**';
        $lines[] = ' * ' . $this->getClassName($metadata);
        $lines[] = ' *  ' . '@ApiResource(';
        $lines[] = ' *     ' . 'attributes={"validation_groups"={"fsr_primary", "fsr_all"}, "fsr_type"="' . $stagingData->getType() . '", "fsr_module"="' . $stagingData->getType() . '", "fsr_singular"="' . $stagingData->getDescriptions()['singular'] . '"},';
        $lines[] = ' *     ' . 'denormalizationContext={"groups"={"fsr_all", "password"}, "fetch_data"=false},';
        $lines[] = ' *     ' . 'normalizationContext={"circular_reference_limit"=2},';
        $lines[] = ' *     ' . 'collectionOperations={';
        $lines[] = ' *          ' . '"get" = { "path"="' . $stagingData->getType() . '", "security_post_denormalize" = "is_granted(\'read.' . $stagingData->getType() . '\', object)" },';
        $lines[] = ' *          ' . '"post" = { "path"="' . $stagingData->getType() . '", "security_post_denormalize" = "is_granted(\'create.' . $stagingData->getType() . '\', object)" }';
        $lines[] = ' *     ' . '},';
        $lines[] = ' *     ' . 'itemOperations={';
        $lines[] = ' *          ' . '"get" = { "path"="' . $stagingData->getType() . '/{uuid}", "security" = "is_granted(\'read.' . $stagingData->getType() . '\', object)" },';
        $lines[] = ' *          ' . '"put" = { "path"="' . $stagingData->getType() . '/{uuid}", "security" = "is_granted(\'update.' . $stagingData->getType() . '\', object)" },';
        $lines[] = ' *          ' . '"delete" = { "path"="' . $stagingData->getType() . '/{uuid}", "security" = "is_granted(\'delete.' . $stagingData->getType() . '\', object)" }';
        $lines[] = ' *     ' . '},';
        $lines[] = ' *     ' . 'subresourceOperations={';
        if ($stagingData->getUserOwnable()) {
            $lines[] = ' *          ' . '"user_owners_get_subresource" = {"path"="' . $stagingData->getType() . '/{uuid}/userOwners"},';
        }
        if ($stagingData->getOwnable()) {
            $lines[] = ' *          ' . '"owners_get_subresource" = {"path"="' . $stagingData->getType() . '/{uuid}/owners"},';
        }
        if ($stagingData->getParentType() == 'filterable_entity') {
            $lines[] = ' *          ' . '"datasets_get_subresource" = {"path"="' . $stagingData->getType() . '/{uuid}/datasets"},';
        }
        $lines[] = ' *     ' . '}';
        $lines[] = ' * ' . ')';

        if ($stagingData->getSoftDeleteable()) {
            $lines[] = ' * ' . '@Gedmo\SoftDeleteable(fieldName="deletedAt")';
        }

        if ($this->generateAnnotations) {
            $lines[] = ' *';

            $methods = [
                'generateTableAnnotation',
                'generateInheritanceAnnotation',
                'generateDiscriminatorColumnAnnotation',
                'generateDiscriminatorMapAnnotation',
                'generateEntityAnnotation',
                'generateEntityListenerAnnotation',
            ];

            foreach ($methods as $method) {
                if ($code = $this->$method($metadata)) {
                    $lines[] = ' * ' . $code;
                }
            }

            if (isset($metadata->lifecycleCallbacks) && $metadata->lifecycleCallbacks) {
                $lines[] = ' * @' . $this->annotationsPrefix . 'HasLifecycleCallbacks';
            }
        }

        $lines[] = ' */';

        return implode("\n", $lines);
    }

    /**
     * Generates a PHP5 Doctrine 2 entity class from the given ClassMetadataInfo instance.
     *
     * @param ClassMetadataInfo $metadata
     *
     * @return string
     */
    public function generateEntityClass(ClassMetadataInfo $metadata, $stagingData = null)
    {
        $placeHolders = [
            '<namespace>',
            '<useStatement>',
            '<entityAnnotation>',
            '<entityClassName>',
            '<entityBody>',
        ];

        $replacements = [
            $this->generateEntityNamespace($metadata),
            $this->generateEntityUse(),
            $this->generateEntityDocBlock($metadata, $stagingData),
            $this->generateEntityClassName($metadata),
            $this->generateEntityBody($metadata, $stagingData),
        ];

        $code = str_replace($placeHolders, $replacements, static::$classTemplate);

        return str_replace('<spaces>', $this->spaces, $code);
    }

    /**
     * @return string
     */
    protected function generateEntityUse()
    {
        if (!$this->generateAnnotations) {
            return '';
        }

        return "\n" .
            'use ApiPlatform\Core\Annotation\ApiResource;' . "\n" .
            'use Symfony\Component\Serializer\Annotation\Groups;' . "\n" .
            'use ApiPlatform\Core\Annotation\ApiProperty;' . "\n" .
            'use Ambta\DoctrineEncryptBundle\Configuration\Encrypted;' . "\n" .
            'use Gedmo\Mapping\Annotation as Gedmo;' . "\n";
    }

    /**
     * Generates and writes entity classes for the given array of ClassMetadataInfo instances.
     *
     * @param array  $metadatas
     * @param string $outputDirectory
     *
     * @return void
     */
    public function generate(array $metadatas, $outputDirectory)
    {
        foreach ($metadatas as $metadata) {
            $this->writeEntityClass($metadata, $outputDirectory);
        }
    }

    /**
     * Generates and writes entity class to disk for the given ClassMetadataInfo instance.
     *
     * @param ClassMetadataInfo $metadata
     * @param string            $outputDirectory
     *
     * @return void
     *
     * @throws \RuntimeException
     */
    public function writeEntityClass(ClassMetadataInfo $metadata, $outputDirectory, $stagingData = null)
    {
        $entityClassPath = str_replace('App\\', '', $metadata->name);
        $path = $outputDirectory . '/' . str_replace('\\', DIRECTORY_SEPARATOR, $entityClassPath) . $this->extension;
        $dir = dirname($path);

        if (!is_dir($dir)) {
            mkdir($dir, 0775, true);
        }

        $this->isNew = !file_exists($path) || $this->regenerateEntityIfExists;

        if (!$this->isNew) {
            $this->parseTokensInEntityFile(file_get_contents($path));
        } else {
            $this->staticReflection[$metadata->name] = ['properties' => [], 'methods' => []];
        }

        if ($this->backupExisting && file_exists($path)) {
            $backupPath = dirname($path) . DIRECTORY_SEPARATOR . basename($path) . "~";
            if (!copy($path, $backupPath)) {
                throw new \RuntimeException("Attempt to backup overwritten entity file but copy operation failed.");
            }
        }

        // If entity doesn't exist or we're re-generating the entities entirely
        if ($this->isNew) {
            file_put_contents($path, $this->generateEntityClass($metadata, $stagingData));
            // If entity exists and we're allowed to update the entity class
        } elseif ($this->updateEntityIfExists) {
            file_put_contents($path, $this->generateUpdatedEntityClass($metadata, $path));
        }
        chmod($path, 0664);
    }
}
