<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use ApiPlatform\Core\Api\IriConverterInterface;
use App\Entity\Authorization\Role;
use App\Entity\ViewTemplate\GenericTemplate;
use App\Entity\ViewTemplate\UserDefaultTemplate;
use Doctrine\ORM\EntityManager;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Symfony\Component\Serializer\SerializerInterface;

class GenericTemplateService
{
    /** @var TokenStorageInterface */
    protected $tokenStorage;

    /** @var IriConverterInterface */
    protected $iriConverter;

    /** @var EntityManager */
    protected $entityManager;

    /** @var SerializerInterface */
    protected $serializer;

    protected $containers;

    public function __construct(TokenStorageInterface $tokenStorage, IriConverterInterface $iriConverter, ContainerInterface $containers)
    {
        $this->tokenStorage = $tokenStorage;
        $this->iriConverter = $iriConverter;
        $this->containers = $containers;
        $this->entityManager = $this->containers->get('doctrine.orm.entity_manager');
    }

    /**
     * @return string
     */
    public function getCurrentUserId()
    {
        return $this->tokenStorage->getToken()->getUser()->getUserId();
    }
    
    public function getCurrentUserAsIri()
    {
        $user = $this->getCurrentUser();
        return $this->iriConverter->getIriFromItem($user);
    }

    public function getCurrentUser()
    {
        return $this->tokenStorage->getToken()->getUser();
    }

    /**
     * @return array string
     */
    public function getCurrentUserRoles()
    {
        /** @var  $user Actor */
        $user = $this->getCurrentUser();
        $roles = $user->getRoles();
        $roleIds = [];
        if (!is_null($roles)) {
            foreach ($roles as $role) {
                array_push($roleIds, $role->getUuid());
            }
        }
        return $roleIds;
    }

    /**
     * @param String $type
     * @return UserDefaultTemplate
     */
    public function getDefaultUserTemplate($type)
    {
        /** @var GenericTemplateRepository $genericTemplateRepository */
        $genericTemplateRepository = $this->entityManager->getRepository(UserDefaultTemplate::class);
        $userId = $this->getCurrentUser()->getUuid();
        return $genericTemplateRepository->findOneBy(["type" => $type, "userId" => $userId]);
    }

    /**
     * @param $type string
     * @return GenericTemplate
     */
    public function getAllGenericTemplates($type, $__selectFields = "")
    {
        /** @var GenericTemplateRepository $genericTemplateRepository */
        $geneTemplate = [];
        $genericTemplateRepository = $this->entityManager->getRepository(GenericTemplate::class);
        $roleIds = $this->getCurrentUserRoles();
        $genericTemplate = $genericTemplateRepository->findBy(["templateType" => $type]);
        foreach ($genericTemplate as $template) {
            if ((sizeof($template->getRoles()) == 0 &&
                $template->getUuid() == '82bc8209-ef8d-463d-bcd8-db3ac08882cb-463d') ||
                $template->getCreateUser() == $this->getCurrentUserAsIri() ||
                sizeof(array_intersect($template->getRoles(), $roleIds)) > 0
            ) {
                array_push($geneTemplate, $template);
            }
        }
        return $geneTemplate;
    }

    /**
     * @inheritdoc
     * @return null
     */
    public function deleteTemplate($template)
    {
        $this->entityManager->remove($template);
        $this->entityManager->flush();
    }

    /**
     * @inheritdoc
     * @return null
     */
    public function saveTemplate($template)
    {
        $this->entityManager->persist($template);
        $this->entityManager->flush();
    }

    /**
     *
     * @param $type string
     * @param array $templateBody
     * @param String $id
     * @return GenericTemplate $genericTemplateObj
     *
     */
    public function saveGenericTemplate($type, $templateBody, $id)
    {
        $genericTemplateObj = new GenericTemplate();
        $genericTemplateObj->setUuid($id);
        $genericTemplateObj->setTemplateType($type);
        $genericTemplateObj->setType('rows');
        if (array_key_exists('config', $templateBody)) {
            $genericTemplateObj->setConfig($templateBody['config']);
        }
        if (array_key_exists('parentTemplateId', $templateBody)) {
            $genericTemplateObj->setParentTemplateId($templateBody['parentTemplateId']);
        }

        if (array_key_exists('displayName', $templateBody)) {
            $genericTemplateObj->setDisplayName($templateBody['displayName']);
        }
        $this->entityManager->persist($genericTemplateObj);
        $this->entityManager->flush();
        return $genericTemplateObj;
    }

    /**
     *
     * @param $type string
     * @param GenericTemplate $genericTemplate
     * @param array $templateBody
     * @param String $id
     * @return GenericTemplate $genericTemplate
     *
     */
    public function updateGenericTemplate($type, $genericTemplate, $templateBody, $id)
    {
        $userId = $templateBody['userId'] ? $templateBody['userId'] : $this->getCurrentUserAsIri();
        $timestamp = $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $genericTemplate->setUpdateDate($timestamp);
        $genericTemplate->setUuid($id);
        $genericTemplate->setTemplateType($type);
        $genericTemplate->setType('rows');
        if (array_key_exists('config', $templateBody)) {
            $genericTemplate->setConfig($templateBody['config']);
        }
        if (array_key_exists('parentTemplateId', $templateBody)) {
            $genericTemplate->setParentTemplateId($templateBody['parentTemplateId']);
        }
        if (array_key_exists('roles', $templateBody)) {
            $genericTemplate->addRole($templateBody['roles']);
        }
        if (array_key_exists('displayName', $templateBody)) {
            $genericTemplate->setDisplayName($templateBody['displayName']);
        }
        $this->entityManager->persist($genericTemplate);
        $this->entityManager->flush();
        return $genericTemplate;
    }

    public function getAllRoles()
    {

        $roles = $this->entityManager->getRepository(Role::class)->findAll();
        $roleIds = [];
        /**@var Role $role*/
        foreach ($roles as $role) {
            array_push($roleIds, '/api/3/roles/' . $role->getUuid());
        }
        return $roleIds;
    }

    /**
     * @param String $type
     * @param String $id
     * @return GenericTemplate $genericTemplateObj
     *
     */
    public function saveDefaultUserTemplate($type, $id, $genericTemplate)
    {
        $defaultUserTemplate = $this->getDefaultUserTemplate($type);
        if ($defaultUserTemplate) {
            $defaultUserTemplate->setTemplateId($genericTemplate);
        } else {
            $defaultTemplateObj = new UserDefaultTemplate();
            $defaultTemplateObj->setUserId($this->getCurrentUser());
            $defaultTemplateObj->setTemplateId($genericTemplate);
            $defaultTemplateObj->setType($type);
            $defaultUserTemplate = $defaultTemplateObj;
        }
        $this->entityManager->persist($defaultUserTemplate);
        $this->entityManager->flush();
        return $defaultUserTemplate;
    }

    /**
     * @param $id string
     * @param $type string
     * @return GenericTemplate
     */
    public function getGenericTemplate($type, $id)
    {
        /** @var GenericTemplateRepository $genericTemplateRepository */
        $genericTemplateRepository = $this->entityManager->getRepository(GenericTemplate::class);
        return $genericTemplateRepository->findBy(["uuid" => $id, "type" => $type]);
    }
}
