<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use Ramsey\Uuid\Uuid;
use App\Traits\Ownable;
use App\Traits\Taggable;
use App\Traits\Indexable;
use App\Traits\Trackable;
use Psr\Log\LoggerInterface;
use Doctrine\Inflector\InflectorFactory;
use Doctrine\ORM\EntityManagerInterface;
use App\Entity\ModelMetadata\ModelMetadata;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Doctrine\DBAL\Exception\ForeignKeyConstraintViolationException;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;


class InsertExpressionBuilder
{
    public const RELATIONSHIP_SQL_GENERIC_QUERY = "INSERT INTO %s (%s) VALUES %s ON CONFLICT DO nothing";

    private $inflector;

    public function __construct(
        private MetadataHelper $metadataHelper,
        private TeamsProvider $teamProvider,
        private EntityManagerInterface $entityManager,
        private TokenStorageInterface $tokenStorage,
        private ContainerInterface $containers,
        private TokenHelper $tokenHelper,
        protected LoggerInterface $logger
    ) {
        $this->inflector = InflectorFactory::create()->build();
    }

    private function isTrackable($resourceClass)
    {
        return in_array(Trackable::class, class_uses($resourceClass));
    }

    private function isIndexable($resourceClass)
    {
        return in_array(Indexable::class, class_uses($resourceClass));
    }

    public function isOwnable($resourceClass)
    {
        return in_array(Ownable::class, class_uses($resourceClass));
    }

    public function isTaggable($resourceClass)
    {
        return in_array(Taggable::class, class_uses($resourceClass));
    }

    private function getSingularName($moduleName)
    {
        // get singular name
        return strtolower($this->inflector->classify($this->inflector->singularize($moduleName)));
    }

    private function classToTableName($className)
    {
        if (strpos($className, '\\') !== false) {
            return substr($className, strrpos($className, '\\') + 1);
        }
        return $className;
    }

    private function getJoinKeyColumnName($module)
    {
        //get class name from module name
        $queryRepository = $this->entityManager->getRepository(ModelMetadata::class);
        $modelMetaData = $queryRepository->findOneBy(array('type' => $module));
        $singular = $modelMetaData->getDescriptions()['singular'];
        $escapedSingular = preg_replace('/[^a-z0-9\s]+/i', '', $singular);
        $escapedSingular = $this->inflector->classify($escapedSingular);
        return strtolower($this->classToTableName($escapedSingular));
    }

    private function isTokenizable($resourceClass)
    {
        return is_subclass_of($resourceClass, '\App\Entity\ThreatIntel\ThreatIntelBase');
    }

    public function handleOwnablity($module, $existingUuids)
    {
        if (empty($existingUuids)) {
            return;
        }
        $sqlValue = null;
        $ownableQueryParams = [];
        $moduleSingularizeName = $this->getSingularName($module);

        $tableName = $moduleSingularizeName . '_team';
        $relationshipFieldname = $moduleSingularizeName . '_uuid';
        $fieldsNames = [$moduleSingularizeName . '_uuid', 'team_uuid'];
        $teams = $this->teamProvider->getDataForCurrentUser();
        if (count($teams) === 0) {
            throw new AccessDeniedException("A user must be on at least one team in order to create records.");
        }
        foreach ($existingUuids as $outerIndex => $uuid) {
            foreach ($teams as $innerIndex => $team) {
                if ($outerIndex != 0 || $innerIndex != 0) {
                    $sqlValue = $sqlValue . ',';
                }
                $sqlValue = $sqlValue . "(:" . $relationshipFieldname . $outerIndex . ",:team_uuid" . $innerIndex . ")";
                $ownableQueryParams[$relationshipFieldname . $outerIndex] = $uuid;
                $ownableQueryParams["team_uuid" . $innerIndex] = $team->getUuid();
            }
        }
        $finalQuery = sprintf(
            self::RELATIONSHIP_SQL_GENERIC_QUERY,
            $tableName,
            implode(',', $fieldsNames),
            $sqlValue,
        );
        try {
            $this->entityManager->getConnection()->executeQuery($finalQuery, $ownableQueryParams);
        } catch (ForeignKeyConstraintViolationException $e) {
            $this->logger->error("Teams Insertion Failed");
        }
    }

    public function handleMultiSelectInsertion($multiselectQueryObject, $existingUuids)
    {
        foreach ($multiselectQueryObject as $tableName => $queryObject) {
            $mergedList = [];
            $fieldNames = $queryObject['fieldName'];
            $multiSelectList = $queryObject['values'];
            $multiSelectValues = array_intersect_key($multiSelectList, array_flip($existingUuids));
            $multiSelectValues = array_values($multiSelectValues);
            foreach ($multiSelectValues as $multiSelectValue) {
                foreach ($multiSelectValue as $multiSelectPicklist) {
                    $mergedList[] = $multiSelectPicklist;
                }
            }
            if (empty($mergedList)) {
                return;
            }
            $sqlValue = [];
            $sqlParams = [];
            foreach ($mergedList as $index => $tagSet) {
                $sqlValue[] = '(:module' . $index . ',:tag' . $index . ')';
                $sqlParams['module' . $index] = $tagSet[0];
                $sqlParams['tag' . $index] = $tagSet[1];
            }

            $finalQuery = sprintf(self::RELATIONSHIP_SQL_GENERIC_QUERY, $tableName, implode(',', $fieldNames), implode(',', $sqlValue));
            $this->entityManager->getConnection()->executeQuery($finalQuery, $sqlParams);
        }
    }

    public function handleFeedSelfLinking($feedSelfLinkingObject, $existingUuids)
    {
        foreach ($feedSelfLinkingObject as $tableName => $queryObject) {
            $fieldNames = $queryObject['fieldName'];
            $feedList = $queryObject['values'];
            $sqlValue = [];
            $sqlParams = [];
            foreach ($feedList as $index => $feedSet) {
                if (!in_array($feedSet[0], $existingUuids)) {
                    continue;
                }
                $sqlValue[] = '(:module' . $index . ',:relatedThreatFeed' . $index . ')';
                $sqlParams['module' . $index] = $feedSet[0];
                $sqlParams['relatedThreatFeed' . $index] = $feedSet[1];
            }
            $finalQuery = sprintf(self::RELATIONSHIP_SQL_GENERIC_QUERY, $tableName, implode(',', $fieldNames), implode(',', $sqlValue));
            $this->entityManager->getConnection()->executeQuery($finalQuery, $sqlParams);

            $fieldNames = array_reverse($fieldNames);
            $finalQuery = sprintf(self::RELATIONSHIP_SQL_GENERIC_QUERY, $tableName, implode(',', $fieldNames), implode(',', $sqlValue));
            $this->entityManager->getConnection()->executeQuery($finalQuery, $sqlParams);
        }
    }

    public function iterateMmd($mmdObject, $resourceClass)
    {
        $sqlInsertParams = ['uuid'];
        $dateTimeField = [];
        $picklistFields = [];
        $lookupFields = [];
        $booleanFields = [];
        $multiselectPickListField = [];
        if ($this->isTrackable($resourceClass)) {
            $sqlInsertParams = array_merge($sqlInsertParams, ['createuser', 'createdate', 'modifyuser', 'modifydate']);
        }
        if ($this->isTokenizable($resourceClass)) {
            $sqlInsertParams = array_merge($sqlInsertParams, ['value_tokens']);
        }
        foreach ($mmdObject->getAttributes() as $attribute) {
            if (in_array($attribute->getFormType(), ['manyToMany', 'oneToMany', null])) {
                continue;
            }
            if ($attribute->getFormType() == "multiselectpicklist") {
                $multiselectPickListField[] = $attribute->getName();
                continue;
            }
            if ($attribute->getFormType() == "checkbox") {
                $booleanFields[] = $attribute->getName();
            }
            if (in_array($attribute->getFormType(), ['lookup', 'picklist'])) {
                $sqlInsertParams[] = strtolower($this->inflector->tableize($attribute->getName()));
            } else {
                $sqlInsertParams[] = strtolower($attribute->getName());
            }

            if ($attribute->getFormType() == 'picklist') {
                $picklistFields[] = $attribute->getName();
            } elseif ($attribute->getFormType() == 'datetime') {
                $dateTimeField[] = $attribute->getName();
            } elseif ($attribute->getFormType() == 'lookup') {
                $lookupFields[] = $attribute->getName();
            }
        }
        return [$sqlInsertParams, $picklistFields, $dateTimeField, $multiselectPickListField, $lookupFields, $booleanFields];
    }

    public function getSqlInsertValues(
        $module,
        $classMeta,
        $objectData,
        $sqlInsertParams,
        $picklistFields,
        $dateTimeField,
        $multiselectPickListField,
        $lookupFields,
        $booleanFields
    ) {
        $fieldNamesMap = $classMeta->fieldNames;
        $associationMappings = $classMeta->associationMappings;
        foreach ($associationMappings as $associationField => $associationMapping) {
            if (array_key_exists('joinColumns', $associationMapping)) {
                $fieldNamesMap[$associationMapping['joinColumns'][0]['name']] = $associationField;
            }
        }
        $sqlValue = null;
        $queryParams = [];
        $multiselectQueries = [];
        $selfLinkingQueries = [];
        $tagQueries = [];
        $tags = [];
        $uuid_list = [];
        $objectCount = count($objectData['data']);
        foreach ($objectData['data'] as $index => $item) {
            $innerSqlValue = [];
            foreach ($sqlInsertParams as $field) {
                if ($field ==  'id') {
                    continue;
                }
                $innerSqlValue[] = ":" . $field . $index;

                switch ($field) {
                    case $field == 'uuid':
                        $queryParams['uuid' . $index] = Uuid::uuid4()->toString();
                        $uuid_list[] = $queryParams['uuid' . $index];
                        break;
                    case in_array($field, ['createuser', 'modifyuser']):
                        $loggedInUserUuid = $this->tokenStorage->getToken()->getUser()->getUuid();
                        $queryParams[$field . $index] = $loggedInUserUuid ?? null;
                        break;
                    case in_array($field, ['createdate', 'modifydate']):
                        $now = new \DateTime('NOW', new \DateTimeZone('UTC'));
                        $queryParams[$field . $index] = $now->format('Y-m-d H:i:s.u') ?? null;
                        break;
                    case in_array($fieldNamesMap[$field], $dateTimeField) && array_key_exists($fieldNamesMap[$field], $item):
                        $fieldVal = date('m/d/Y H:i:s', $item[$fieldNamesMap[$field]]) ?? null;
                        $queryParams[$field . $index] = $fieldVal ?? null;
                        break;
                    case in_array($fieldNamesMap[$field], $booleanFields) && array_key_exists($fieldNamesMap[$field], $item):
                        $queryParams[$field . $index] = $item[$fieldNamesMap[$field]] ? 1: 0;
                        break;
                    case in_array($fieldNamesMap[$field], $picklistFields) && array_key_exists($fieldNamesMap[$field], $item):
                        if (is_array($item[$fieldNamesMap[$field]]) && array_key_exists("picklistValue", $item[$fieldNamesMap[$field]])) {
                            $fieldVal = $this->metadataHelper->handlePicklistNameToIriConversion('/api/3/picklists', $item[$fieldNamesMap[$field]]);
                            $queryParams[$field . $index] = $fieldVal ? $fieldVal->getUuid() : null;
                            break;
                        }
                        if (is_array($item[$fieldNamesMap[$field]]) && array_key_exists("uuid", $item[$fieldNamesMap[$field]])) {
                            $queryParams[$field . $index] = $item[$fieldNamesMap[$field]]["uuid"];
                            break;
                        }
                        if (strpos($item[$fieldNamesMap[$field]], 'api/3/picklists')) {
                            $explodedValue = explode('/', $item[$fieldNamesMap[$field]]);
                            $item[$fieldNamesMap[$field]] = end($explodedValue);
                        }
                        $fieldVal = $item[$fieldNamesMap[$field]];
                        $queryParams[$field . $index] = $fieldVal ?? null;
                        break;
                    case in_array($fieldNamesMap[$field], $lookupFields) && $item[$fieldNamesMap[$field]]:
                        if (strpos($item[$fieldNamesMap[$field]], 'api/3/')) {
                            $explodedValue = explode('/', $item[$fieldNamesMap[$field]]);
                            $item[$fieldNamesMap[$field]] = end($explodedValue);
                        }
                        $fieldVal = $item[$fieldNamesMap[$field]];
                        $queryParams[$field . $index] = $fieldVal ?? null;
                        break;
                    case $field == 'value_tokens':
                        $value = $item['value'] ?? null;
                        $fieldVal = $this->tokenHelper->extractThreatFeedTokens($value);
                        $queryParams[$field . $index] = $fieldVal;
                        break;
                    default:
                        $queryParams[$field . $index] = $item[$fieldNamesMap[$field]] ?? null;
                }
            }
            $sqlValue = $sqlValue . "(";
            $sqlValue = $sqlValue . implode(",", $innerSqlValue) . ")";
            if ($index < $objectCount - 1) {
                $sqlValue = $sqlValue . ',';
            }
            $multiselectQueries = $this->iterateMultiselectFields(
                $multiselectQueries,
                $multiselectPickListField,
                $item,
                $module,
                $queryParams,
                $index
            );
            if ($module === 'threat_intel_feeds' && !empty($item['relatedThreatIntelFeeds'])) {
                $selfLinkingQueries = $this->iteratFeedSelfLinking($item, $index, $queryParams, $selfLinkingQueries);
            }
            if (array_key_exists('recordTags', $item) && $item['recordTags']) {
                [$tagQueries, $tags] = $this->iterateTagFields(
                    $tagQueries,
                    $tags,
                    $item['recordTags'],
                    $module,
                    $queryParams,
                    $index
                );
            }
        }
        return [$sqlValue, $queryParams, $multiselectQueries, $tagQueries, $selfLinkingQueries, $tags, $uuid_list];
    }

    private function iterateMultiselectFields(
        $multiselectQueries,
        $multiselectPickListField,
        $item,
        $module,
        $queryParams,
        $index
    ) {
        foreach ($multiselectPickListField as $field) {
            if (is_array($item[$field]) && $item[$field]) {
                $tableName = $module . '_picklist_' . $this->inflector->tableize($field);
                if (!array_key_exists($tableName, $multiselectQueries)) {
                    $moduleRefFieldName = $this->getJoinKeyColumnName($module);
                    $fieldsNames = [$moduleRefFieldName . '_uuid', 'picklist_uuid'];
                    $multiselectQueries[$tableName]['fieldName'] = $fieldsNames;
                    $multiselectQueries[$tableName]['values'] = [];
                }
                foreach ($item[$field] as $picklist) {
                    if (is_array($picklist) && array_key_exists("picklistValue", $picklist)) {
                        $fieldVal = $this->metadataHelper->handlePicklistNameToIriConversion('/api/3/picklists', $picklist);
                        $multiselectQueries[$tableName]['values'][$queryParams['uuid' . $index]][] = [$queryParams['uuid' . $index], $fieldVal->getUuid()];
                    } else {
                        $multiselectQueries[$tableName]['values'][$queryParams['uuid' . $index]][] = [$queryParams['uuid' . $index], $picklist];
                    }
                }
            }
        }
        return $multiselectQueries;
    }

    private function iteratFeedSelfLinking($item, $index, $queryParams, $selfLinkingQueries)
    {
        $field = 'relatedThreatIntelFeeds';
        $tableName = 'threatintel_threatintel';
        $fieldNames = ['threatintel_source', 'threatintel_target'];
        if ($item[$field] && is_array($item[$field]) && empty($selfLinkingQueries[$tableName]['fieldName'])) {
            $selfLinkingQueries[$tableName]['fieldName'] = $fieldNames;
            $selfLinkingQueries[$tableName]['values'] = [];
        }
        foreach ($item[$field] as $feed) {
            if (is_array($feed) && array_key_exists("value", $feed)) {
                $selfLinkingQueries[$tableName]['values'][] = [$queryParams['uuid' . $index], $feed['uuid']];
            } else {
                $selfLinkingQueries[$tableName]['values'][] = [$queryParams['uuid' . $index], $feed];
            }
        }
        return $selfLinkingQueries;
    }

    private function iterateTagFields(
        $tagQueries,
        $tags,
        $recordTags,
        $module,
        $queryParams,
        $index
    ) {
        $moduleSingularizeName = $this->getJoinKeyColumnName($module);
        foreach ($recordTags as $tag) {
            $tableName = $moduleSingularizeName . '_tag';
            if (!array_key_exists($tableName, $tagQueries)) {
                $fieldsNames = [$moduleSingularizeName . '_uuid', 'tag_uuid'];
                $tagQueries[$tableName]['fieldName'] = $fieldsNames;
                $tagQueries[$tableName]['values'] = [];
            }
            if (strpos($tag, 'api/3/tags')) {
                $tag = end(explode('/', $tag));
            }
            $tags[] = $tag;
            $tagQueries[$tableName]['values'][$queryParams['uuid' . $index]][] = [$queryParams['uuid' . $index], $tag];
        }
        return [$tagQueries, $tags];
    }

    public function insertTags($tagQueries, $tags, $existingUuids)
    {
        if (!empty($tags)) {
            // Insert Tags in Tag Table
            $sqlParams = [];
            $tagsKeys = array_keys($tags);
            foreach ($tagsKeys as $index) {
                $sqlValue[] = '(:tags' . $index . ')';
                $sqlParams['tags' . $index] = $tags[$index];
            }
            $finalQuery = sprintf(self::RELATIONSHIP_SQL_GENERIC_QUERY, 'tags', 'uuid', implode(',', $sqlValue));
            $this->entityManager->getConnection()->executeQuery($finalQuery, $sqlParams);

            // Insert Tag Value into respective relationship tables.
            $mergedList = [];
            $sqlValue = [];
            $sqlParams = [];
            $tableName = array_keys($tagQueries)[0];
            $fieldNames = $tagQueries[$tableName]['fieldName'];
            $tagValues = $tagQueries[$tableName]['values'];
            if (empty($existingUuids)) {
                return;
            }
            $tagValues = array_intersect_key($tagValues, array_flip($existingUuids));
            $tagValuesSet = array_values($tagValues);
            foreach ($tagValuesSet as $tagValueSet) {
                foreach ($tagValueSet as $tagValue) {
                    $mergedList[] = $tagValue;
                }
            }
            if (empty($mergedList)) {
                return;
            }
            foreach ($mergedList as $index => $tagSet) {
                $sqlValue[] = '(:module' . $index . ',:tag' . $index . ')';
                $sqlParams['module' . $index] = $tagSet[0];
                $sqlParams['tag' . $index] = $tagSet[1];
            }

            $finalQuery = sprintf(self::RELATIONSHIP_SQL_GENERIC_QUERY, $tableName, implode(',', $fieldNames), implode(',', $sqlValue));
            $this->entityManager->getConnection()->executeQuery($finalQuery, $sqlParams);
        }
    }
}
