<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use App\Entity\Custom\Queue;
use Symfony\Component\Serializer\SerializerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use App\Service\WorkflowHelper;
use Ramsey\Uuid\Uuid;

class RuleService
{

    private const QUEUE_CHANNEL_UUID = '44fd54ca-f82d-431e-a6fa-dfde019a1ed6';

    private $container;
    private $workflowHelper;

    public function __construct(WorkflowHelper $workflowHelper, ContainerInterface $container)
    {
        $this->container = $container;
        $this->workflowHelper = $workflowHelper;
    }

    /** @var SerializerInterface */
    protected $serializer;

    public function upsertRulesForQueueManagement($queueView)
    {
        if (!$queueView->getConfig()) {
            return;
        }
        foreach ($queueView->getConfig() as $config) {
            if ($config['type'] === 'fsQueueSettings') {
                foreach ($config['clearRules'] as $recordType => $clearRule) {
                    $existingRules = $this->loadRules($recordType);
                    $ruleData = [
                        'name' => sprintf('%s - queue clear', $recordType),
                        'source' => $recordType,
                        'entity_type' => $recordType,
                        'event_type' => 'update',
                        'trigger_condition' => $this->addIsChangedConditions($clearRule['isChangedFields'], $clearRule['condition']),
                        'actions' => [[
                            'type' => 'queue',
                            'channel_uuid' => '44fd54ca-f82d-431e-a6fa-dfde019a1ed6',
                            'sub_action' => 'clear',
                            'enabled' => true
                        ]],
                        'is_active' => $clearRule['isActive'],
                        'is_system' => true,
                        'event_source' => 'crudhub',
                        'visible' => false,
                        'priority' => 10
                    ];
                    $existingRule = $this->findRuleWithConditions($existingRules, 'update', 'queue clear', $recordType);
                    if ($existingRule) {
                        $ruleData['uuid'] = $existingRule['uuid'];
                    }

                    $this->saveRule($ruleData);
                }
            }
        }
    }

    public function upsertRulesForQueue(Queue $queue)
    {
        if (!$queue->getRecordTypeRules()) {
            return;
        }
        if ($queue->getUuid()) {
            $existingRules = $this->loadRules($queue->getUuid());
            foreach ($existingRules as $rule) {
                $this->deleteRule($rule);
            }
        }
        foreach ($queue->getRecordTypeRules() as $recordTypeRule) {
            $queueRuleType = 'primary';
            $action = [
                'type' => 'queue',
                'channel_uuid' => self::QUEUE_CHANNEL_UUID,
                'queue' => $queue->getUuid(),
                'sub_action' => 'set',
                'enabled' => true
            ];
            $isActive = $queue->getActive() && !$queue->getDraft();
            $newUuid = Uuid::uuid4()->toString();
            $useIsChangedConditions = $recordTypeRule['eventType'] === 'update' && array_key_exists('isChangedFields', $recordTypeRule) && !empty($recordTypeRule['isChangedFields']);
            $ruleData = [
                'name' => sprintf('%s - %s - %s - %s - %s', $queue->getUuid(), $recordTypeRule['recordType'], $recordTypeRule['eventType'], $queueRuleType, $newUuid),
                'source' => $queue->getUuid(),
                'entity_type' => $recordTypeRule['recordType'],
                'event_type' => $recordTypeRule['eventType'],
                'trigger_condition' => $useIsChangedConditions ? $this->addIsChangedConditions($recordTypeRule['isChangedFields'], $recordTypeRule['condition']) : $recordTypeRule['condition'],
                'actions' => [$action],
                'is_active' => $isActive,
                'is_system' => true,
                'event_source' => 'crudhub',
                'visible' => false,
                'priority' => $recordTypeRule['priority']
            ];

            $this->saveRule($ruleData);
        }
    }

    public function deleteRulesForQueue($uuid)
    {
        $existingRules = $this->loadRules($uuid);
        foreach ($existingRules as $rule) {
            $this->deleteRule($rule);
        }
    }

    private function findRuleWithConditions($existingRules, $eventType, $queueRuleType, $recordType)
    {
        foreach ($existingRules as $existingRule) {
            if ($existingRule['entity_type'] == $recordType && $existingRule['event_type'] == $eventType && strpos($existingRule['name'], '- ' . $queueRuleType) > -1) {
                return $existingRule;
            }
        }
        return null;
    }

    private function saveRule(array $ruleData)
    {
        if (array_key_exists("uuid", $ruleData) && $ruleData['uuid']) {
            $uri = $this->container->getParameter('rule.uri') . '/api/rules/' . $ruleData['uuid'] . '/';
            $method = 'PUT';
        } else {
            $uri = $this->container->getParameter('rule.uri') . '/api/rules/';
            $method = 'POST';
        }
        $this->workflowHelper->sealabRequest($uri, $method, $ruleData);
    }

    private function deleteRule(array $ruleData)
    {
        $uri = $this->container->getParameter('rule.uri') . '/api/rules/' . $ruleData['uuid'] . '/';
        $method = 'DELETE';
        $this->workflowHelper->sealabRequest($uri, $method);
    }

    private function loadRules(string $source)
    {
        $uri = sprintf('%s/api/rules/?ordering=name&source=%s', $this->container->getParameter('rule.uri'), $source);
        $method = 'GET';
        return $this->workflowHelper->sealabRequest($uri, $method);
    }

    private function addIsChangedConditions($isChangedFields, $condition)
    {
        $changedCondition = [
            'filters' => [],
            'logic' => 'OR'
        ];
        foreach ($isChangedFields as $isChangedField) {
            $changedCondition['filters'][] = [
                'field' => $isChangedField,
                'operator' => 'changed',
                'value' => true
            ];
        }
        if (!empty($changedCondition['filters'])) {
            return [
                'filters' => [
                    $condition,
                    $changedCondition
                ],
                'logic' => 'AND'
            ];
        } else {
            return $condition;
        }
    }
}
