<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use App\Entity\Actors\Actor;
use Psr\Log\LoggerInterface;
use Doctrine\ORM\EntityManager;
use App\Security\HmacAuthenticator;
use App\Entity\Settings\UserSettings;
use App\Entity\Settings\SystemSettings;
use App\Service\AbstractUserDataProvider;
use App\Repository\UserSettingsRepository;
use App\Repository\SystemSettingsRepository;
use Symfony\Component\HttpFoundation\Request;
use ApiPlatform\Core\Api\IriConverterInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Lexik\Bundle\JWTAuthenticationBundle\Security\Authenticator\JWTAuthenticator;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

/**
 * Data provider for the Settings document.
 *
 */
class SettingsHelper extends AbstractUserDataProvider
{

    protected $iriConverter;

    /** @var EntityManager */
    protected $entityManager;

    protected $containers;

    private $hmacAuthenticator;

    private $logger;

    public function __construct(
        IriConverterInterface $iriConverter,
        ContainerInterface $containers,
        TokenStorageInterface $tokenStorage,
        HmacAuthenticator $hmacAuthenticator,
        protected JWTAuthenticator $jWTAuthenticator,
        LoggerInterface $logger
    ) {
        $this->iriConverter = $iriConverter;
        $this->containers = $containers;
        $this->entityManager = $this->containers->get('doctrine.orm.entity_manager');
        $this->hmacAuthenticator = $hmacAuthenticator;
        $this->jWTAuthenticator = $jWTAuthenticator;
        $this->logger = $logger;
        parent::__construct($tokenStorage);
    }

    /**
     * @inheritdoc
     * @return array
     */
    public function getDataForActor(Actor $actor)
    {
        $userIri = $this->iriConverter->getIriFromItem($actor);
        $userSettings = $this->getUserSettingsFromIRI($userIri);
        /** @var SystemSettingsRepository $systemSettingsRepo */
        $systemSettingsRepo = $this->entityManager->getRepository(SystemSettings::class);
        $systemSettings = $systemSettingsRepo->findAll()[0];
        return [$userSettings, $systemSettings];
    }

    public function getUserSettingsFromIRI($userIri)
    {
        /** @var UserSettingsRepository $userSettingsRepo */
        $userSettingsRepo = $this->entityManager->getRepository(UserSettings::class);
        $userSettings = $userSettingsRepo->findBy(['owner' => $userIri]);
        if (count($userSettings)) {
            $userSettings = $userSettings[0];
        }

        if (!$userSettings || is_null($userSettings)) {
            $userSettings = new UserSettings();
            $userSettings->setOwner($userIri);
            $userSettings->setPublicValues([]);
        }
        return $userSettings;
    }

    public function getCurrentUser()
    {
        if ($this->tokenStorage->getToken()) {
            return $this->tokenStorage->getToken()->getUser();
        }
        return null;
    }

    /**
     * @return string
     */
    public static function extractUserName($user)
    {
        if (method_exists($user, "getfirstname")) {
            if (method_exists($user, "getlastname")) {
                $userName = $user->getfirstname() . ' ' . $user->getlastname();
            } else {
                $userName = $user->getfirstname();
            }
        } else {
            $userName = $user->getname();
        }
        return $userName;
    }

    /**
     * @param UserSettings $userSettings
     * @param array $newValues
     */
    public function saveUserSettings(UserSettings $userSettings, $newValues = [])
    {
        $userSettings->setPublicValues($newValues);

        $this->entityManager->persist($userSettings);
        $this->entityManager->flush();
    }

    /**
     * @param EntityManager $entityManager
     */
    public function setEntityManager($entityManager)
    {
        $this->entityManager = $entityManager;
    }

    /**
     * @param IriConverterInterface $iriConverter
     */
    public function setIriConverter($iriConverter)
    {
        $this->iriConverter = $iriConverter;
    }

    /**
     * @param UserSettings $userSettings
     * @param SystemSettings $systemSettings
     *
     * @return array
     */
    public function getMergedSettings(UserSettings $userSettings, SystemSettings $systemSettings)
    {
        return array_replace_recursive(
            is_null($systemSettings) || is_null($systemSettings->getPublicValues()) ? [] : $systemSettings->getPublicValues(),
            is_null($userSettings) || is_null($userSettings->getPublicValues()) ? [] : $userSettings->getPublicValues(),
            is_null($systemSettings) || is_null($systemSettings->getPrivateValues()) ? [] : $systemSettings->getPrivateValues()
        );
    }

    public function setRefreshedToken($rawToken, $tokenType = 'JWT')
    {
        $subRequest = Request::create('/api/3', 'GET', [], [], [], [], null);
        if ($tokenType == 'JWT') {
            $headers['Authorization'] = 'Bearer ' . $rawToken;
            $subRequest->headers->replace($headers);
            $passport = $this->jWTAuthenticator->authenticate($subRequest);
            $token = $this->jWTAuthenticator->createToken($passport, 'CS');
        } else {
            $passport = $this->hmacAuthenticator->authenticateToken($rawToken);
            $token = $this->hmacAuthenticator->createToken($passport, 'CS');
        }
        $this->tokenStorage->setToken($token);
    }
}
