<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use ZipArchive;
use Ramsey\Uuid\Uuid;
use App\Entity\Core\Tag;
use App\Entity\Core\File;
use App\Entity\Core\PgFile;
use Psr\Log\LoggerInterface;
use App\Entity\Core\Picklist;
use InvalidArgumentException;
use App\Entity\Core\ImportJob;
use App\Constants\AppConstants;
use App\Constants\CacheConstants;
use App\Constants\ErrorConstants;
use App\Entity\Core\AbstractFile;
use App\Entity\Core\SolutionPack;
use App\Entity\Authorization\Team;
use App\Repository\PicklistRepository;
use Doctrine\ORM\EntityManagerInterface;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ConnectException;
use Symfony\Component\Filesystem\Filesystem;
use App\Security\HmacPostAuthenticationToken;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use App\EventSubscriber\MqMessagebroadcastSubscriber;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use Symfony\Component\HttpKernel\Exception\HttpException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Lexik\Bundle\JWTAuthenticationBundle\Security\Authenticator\Token\JWTPostAuthenticationToken;


class SolutionPackUtilityService
{
    private $solutionTempDir;
    protected $containers;
    private  $fileSystem;

    private $entityManager;

    private $tokenStorage;
    private $cache;

    public function __construct(
        ContainerInterface $containers,
        EntityManagerInterface $entityManager,
        TokenStorageInterface $tokenStorage,
        TagAwareCacheInterface $cache,
        protected LoggerInterface $logger
    ) {

        $this->containers = $containers;
        $this->entityManager = $entityManager;
        $this->tokenStorage = $tokenStorage;
        $this->solutionTempDir = sys_get_temp_dir() . '/solutionpacks/';
        $this->fileSystem = new Filesystem();
        $this->cache = $cache;
    }

    public function downloadAndDumpToTemp($contentPackName, $version = null, $buildNumber = "latest")
    {
        $client = new \GuzzleHttp\Client();
        if (!$version) {
            $record = $this->getRecord($contentPackName);
            if ($record) {
                $version = $record->getVersion();
            } else {
                throw new NotFoundHttpException("Missing required version in request body");
            }
        }
        $repoServer = getenv("REPOSERVER") ? getenv("REPOSERVER") : getenv("product_yum_server");
        if (!$buildNumber) {
            $buildNumber = "latest";
        }

        try {
            // Get Zip data from repo
            $ContentName = $contentPackName . '-' . $version;
            $solutionTempPath = $this->solutionTempDir . $ContentName . "/";
            $ContentTgzRepoLink = sprintf('https://%s%s%s/%s/%s.zip', $repoServer, AppConstants::CONTENTHUB_URL, $ContentName, $buildNumber, $ContentName);
            $responseObject = $client->request('GET', $ContentTgzRepoLink, ['verify' => false]);
            $archiveFolder = $this->unzipFile($responseObject->getBody()->getContents(), $ContentName . '.zip', $solutionTempPath);
            return $this->extractZipNameFromExtractedPath($archiveFolder, $solutionTempPath, $ContentName);
        } catch (ConnectException |  ClientException $e) {
            $this->logger->error($e->getMessage());
            $errorMessage = "Unable to download " . $ContentName . " file. Please check the network connection to the " . $repoServer;
            throw new BadRequestHttpException($errorMessage);
        } catch (NotFoundHttpException | \Exception $e) {
            $this->logger->error($e->getMessage());
            throw new NotFoundHttpException($e->getMessage());
        }
    }

    public function extractZipNameFromExtractedPath($archiveFolder, $solutionTempPath, $ContentName)
    {
        $sDir = scandir($archiveFolder);
        foreach ($sDir as $value) {
            if (!in_array($value, array(".", "..")) && !str_starts_with($value, "__") && is_dir($solutionTempPath . $ContentName)) {
                return [$solutionTempPath, $solutionTempPath . $ContentName . "/", $ContentName . ".zip"];
            }
        }
        foreach ($sDir as $value) {
            if (!in_array($value, array(".", "..")) && !str_starts_with($value, "__") && is_dir($solutionTempPath  . $value)) {
                return [$solutionTempPath, $solutionTempPath . $value . "/", $ContentName . ".zip"];
            }
        }
    }
    public function unzipFile($tgzFileData, $contentFileName, $contentFolderPath)
    {
        try {
            if (!file_exists($this->solutionTempDir)) {
                mkdir($this->solutionTempDir);
            }
            if (!file_exists($contentFolderPath)) {
                mkdir($contentFolderPath);
            }
            $zipPath = sprintf('%s%s', $contentFolderPath, $contentFileName);
            $this->checkFileDumpPath($zipPath);
            $this->fileSystem->dumpFile($zipPath, $tgzFileData);
            $zip = new ZipArchive();
            $zip->open($zipPath);
            $zip->extractTo($contentFolderPath);
            $zip->close();
        } catch (BadRequestHttpException | \Exception $e) {
            throw new HttpException(Response::HTTP_BAD_REQUEST, $e->getMessage(), $e);
        } finally {
            $this->fileSystem->chown($contentFolderPath, "nginx", true);
            return $contentFolderPath;
        }
    }

    private function checkFileDumpPath($dirPath, $path = null)
    {
        if (is_null($path) && strpos(realpath(dirname($dirPath)), AppConstants::TMP_SAFE_PATH) !== 0) {
            throw new BadRequestHttpException(ErrorConstants::INVALID_FILE_NAME);
        } else if (!is_null($path)) {
            $path = explode('/', $path);
            if (strpos(realpath(dirname($dirPath)), '/' . $path[1] . '/') !== 0) {
                throw new BadRequestHttpException(ErrorConstants::INVALID_FILE_NAME);
            }
        }
    }

    public function checkTempFreeSpace($file)
    {
        $freeSpace = disk_free_space(sys_get_temp_dir());
        $buffer = 10000000; // 10Mb buffer
        if ($freeSpace - $file->getSize() < $buffer) {
            throw new \Exception(sprintf('Not enough free space in /tmp to import this file. Available: %.1fMb', $freeSpace / 1000000));
        }
    }

    public function extractPgFileTo($folderPath, $fileName, $pgFile)
    {
        $this->fileSystem->mkdir($folderPath);
        $zipFilePath = $folderPath . $fileName;
        $this->fileSystem->dumpFile($zipFilePath, base64_decode(stream_get_contents($pgFile->getFile())));
        $zip = new ZipArchive();
        $zip->open($zipFilePath);
        $zip->extractTo($folderPath);
        $zip->close();
    }

    public function loadFolderIntoTmp($record)
    {
        /** @var string $fileIri */
        $fileIri = $record->getImportJob() ? $record->getImportJob()->getFile() : null;
        if (!$fileIri) {
            return $this->downloadAndDumpToTemp($record->getName(), $record->getVersion());
        } else {
            $explodedIRI = explode('/', $fileIri);
            $fileUuid = end($explodedIRI);
            $repository = $this->entityManager->getRepository(File::class);
            /** @var AbstractFile $file */
            $file =  $repository->findOneBy(['uuid' => $fileUuid]);
            $pgFileUuid = $file->getFile();
            $pgFileRepo = $this->entityManager->getRepository(PgFile::class);
            /** @var AbstractFile $file */
            $pgFile = $pgFileRepo->findOneBy(['uuid' => $pgFileUuid]);
            $folderPath = sprintf('%s%s/', $this->solutionTempDir, $record->getName() . "-" . $record->getVersion());
            $contentName = $record->getName() . "-" . $record->getVersion();
            $zipFileName = $record->getName() . "-" . $record->getVersion() . ".zip";
            $this->checkTempFreeSpace($file);
            $this->extractPgFileTo($folderPath, $zipFileName, $pgFile);
            $sDir = scandir($folderPath);
            foreach ($sDir as $value) {
                if (!in_array($value, array(".", "..")) && !str_starts_with($value, "__") && is_dir($folderPath . $record->getName() . "-" . $record->getVersion())) {
                    return [$folderPath, $folderPath . $contentName . '/', $zipFileName];
                }
            }
            foreach ($sDir as $value) {
                if (!in_array($value, array(".", "..")) && !str_starts_with($value, "__") && is_dir($folderPath  . $value)) {
                    return [$folderPath, $folderPath . $value . '/', $zipFileName];
                }
            }
        }
    }

    public function getRecord($name, $version = null, $type = null, $source=null)
    {
        $repository = $this->entityManager->getRepository(SolutionPack::class);
        $filter = [];
        $filter['name'] = $name;
        if ($version) {
            $filter['version'] = $version;
        }
        if ($type) {
            $filter['type'] = $type;
        }
        $existingSolutionPack = $repository->findBy($filter, array("version" => "desc"));
        if (!empty($existingSolutionPack) && is_null($version) && $source == 'repo') {
            foreach ($existingSolutionPack as $sp) {
                if (str_contains($sp->getVersion(), '_dev')) {
                     continue;
                }
                return $sp;
            }
        }
        return empty($existingSolutionPack) ?  false : $existingSolutionPack[0];
    }


    public function uploadFile($filePath, $fileName)
    {
        $fileUuid =  Uuid::uuid4()->toString();
        /** @var UploadedFile $upload */
        $upload = new uploadedFile($filePath . $fileName, $fileName);
        $teamsRepo = $this->entityManager->getRepository(Team::class);
        $teams = $teamsRepo->findAll();
        $file = new File();
        $file->setFilename($upload->getClientOriginalName());
        $file->setMimeType($upload->getClientMimeType());
        $file->setSize($upload->getSize());
        $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $file->setUploadDate($timestamp);

        $pgFile = new PgFile();
        /** @var AbstractFile $file */
        $strm = fopen($upload->getPathname(), 'rb');
        $pgFile->setFile(base64_encode(stream_get_contents($strm)));

        $this->entityManager->persist($pgFile);
        $this->entityManager->flush();
        $file->setFile($pgFile->getUuid());
        $file->setUuid($fileUuid);
        $file->setId($fileUuid);
        foreach ($teams as $team) {
            $file->addOwner($team);
        }
        $this->entityManager->persist($file);
        $this->entityManager->flush();
        return $file;
    }

    public function getCurrentUser()
    {
        if ($this->tokenStorage->getToken()) {
            return $this->tokenStorage->getToken()->getUser();
        }
        return null;
    }

    public function getCurrentUserType()
    {
        if ($this->tokenStorage->getToken() && $this->tokenStorage->getToken() instanceof JWTPostAuthenticationToken) {
            return "JWT";
        } else if ($this->tokenStorage->getToken() && $this->tokenStorage->getToken() instanceof HmacPostAuthenticationToken) {
            return "HMAC";
        }
        return null;
    }

    public function getTokenFromStorage()
    {
        if ($this->tokenStorage->getToken()) {
            return $this->tokenStorage->getToken()->getCredentials();
        }
        return null;
    }

    public static function extractUserName($user)
    {
        if (method_exists($user, "getfirstname")) {
            if (method_exists($user, "getlastname")) {
                $userName = $user->getfirstname() . ' ' . $user->getlastname();
            } else {
                $userName = $user->getfirstname();
            }
        } else {
            $userName = $user->getname();
        }
        return $userName;
    }

    /**
     * @return array
     */
    public function prepareChangeData(array $oldData, array $newData, string $entityId, $method = "Update")
    {
        $deltaData = [];
        $deltaData['oldData'] = $oldData;
        $deltaData['newData'] = $newData;

        try {
            $user = $this->getCurrentUser();
            if ($user) {
                $userUUID = $user->getUuid();
                $userName = $this->extractUserName($user);
            } else {
                $userUUID = 'SYSTEM';
                $userName = 'SYSTEM';
            }
        } catch (InvalidArgumentException | \Exception $e) {
            $userUUID = 'SYSTEM';
            $userName = 'SYSTEM';
        }
        $sourceIp = 'localhost';
        if (isset($_SERVER['HTTP_X_FORWARDED_FOR'])) {
            $sourceIp = $_SERVER['HTTP_X_FORWARDED_FOR'];
        } elseif (isset($_SERVER['REMOTE_ADDR'])) {
            $sourceIp = $_SERVER['REMOTE_ADDR'];
        }
        return [
            AppConstants::COMPONENT => 'crudhub',
            AppConstants::ENTITY_TYPE => 'SolutionPack',
            AppConstants::ENTITY_UUID => $entityId,
            AppConstants::HTTP_METHOD => $method,
            AppConstants::USER_UUID => $userUUID,
            AppConstants::USER => $userName,
            AppConstants::TRANSACTION_DATE => (int)($_SERVER['REQUEST_TIME_FLOAT'] * 1000),
            AppConstants::DELTA_DATA => $deltaData,
            AppConstants::REMOTE_ADDR => $sourceIp
        ];
    }

    public function checkPrimaryNode()
    {
        $instanceClusterDetails = $this->getClusterDetails();
        if (isset($instanceClusterDetails['role'])) {
            return $instanceClusterDetails["role"] == 'primary';
        }
        return false;
    }

    public function getClusterDetails()
    {
        $item = $this->cache->getItem(CacheConstants::CLUSTER_INSTANCE_SETTINGS);
        if ($item->isHit()) {
            return $item->get();
        } else {
            $dasUri = $this->containers->getParameter('das.uri');
            $dasClient = $this->containers->get("app.proxy.client.das");
            $headers = array(
                "Content-type" => "application/json"
            );
            $dasRequests = new GuzzleRequest('get', $dasUri . '/cluster/?self_only=true', $headers);
            try {
                $dasResponse = $dasClient->send($dasRequests);
                $dasResponseData = $dasResponse->getBody()->getContents();
                $jsonDecodeResponse = json_decode($dasResponseData, true);
                $instanceClusterDetails = is_array($jsonDecodeResponse["nodes"]) && isset($jsonDecodeResponse["nodes"][0]) ? $jsonDecodeResponse["nodes"][0] : null;
                // Expected Roles to be  'active', 'passive', 'faulted', 'primary', 'secondary'
                if (isset($instanceClusterDetails)) {
                    $item->set($instanceClusterDetails);
                }
                $this->cache->save($item);
            } catch (BadRequestHttpException | \Exception $error) {
                $this->logger->error($error->getMessage());
            }
            return $instanceClusterDetails;
        }
    }

    public function updateSolutionPack(SolutionPack $solutionPack, $status = null, $installed = null, $latestAvailableVersion = null, $latestCompatibilityVersion = null)
    {
        $oldData = [
            '@id' => '/api/3/solutionpacks/' . $solutionPack->getUuid()
        ];
        $newData = [
            '@id' => '/api/3/solutionpacks/' . $solutionPack->getUuid()
        ];
        $changeData = [];

        if ($status !== null) {
            $oldStatus = $solutionPack->getStatus();
            $solutionPack->setStatus($status);
            $oldData['status'] = $oldStatus;
            $newData['status'] = $status;
            if ($oldStatus != $status) {
                $changeData['status'] = $status;
            }
        }

        if ($latestAvailableVersion !== null) {
            $oldLatestAvailableVersion = $solutionPack->getLatestAvailableVersion();
            $solutionPack->setLatestAvailableVersion($latestAvailableVersion);
            $oldData['latestAvailableVersion'] = $oldLatestAvailableVersion;
            $newData['latestAvailableVersion'] = $latestAvailableVersion;
        } else {
            $solutionPack->setLatestAvailableVersion($latestAvailableVersion);
        }

        if ($latestCompatibilityVersion !== null) {
            $oldLatestCompatibilityVersion = $solutionPack->getLatestCompatibleVersion();
            $solutionPack->setLatestCompatibleVersion($latestCompatibilityVersion);
            $oldData['latestCompatibilityVersion'] = $oldLatestCompatibilityVersion;
            $newData['latestCompatibilityVersion'] = $latestCompatibilityVersion;
        } else {
            $solutionPack->setLatestCompatibleVersion($latestCompatibilityVersion);
        }

        if ($installed !== null) {
            $oldInstalled = $solutionPack->getInstalled();
            if ($solutionPack->getType() != 'connector') {
                $solutionPack->setInstalled($installed);
            }
            $oldData['installed'] = $oldInstalled;
            $newData['installed'] = $installed;
            if ($oldInstalled != $installed) {
                $changeData['installed'] = $installed;
            }
        }
        $newData['type'] = $solutionPack->getType();
        $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $solutionPack->setModifyDate($timestamp);
        try {
            if (!$this->entityManager->isOpen()) {
                $this->entityManager = $this->containers->get('doctrine')->resetManager();
            }
            $this->entityManager->persist($solutionPack);
            $this->entityManager->flush();
        } catch (\Exception $e) {
            $message = sprintf('Error while updating solution pack : %s', $e->getMessage());
            $this->logger->error($message);
        }
        $body = $this->prepareChangeData($oldData, $newData, $solutionPack->getUuid());
        $this->containers->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->setContentType('application/json');
        $this->containers->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->publish(json_encode([$body]), MqMessagebroadcastSubscriber::DEFALUT_KEY);
    }

    public function updateImportJob(ImportJob $importJob, $status = null, $progressPercent = null, $options = null, $errorMessage = null, $currentlyImporting = null, $logMessage = null)
    {
        $oldData = [
            '@id' => '/api/3/import_jobs/' . $importJob->getUuid()
        ];
        $newData = [
            '@id' => '/api/3/import_jobs/' . $importJob->getUuid()
        ];
        $oldData['id'] = $importJob->getId();
        $newData['id'] = $importJob->getId();
        if ($status !== null) {
            $oldStatus = $importJob->getStatus();
            $importJob->setStatus($status);
            $oldData['status'] = $oldStatus;
            $newData['status'] = $status;
        }
        if ($progressPercent !== null) {
            $oldProgressPercent = $importJob->getProgressPercent();
            $importJob->setProgressPercent($progressPercent);
            $oldData['progressPercent'] = $oldProgressPercent;
            $newData['progressPercent'] = $progressPercent;
        }
        if ($errorMessage !== null) {
            $oldErrorMessage = $importJob->getErrorMessage();
            $importJob->setErrorMessage($errorMessage);
            $oldData['errorMessage'] = $oldErrorMessage;
            $newData['errorMessage'] = $errorMessage;
        }
        if ($options !== null) {
            $oldOptions = $importJob->getOptions();
            $importJob->setOptions($options);
            $oldData['options'] = $oldOptions;
            $newData['options'] = $options;
        }
        if ($currentlyImporting !== null) {
            $old = $importJob->getCurrentlyImporting();
            $importJob->setCurrentlyImporting($currentlyImporting);
            $oldData['currentlyImporting'] = $old;
            $newData['currentlyImporting'] = $currentlyImporting;
        }
        if ($logMessage !== null) {
            if ($logMessage === []) {
                $importJob->clearLogMessages();
            } else {
                $importJob->addLogMessage($logMessage);
            }
            // Don't audit logMessages
        }
        $timestamp = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $importJob->setModifyDate($timestamp);
        try {
            if (!$this->entityManager->isOpen()) {
                $this->entityManager = $this->containers->get('doctrine')->resetManager();
            }
            $this->entityManager->persist($importJob);
            $this->entityManager->flush();
        } catch (\Exception $e) {
            $message = sprintf('Error while updating import job : %s', $e->getMessage());
            $this->logger->error($message);
        }
        $body = $this->prepareChangeData($oldData, $newData, $importJob->getUuid());
        $this->containers->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->setContentType('application/json');
        $this->containers->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->publish(json_encode([$body]), MqMessagebroadcastSubscriber::DEFALUT_KEY);
    }

    public function removeFileObj($solutionPack, $removeImportJob = true, $removeExportJob = false, $removeExportTemplate = false)
    {
        $importJob = $exportJob = $template = null;
        $fileToDelete = [];
        if ($removeImportJob && $solutionPack->getImportJob()) {
            $importJob = $solutionPack->getImportJob();
            array_push($fileToDelete, $solutionPack->getImportJob()->getFile());
            $solutionPack->setImportJob(null);
        }
        if ($removeExportJob && $solutionPack->getExportJob()) {
            $exportJob = $solutionPack->getExportJob();
            array_push($fileToDelete, $solutionPack->getExportJob()->getFile());
            $solutionPack->setExportJob(null);
        }
        if ($removeExportTemplate && $solutionPack->getTemplate()) {
            $template = $solutionPack->getTemplate();
            $solutionPack->setTemplate(null);
        }
        $this->cleanUpFile($fileToDelete);
        $this->entityManager->persist($solutionPack);
        $this->entityManager->flush();

        if ($importJob) {
            $this->entityManager->remove($importJob);
        }
        if ($exportJob) {
            $this->entityManager->remove($exportJob);
        }
        if ($template) {
            $this->entityManager->remove($template);
        }
    }

    private function cleanUpFile(array $fileIRIs)
    {
        foreach ($fileIRIs as $fileIRI) {
            $explodedIRI = explode('/', $fileIRI);
            $fileUuid = end($explodedIRI);
            $fileRepo = $this->entityManager->getRepository(File::class);
            $fileObj = $fileRepo->findOneBy(array("uuid" => $fileUuid));
            if ($fileObj && $fileObj->getFile()) {
                $pgFileRepo = $this->entityManager->getRepository(PgFile::class);
                $pgFile = $pgFileRepo->findOneBy(array("uuid" => $fileObj->getFile()));
                $this->entityManager->remove($pgFile);
                $this->entityManager->remove($fileObj);
            }
        }
    }

    public function prepareCategoryData($category)
    {
        if (is_string($category)) {
            $category = explode(',', $category);
        }
        if (empty($category) || $category == "") {
            return [];
        }
        $picklistIRI = [];
        $picklistItems = [];
        $list = [];
        /**@var PicklistRepository $repository */
        $repository = $this->entityManager->getRepository(Picklist::class);
        $allCategoryList =  $repository->findBy(["listName" => '9ac56bf5-321f-4602-bbd6-e5a14338df2a']);

        foreach ($allCategoryList as $value) {
            $list[strtolower(trim($value->getItemValue(), " "))] = $value->getUuid();
        }
        foreach ($category as $value) {
            if ($value == "") {
                continue;
            }
            if (!array_key_exists(strtolower(trim($value, " ")), $list)) {
                $uuid =  Uuid::uuid4()->toString();
                $picklistItems[] = ["uuid" => $uuid, "itemValue" => trim($value, " "), "orderIndex" => 0, "color" => null, "icon" => null, "listName" => "9ac56bf5-321f-4602-bbd6-e5a14338df2a"];
                $picklistIRI[] = '/api/3/picklists/' . $uuid;
            } else {
                $picklistIRI[] = '/api/3/picklists/' . $list[strtolower(trim($value, " "))];
            }
        }
        $repository->upsertPicklist($picklistItems);
        return $picklistIRI;
    }

    public function getInfoContent($options)
    {
        $configTypes = AppConstants::CONFIG_TYPES;
        $contents = [];

        foreach ($options as $key => $value) {
            if (!empty($value)) {
                /** @var PortingConfig $portingConfig */
                $portingConfig = $this->containers->get($configTypes[$key]);
                $contentData = $portingConfig->generateInfoContent($value);
                if ($key == "playbooks") {
                    $contents["playbooks"] = $contentData["playbooks"];
                    $contents["globalVariables"] = $contentData["globalVariables"];
                } else {
                    $contents[$key] = $contentData;
                }
            }
        }

        return $contents;
    }

    public function getLatestSolutionPack($name)
    {
        /** @var EntityManager $manager */
        $managerRegistry = $this->containers->get('doctrine');
        $manager = $managerRegistry->getManagerForClass(SolutionPack::class);
        $repository = $manager->getRepository(SolutionPack::class);
        $queryBuilder = $repository->createQueryBuilder('s');
        $solutionPackRecords = $queryBuilder->where('s.name = :name')
            ->setParameter('name', $name)
            ->addOrderBy('s.version', 'DESC')
            ->getQuery()
            ->execute();
        if (!empty($solutionPackRecords)) {
            foreach ($solutionPackRecords as $sp) {
                if ($sp->getInstalled()) {
                    return $sp;
                }
            }
            return $solutionPackRecords[0];
        } else {
            $solutionPackRecord = null;
            $repoSolutionsPack = $this->getSolutionPackFromRepo($name);
            if ($repoSolutionsPack) {
                $solutionPackRecord = new SolutionPack();
                $solutionPackRecord->setName($repoSolutionsPack->name);
                $solutionPackRecord->setVersion($repoSolutionsPack->version);
                $solutionPackRecord->setBuildNumber($repoSolutionsPack->buildNumber);
                $solutionPackRecord->setType('connector');
                $solutionPackRecord->setLatestAvailableVersion(null);
                $solutionPackRecord->setFsrMinCompatibility($repoSolutionsPack->fsrMinCompatibility);
                $solutionPackRecord->setLatestCompatibleVersion(null);
                $solutionPackRecord->setInstalled(false);
                $solutionPackRecord->setLocal(false);
            }
            return $solutionPackRecord;
        }
    }

    private function validateAndGetTags($tagList)
    {
        if (!is_array($tagList)) {
            return [];
        }
        $tags = [];
        foreach ($tagList as $tag) {
            if (preg_match("/[#,?'\"]/", $tag)) {
                continue;
            }
            $tags[] = $tag;
        }
        return $tags;
    }

    public function importTags($tagList)
    {
        $tags = $this->validateAndGetTags($tagList);
        $tagRepository = $this->entityManager->getRepository(Tag::class);
        if (!empty($tags)) {
            $tagRepository->upsertTag($tags);
        }
        return $tags;
    }

    public function getSolutionPackFromRepo($name = null, $version = null)
    {
        try {
            $client = new \GuzzleHttp\Client();
            $repoServer = getenv("REPOSERVER") ? getenv("REPOSERVER") : getenv("product_yum_server");
            if ($version && $name) {
                $repoLink = sprintf('https://' . $repoServer .  AppConstants::CONTENTHUB_URL . $name . "-" . $version . "/latest/info.json");
                $marketplaceJsonObject = $client->request('GET', $repoLink, ['verify' => false]);
                return json_decode($marketplaceJsonObject->getBody()->getContents());
            }
            $allRepoLink = sprintf('https://' . $repoServer .  AppConstants::CONTENTHUB_URL . AppConstants::CONTENTHUB_METADATA_JSON);
            $marketplaceJsonObject = $client->request('GET', $allRepoLink, ['verify' => false]);
            $allSolutions = json_decode($marketplaceJsonObject->getBody()->getContents());
            if ($name) {
                foreach ($allSolutions as $value) {
                    if ($value->name == $name) {
                        return $value;
                    }
                }
            } else {
                return $allSolutions;
            }
        } catch (\Exception $e) {
            $this->logger->error($e->getMessage());
            $errorMessage = "Unable to download content hub metadata json file. Please check the network connection to the " . $repoServer;
            $this->logger->error($errorMessage);
        }
        return null;
    }
}
