<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use DateTime;
use App\Service\SKUService;
use Psr\Log\LoggerInterface;
use App\Entity\Core\Picklist;
use InvalidArgumentException;
use App\Constants\AppConstants;
use Symfony\Component\Yaml\Yaml;
use App\Constants\CacheConstants;
use App\Constants\ErrorConstants;
use App\Entity\Core\EmailTemplate;
use App\Entity\Settings\SystemSettings;
use Doctrine\ORM\EntityManagerInterface;
use App\Service\AccessibleTeamIdsProvider;
use App\Entity\ModelMetadata\ModelMetadata;
use Symfony\Component\Filesystem\Filesystem;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use ApiPlatform\Core\Api\IriConverterInterface;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use ApiPlatform\Core\DataProvider\ItemDataProviderInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;

class UtilityService
{

    public const WIDGET_PREVIEW_PATH = '/opt/cyops-ui/widgets_preview/';
    public  const WIDGET_PATH = '/opt/cyops-ui/widgets/';

    private $headers = array("Content-type" => "application/json");

    protected $dasClient;
    protected $dasUri;
    private  $fileSystem;

    public function __construct(
        protected AccessibleTeamIdsProvider $accessibleTeamIdsProvider,
        protected TokenStorageInterface $tokenStorage,
        protected ContainerInterface $containers,
        protected TagAwareCacheInterface $cache,
        protected SKUService $skuService,
        protected ItemDataProviderInterface $itemDataProvider,
        protected MetadataHelper $metadataHelper,
        protected IriConverterInterface $iriConverter,
        protected EntityManagerInterface $entityManager,
        protected LoggerInterface $logger
    ) {
        $this->dasClient = $this->containers->get("app.proxy.client.das");
        $this->dasUri = $this->containers->getParameter('das.uri');
        $this->fileSystem = new Filesystem();
    }

    public function getUniqueResourceFields($request_data, $resourceShortName)
    {
        $permission = $this->getModulePermission($resourceShortName);
        if (isset($request_data) && isset($request_data->{'__unique'}) && $request_data->{'__unique'}) {
            return [$request_data->{'__unique'}, $permission];
        }
        $cacheName = 'constraintField_' . $resourceShortName;
        $item = $this->cache->getItem($cacheName);
        if ($item->isHit()) {
            return $item->get();
        }
        $managerRegistry = $this->containers->get('doctrine');
        $manager = $managerRegistry->getManagerForClass(ModelMetadata::class);
        $modelMetaDataRepository = $manager->getRepository(ModelMetadata::class);
        try {
            $modelMetaData = $modelMetaDataRepository->findOneBy(['type' => $resourceShortName]);
        } catch (InvalidArgumentException | \Exception $e) {
            $this->logger->error($e->getMessage());
        }
        if (!$modelMetaData) {
            return [[], $permission];
        }
        $uniqueConstraint = count($modelMetaData->getUniqueConstraint()) ? $modelMetaData->getUniqueConstraint()[0] : [];
        $module = $modelMetaData->getType();
        $columns = [];
        foreach ($uniqueConstraint as $key => $value) {
            $columns = isset($value['columns']) ? $value['columns'] : [];
        }
        $item->set([$columns, $module]);
        $item->expiresAfter(3000);
        $item->tag(CacheConstants::UPSERT_MODULE_FIELDS);
        $this->cache->save($item);
        return [$columns, $module];
    }

    public function getModulePermission($resourceShortName)
    {
        $typeClassMap = $this->cache->getItem(CacheConstants::ENTITY_TYPE_CLASS_MODEL_MAP);
        if (!$typeClassMap->isHit()) {
            $this->metadataHelper->createModuleMapCache();
            $typeClassMap = $this->cache->getItem(CacheConstants::ENTITY_TYPE_CLASS_MODEL_MAP);
        }
        [$classPath, $module] = $typeClassMap->get()[$resourceShortName];

        return $module;
    }

    private function handleDefaultValue($item, $filter, $resource)
    {
        if ($item === 'tenant') {
            $filter[$item] = 'b3a700f7-00be-4ef9-90c6-3c8fe6e1be63';
        } else {
            $entityFieldMap = $this->metadataHelper->getEntityFields($resource) ?? null;
            $defaultValues = $entityFieldMap['defaultValue'] ?? [];
            if (!empty($defaultValues) && array_key_exists($item, $defaultValues)) {
                $filter[$item] = $defaultValues[$item];
            }
        }
        return $filter;
    }

    public function prepareFilterData($requestBody, $filterItems, $resource = null)
    {
        $filter = [];
        foreach ($filterItems as $item) {
            if (is_string($item)) {
                if (is_array($requestBody)) {
                    $value = $requestBody[$item] ?? null;
                    if (!is_null($value) && is_array($value) && isset($value['picklistName']) && isset($value['picklistValue'])) {
                        $iri = $this->iriConverter->getIriFromResourceClass(Picklist::class);
                        $filter[$item] = $this->metadataHelper->handlePicklistNameToIriConversion($iri, $value);
                    } else {
                        $filter[$item] = $value;
                    }
                } else {
                    $value = $requestBody->$item ?? null;
                    if (!is_null($value) && is_object($value) && isset($value->picklistName) && isset($value->picklistValue)) {
                        $iri = $this->iriConverter->getIriFromResourceClass(Picklist::class);
                        $filter[$item] = $this->metadataHelper->handlePicklistNameToIriConversion($iri, json_decode(json_encode($value), true));
                    } else {
                        $filter[$item] = $value;
                    }
                }
                if (is_null($filter[$item]) && $resource) {
                    $filter = $this->handleDefaultValue($item, $filter, $resource);
                }
            } else {
                $filter = array_merge($filter, get_object_vars($item));
            }
        }
        return $filter;
    }

    protected function updatedUuidList($iriList)
    {
        foreach ($iriList as $iri) {
            if (is_string($iri) && str_starts_with($iri, '/api/3/')) {
                $result[] = basename($iri);
            } else {
                $result[] = $iri; // can check if its uuid format
            }
        }
        $result = isset($result) ? $result : [];
        return $result;
    }

    public function getItemWithWhere($repository, $filters, array $orderBy = [], array $selectFields = [])
    {
        // OrderBy sample pattern
        // [{"field": "o.fieldName", "direction" : "ASC"}]
        $queryBuilder = $repository->createQueryBuilder('o');

        foreach ($filters as $key => $value) {
            if (gettype($value) == 'array') {
                $value = $this->updatedUuidList($value);
                $queryBuilder->andwhere('o.' . $key . ' in (:' . $key . ')');
                $queryBuilder->setParameter($key, $value);
                continue;
            }

            if (is_object($value) && isset($value->{"@id"})) {
                $value = $value->{"@id"};
            }
            if (gettype($value) != 'boolean') {
                if (is_string($value) && str_starts_with($value, '/api/3/')) {
                    $explodedValue = explode('/', $value);
                    $value = end($explodedValue);
                }
                $queryBuilder->andwhere('o.' . $key . '= :' . $key);
                $queryBuilder->setParameter($key, $value);
            } else {
                $queryBuilder->andwhere('o.' . $key . '= :value');
                $queryBuilder->setParameter('value', $value);
            }
        }
        if (!empty($orderBy)) {
            foreach ($orderBy as $order) {
                $queryBuilder->addOrderBy($order['field'], $order['direction']);
            }
        }
        if (!empty($selectFields)) {
            $queryBuilder->select(implode(', ', $selectFields));
        }
        return $queryBuilder->getQuery()->execute();
    }

    /**
     * @param $item
     * @param $normalizedTeamData
     * @param $normalizedItemData
     */
    protected function isRecordInList($accessibleRecordTeamUuid, $loggedInUserTeamUuid)
    {
        $returnFlag = true;
        if (count($accessibleRecordTeamUuid) == 0) {
            // Public record
            $returnFlag =  true;
        }
        if (count($loggedInUserTeamUuid) == 0) {
            // User not on any teams
            $returnFlag =  false;
        }
        if (count($accessibleRecordTeamUuid) && count($loggedInUserTeamUuid) && (!count(array_intersect($accessibleRecordTeamUuid, $loggedInUserTeamUuid)))) {
            // Checking if Record owning teams and logged in user teams have common teams
            $returnFlag =  false;
        }
        return $returnFlag;
    }

    protected function getRecordTeamUuid($object)
    {
        $recordTeam = [];
        if (method_exists($object, 'getOwners') && $object->getOwners() != null && $object->getOwners() != []) {
            $recordTeam = $this->accessibleTeamIdsProvider->getParentDataForTeam($object->getOwners());
        }
        return $recordTeam;
    }

    protected function getLoggedInUserTeamUuid($object)
    {
        $recordTeam = [];
        if (method_exists($object, 'getTeams') && $object->getTeams() != null && $object->getTeams() != []) {
            foreach ($object->getTeams() as $team) {
                array_push($recordTeam, $team->getUuid());
            }
        }
        return $recordTeam;
    }

    private function getUser()
    {
        if (!is_null($this->tokenStorage->getToken())) {
            return $this->tokenStorage->getToken()->getUser();
        } else {
            $this->logger->error("Token not found in token storage");
        }
    }

    public function validateOwnerShip($request, $item)
    {
        if ($request && $item) {
            $accessibleRecordTeamUuid = $this->getRecordTeamUuid($item);
            $loggedInUser = $this->getUser();
            $loggedInUserTeamUuid = $this->getLoggedInUserTeamUuid($loggedInUser);
            if (!count($loggedInUserTeamUuid)) {
                // User with no team should not be allowed to view the resource
                throw new AccessDeniedException();
            }
            // Check if Record is ownable
            $teamOwned = $this->isRecordInList($accessibleRecordTeamUuid, $loggedInUserTeamUuid);
            $userOwned = $this->doesUserOwnRecord($item);
            if (!$teamOwned && !$userOwned) {
                throw new AccessDeniedException();
            }
        }
    }

    protected function doesUserOwnRecord($item)
    {
        if (
            method_exists($item, 'getUserOwners')
            && $item->getUserOwners() != null
            && $item->getUserOwners() != []
            && $item->getUserOwners()->contains($this->getUser())
        ) {
            return true;
        }
        return false;
    }

    public function filePutContent($filePath, $fileContent)
    {
        file_put_contents($filePath, $fileContent);
    }

    protected function createWidgetControllerFiles($src, $dst, $matchString, $replaceString)
    {
        $fileContent = preg_replace('/' . $matchString . '/', $replaceString, file_get_contents($src));

        $file = fopen($dst, "w");
        fwrite($file, $fileContent);
        fclose($file);
        if (strpos($src, "widgets/installed/") && strpos($dst, "widgets/development/")) {
            $this->filePutContent($dst, str_replace("widgets/installed/", "widgets/development/", file_get_contents($dst)));
        }
        if (strpos(file_get_contents($src), "widgets/development/") && strpos(file_get_contents($src), "widgets/installed/")) {
            $this->filePutContent($dst, str_replace("widgets/development/", "widgets/installed/", file_get_contents($dst)));
        }
    }

    protected function widgetCustomCopy($src, $dst, $widgetName, $widgetVersion, $removeDevString)
    {
        $filePathArray = explode('/', $src);
        $filePathName = end($filePathArray);

        if (count($filePathArray) == 7 && $filePathName == 'view.controller.js') {
            $widgetVersionString = str_replace('.', '', $widgetVersion);
            $matchString = $removeDevString ? ($widgetName . $widgetVersionString . 'DevCtrl') : ($widgetName . $widgetVersionString . 'Ctrl');
            $replaceString = $removeDevString ? ($widgetName . $widgetVersionString . 'Ctrl') : ($widgetName . $widgetVersionString . 'DevCtrl');

            $this->createWidgetControllerFiles($src, $dst, $matchString, $replaceString);
        } elseif (count($filePathArray) == 7 && $filePathName == 'edit.controller.js') {
            $widgetVersionString = str_replace('.', '', $widgetVersion);
            $matchString = $removeDevString ? ('edit' . ucfirst($widgetName) . $widgetVersionString . 'DevCtrl') : ('edit' . ucfirst($widgetName) . $widgetVersionString . 'Ctrl');
            $replaceString = $removeDevString ? ('edit' . ucfirst($widgetName) . $widgetVersionString . 'Ctrl') : ('edit' . ucfirst($widgetName) . $widgetVersionString . 'DevCtrl');

            $this->createWidgetControllerFiles($src, $dst, $matchString, $replaceString);
        } else {
            $this->fileSystem->copy($src, $dst);

            if (count($filePathArray) == 7 && $filePathName == 'info.json') {
                $widget_json = json_decode(file_get_contents($dst), true);
                if (array_key_exists('metadata', $widget_json) && array_key_exists('publisher', $widget_json['metadata']) && $widget_json['metadata']['publisher'] == 'Fortinet') {
                    $widget_json['metadata']['publisher'] = null;
                    file_put_contents($dst, json_encode($widget_json, JSON_PRETTY_PRINT));
                }
            }
        }
    }

    public function copyWidgetFiles($src, $dst, $widgetName, $widgetVersion, $removeDevString)
    {
        if (!is_dir($dst)) {
            $this->fileSystem->mkdir($dst);
        }
        if (is_dir($src)) {
            $dir = opendir($src);
            while ($file = readdir($dir)) {
                if ($file != "." && $file != "..") {
                    if (is_dir($src . "/" . $file)) {
                        $this->copyWidgetFiles($src . "/" . $file, $dst . "/" . $file, $widgetName, $widgetVersion, $removeDevString);
                    } else {
                        $this->widgetCustomCopy($src . "/" . $file, $dst . "/" . $file, $widgetName, $widgetVersion, $removeDevString);
                    }
                }
            }
            closedir($dir);
        } else {
            $this->widgetCustomCopy($src, $dst, $widgetName, $widgetVersion, $removeDevString);
        }

        return true;
    }

    public function removeWidgetDir($widgetPath)
    {
        if (!file_exists($widgetPath)) {
            return;
        }
        try {
            $files = scandir($widgetPath);
            foreach ($files as $file) {
                if ($file === '.' || $file === '..') {
                    continue;
                }
                $path = $widgetPath . '/' . $file;
                if (is_dir($path)) {
                    $this->removeWidgetDir($path);
                } else {
                    unlink($path);
                }
            }
            $this->fileSystem->remove($widgetPath);
        } catch (\Exception $th) {
            $this->logger->error("Error: Widget directory cleanup operation not completed.");
        }
    }

    public function removeWidgetFiles($widgetPath)
    {
        if ($this->fileSystem->exists($widgetPath)) {
            $this->fileSystem->remove($widgetPath);
        }
        return true;
    }

    public function syncHANodes($path, $haSynCommandType)
    {
        if ($haSynCommandType === 'delete') {
            $payload =  ['paths' => [$path]];
        } elseif ($haSynCommandType === 'copy') {
            $payload = ['source' => ['paths' => [$path]]];
        } elseif ($haSynCommandType === 'rename') {
            $payload = [
                'source' => $path['source'],
                'target' => $path['target']
            ];
        }
        $objectData = [
            'handlerType' => 'file_handler',
            'commandType' => $haSynCommandType,
            'nodeIds' => 'OTHER_NODES',
            'payload' => $payload
        ];
        try {
            $dasRequests = new GuzzleRequest('post', $this->dasUri . '/clustercommand/', $this->headers, json_encode($objectData));
            $dasResponse = $this->dasClient->send($dasRequests);
            $resCode = $dasResponse->getStatusCode();
            if ($resCode != 200) {
                return -1;
            }
        } catch (\Exception $error) {
            $this->logger->error($error->getMessage());
            return -1;
        }
    }

    public function broadcastToHA($command)
    {
        $payload = [
            "command" => $command,
        ];
        $objectData = [
            "handlerType" => "cmd_handler",
            "commandType" => "clear-cache",
            "payload" => $payload,
            "nodeIds" => 'OTHER_ACTIVE_NODES',
        ];

        $dasRequests = new GuzzleRequest('post', $this->dasUri . '/clustercommand/', $this->headers, json_encode($objectData));
        try {
            $dasResponse = $this->dasClient->send($dasRequests);
            $resCode = $dasResponse->getStatusCode();
            if ($resCode > 299) {
                return 0;
            }
        } catch (BadRequestHttpException | \Exception $error) {
            $this->logger->error($error->getMessage());
        }
        return true;
    }

    public static function identifyChangeFields($requestBody)
    {
        $keys = ["fsr_primary", "password"];
        foreach ($requestBody as $key => $value) {
            $keys[] = $key;
            if ($key == '__link') {
                foreach ($requestBody->{'__link'} as $innerKey => $innerValue) {
                    $keys[] = $innerKey;
                }
            }
            if ($key == '__unlink') {
                foreach ($requestBody->{'__unlink'} as $innerKey => $innerValue) {
                    $keys[] = $innerKey;
                }
            }
        }
        return $keys;
    }

    public function getRequiredLicenseDetails($requiredDetails)
    {
        $item = $this->cache->getItem($requiredDetails);
        if ($item->isHit()) {
            return $item->get();
        } else {
            [$isDistributedInstance, $instanceRole, $timEnabled, $supportedIngestionLimit, $supportedQueryLimit] = $this->skuService->getDetailsFromDas();
        }
        return match ($requiredDetails) {
            CacheConstants::IS_DISTRIBUTED_INSTANCE => $isDistributedInstance,
            CacheConstants::INSTANCE_ROLE => $instanceRole,
            CacheConstants::IS_TIM_ENABLED => $timEnabled,
            CacheConstants::TIM_SUPPORTED_INGESTION_LIMIT => $supportedIngestionLimit,
            CacheConstants::TIM_SUPPORTED_QUERY_LIMIT => $supportedQueryLimit,
            default => false
        };
    }

    /**
     * @param $val
     * @return DateTime
     */
    public function parseDateField($val)
    {
        if (strpos($val, '{{') !== false) {
            $regex = '/{{\s*getRelativeDate\(\s*(?\'years\'[^,]*)\s*,\s*(?\'months\'[^,]*)\s*,\s*(?\'days\'[^,]*)\s*,\s*(?\'hours\'[^,]*)\s*,\s*(?\'minutes\'[^,]*)\s*,\s*(?\'seconds\'[^,]*)\s*\)\s*}}/m';
            preg_match_all($regex, $val, $matches, PREG_SET_ORDER, 0);
            if (count($matches) != 1) {
                return null;
            }
            $years = $matches[0]['years'];
            $relativeDate = new DateTime();
            $relativeDate->modify($years . ' years');

            $months = $matches[0]['months'];
            if ($months == '"start"') {
                $relativeDate->setDate($relativeDate->format('Y'), 1, $relativeDate->format('d'));
            } elseif ($months == '"end"') {
                $relativeDate->setDate($relativeDate->format('Y'), 12, $relativeDate->format('d'));
            } elseif (intval($months)) {
                $relativeDate->modify($months . ' months');
            }

            $days = $matches[0]['days'];
            if ($days == '"start"') {
                $relativeDate->setDate($relativeDate->format('Y'), $relativeDate->format('m'), 1);
            } elseif ($days == '"end"') {
                $relativeDate->modify('last day of this month');
            } elseif (intval($days)) {
                $relativeDate->modify($days . ' days');
            }

            $hours = $matches[0]['hours'];
            if ($hours == '"start"') {
                $relativeDate->setTime(0, $relativeDate->format('i'));
            } elseif ($hours == '"end"') {
                $relativeDate->setTime(23, $relativeDate->format('i'));
            } elseif (intval($hours)) {
                $relativeDate->modify($hours . ' hours');
            }

            $minutes = $matches[0]['minutes'];
            if ($minutes == '"start"') {
                $relativeDate->setTime($relativeDate->format('H'), 0);
            } elseif ($minutes == '"end"') {
                $relativeDate->setTime($relativeDate->format('H'), 59);
            } elseif (intval($minutes)) {
                $relativeDate->modify($minutes . ' minutes');
            }

            $seconds = $matches[0]['seconds'];
            if ($seconds == '"start"') {
                $relativeDate->setTime($relativeDate->format('H'), $relativeDate->format('i'), 0);
            } elseif ($seconds == '"end"') {
                $relativeDate->setTime($relativeDate->format('H'), $relativeDate->format('i'), 59);
            } elseif (intval($seconds)) {
                $relativeDate->modify($seconds . ' seconds');
            }
            return $relativeDate;
        }
        return new DateTime($val);
    }

    public function sendMail($templateId, $dataToReplace, $mailDetails, $integrationClient, $defaultMessage)
    {
        $resource = EmailTemplate::class;
        $email_template = [];
        try {

            $object = $this->itemDataProvider->getItem($resource, ["uuid" => $templateId]);
            $subject = $object->getSubject();
            $content = $object->getContent();
        } catch (InvalidArgumentException | \Exception $error) {
            $object = false;
        }

        if (!$object) {
            $email_options = Yaml::parse(file_get_contents('/opt/cyops/configs/cyops-api/mailtemplate.yml'));
            $email_template = $email_options[$defaultMessage];
            $subject = $email_template['subject'];
            $content = $email_template['content'];
        }

        if (strpos($content, 'https://{{@Server_fqhn}}')) {
            $content = str_replace('https://{{@Server_fqhn}}', '{{@Server_fqhn}}', $content);
        }
        if (strpos($content, 'https://{{@cyops_hostname}}')) {
            $content = str_replace('https://{{@cyops_hostname}}', '{{@cyops_hostname}}', $content);
        }
        foreach ($dataToReplace as $key => $value) {
            $content = str_replace($key, $value, $content);
        }

        $smtpObject = '{
                            "connector": "smtp",
                            "operation": "send_richtext_email",
                            "params": {
                                "ccc": "",
                                "bcc": "",
                                "content_type":"text/html"
                            }
                        }';

        $jsonDecodedSmtp = json_decode($smtpObject, true);
        $jsonDecodedSmtp['params']['subject'] = $subject;
        $jsonDecodedSmtp['params']['content'] = $content;
        $jsonDecodedSmtp['params']['to'] = $mailDetails['to'];
        $smtpUri = $this->containers->getParameter('integrations.uri');
        $smtpUri = $smtpUri . '/execute/?format=json';
        $smtpRequest = new GuzzleRequest('post', $smtpUri, $this->headers, json_encode($jsonDecodedSmtp));
        try {
            $integrationClient->send($smtpRequest);
        } catch (BadRequestHttpException | \Exception $exception) {
            $this->logger->error($exception->getMessage());
            $this->logger->error($exception->getCode());
            return false;
        }
        return true;
    }

    // https://stackoverflow.com/questions/20522605/what-is-the-best-way-to-resolve-a-relative-path-like-realpath-for-non-existing
    // Adding custom Function to resolve as realpath might not work with nonexisting file paths.
    public function normalize_path($str)
    {
        $N = 0;
        $A = explode("/", preg_replace("/\/\.\//", '/', $str));  // remove current_location
        $B = [];
        for ($i = sizeof($A) - 1; $i >= 0; --$i) {
            if (trim($A[$i]) === "..") {
                $N++;
            } else {
                if ($N > 0) {
                    $N--;
                } else {
                    $B[] = $A[$i];
                }
            }
        }
        return implode("/", array_reverse($B));
    }

    public function getRestrictedMimeTypeList()
    {
        $restrictedMimeType = [];
        $systemSettingsRepo = $this->entityManager->getRepository(SystemSettings::class);
        $systemSettingsObj = $systemSettingsRepo->findOneBy(["name" => "Restricted File Mime-types"]);
        if ($systemSettingsObj) {
            $restrictedMimeType = $systemSettingsObj->getPublicValues()['values'];
        }
        return $restrictedMimeType;
    }

    public function checkFileDumpPath($dirPath, $path = null)
    {
        if (is_null($path) && strpos(realpath(dirname($dirPath)), AppConstants::TMP_SAFE_PATH) !== 0) {
            throw new BadRequestHttpException(ErrorConstants::INVALID_FILE_NAME);
        } elseif (!is_null($path)) {
            $path = explode('/', $path);
            if (strpos(realpath(dirname($dirPath)), '/' . $path[1] . '/') !== 0) {
                throw new BadRequestHttpException(ErrorConstants::INVALID_FILE_NAME);
            }
        }
    }

    public function isTimSkuEnabled()
    {
        $item = $this->cache->getItem(CacheConstants::IS_TIM_ENABLED);
        if ($item->isHit()) {
            return $item->get();
        } else {
            return $this->getRequiredLicenseDetails(CacheConstants::IS_TIM_ENABLED);
        }
    }

    /*
    This function validate for given relationships fields, if all the related items are existing in database or not.
    Any Item if not existing it will pop from collection, ensuring replication are never failed on MSSP Environment for missing relationships.
    currently we are using this feature, for workflow steps and routes and hence put a check for only validating same for custom modules.
    This validation will only run if we have passed replicated true as part of query params.
    This is done to ensure regular flow keep running and failing for missing relationships.

    * @param string  $className
    * @param array  $value
    * @return array
    */
    public function validateRelationFieldsValues(string $className, array $values)
    {
        $uuidList = array();
        /*
            Checked all possible combination
            1. Directly an IRI is sent
            2. Directly UUID is sent
            3. An Object containing uuid is been sent
            4. An object containing iri is being sent
            5. An array containing uuid is sent
            6. An Array containing iri is sent
            */
        foreach ($values as $item) {
            if (is_string($item) && str_contains($item, '/api/3')) {
                $uuidList[] = explode('/', $item)[4];
            } elseif (is_string($item)) {
                $uuidList[] = $item;
            } elseif (gettype($item) == 'object' && isset($item->{'uuid'})) {
                $uuidList[] = explode('/', $item->{"uuid"});
            } elseif (gettype($item) == 'object' && isset($item->{'@id'})) {
                $uuidList[] = explode('/', $item->{"@id"})[4];
            } elseif (gettype($item) == 'array' && array_key_exists('uuid', $item)) {
                $uuidList[] = $item["uuid"];
            } elseif (gettype($item) == 'array' && array_key_exists('@id', $item)) {
                $uuidList[] = explode('/', $item["@id"])[4];
            }
        }
        $entityRepo = $this->entityManager->getRepository($className);
        $persistedList = $this->getItemWithWhere($entityRepo, ['uuid' => $uuidList], selectFields: ['o.uuid']);
        if (count($persistedList) == count($values)) {
            // All values being passed are present in database.
            return $values;
        } else {
            // Some records are missing in database.
            // Ignore such records.
            $this->logger->warning("Few related records added for linking are not present on system");
            $persistedUuid = array();
            foreach ($persistedList as $item) {
                $persistedUuid[] = strval($item['uuid']);
            }
            return array_intersect($uuidList, $persistedUuid);
        }
    }
}
