<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Utility;

use App\Service\UtilityService;

class FilterUtility
{
    const FIELD_OR_OPERATOR = 'OR';
    const FIELD_AND_OPERATOR = 'AND';
    const LIKE_OPERATOR = 'like';
    const NOT_LIKE_OPERATOR = 'notlike';
    const IN_ALL_OPERATOR = 'in_all';
    const NIN_ALL_OPERATOR = 'nin_all';
    const IN_OPERATOR = 'in';
    const NIN_OPERATOR = 'nin';
    const ADDED_VALUE_IS = 'addedvalueis';
    const EQUALS_OPERATOR = 'eq';
    const NOT_EQUALS_OPERATOR = 'neq';
    const GREATER_THAN_EQUALS_OPERATOR = 'gte';
    const LESS_THAN_EQUALS_OPERATOR = 'lte';
    const GREATER_THAN_OPERATOR = 'gt';
    const LESS_THAN_OPERATOR = 'lt';
    const NULL_OPERATOR = 'isnull';
    const CHANGED_OPERATOR = 'changed';

    const NON_DATETIME_CONVERT_OPERATOR = [self::NULL_OPERATOR, self::CHANGED_OPERATOR];
    // field Constants
    const FIELD = 'field';
    const VALUE = 'value';
    const FIELD_TYPE = 'type';

    public function __construct(private UtilityService $utilityService) {
    }

    /**
     * @param $queryObject
     * @param $currentData
     * @param $previousData
     * @return boolean
     */
    public function evaluateObject($queryObject, $currentData, $previousData, $max_relation_count=100)
    {
        $fieldExpressions = $queryObject != null && array_key_exists('filters', $queryObject) ? $queryObject['filters'] : [];
        $fieldOperator = $queryObject != null && array_key_exists('logic', $queryObject) ? $queryObject['logic'] : self::FIELD_AND_OPERATOR;
        $evaluation = false;
        if (empty($fieldExpressions)) {
            return true;
        }
        if ($fieldOperator === self::FIELD_AND_OPERATOR) {
            $evaluation = true;
        }
        foreach ($fieldExpressions as $fieldExpression) {
            if (array_key_exists('filters', $fieldExpression)) {
                $currEval = $this->evaluateObject($fieldExpression, $currentData, $previousData, $max_relation_count);
                $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                continue;
            }
            if (($currentData && !array_key_exists($fieldExpression[self::FIELD], $currentData)) && ($previousData && !array_key_exists($fieldExpression[self::FIELD], $previousData))) {
                // This will be the case for M2M, For performance optimization, if M2M fields is not changed, it is not normalized.
                // M2M only supports 'isChanged' and 'added' operator. So field not present means there has been no change,
                // and condition should evaluate to False
                if ($fieldOperator === self::FIELD_AND_OPERATOR) {
                    return false;
                }
                continue;
            }
            $currDataVal = $currentData[$fieldExpression[self::FIELD]];
            $fieldExprVal = $fieldExpression[self::VALUE];
            $fieldType = isset($fieldExpression[self::FIELD_TYPE]) ? $fieldExpression[self::FIELD_TYPE] : null;
            $fieldExpressionOperator = $fieldExpression['operator'];

            if ($fieldType == 'datetime' && !in_array($fieldExpressionOperator, self::NON_DATETIME_CONVERT_OPERATOR)) {
                $dateFieldResult = $this->utilityService->parseDateField($fieldExprVal);
                $fieldExprVal = $dateFieldResult ? $dateFieldResult->getTimestamp() : $dateFieldResult;
            }
            if ($fieldExpression[self::FIELD] == 'recordTags') {
                $tags = [];
                foreach ($fieldExprVal as $item) {
                    $exploded_item = explode('/', $item);
                    array_push($tags, end($exploded_item));
                }
                $fieldExprVal = $tags;
            }
            if ($previousData && is_array($fieldExpression) && isset($fieldExpression[self::FIELD])) {
                $prevDataVal = $previousData[$fieldExpression[self::FIELD]];
            } else {
                $prevDataVal = [];
            }
            if (is_array($currDataVal) && isset($currDataVal['@id'])) {
                $currDataVal = $currDataVal['@id'];
            } elseif (is_array($currDataVal)) {
                $finalVal = [];
                foreach ($currDataVal as $val) {
                    $item = isset($val['@id']) ? $val['@id'] : $val;
                    array_push($finalVal, $item);
                }
                $currDataVal = $finalVal;
            }
            if (is_array($prevDataVal) && isset($prevDataVal['@id'])) {
                $prevDataVal = $prevDataVal['@id'];
            } elseif (is_array($prevDataVal)) {
                $finalVal = [];
                foreach ($prevDataVal as $val) {
                    $item = isset($val['@id']) ? $val['@id'] : $val;
                    array_push($finalVal, $item);
                }
                $prevDataVal = $finalVal;
            }
            if (is_array($fieldExprVal) && isset($fieldExprVal['@id'])) {
                $fieldExprVal = $fieldExprVal['@id'];
            } elseif (is_array($fieldExprVal)) {
                $finalVal = [];
                foreach ($fieldExprVal as $val) {
                    $item = isset($val['@id']) ? $val['@id'] : $val;
                    array_push($finalVal, $item);
                }
                $fieldExprVal = $finalVal;
            }
            // Backwards compatibility after tags operator change
            // TODO: Remove after a few releases
            if ($fieldExpression['field'] == 'recordTags' && $fieldExpression['operator'] == self::IN_OPERATOR && !$fieldExpression['useInOperator']) {
                $fieldExpression['operator'] = self::IN_ALL_OPERATOR;
            }
            switch ($fieldExpression['operator']) {

                case self::EQUALS_OPERATOR:
                    $currEval = strtolower($currDataVal) == strtolower($fieldExprVal);
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::NOT_EQUALS_OPERATOR:
                    $currEval = strtolower($currDataVal) !== strtolower($fieldExprVal);
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::LIKE_OPERATOR:
                    $currEval = false;
                    $fieldExprVal = preg_quote($fieldExprVal);
                    $expression = str_replace('%', '.*', $fieldExprVal);
                    $expression = str_replace('_', '.{1}', $expression);
                    $regex = sprintf('/^%s$/mi', $expression);
                    preg_match_all($regex, $currDataVal, $matches, PREG_SET_ORDER, 0);
                    $currEval = count($matches) > 0;
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::NOT_LIKE_OPERATOR:
                    $currEval = false;
                    $fieldExprVal = preg_quote($fieldExprVal);
                    $expression = str_replace('%', '.*', $fieldExprVal);
                    $expression = str_replace('_', '.{1}', $expression);
                    $regex = sprintf('/%s/mi', $expression);
                    preg_match_all($regex, $currDataVal, $matches, PREG_SET_ORDER, 0);
                    $currEval = count($matches) == 0;
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::NULL_OPERATOR:
                    $currEval = (is_null($currDataVal) ? 'true' : 'false') === $fieldExprVal;
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::IN_ALL_OPERATOR:
                    $currDataVal = $currDataVal ? $currDataVal : [];
                    $currEval = $this->in_array($fieldExprVal, $currDataVal, true);
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::NIN_ALL_OPERATOR:
                    $currDataVal = $currDataVal ? $currDataVal : [];
                    $currEval = !($this->in_array($fieldExprVal, $currDataVal, true));
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::IN_OPERATOR:
                    $currDataVal = $currDataVal ? $currDataVal : [];
                    $currEval = $this->in_array($fieldExprVal, $currDataVal);
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::NIN_OPERATOR:
                    $currDataVal = $currDataVal ? $currDataVal : [];
                    $currEval = !($this->in_array($fieldExprVal, $currDataVal));
                    $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    break;
                case self::CHANGED_OPERATOR:
                    if (is_array($currDataVal) || is_array($prevDataVal)) {
                        $prevDataVal = $prevDataVal ? $prevDataVal : [];
                        # no of related records is large. since the key was present in data, detect as changed
                        if (
                            !is_null($currDataVal) && count($currDataVal) >= $max_relation_count &&
                            !is_null($prevDataVal) && count($prevDataVal) >= $max_relation_count
                        ) {
                            $currEval = true;
                        } elseif (!is_null($currDataVal)) {
                            sort($currDataVal);
                            sort($prevDataVal);
                            $currEval = $currDataVal !== $prevDataVal;
                        } else {
                            $currEval = !is_null($prevDataVal);
                        }
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                        break;
                    } else {
                        $currEval = strtolower($currDataVal) !== strtolower($prevDataVal);
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                        break;
                    }
                case self::ADDED_VALUE_IS:
                    if (is_array($currDataVal)) {
                        $prevDataVal = $prevDataVal ? $prevDataVal : [];
                        $fieldExprVal = is_array($fieldExprVal) ? $fieldExprVal : [$fieldExprVal];
                        $currentMatch = array_intersect($fieldExprVal, $currDataVal);
                        $prevMatch = array_intersect($fieldExprVal, $prevDataVal);
                        if (count($currentMatch) > 0 && count($prevMatch) == 0) {
                            $currEval = true;
                        } else {
                            $currEval = false;
                        }
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                        break;
                    } else {
                        $evaluation = false;
                        break;
                    }
                case self::GREATER_THAN_EQUALS_OPERATOR:
                    if (is_int($fieldExprVal) && is_int($currDataVal)) {
                        $currEval = $currDataVal >= $fieldExprVal;
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    } else {
                        $evaluation = false;
                    }
                    break;
                case self::LESS_THAN_EQUALS_OPERATOR:
                    if (is_int($fieldExprVal) && is_int($currDataVal)) {
                        $currEval = $currDataVal <= $fieldExprVal;
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    } else {
                        $evaluation = false;
                    }
                    break;
                case self::GREATER_THAN_OPERATOR:
                    if (is_int($fieldExprVal) && is_int($currDataVal)) {
                        $currEval = $currDataVal > $fieldExprVal;
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    } else {
                        $evaluation = false;
                    }
                    break;
                case self::LESS_THAN_OPERATOR:
                    if (is_int($fieldExprVal) && is_int($currDataVal)) {
                        $currEval = $currDataVal < $fieldExprVal;
                        $evaluation = $this->evaluateOperator($fieldOperator, $currEval, $evaluation);
                    } else {
                        $evaluation = false;
                    }
                    break;
                default:
                    return $this->evaluateOperator($fieldOperator, false, $evaluation);
            }
        }
        return $evaluation;
    }

    /**
     * @param $operator
     * @param $prevEval
     * @param $currEval
     * @return boolean
     */
    private function evaluateOperator($operator, $prevEval, $currEval)
    {
        if ($operator === self::FIELD_OR_OPERATOR) {
            return $prevEval || $currEval;
        } elseif ($operator === self::FIELD_AND_OPERATOR) {
            return $prevEval && $currEval;
        } else {
            return false;
        }
    }

    public function in_array($needles, $haystack, $all = false)
    {
        if (!is_array($haystack)) {
            $haystack = [$haystack];
        }
        $intersect = array_uintersect($needles, $haystack, 'strcasecmp');
        return $all ? count($intersect) == count($needles) : !empty($intersect);
    }
}
