<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Utility;

use stdClass;
use App\Entity\Custom\Tenant;
use App\Traits\PeerReplicable;
use App\Service\MetadataHelper;
use App\Service\UtilityService;
use App\Entity\Workflow\Workflow;
use Doctrine\ORM\EntityManagerInterface;
use App\Entity\Workflow\WorkflowCollection;
use App\Entity\ModelMetadata\StagingModelMetadata;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

class MsspUtility
{

    public const DATA_ROUTING_KEY = 'route.crudhub.data.remoterequest';
    public const PLAYBOOK_ALIAS_ROUTING_KEY = 'route.crudhub.instructions.remoterequest.playbookaliaspush';
    public const MMD_ROUTING_KEY = 'route.crudhub.instructions.remoterequest.stagingmmdupdate';
    public const IS_DISTRIBUTED_INSTANCE = 'isDistributedInstance';

    public function __construct(
        protected UtilityService $utilityService,
        protected MetadataHelper $metadataHelper,
        protected EntityManagerInterface $entityManager,
        protected ContainerInterface $container,
        protected TagAwareCacheInterface $cache,
        protected FilterUtility $filterUtility
    ) {
    }

    public function isPeerReplicable($entity_class)
    {
        return in_array(PeerReplicable::class, class_uses($entity_class));
    }

    public function prepareReplicationDiffData(
        $currentData,
        $replicationOp,
        $previousData = [],
        $changeData = [],
        $isRestore = false
    ) {
        $obj = new stdClass();
        if ($replicationOp == 'PUT' && isset($previousData['tenant']) && $previousData['tenant'] != $currentData['tenant']) {
            $currentData['__sync'] = true;
        }
        if ($replicationOp == 'PUT') {
            if ($isRestore) {
                $currentData['__sync'] = true;
            }
        }
        $obj->{"old_data"} = $previousData;
        $obj->{"new_data"} = $currentData;
        $obj->{"change_data"} = $changeData;
        $obj->{"method"} = $replicationOp;
        return $obj;
    }

    public function replicateData($entity, $requestBody, $objects, $method)
    {
        $item = $this->cache->getItem(self::IS_DISTRIBUTED_INSTANCE);
        $isDistributedInstance = $item->get();
        if (!$isDistributedInstance) {
            // Instance is not multi tenant no replication required.
            return;
        }
        $replicateDataList = [];
        $sync = false;
        $replicateData = true;
        $routingKey = '';
        $entityClass = $entity;
        $isReplicationEnabled = $this->metadataHelper->getReplicationEnabledStatus($entityClass) ?? null;
        $isCheckReplicationFilter = false;

        if (isset($requestBody->data) && gettype($requestBody->data) == gettype([])) {
            if (isset($requestBody->data[0]->{'__sync'})) {
                $sync = $requestBody->data[0]->{'__sync'};
            }
        } elseif (isset($requestBody->{'__sync'})) {
            $sync = $requestBody->{'__sync'};
        }

        if (isset($requestBody->{'__replicateData'})) {
            $replicateData = $requestBody->{'__replicateData'};
            if (is_object($objects[0]->old_data) && isset($objects[0]->old_data->{'conflict'}) && $objects[0]->old_data->{'conflict'}) {
                $replicateData = true;
            }
        }
        if ($this->isPeerReplicable($entityClass) && $isReplicationEnabled) {
            $routingKey = self::DATA_ROUTING_KEY;
            $replicateDataList = $objects;
            if ($sync) {
                $replicateDataList = $this->processSyncRequest($objects);
            }
            $isCheckReplicationFilter = true;
        } elseif ($entityClass == Workflow::class) {
            $routingKey = self::PLAYBOOK_ALIAS_ROUTING_KEY;
            $replicateDataList = $this->processPlaybook($objects, $replicateDataList);
        } elseif ($entityClass == WorkflowCollection::class) {
            $routingKey = self::PLAYBOOK_ALIAS_ROUTING_KEY;
            $replicateDataList = $this->processPlaybookCollections($objects, $replicateDataList, $method);
        } elseif ($entityClass == StagingModelMetadata::class) {
            $routingKey = self::MMD_ROUTING_KEY;
            $replicateDataList = $objects;
        }
        if ($replicateData && count($replicateDataList) > 0) {
            $this->sendReplicateData($replicateDataList, $method, $routingKey, $isCheckReplicationFilter, $entityClass);
        }
    }

    private function sendReplicateData($object, $method, $routingKey, $isCheckReplicationFilter, $entityClass)
    {
        $groupedObj = (object) null;
        foreach ($object as $obj) {
            $objData = is_array($obj->new_data) ? $obj->new_data : $obj->old_data;
            if ($isCheckReplicationFilter) {
                $entityFieldMap = $this->metadataHelper->getEntityFields($entityClass) ?? null;
                $replicationFilters = $entityFieldMap['replicationFilters'];
                if (!$this->filterUtility->evaluateObject($replicationFilters, $objData, null)) {
                    continue;
                }
            }
            $destinationIds = $this->getDestinationIds($objData);
            if (!$destinationIds) {
                continue;
            }
            foreach ($destinationIds as $id) {
                if (isset($groupedObj->$id)) {
                    $merge = array_merge($groupedObj->$id, [$obj]);
                    $groupedObj->$id = $merge;
                } else {
                    $groupedObj->$id = [$obj];
                }
            }
        }
        if ($groupedObj) {
            foreach ($groupedObj as $key => $value) {
                $replicationData = $this->prepareReplicationData($value, $method, $this->getTenantObject()->getTenantId(), $key);
                $this->sendReplicationData($replicationData, $routingKey);
            }
        }
    }

    public function getTenantObject($tenantUuid = null)
    {
        if ($tenantUuid) {
            $tenantFilter = ['uuid' => $tenantUuid];
        } else {
            $tenantFilter = ['role' => 'self'];
        }
        $tenantResource = Tenant::class;
        $entityRepo = $this->entityManager->getRepository($tenantResource);
        $tenantObj = $this->utilityService->getItemWithWhere($entityRepo, $tenantFilter);
        if ($tenantObj) {
            return $tenantObj[0];
        }
        return null;
    }

    /**
     * @param $contextRef
     * @param $Object
     * @param $method
     * @return array
     */
    protected function getDestinationIds($Object)
    {
        $ids = [];
        $tenantUuid = null;
        $tenantResource = Tenant::class;
        if (isset($Object['tenant']) && $Object['tenant']) {
            $tenant = $Object['tenant'];
            $tenantArray = explode('/', $tenant);
            $tenantUuid = end($tenantArray);
        }
        $tenantObj = $this->getTenantObject($tenantUuid);
        if ($tenantObj) {
            if ($tenantObj->getRole() == 'self' || $Object['@type'] == 'Workflow') {
                $tenantFilter = ['role' => 'master', 'isDedicated' => true];
                $entityRepo = $this->entityManager->getRepository($tenantResource);
                $destinationTenants = $this->utilityService->getItemWithWhere($entityRepo, $tenantFilter);
                if ($destinationTenants) {
                    foreach ($destinationTenants as $destinationTenant) {
                        $masterId = $destinationTenant->getTenantId();
                        array_push($ids, $masterId);
                    }
                }
            } elseif ($tenantObj->getTenantId() && $tenantObj->getIsDedicated()) {
                array_push($ids, $tenantObj->getTenantId());
            }
        }
        return $ids;
    }

    public function prepareReplicationData($object, $method, $sourceId, $destinationId)
    {
        return [
            'data' => $object,
            'method' => $method,
            'sourceId' => $sourceId,
            'destinationId' => $destinationId,
        ];
    }

    protected function processSyncRequest($objects)
    {
        foreach ($objects as $obj) {
            $obj->new_data['__sync'] = true;
            $obj->new_data['__replicationLevel'] = 1;
        }
        return $objects;
    }

    protected function processPlaybook($objects, $replicateDataList)
    {
        $remoteExecutableNew = false;
        $remoteExecutableOld = false;
        foreach ($objects as $obj) {
            if (is_array($obj->new_data)) {
                $remoteExecutableNew = $obj->new_data['remoteExecutableFlag'];
            }
            if (is_array($obj->old_data)) {
                $remoteExecutableOld = $obj->old_data['remoteExecutableFlag'];
            }
            if ($remoteExecutableNew || $remoteExecutableOld) {
                if (is_array($obj->new_data)) {
                    unset($obj->new_data['steps']);
                    unset($obj->new_data['routes']);
                    unset($obj->new_data['versions']);
                }
                if (is_array($obj->old_data)) {
                    unset($obj->old_data['steps']);
                    unset($obj->old_data['routes']);
                    unset($obj->new_data['versions']);
                }
                array_push($replicateDataList, $obj);
            }
        }
        return $replicateDataList;
    }

    protected function processPlaybookCollections($objects, $replicateDataList, $method)
    {
        if ($method != 'PUT') {
            foreach ($objects as $wf_collection_obj) {
                if (is_array($wf_collection_obj->new_data) && count($wf_collection_obj->new_data)) {
                    $processObject = $wf_collection_obj->new_data;
                } elseif (is_array($wf_collection_obj->old_data) && count($wf_collection_obj->old_data)) {
                    $processObject = $wf_collection_obj->old_data;
                } else {
                    return [];
                }
                if (isset($processObject['workflows']) && count($processObject['workflows'])) {
                    foreach ($processObject['workflows'] as $wf_object) {
                        if ($wf_object['remoteExecutableFlag']) {
                            unset($wf_object['steps']);
                            unset($wf_object['routes']);
                            unset($wf_object['versions']);
                            if ($method == 'POST') {
                                $diffObj = $this->prepareReplicationDiffData($wf_object, $method);
                            } elseif ($method == 'DELETE') {
                                $diffObj = $this->prepareReplicationDiffData([], $method, $wf_object);
                            }
                            array_push($replicateDataList, $diffObj);
                        }
                    }
                }
            }
        }
        return $replicateDataList;
    }

    public function sendReplicationData(array $data, $routingKey)
    {
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->setContentType('application/json');
        $this->container->get('old_sound_rabbit_mq.cyops_crud_auditing_producer')->publish(json_encode($data), $routingKey);
    }
}
