<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Utility;

use Ramsey\Uuid\Uuid;
use App\Entity\Core\ConfigProp;
use App\Service\UtilityService;
use Doctrine\ORM\EntityManagerInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\HttpFoundation\Exception\BadRequestException;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class PayloadUtils
{

    protected $utilityService;
    protected $entityManager;

    public const STATIC_STRING = 'AK*&^%$#@!AG';
    public const MAX_TIME_GAP_ALLOWED = 60;
    public const LICENSE_TAMPERING_DETECTED = 'Invalid License Information';
    public const REQUIRED_MANDATE_CONFIG_MISSING = 'Mandatory config required is missing';
    public const FEED_INGESTION_LIMIT_EXHAUSTED = "Daily Feed Ingestion Limit reached";
    public const FEED_COUNT = 'feed_count';
    public const TIM_SUPPORTED_INGESTION_LIMIT = 'tim_supported_ingestion_limit';

    public function __construct(
        UtilityService $utilityService,
        EntityManagerInterface $entityManager,
        private LoggerInterface $logger
    )
    {
        $this->utilityService = $utilityService;
        $this->entityManager = $entityManager;
    }

    public function validatePayload($resource, &$objectData)
    {
        if ($resource != 'App\\Entity\\Custom\\ThreatIntel' || $this->utilityService->isTimSkuEnabled()) {
            return;
        }
        $permissibleIngestionCount = $this->utilityService->getRequiredLicenseDetails(self::TIM_SUPPORTED_INGESTION_LIMIT);
        $configPropRepo = $this->entityManager->getRepository(ConfigProp::class);
        $threatIntelRepo = $this->entityManager->getRepository('\\App\\Entity\\Custom\\ThreatIntel');
        $feedInfo = $configPropRepo->findOneBy(['key' => self::FEED_COUNT]);
        if (empty($feedInfo)) {
            $this->logger->warning(self::LICENSE_TAMPERING_DETECTED);
            throw new BadRequestHttpException(self::REQUIRED_MANDATE_CONFIG_MISSING);
        }
        $lastCreatedFeed = $threatIntelRepo->createQueryBuilder('o')->orderBy('o.createDate', 'DESC')->setMaxResults(1)->getQuery()->getOneOrNullResult();
        
        if ($lastCreatedFeed) {
            self::validateModifyDate($lastCreatedFeed->getCreateDate(), $feedInfo->getModifiedAt());
        }
        $usedLimit = $feedInfo->getValue();
        if (!self::validateHash($feedInfo->getChecksum(), $usedLimit, $feedInfo->getModifiedAt())) {
            throw new BadRequestException(self::LICENSE_TAMPERING_DETECTED);
        }
        $usedLimit = self::identifyCurrentDayLimit($feedInfo->getModifiedAt(), $usedLimit);

        $allowedIngestionSize = $permissibleIngestionCount - $usedLimit;
        if ($allowedIngestionSize <= 0) {
            throw new BadRequestHttpException(self::FEED_INGESTION_LIMIT_EXHAUSTED);
        }

        // Slice the incoming data with permissible limit, ignore the remaining data
        $objectData['data'] = array_slice($objectData['data'], 0, $allowedIngestionSize);
        $currentIngestionDataCount = count($objectData['data']);

        $newLimit = $usedLimit + $currentIngestionDataCount;
        $newLimit = ($newLimit <= $permissibleIngestionCount) ? $newLimit : $permissibleIngestionCount;
        $now = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $feedInfo->setValue($newLimit);
        $feedInfo->setModifiedAt($now);
        $randomUuid = Uuid::uuid4()->toString();
        $feedInfo->setChecksum(self::computeHash($randomUuid, $newLimit, $now));
        $feedInfo->setSalt($randomUuid);
        $this->entityManager->persist($feedInfo);
    }

    public static function computeHash(string $secret_key, int $usedLimit, \Datetime $now)
    {
        $jsonData = [
            'date' => $now->format('U'),
            'feedUsedLimit' => self::STATIC_STRING . strval($usedLimit)
        ];
        $data = json_encode($jsonData);
        return $secret_key . "#" . hash("sha512", $secret_key . $data);
    }

    public static function validateHash(string $checksum, int $usedLimit, $modifiedAt)
    {
        $jsonData = [
            'date' => $modifiedAt->format('U'),
            'feedUsedLimit' => self::STATIC_STRING . strval($usedLimit)
        ];
        $data = json_encode($jsonData);
        if (hash("sha512", substr($checksum, 0, strpos($checksum, "#")) . $data) == substr($checksum, strpos($checksum, "#") + 1)) {
            return true;
        }
        return false;
    }

    public static function identifyCurrentDayLimit(\DateTime $modifiedAt, int $usedLimit)
    {
        $today = new \DateTime('NOW', new \DateTimeZone('UTC'));
        $diff = ($today->setTime(00, 00))->diff($modifiedAt->setTime(00, 00));
        $diffDays = (int)$diff->format("%R%a");
        if ($diffDays != 0) {
            return 0;
        }
        return $usedLimit;
    }


    public static function validateModifyDate(\DateTime $lastCreatedFeed, \DateTime $modifiedAt)
    {
        $dateDiff = $lastCreatedFeed->diff($modifiedAt);
        if ($dateDiff->i > self::MAX_TIME_GAP_ALLOWED) {
            throw new BadRequestHttpException(self::LICENSE_TAMPERING_DETECTED);
        }
    }
}