<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */


namespace App\Utility;

use Symfony\Component\HttpClient\Exception\InvalidArgumentException;


/**
 * Class Resolver
 *
 * Provides functions for array manipulation with delimited notation.
 *
 * If making changes, please update the tests!
 *
 * @package App\Utility
 */
class Resolver
{
    const ESCAPE_REGEX = "/(?<![^\\\\]\\\\)\\%s/";

    /**
     * This function will get or set a resolved path against an array.
     * If the path cannot be resolved, the array is returned unmodified.
     *
     * Escaping delimiters (with '\') is also supported, as shown with the example:
     *
     * $test = [
     *   "a.d" => [
     *     "b" => "c"
     *   ]
     * ];
     *
     * Set:
     * set($test, "a\.d.b", 'test');
     *
     * Get:
     * get($test, "a\.d.b");
     *
     * $test now is:
     * [
     *   "a.d" => [
     *     "b" => "test"
     *   ]
     * ];
     *
     * @param array $array The array to resolve
     * @param string $path The path to resolve against the array
     * @param null $value The value to set the resolved path to
     * @param string $delimiter The delimiter of the path tokens
     */
    static public function set(array &$array, $path, $value = null, $delimiter = '.')
    {
        $path = self::sanitizePath($path, $delimiter);
        $regex = sprintf(self::ESCAPE_REGEX, $delimiter);
        $tokens = preg_split($regex, $path);

        while(count($tokens) > 1) {
            $key = array_shift($tokens);
            if(!isset($array[$key])) {
                $array[$key] = [];
            }
            $array = &$array[$key];
        }

        $key = reset($tokens);
        $array[$key] = $value;
    }

    /**
     * This works like set, but will not change any values.
     *
     * @see Resolver::set
     *
     * @param array $array
     * @param string $path
     * @param string $delimiter
     *
     * @return mixed
     *
     * @throws InvalidArgumentException
     */
    static public function get(array &$array, $path, $delimiter = '.')
    {
        $path = self::sanitizePath($path, $delimiter);
        $regex = sprintf(self::ESCAPE_REGEX, $delimiter);
        $tokens = preg_split($regex, $path);

        while(count($tokens) > 1) {
            $key = array_shift($tokens);
            if(!isset($array[$key])) {
                break;
            }
            $array = &$array[$key];
        }

        $key = reset($tokens);
        if (is_array($array) && !array_key_exists($key, $array)) {
            throw new \InvalidArgumentException(sprintf('The "%s" does not exist in Resolver class.', $key));
        }

        return $array[$key];
    }

    /**
     * Unsets a key by path
     *
     * @param array $array
     * @param $path
     * @param string $delimiter
     *
     * @throws InvalidArgumentException
     */
    static public function uset(array &$array, $path, $delimiter = '.')
    {
        $path = self::sanitizePath($path, $delimiter);
        $regex = sprintf(self::ESCAPE_REGEX, preg_quote($delimiter));
        $tokens = preg_split($regex, $path);

        while(count($tokens) > 1) {
            $key = array_shift($tokens);
            if(!isset($array[$key])) {
                break;
            }
            $array = &$array[$key];
        }

        $key = reset($tokens);
        if (is_array($array) && !array_key_exists($key, $array)) {
            throw new \InvalidArgumentException(sprintf('The "%s" does not exist in Resolver class.', $key));
        }

        unset($array[$key]);
    }

    private static function sanitizePath($path, $delimiter)
    {
        //Remove first delimiter from path
        return preg_replace(sprintf('/(^[\%s])/', preg_quote($delimiter)), '', $path);
    }
}