<?php
/* Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Utility;

use Ramsey\Uuid\Uuid;
use Doctrine\ORM\EntityManagerInterface;
use App\Entity\ActionSuggestion\Statistic;
use App\Entity\Core\Picklist;

class StatsUtility
{

  public const WORKFLOW_IRI = '/api/3/workflows/';

  public function __construct(
    protected EntityManagerInterface $entityManager
  ) {
  }

  public const STATS_TYPE = [
    'Workflow' => '017a3923-62a5-4e61-8c86-7ac98e2da3df',
    'Connector Action' => 'b39f2d5e-c4e3-4f0a-aa24-25010d9f639c',
    'Others' => 'eea2dd46-b148-4aaf-9354-dbef37b90a72',
    'Misc' => '7f694fe6-26f0-46d3-afe2-5673a954bca8'
  ];

  /**
   * Converts IRIs to UUIds
   *
   * @param string $iri
   * @return string
   */
  protected static function getUuidFromIri($iri)
  {
    $iriParts = explode('/', $iri);
    return end($iriParts);
  }

  public function saveStats(string $resource, array $recordIris, string $statsType, string $actionId, string $actionName, array $taskIds)
  {
    $picklistsRepo = $this->entityManager->getRepository(Picklist::class);
    foreach ($recordIris as $index => $recordIri) {
      $stats = new Statistic();
      $stats->setUuid(Uuid::uuid4()->toString());
      $stats->setRecordUuid($this->getUuidFromIri($recordIri));
      $stats->setRecordType($resource);
      $stats->setStatsType($picklistsRepo->findOneBy(['uuid' => [self::STATS_TYPE[$statsType]]]));
      $stats->setActionUuid($actionId);
      $stats->setAction($actionName);
      $stats->setTaskId($taskIds[$index] ?? $taskIds[0]);
      $this->entityManager->persist($stats);
    }
    $this->entityManager->flush();
  }

  private function checkAndWeightSequence(array &$weightedGraph, $startNode, $endNode, $similarityScore)
  {
    foreach ($weightedGraph as &$subWeightedGraph) {
      if (@$subWeightedGraph[0] === $startNode && @$subWeightedGraph[1] === $endNode) {
        $subWeightedGraph = array($startNode, $endNode, @$subWeightedGraph[2] + $similarityScore);
        return;
      }
    }
    $weightedGraph[] = array($startNode, $endNode, $similarityScore);
  }

  private function prepareActionSeqWeightMap($weightedGraph, $actionTriggerSeq, $similarityScore = 10)
  {
    $counter = 0;
    if (count($actionTriggerSeq) == 1) {
      $this->checkAndWeightSequence($weightedGraph, $actionTriggerSeq[$counter], $actionTriggerSeq[$counter], $similarityScore);
    }

    while ($counter  < count($actionTriggerSeq) - 1) {
      $this->checkAndWeightSequence($weightedGraph, $actionTriggerSeq[$counter], $actionTriggerSeq[$counter + 1], $similarityScore);
      $counter++;
    }
    array_multisort(array_column($weightedGraph, 2), SORT_DESC, $weightedGraph);
    return $weightedGraph;
  }

  // Recursive function to find a path sequence through the graph
  private function findPath($adjList, $weights, $current, $visited, $path, $unvisited)
  {
    $visited[$current] = true;
    $unvisited = array_diff($unvisited, [$current]);
    $path[] = $current;
    if (empty($unvisited)) {
      return $path;
    }
    $maxWeight = -1;
    $maxPath = null;
    foreach ($adjList[$current] as $neighbor => $weight) {
      if (!array_key_exists($neighbor, $visited)) {
        $result = $this->findPath($adjList, $weights, $neighbor, $visited, $path, $unvisited);
        if ($result !== null) {
          $weight += $weights["$current,$neighbor"];
          if ($weight > $maxWeight) {
            $maxWeight = $weight;
            $maxPath = $result;
          }
        }
      }
    }
    array_pop($path);
    unset($visited[$current]);
    return $maxPath;
  }

  private function recommendSequence($weightedGraph)
  {
    $nodes = array_unique(array_merge(array_column($weightedGraph, 0), array_column($weightedGraph, 1)));

    $adjList = [];
    $weights = [];

    foreach ($weightedGraph as $edge) {
      $start = $edge[0];
      $end = $edge[1];
      $weight = $edge[2];

      if (!array_key_exists($start, $adjList)) {
        $adjList[$start] = [];
      }
      $adjList[$start][$end] = $weight;
      $weights["$start,$end"] = $weight;
    }

    $path = [];

    foreach ($nodes as $node) {
      if (!in_array($node, $path)) {
        $visited = [];
        $result = $this->findPath($adjList, $weights, $node, null, $visited, []);
        if ($result !== null && !in_array($result, $path)) {
          $path = array_merge($path, $result);
        }
      }
    }

    // Check if all nodes are visited
    if (count($path) < count($nodes)) {
      $unvisited = array_diff($nodes, $path);
      $path = array_merge($path, $unvisited);
    }
    return $path;
  }

  public function computeActionSequence(array $statsMap, array $similarRecords)
  {
    $weightedGraph = [];

    foreach ($similarRecords as $similarRecord) {
      $actionTriggerSeq = $statsMap[$similarRecord['uuid']] ?? [];
      if (empty($actionTriggerSeq)) {
        continue;
      }
      $weightedGraph = $this->prepareActionSeqWeightMap($weightedGraph, $actionTriggerSeq, $similarRecord['score']);
    }
    // $weightedGraph = [
    //   ['A', 'B', 1],
    //   ['B', 'C', 2],
    //   ['C', 'D', 3],
    //   ['D', 'E', 2],
    //   ['A', 'C', 1]
    // ];
    return [$this->recommendSequence($weightedGraph), $weightedGraph];
  }

  public function prepareWorkflowNameUuidObject($sequences, $actionUuidToNameMap)
  {
    $response = [];
    foreach ($sequences as $sequence) {
      $response[] = ['actionUuid' => self::WORKFLOW_IRI . $sequence, 'actionName' => $actionUuidToNameMap[$sequence]];
    }
    return $response;
  }

  public function getCurrentExecuteWorkflow($recordUuid)
  {
    $preExecutedWorkflow = [];
    $statsRepo = $this->entityManager->getRepository(Statistic::class);
    $statsObjects = $statsRepo->findBy(['recordUuid' => $recordUuid]);
    foreach ($statsObjects as $statsObject) {
      if (!in_array($statsObject->getActionUuid(), $preExecutedWorkflow)) {
        $preExecutedWorkflow[] = $statsObject->getActionUuid();
      }
    }
    return $preExecutedWorkflow;
  }
}
